// Find a list of MAP configurations by recursively setting values of the variables with high connection to undetermined variables.
// Difference from findmaxconfig8: operate on mix_facgraph data structure.
// Difference from findmaxconfig9: use maxsumproduct3 function.
// Notice path selection variables are not fixed during the iteration, even though the beliefs are uneven.
// This is because there is only one factor node connect to each path selection variable.
// Difference from findmaxconfig10: ignore isolated variables when setting values.  An isolated variable does not connect to any factor node, i.e., is not associated with ko effects.
// Difference from previous findmaxconfig functions: can choose enumerating all MAP configs or narrowing down to one MAP config.
// mode=0: enumerate all MAP configs.  mode=1: choose one MAP config.

static int nodeind=0;

//int * findmaxconfig15(struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int mode, int *Nmapconfigs)
int * findmaxconfig15(struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int mode, int *Nmapconfigs, struct csparsegraph *model, int *sel2ref, char *anothergenenames)
{
  int i, j, k, *configuration, *mapconfigs, *isolated;
  struct confignode *configtree;
  
  configuration=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(configuration+i)=-1;

  // Find the max configs of variables.
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      int Nv=0, v[(maxbelief+i)->Nconfigs];
      double sumval, val, maxval=-1;
      
      sumval=0;
      for (j=0; j<=((maxbelief+i)->Nconfigs-1); j++)
	sumval+=*((maxbelief+i)->probs+j);
      for (j=0; j<=((maxbelief+i)->Nconfigs-1); j++)
	if ((val=*((maxbelief+i)->probs+j)/sumval)>maxval)
	  maxval=val;
      for (j=0; j<=((maxbelief+i)->Nconfigs-1); j++)
	{
	  if ((maxval-(val=*((maxbelief+i)->probs+j)/sumval))<1e-20)
	    {
	      Nv++; v[Nv-1]=j;
	    }
	}
      if (Nv==1)
	*(configuration+i)=v[0];
    }
  
  // Set the variables which are already fixed.
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      if (*(configuration+i)!=-1)
	{
	  int newfix=i;

	  for (j=0; j<=(fgraph->Ne*2-1); j++)
	    {
	      if (((maxmsg+j)->dir==0)&&((maxmsg+j)->from==newfix))
		{
		  int ind=*(configuration+newfix);
		  
		  for (k=0; k<=((maxmsg+j)->phi->Nconfigs-1); k++)
		    *((maxmsg+j)->phi->probs+k)=0.0;
		  *((maxmsg+j)->phi->probs+ind)=1.0;
		}
	    }
	}
    }

  // Identify isolated variables.
  // An isolated variable does not connect to any factor node.
  // Since path selection variables should not be fixed, they are also considered as isolated variables. 
  isolated=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(isolated+i)=1;
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    for (j=0; j<=((fgraph->facnodes+i)->phivi->Nvars-1); j++)
      *(isolated+*((fgraph->facnodes+i)->phivi->varinds+j))=0;
  for (i=0; i<=(fgraph->Nvars-1); i++)
    if ((fgraph->varnodes+i)->type==PATHSEL)
      *(isolated+i)=1;


  // Fix one variable each time until all configuration values are determined.
  // Build the root configuration node.
  configtree=malloc(sizeof(struct confignode));
  configtree->Nvars=fgraph->Nvars; configtree->level=0;
  configtree->config=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(configtree->config+i)=*(configuration+i);
  configtree->Nchildren=0;

  // Debug
  configtree->ind=nodeind; nodeind++;

  
  // mode=0.  Recursively fix the values of variables and spin off children nodes until all edge existence and sign variables are determined.
  if (mode==0)
    recurse_fixval9(configtree,fgraph,maxmsg,maxbelief,isolated);

  // mode=1.  Recursively fix variables to one value.  Create one MAP config.
  else if (mode==1)
    //recurse_fixval10(configtree,fgraph,maxmsg,maxbelief,isolated);
    recurse_fixval10(configtree,fgraph,maxmsg,maxbelief,isolated,model,sel2ref,anothergenenames);


  // Obtain the MAP configurations from the configtree.
  *Nmapconfigs=0;
  mapconfigs=malloc(sizeof(int)*fgraph->Nvars);
  mapconfigs=recurse_getmapconfigs(configtree,Nmapconfigs,mapconfigs);
  
  // Free the space allocated to configtree.
  recurse_freeconfignode(configtree);
  free(configtree); free(configuration); free(isolated);

  return mapconfigs;
}


// Recursively fix values of the variables with the highest connection to undetermined variables.
// Difference from recurse_fixval: operates on mix_facgraph structure and does not require sumproduct.
// Difference from recurse_fixval2: use maxsumproduct3 function to perform max product or sum product.
// Difference from recurse_fixval3: ignore isolated variables when setting values.
void recurse_fixval9(struct confignode *configtree, struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int *isolated)
{
  int i, j, k, *influencecnt, undetermined, maxconnnode, *configuration, *fixedvals;
  struct message *tmpmsg;
  struct potential *tmpbelief;
  struct confignode *child1, *child2; 
  
  
  configuration=configtree->config;

  // Debug
  //printf("level=%d, \n",configtree->level);

  // If all non-isolated variables are fixed, then return.
  undetermined=0; i=0;
  while ((i<=(fgraph->Nvars-1))&&(undetermined==0))
    {
      if ((*(isolated+i)==0)&&(*(configuration+i)==-1))
	undetermined=1;
      i++;
    }
  if (undetermined==0)
    return;


  // Perform maxproduct.
  tmpmsg=malloc(sizeof(struct message)*fgraph->Ne*2);
  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      (tmpmsg+i)->from=(maxmsg+i)->from;
      (tmpmsg+i)->to=(maxmsg+i)->to;
      (tmpmsg+i)->dir=(maxmsg+i)->dir;
      (tmpmsg+i)->phi=malloc(sizeof(struct potential));
      (tmpmsg+i)->phi->Nvars=1; 
      (tmpmsg+i)->phi->varinds=malloc(sizeof(int));
      *((tmpmsg+i)->phi->varinds)=*((maxmsg+i)->phi->varinds);
      (tmpmsg+i)->phi->Nconfigs=(maxmsg+i)->phi->Nconfigs;
      (tmpmsg+i)->phi->Nstates=malloc(sizeof(int));
      *((tmpmsg+i)->phi->Nstates)=*((maxmsg+i)->phi->Nstates);
      (tmpmsg+i)->phi->probs=malloc(sizeof(double)*(tmpmsg+i)->phi->Nconfigs);
      for (j=0; j<=((tmpmsg+i)->phi->Nconfigs-1); j++)
	*((tmpmsg+i)->phi->probs+j)=*((maxmsg+i)->phi->probs+j);

      // Reset the messages going out of path sel vars to be even.
      if (((tmpmsg+i)->dir==0)&&((fgraph->varnodes+(tmpmsg+i)->from)->type==PATHSEL))
	for (j=0; j<=((tmpmsg+i)->phi->Nconfigs-1); j++)
	  *((tmpmsg+i)->phi->probs+j)=1.0;    
    }

  tmpbelief=malloc(sizeof(struct potential)*fgraph->Nvars);
  
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      (tmpbelief+i)->Nvars=1;
      (tmpbelief+i)->varinds=malloc(sizeof(int));
      *((tmpbelief+i)->varinds)=*((maxbelief+i)->varinds);
      (tmpbelief+i)->Nconfigs=(maxbelief+i)->Nconfigs;
      (tmpbelief+i)->Nstates=malloc(sizeof(int));
      *((tmpbelief+i)->Nstates)=*((maxbelief+i)->Nstates);
      (tmpbelief+i)->probs=malloc(sizeof(double)*(tmpbelief+i)->Nconfigs);
      for (j=0; j<=((tmpbelief+i)->Nconfigs-1); j++)
	*((tmpbelief+i)->probs+j)=*((maxbelief+i)->probs+j);
    }

  // Do not fix the path selection variables.
  fixedvals=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)  
    *(fixedvals+i)=-1;
  for (i=0; i<=(fgraph->Nvars-1); i++) 
    if ((fgraph->varnodes+i)->type!=PATHSEL)
      *(fixedvals+i)=*(configuration+i);

  maxsumproduct3(fgraph,tmpmsg,fixedvals,tmpbelief,0,0);

  // Determine the configuration value of previously undetermined variables or path sel variables.
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      if ((*(configuration+i)==-1)||((fgraph->varnodes+i)->type==PATHSEL))
	{
	  int Nv=0, v[(maxbelief+i)->Nconfigs];
	  double sumval, val, maxval=-1;
      
	  sumval=0;
	  for (j=0; j<=((tmpbelief+i)->Nconfigs-1); j++)
	    sumval+=*((tmpbelief+i)->probs+j);
	  for (j=0; j<=((tmpbelief+i)->Nconfigs-1); j++)
	    if ((val=*((tmpbelief+i)->probs+j)/sumval)>maxval)
	      maxval=val;
	  for (j=0; j<=((tmpbelief+i)->Nconfigs-1); j++)
	    {
	      //if ((maxval-(val=*((tmpbelief+i)->probs+j)/sumval))<1e-20)
	      if ((maxval-(val=*((tmpbelief+i)->probs+j)/sumval))<1e-6)
		{
		  Nv++; v[Nv-1]=j;
		}
	    }
	  if (Nv==1)
	    *(configuration+i)=v[0];
	}
    }

  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      free((tmpmsg+i)->phi->varinds);
      free((tmpmsg+i)->phi->Nstates);
      free((tmpmsg+i)->phi->probs);
      free((tmpmsg+i)->phi);
    }
  free(tmpmsg);
  
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      free((tmpbelief+i)->varinds);
      free((tmpbelief+i)->Nstates);
      free((tmpbelief+i)->probs);
    }
  
  free(tmpbelief); free(fixedvals); 


  // Again, if all non-isolated variables are fixed, then return.
  undetermined=0; i=0;
  while ((i<=(fgraph->Nvars-1))&&(undetermined==0))
    {
      if ((*(isolated+i)==0)&&(*(configuration+i)==-1))
	undetermined=1;
      i++;
    }
  if (undetermined==0)
    return;


  // Otherwise choose the (edge) variable which has the highest connection to other undetermined edge variables.
  influencecnt=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(influencecnt+i)=0; 

  // Find the number of one-neighbors of each variable nodes.
  // Do not count the neighbors of variable nodes if the path corresponding to the factor is not selected. 
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    {
      int pathselind=(fgraph->facnodes+i)->sigma;

      if ((pathselind<0)||((pathselind>=0)&&(*(configuration+pathselind)==1)))
	{
	  for (j=0; j<=((fgraph->facnodes+i)->phivi->Nvars-2); j++)
	    {
	      for (k=(j+1); k<=((fgraph->facnodes+i)->phivi->Nvars-1); k++)
		{
		  int ind1, ind2;
		  ind1=*((fgraph->facnodes+i)->phivi->varinds+j);
		  ind2=*((fgraph->facnodes+i)->phivi->varinds+k);
		  if ((*(configuration+ind1)==-1)&&(*(configuration+ind2)==-1)&&(*(isolated+ind1)==0)&&(*(isolated+ind2)==0))
		    {
		      *(influencecnt+ind1)+=1;
		      *(influencecnt+ind2)+=1;
		    }
		}
	    }
	}
    }
    
  // Fix the variable of the maximum influence and determine other configuration values.
  // Based on sumproduct but check if maxproduct is consistent.
  j=-1; k=0;
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      if (*(influencecnt+i)>k)
	{
	  k=*(influencecnt+i); j=i;
	}
    }
  maxconnnode=j; free(influencecnt);

  if (maxconnnode<0)
    return;

  
  // Debug
  //printf("maxconnnode=%d \n",maxconnnode);

  
  // Create two child nodes, each one corresponds to a particular value of the maxconnnode variable.
  configtree->Nchildren=2; configtree->children=malloc(sizeof(struct confignode)*2);
  child1=configtree->children; child2=configtree->children+1;
  child1->Nvars=configtree->Nvars; child1->level=configtree->level+1;
  child1->config=malloc(sizeof(int)*child1->Nvars);
  for (i=0; i<=(child1->Nvars-1); i++)
    *(child1->config+i)=*(configtree->config+i);
  child1->Nchildren=0;
  *(child1->config+maxconnnode)=0;
  
  // Debug
  child1->ind=nodeind; nodeind++;

  // Go to the children nodes.
  recurse_fixval9(child1,fgraph,maxmsg,maxbelief,isolated);

  child2->Nvars=configtree->Nvars; child2->level=configtree->level+1;
  child2->config=malloc(sizeof(int)*child2->Nvars);
  for (i=0; i<=(child2->Nvars-1); i++)
    *(child2->config+i)=*(configtree->config+i);
  child2->Nchildren=0;
  *(child2->config+maxconnnode)=1;

  // Debug
  child2->ind=nodeind; nodeind++;

  // Go to the children nodes.
  recurse_fixval9(child2,fgraph,maxmsg,maxbelief,isolated);
  
  return;

}

