// Functions of determining edge sign configurations and explanation.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "connect.h"
# include "integration.h"
# include "cluster.h"
# include "matrix.h"
# include "integration3.h"
# define MAXLINEWIDTH 1000
# define GENENAMELENGTH 9
# define PHASELENGTH 4
# define INFINITY 1e10
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)

extern char * getitemval(char *s, int ind, int *length);
extern char * getnextitemval(char *s, int *length, int curpos, int *nextpos);
extern char * getnextitemval2(char *s, int *length, int curpos, int *nextpos, int *end);
extern double atof2(char *s, int length);
extern long * loadmatrix(char *filename);
extern double * loadmatrix2(char *filename);
extern int inlist(int Nv, int *v, int q);
extern struct compresscpath * findvalidcpaths(struct csparsegraph *cgraph, int pathlength, int *Nvalidcpaths);
extern void number2config(int Nvars, int val, int *configuration);


void determinevalidpaths(int Ncpaths, struct compresscpath *compresscpaths, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, struct csparsegraph *model, int N, int *isknocked, int *isvalid, int *kopairs);


extern int maxpathlength, submaxpathlength;
int FUS3ind=121, KSS1ind=2192;
double epsilon=0.001;
//double ptrue_positive=0.005;
double ptrue_positive=0.02;


	      
// Determine valid paths from compressed paths.
// A path is valid if it connects knockout pairs, the last edge is type 0, and intermediate nodes either do not have ko exp or have significant ko effect, or intermediate nodes have double ko effects.
void determinevalidpaths(int Ncpaths, struct compresscpath *compresscpaths, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, struct csparsegraph *model, int N, int *isknocked, int *isvalid, int *kopairs)
{
  int i, j, k, flag;

  for (i=0; i<=(Ncpaths-1); i++)
    {
      *(isvalid+i)=0; *(kopairs+i)=-1;
    }

  for (i=0; i<=(Ncpaths-1); i++)
    {
      if ((compresscpaths+i)->lasttype==0)
	{
	  j=0; flag=1;
	  while ((j<=(Nkos-1))&&(flag==1))
	    {
	      if ((*(ref2sel+(ko+j)->i)==(compresscpaths+i)->start)&&(*(ref2sel+(ko+j)->j)==(compresscpaths+i)->end))
		{
		  *(isvalid+i)=1; flag=0; *(kopairs+i)=j;
		}
	      j++;
	    }
	}
    }


  for (i=0; i<=(Ncpaths-1); i++)
    {
      if (*(isvalid+i)==1)
	{
	  struct cpath *path;
	  int ind, Nnodes, *nodes;

	  path=decompresscpath(model,compresscpaths,i);

	  // An edge can have index -1 if the p-p edge is overwritten by an p-d edge.
	  j=0;
	  while ((j<=(path->Ne-1))&&(*(isvalid+i)==1))
	    {
	      if (*(path->e+j)<0)
		*(isvalid+i)=0;
	      j++;
	    }

	  if (*(isvalid+i)==1)
	    {  
	      /*
	      // This routine does not consider the case when the directions of two adjacent edges are reversed.
	      for (j=1; j<=(path->Ne-1); j++)
		{
		  ind=*(sel2ref+(model->e+*(path->e+j))->from);	      

		  if ((*(isknocked+ind)==1)&&(ind!=FUS3ind)&&(ind!=KSS1ind))
		    {		  
		      flag=1; k=0;
		      while ((k<=(Nkos-1))&&(flag==1))
			{
			  if (((ko+k)->i==ind)&&((ko+k)->j==*(sel2ref+(compresscpaths+i)->end)))
			    flag=0;
			  k++;
			}

		      if (flag==1)
			*(isvalid+i)=0;		      
		    }
		
		  if (*(isvalid+i)==0)
		    break;
		}
	      */

	      
	      // If intermediate gene is knocked out, then check whether its deletion affects the most downstream gene.  If not then this path is not valid.
	      // Exclude FUS3 and KSS1.
	      nodes=malloc(sizeof(int)); Nnodes=0;
	      for (j=0; j<=(path->Ne-1); j++)
		{
		  int ind;
		  if (((ind=(model->e+*(path->e+j))->from)!=(compresscpaths+i)->start)&&(ind!=(compresscpaths+i)->end)&&(*(sel2ref+ind)!=FUS3ind)&&(*(sel2ref+ind)!=KSS1ind)&&(*(isknocked+*(sel2ref+ind))==1)&&(inlist(Nnodes,nodes,ind)==0))
		    {
		      Nnodes++; nodes=realloc(nodes,sizeof(int)*Nnodes);
		      *(nodes+Nnodes-1)=ind;
		    }

		  if (((ind=(model->e+*(path->e+j))->to)!=(compresscpaths+i)->start)&&(ind!=(compresscpaths+i)->end)&&(*(sel2ref+ind)!=FUS3ind)&&(*(sel2ref+ind)!=KSS1ind)&&(*(isknocked+*(sel2ref+ind))==1)&&(inlist(Nnodes,nodes,ind)==0))
		    {
		      Nnodes++; nodes=realloc(nodes,sizeof(int)*Nnodes);
		      *(nodes+Nnodes-1)=ind;
		    }
		}
	      
	      j=0;
	      while ((*(isvalid+i)==1)&&(j<=(Nnodes-1)))
		{
		  flag=1; k=0;
		  while ((k<=(Nkos-1))&&(flag==1))
		    {
		      if (((ko+k)->i==*(sel2ref+*(nodes+j)))&&((ko+k)->j==*(sel2ref+(compresscpaths+i)->end)))
			flag=0;
		      k++;
		    }
		  if (flag==1)
		    *(isvalid+i)=0;
		  j++;
		}
	      if (Nnodes>0)
		free(nodes);
	      

	    }
	  
	  free(path->e); free(path);
	}  
    }
 
}


