// Utility functions regarding the operations of potential functions defined by both functional forms and lookup tables.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "integration.h"
# include "integration3.h"
# include "connect.h"
# include "cluster.h"
# include "integration4.h"
# include "integration5.h"
# define MAXLINEWIDTH 1000
# define INFINITY 1e38
# define POSINF 100000
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)
# define GENENAMELENGTH 9


extern void maxsumproduct3(struct mix_facgraph *fgraph, struct message *msg, int *fixedvals, struct potential *belief, int mode, int updatemode);
extern void clear_mix_facgraph(struct mix_facgraph *fgraph);;
extern struct cpath * decompresscpath(struct csparsegraph *cgraph, struct compresscpath *path, int ind);
extern int inlist(int Nv, int *v, int q);
extern int inlistind(int Nv, int *v, int q);

extern int maxpathlength, submaxpathlength;
extern double epsilon, ptrue_positive;
extern double v1, v2, v3, koinv;


// Construct the factor graph of model inference.
// Input: model, valid paths, knockout data.
// Basically identical to the major part of computeconfigs2.  Except adding the info about variable types to the data structure.
// Difference from construct_fac_graph1: does not create lookup tables for SINGLEPATHEXPLAIN and PATHSELOR type functions.
// Difference from construct_fac_graph2: use likelihood ratios instead of p-values in the potential terms of edge presences, discard potential terms generated by repeated paths.
struct mix_facgraph * construct_fac_graph4(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko)
{
  int i, j, k, l, Nvalidpaths, **kopaths;
  int Nvars, Ninteractions;
  int Nexplainedkos, *explainedkos, *explainedko2var;
  int Nsigedges, *sigdirvar2edge, *edge2sigdirvar;
  struct mix_facgraph *fgraph;


  Nvalidpaths=0;
  for (i=0; i<=(Ncpaths-1); i++)
    if (*(isvalid+i)==1)
      Nvalidpaths++;
  
  Nsigedges=0;
  
  for (i=0; i<=(model->Ne-1); i++)
    if ((model->e+i)->type==1)
      Nsigedges++;
  
  Nvars=model->Ne*2+Nsigedges;
  Ninteractions=Nvalidpaths;
  fgraph=malloc(sizeof(struct mix_facgraph));
  fgraph->Nvars=Nvars;
  fgraph->vars=malloc(sizeof(int)*Nvars);
  fgraph->Ninteractions=Ninteractions;
  fgraph->interactions=malloc(sizeof(int)*Ninteractions);
  for (i=0; i<=(Ninteractions-1); i++)
    *(fgraph->interactions+i)=i;
  for (i=0; i<=(Nvars-1); i++)
    *(fgraph->vars+i)=i;
  fgraph->Ne=0; fgraph->e=malloc(sizeof(struct edge));  
  fgraph->facnodes=malloc(sizeof(struct mix_path_fac_node)*fgraph->Ninteractions); 
  Ninteractions=0;

  // Debug
  //printf("sub breakpnt1.1\n");
 
  
  // Build the lookup table from variables to signal edge dirs.
  sigdirvar2edge=malloc(sizeof(int)*Nsigedges);
  edge2sigdirvar=malloc(sizeof(int)*model->Ne);
  for (i=0; i<=(model->Ne-1); i++)
    *(edge2sigdirvar+i)=-1;
  j=0;
  for (i=0; i<=(model->Ne-1); i++)
    {
      if ((model->e+i)->type==1)
	{
	  *(sigdirvar2edge+j)=i;
	  *(edge2sigdirvar+i)=j;
	  j++;
	}
    }

  // Debug
  //printf("sub breakpnt1.2\n");
 

  // Add information about edge presence, edge sign and edge direction vars.
  fgraph->varnodes=malloc(sizeof(struct mix_path_var_node)*Nvars);
  for (i=0; i<=(model->Ne-1); i++)
    {
      (fgraph->varnodes+i)->type=EDGEPRESENCE;
      (fgraph->varnodes+i)->edgeind=i;
      (fgraph->varnodes+i)->geneind=-1;
      (fgraph->varnodes+i)->koind=-1;
    }
  for (i=0; i<=(model->Ne-1); i++)
    {
      (fgraph->varnodes+i+model->Ne)->type=EDGESIGN;
      (fgraph->varnodes+i+model->Ne)->edgeind=i;
      (fgraph->varnodes+i+model->Ne)->geneind=-1;
      (fgraph->varnodes+i+model->Ne)->koind=-1;
    }
  for (i=0; i<=(Nsigedges-1); i++)
    {
      //(fgraph->varnodes+i+model->Ne*2)->type=2;
      (fgraph->varnodes+i+model->Ne*2)->type=EDGEDIR;
      (fgraph->varnodes+i+model->Ne*2)->edgeind=*(sigdirvar2edge+i);
      (fgraph->varnodes+i+model->Ne*2)->geneind=-1;
      (fgraph->varnodes+i+model->Ne*2)->koind=-1;
    }

  // Debug
  //printf("sub breakpnt1.3\n");

  Nexplainedkos=0; explainedkos=malloc(sizeof(int));
  explainedko2var=malloc(sizeof(int));
  
  kopaths=malloc(sizeof(int)*Nkos);
  for (i=0; i<=(Nkos-1); i++)
    {
      *(kopaths+i)=malloc(sizeof(int));
      **(kopaths+i)=0;
    }  

  // Debug
  //printf("sub breakpnt1.4\n");


  // Create potential function terms of single path explanations.
  for (i=0; i<=(Nkos-1); i++)
    {
      int Npairpaths=0, *pairpaths;
      int Nkept, *kept;

      /*
      pairpaths=malloc(sizeof(int));
      for (j=0; j<=(Ncpaths-1); j++)
	{
	  //if ((*(isvalid+j)==1)&&((compresscpaths+j)->start==*(ref2sel+(ko+i)->i))&&((compresscpaths+j)->end==*(ref2sel+(ko+i)->j)))

	  if ((*(isvalid+j)==1)&&((compresscpaths+j)->start==*(ref2sel+(ko+i)->i))&&((compresscpaths+j)->end==*(ref2sel+(ko+i)->j))&&((compresscpaths+j)->length<=maxpathlength))

	    {
	      Npairpaths++;
	      pairpaths=realloc(pairpaths,sizeof(int)*Npairpaths);
	      *(pairpaths+Npairpaths-1)=j;
	    }
	}
      */


      // Find the paths which connect ko pairs and are not repeated.
      pairpaths=malloc(sizeof(int));
      for (j=0; j<=(Ncpaths-1); j++)
	{
	  // Exclude the paths whose last edges are pp.  This step should be done in the augment_validcpaths function but the procedures are commented.
	  if ((*(isvalid+j)==1)&&((compresscpaths+j)->start==*(ref2sel+(ko+i)->i))&&((compresscpaths+j)->end==*(ref2sel+(ko+i)->j))&&((compresscpaths+j)->length<=maxpathlength))
	    //if ((*(isvalid+j)==1)&&((compresscpaths+j)->start==*(ref2sel+(ko+i)->i))&&((compresscpaths+j)->end==*(ref2sel+(ko+i)->j))&&((compresscpaths+j)->length<=maxpathlength)&&((compresscpaths+j)->lasttype==0))
	    {
	      struct cpath *path1, *path2;
	      int flag2=0;

	      path1=decompresscpath(model,compresscpaths,j); k=0;

	      while ((k<=(Npairpaths-1))&&(flag2==0))
		{
		  if ((compresscpaths+*(pairpaths+k))->length==path1->Ne)
		    {
		      path2=decompresscpath(model,compresscpaths,*(pairpaths+k));
		      flag2=1; l=0;
		      while ((l<=(path1->Ne-1))&&(flag2==1))
			{
			  if (*(path1->e+l)!=*(path2->e+l))
			    flag2=0;
			  l++;
			}
		      free(path2->e); free(path2);
		    }
		  k++;		  
		}
	      free(path1->e); free(path1);

	      if (flag2==0)
		{
		  Npairpaths++;
		  pairpaths=realloc(pairpaths,sizeof(int)*Npairpaths);
		  *(pairpaths+Npairpaths-1)=j;
		}
	    }
	}



      // If a knockout pair can only be explained by paths of length > submaxlength, then keep them.  Otherwise remove the paths of length > submaxpathlength.
      Nkept=0; kept=malloc(sizeof(int)*Npairpaths);
      for (j=0; j<=(Npairpaths-1); j++)
	{
	  struct cpath *path;
	  path=decompresscpath(model,compresscpaths,*(pairpaths+j));
	  if (path->Ne<=submaxpathlength)
	    {
	      Nkept++; *(kept+Nkept-1)=*(pairpaths+j);
	    }
	  free(path->e); free(path);
	}
      if (Nkept>0)
	{
	  Npairpaths=Nkept;
	  for (j=0; j<=(Nkept-1); j++)
	    *(pairpaths+j)=*(kept+j);
	  pairpaths=realloc(pairpaths,sizeof(int)*Npairpaths);
	}
      free(kept);

      
      // If Npairpaths>0, then the knockout effect is possible to be explained.
      if (Npairpaths>0)
	{
	  Nexplainedkos+=1; 
	  explainedkos=realloc(explainedkos,sizeof(int)*Nexplainedkos);
	  *(explainedkos+Nexplainedkos-1)=i;
	  
	  // Create a variable for knockout effect.
	  Nvars++;
	  fgraph->Nvars=Nvars; 
	  fgraph->vars=realloc(fgraph->vars,sizeof(int)*Nvars);
	  *(fgraph->vars+Nvars-1)=Nvars-1;

	  explainedko2var=realloc(explainedko2var,sizeof(int)*Nexplainedkos);
	  *(explainedko2var+Nexplainedkos-1)=Nvars-1;
	  
	  **(kopaths+Nexplainedkos-1)=Npairpaths;
	  *(kopaths+Nexplainedkos-1)=realloc(*(kopaths+Nexplainedkos-1),sizeof(int)*(Npairpaths+1));
	  for (j=0; j<=(Npairpaths-1); j++)
	    *(*(kopaths+Nexplainedkos-1)+j+1)=Nvars+j;

	  // Add the variable type info.
	  fgraph->varnodes=realloc(fgraph->varnodes,sizeof(struct mix_path_var_node)*Nvars);
	  (fgraph->varnodes+Nvars-1)->type=KNOCKOUT;
	  (fgraph->varnodes+Nvars-1)->koind=i;
	  (fgraph->varnodes+Nvars-1)->geneind=-1;
	  (fgraph->varnodes+Nvars-1)->edgeind=-1;

	  /*
	  // Debug
	  if (i==(Nkos-1))
	    {
	      struct cpath *path;

	      printf("Npairpaths=%d\n",Npairpaths);
	      for (j=0; j<=(Npairpaths-1); j++)
		{
		  path=decompresscpath(model,compresscpaths,*(pairpaths+j));
		  printf("path %d: ",*(pairpaths+j));
		  for (k=0; k<=(path->Ne-1); k++)
		    printf("(%d,%d)",(model->e+*(path->e+k))->from,(model->e+*(path->e+k))->to);
		  printf("\n");
		}
	      free(path->e); free(path);
	    }
	  */

	  for (j=0; j<=(Npairpaths-1); j++)
	    {
	      struct cpath *path;
	      int Nv, Nsignals;

	      // For each candidate path on each ko effect, create a path selection var.
	      Nvars++;
	      fgraph->Nvars=Nvars;
	      fgraph->vars=realloc(fgraph->vars,sizeof(int)*Nvars);
	      *(fgraph->vars+Nvars-1)=Nvars-1;

	      // Add variable type info.
	      fgraph->varnodes=realloc(fgraph->varnodes,sizeof(struct mix_path_var_node)*Nvars);
	      (fgraph->varnodes+Nvars-1)->type=PATHSEL;
	      (fgraph->varnodes+Nvars-1)->pathinfo=malloc(sizeof(struct pair));
	      (fgraph->varnodes+Nvars-1)->pathinfo->i=i;
	      (fgraph->varnodes+Nvars-1)->pathinfo->j=Nvars-1;
	      (fgraph->varnodes+Nvars-1)->pathinfo->k=*(pairpaths+j);
	      (fgraph->varnodes+Nvars-1)->geneind=-1;
	      (fgraph->varnodes+Nvars-1)->edgeind=-1;
	      (fgraph->varnodes+Nvars-1)->koind=-1;	      

	      path=decompresscpath(model,compresscpaths,*(pairpaths+j));  


	      Ninteractions++;
	      fgraph->Ninteractions=Ninteractions;
	      *(fgraph->interactions+Ninteractions-1)=Ninteractions-1;
	      
	      // The factor node connects to presence and sign variables of the edges on path,
	      // the signal edge dir variables, and the path sel and ko vars.
	      for (k=0; k<=(path->Ne-1); k++)
		{
		  fgraph->Ne+=2;
		  fgraph->e=realloc(fgraph->e,sizeof(struct edge)*fgraph->Ne);
		  (fgraph->e+fgraph->Ne-2)->from=*(path->e+k);
		  (fgraph->e+fgraph->Ne-2)->to=Ninteractions-1;
		  (fgraph->e+fgraph->Ne-1)->from=*(path->e+k)+model->Ne;
		  (fgraph->e+fgraph->Ne-1)->to=Ninteractions-1;

		  if (*(edge2sigdirvar+*(path->e+k))>=0)
		    {
		      fgraph->Ne+=1;
		      fgraph->e=realloc(fgraph->e,sizeof(struct edge)*fgraph->Ne);
		      (fgraph->e+fgraph->Ne-1)->from=*(edge2sigdirvar+*(path->e+k))+model->Ne*2;
		      (fgraph->e+fgraph->Ne-1)->to=Ninteractions-1;
		    }
		  		  
		}
	      
	      fgraph->Ne+=1;
	      fgraph->e=realloc(fgraph->e,sizeof(struct edge)*fgraph->Ne);
	      (fgraph->e+fgraph->Ne-1)->from=Nvars-1;
	      (fgraph->e+fgraph->Ne-1)->to=Ninteractions-1;
	      fgraph->Ne+=1;
	      fgraph->e=realloc(fgraph->e,sizeof(struct edge)*fgraph->Ne);
	      (fgraph->e+fgraph->Ne-1)->from=*(explainedko2var+Nexplainedkos-1);
	      (fgraph->e+fgraph->Ne-1)->to=Ninteractions-1;

	      // Establish mix_path_fac_node structure on each interaction node.
	      fgraph->facnodes=realloc(fgraph->facnodes,sizeof(struct mix_path_fac_node)*Ninteractions);
	      //(fgraph->facnodes+Ninteractions-1)->type=LOGICAL;
	      
	      (fgraph->facnodes+Ninteractions-1)->type=SINGLEPATHEXPLAIN;

	      (fgraph->facnodes+Ninteractions-1)->Nedges=path->Ne;
	      (fgraph->facnodes+Ninteractions-1)->presences=malloc(sizeof(int)*path->Ne);
	      (fgraph->facnodes+Ninteractions-1)->signs=malloc(sizeof(int)*path->Ne);
	      (fgraph->facnodes+Ninteractions-1)->Nsigedges=0;
	      
	      for (k=0; k<=(path->Ne-1); k++)
		{
		  *((fgraph->facnodes+Ninteractions-1)->presences+k)=*(path->e+k);
		  *((fgraph->facnodes+Ninteractions-1)->signs+k)=*(path->e+k)+model->Ne;

		  if (*(edge2sigdirvar+*(path->e+k))>=0)
		    (fgraph->facnodes+Ninteractions-1)->Nsigedges+=1;
		}	      

	      
	      // Start from the last edge on the path, which is always non-signal.  The preferred direction of a signal edge is to the node which is the source node of the next edge.
	      Nsignals=(fgraph->facnodes+Ninteractions-1)->Nsigedges;
	      (fgraph->facnodes+Ninteractions-1)->sigdirvars=malloc(sizeof(int)*Nsignals);
	      (fgraph->facnodes+Ninteractions-1)->sigdirvarvals=malloc(sizeof(int)*Nsignals);
	      l=Nsignals-1;

	      for (k=(path->Ne-2); k>=0; k--)
		{
		  if (*(edge2sigdirvar+*(path->e+k))>=0)
		    {
		      int thisedgeind, nextedgeind, tmp;
		      int thisfrom, thisto, nextfrom, nextto;
		      
		      thisedgeind=*(path->e+k); nextedgeind=*(path->e+k+1);
		      nextfrom=(model->e+nextedgeind)->from;
		      nextto=(model->e+nextedgeind)->to;

		      // If the next edge is a signal edge and the preferred dir is -1, then swap nextfrom and nextto.
		      if ((*(edge2sigdirvar+nextedgeind)>=0)&&(*((fgraph->facnodes+Ninteractions-1)->sigdirvarvals+l+1)==0))
			{
			  tmp=nextfrom; nextfrom=nextto; nextto=tmp;
			}
		      *((fgraph->facnodes+Ninteractions-1)->sigdirvars+l)=*(edge2sigdirvar+thisedgeind)+model->Ne*2;
		      thisfrom=(model->e+thisedgeind)->from;
		      thisto=(model->e+thisedgeind)->to;

		      // If thisto node is nextfrom node, then the preferred dir value is 1, otherwise -1 (0).
		      if (thisto==nextfrom)
			*((fgraph->facnodes+Ninteractions-1)->sigdirvarvals+l)=1;
		      else
			*((fgraph->facnodes+Ninteractions-1)->sigdirvarvals+l)=0;
		      l--;
		    }
		}
	      	      
	      (fgraph->facnodes+Ninteractions-1)->sigma=Nvars-1;
	      (fgraph->facnodes+Ninteractions-1)->ko=*(explainedko2var+Nexplainedkos-1);
	      
	      // Establish the lookup table of the potential function.
	      (fgraph->facnodes+Ninteractions-1)->phivi=malloc(sizeof(struct potential));
	      Nv=path->Ne*2+Nsignals+2;
	      (fgraph->facnodes+Ninteractions-1)->phivi->Nvars=Nv;
	      (fgraph->facnodes+Ninteractions-1)->phivi->varinds=malloc(sizeof(int)*Nv);
	      for (k=0; k<=(path->Ne-1); k++)
		{
		  *((fgraph->facnodes+Ninteractions-1)->phivi->varinds+k)=*(path->e+k);
		  *((fgraph->facnodes+Ninteractions-1)->phivi->varinds+k+path->Ne)=*(path->e+k)+model->Ne;
		}
	      for (k=0; k<=(Nsignals-1); k++)
		*((fgraph->facnodes+Ninteractions-1)->phivi->varinds+k+path->Ne*2)=*((fgraph->facnodes+Ninteractions-1)->sigdirvars+k);
	      *((fgraph->facnodes+Ninteractions-1)->phivi->varinds+path->Ne*2+Nsignals)=Nvars-1;
	      *((fgraph->facnodes+Ninteractions-1)->phivi->varinds+path->Ne*2+Nsignals+1)=*(explainedko2var+Nexplainedkos-1);
	      
	      // Does not allocate memory for Nstates and probability table.
	      (fgraph->facnodes+Ninteractions-1)->phivi->Nconfigs=0;

	      free(path->e); free(path);
	    }
	}
      free(pairpaths);

    }

  // Debug
  //printf("sub breakpnt2\n");
  
  // Potential function terms of edge presence from physical data.
  // phi(x=0)=1/(1+likelihood ratio), phi(x=1)=likelihood ratio/(1+likelihoodratio).
  for (i=0; i<=(model->Ne-1); i++)
    {
      Ninteractions++;
      fgraph->Ninteractions=Ninteractions;
      fgraph->interactions=realloc(fgraph->interactions,sizeof(int)*Ninteractions);
      *(fgraph->interactions+Ninteractions-1)=Ninteractions-1;
      fgraph->facnodes=realloc(fgraph->facnodes,sizeof(struct mix_path_fac_node)*Ninteractions);      
      (fgraph->facnodes+Ninteractions-1)->type=SINGLEVAR;
      (fgraph->facnodes+Ninteractions-1)->Nedges=1;
      (fgraph->facnodes+Ninteractions-1)->Nsigedges=0;
      (fgraph->facnodes+Ninteractions-1)->presences=malloc(sizeof(int));
      (fgraph->facnodes+Ninteractions-1)->signs=malloc(sizeof(int));
      *((fgraph->facnodes+Ninteractions-1)->presences)=i;
      (fgraph->facnodes+Ninteractions-1)->sigma=-1;
      (fgraph->facnodes+Ninteractions-1)->ko=-1;
      (fgraph->facnodes+Ninteractions-1)->phivi=malloc(sizeof(struct potential));
      (fgraph->facnodes+Ninteractions-1)->phivi->Nvars=1;
      (fgraph->facnodes+Ninteractions-1)->phivi->varinds=malloc(sizeof(int));
      *((fgraph->facnodes+Ninteractions-1)->phivi->varinds)=i;
      (fgraph->facnodes+Ninteractions-1)->phivi->Nstates=malloc(sizeof(int));
      *((fgraph->facnodes+Ninteractions-1)->phivi->Nstates)=2;
      (fgraph->facnodes+Ninteractions-1)->phivi->Nconfigs=2;
      (fgraph->facnodes+Ninteractions-1)->phivi->probs=malloc(sizeof(double)*2);
      *((fgraph->facnodes+Ninteractions-1)->phivi->probs+0)=1/(1+(model->e+i)->priorp);
      *((fgraph->facnodes+Ninteractions-1)->phivi->probs+1)=(model->e+i)->priorp/(1+(model->e+i)->priorp);
      fgraph->Ne+=1;
      fgraph->e=realloc(fgraph->e,sizeof(struct edge)*fgraph->Ne);
      (fgraph->e+fgraph->Ne-1)->from=i;
      (fgraph->e+fgraph->Ne-1)->to=Ninteractions-1;
    }

  // Debug
  //printf("sub breakpnt3\n");

  // Potential function terms of ko effects.
  // phi(ko=0)=1/(1+likelihood ratio+koinv), phi(ko=observed dir)=likelihood ratio/(1+likelihood ratio+koinv), phi(ko=inverse observed dir)=koinv/(1+likelihood ratio+koinv).
  for (i=0; i<=(Nexplainedkos-1); i++)
    {
      Ninteractions++;
      fgraph->Ninteractions=Ninteractions;
      fgraph->interactions=realloc(fgraph->interactions,sizeof(int)*Ninteractions);
      *(fgraph->interactions+Ninteractions-1)=Ninteractions-1;
      fgraph->facnodes=realloc(fgraph->facnodes,sizeof(struct mix_path_fac_node)*Ninteractions);
      (fgraph->facnodes+Ninteractions-1)->type=SINGLEVAR;
      (fgraph->facnodes+Ninteractions-1)->Nedges=0;
      (fgraph->facnodes+Ninteractions-1)->Nsigedges=0;
      (fgraph->facnodes+Ninteractions-1)->ko=*(explainedko2var+i);
      (fgraph->facnodes+Ninteractions-1)->sigma=-1;
      (fgraph->facnodes+Ninteractions-1)->phivi=malloc(sizeof(struct potential));
      (fgraph->facnodes+Ninteractions-1)->phivi->Nvars=1;
      (fgraph->facnodes+Ninteractions-1)->phivi->varinds=malloc(sizeof(int));
      *((fgraph->facnodes+Ninteractions-1)->phivi->varinds)=*(explainedko2var+i);
      (fgraph->facnodes+Ninteractions-1)->phivi->Nstates=malloc(sizeof(int));
      *((fgraph->facnodes+Ninteractions-1)->phivi->Nstates)=3;
      (fgraph->facnodes+Ninteractions-1)->phivi->Nconfigs=3;
      (fgraph->facnodes+Ninteractions-1)->phivi->probs=malloc(sizeof(double)*3);
      
      *((fgraph->facnodes+Ninteractions-1)->phivi->probs+0)=1/(1+(ko+*(explainedkos+i))->s+koinv);
      if ((ko+*(explainedkos+i))->k>0)
	{
	  *((fgraph->facnodes+Ninteractions-1)->phivi->probs+1)=(ko+*(explainedkos+i))->s/(1+(ko+*(explainedkos+i))->s+koinv);
	  *((fgraph->facnodes+Ninteractions-1)->phivi->probs+2)=koinv/(1+(ko+*(explainedkos+i))->s+koinv);
	}
      else
	{
	  *((fgraph->facnodes+Ninteractions-1)->phivi->probs+2)=(ko+*(explainedkos+i))->s/(1+(ko+*(explainedkos+i))->s+koinv);
	  *((fgraph->facnodes+Ninteractions-1)->phivi->probs+1)=koinv/(1+(ko+*(explainedkos+i))->s+koinv);
	}
      fgraph->Ne+=1;
      fgraph->e=realloc(fgraph->e,sizeof(struct edge)*fgraph->Ne);
      (fgraph->e+fgraph->Ne-1)->from=*(explainedko2var+i);
      (fgraph->e+fgraph->Ne-1)->to=Ninteractions-1;
    }

  // Debug
  //printf("sub breakpnt4\n");
 
  // Potential function terms linking path selection vars of the same ko var.
  // The potential function table of path sel vars: 1 if any sigma=1, 0 if all sigmas are 0.
  for (i=0; i<=(Nexplainedkos-1); i++)
    {
      Ninteractions++;
      fgraph->Ninteractions=Ninteractions; 
      fgraph->interactions=realloc(fgraph->interactions,sizeof(int)*Ninteractions);
      *(fgraph->interactions+Ninteractions-1)=Ninteractions-1;	  
      fgraph->facnodes=realloc(fgraph->facnodes,sizeof(struct mix_path_fac_node)*Ninteractions);
      (fgraph->facnodes+Ninteractions-1)->type=PATHSELOR;
      (fgraph->facnodes+Ninteractions-1)->Nedges=0;
      (fgraph->facnodes+Ninteractions-1)->Nsigedges=0;
      (fgraph->facnodes+Ninteractions-1)->sigma=-1;
      (fgraph->facnodes+Ninteractions-1)->ko=-1;
      (fgraph->facnodes+Ninteractions-1)->phivi=malloc(sizeof(struct potential));
      (fgraph->facnodes+Ninteractions-1)->phivi->Nvars=**(kopaths+i);	  
      (fgraph->facnodes+Ninteractions-1)->phivi->varinds=malloc(sizeof(int)*(**(kopaths+i)));
      for (k=0; k<=(**(kopaths+i)-1); k++)
	*((fgraph->facnodes+Ninteractions-1)->phivi->varinds+k)=*(*(kopaths+i)+k+1);
      (fgraph->facnodes+Ninteractions-1)->phivi->Nconfigs=0;

      for (k=0; k<=(**(kopaths+i)-1); k++)
	{
	  fgraph->Ne+=1;
	  fgraph->e=realloc(fgraph->e,sizeof(struct edge)*fgraph->Ne);
	  (fgraph->e+fgraph->Ne-1)->from=*(*(kopaths+i)+k+1);
	  (fgraph->e+fgraph->Ne-1)->to=Ninteractions-1;
	}
    }
  
  // Debug
  //printf("sub breakpnt5, Ninteractions=%d\n",Ninteractions);
  
  for (i=0; i<(Nkos-1); i++)
    free(*(kopaths+i));
  free(kopaths); free(explainedkos); free(explainedko2var);
  free(sigdirvar2edge); free(edge2sigdirvar);

  return fgraph;

}
