// Utility functions regarding the operations of potential functions defined by both functional forms and lookup tables.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include "integration.h"
# include "integration3.h"
# include "connect.h"
# include "cluster.h"
# include "integration4.h"
# include "integration5.h"
# define MAXLINEWIDTH 1000
# define INFINITY 1e38
# define POSINF 100000
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)


extern int inlistind(int Nv, int *v, int q);
extern struct cpath * decompresscpath(struct csparsegraph *cgraph, struct compresscpath *path, int ind);
extern long config2number2(int Nvars, int *Nstates, int *configuration);
extern void number2config2(int Nvars, int *Nstates, long val, int *configuration);
extern void number2config(int Nvars, int val, int *configuration);


void clear_mix_facgraph(struct mix_facgraph *fgraph);
struct mix_facgraph * retract_mix_facgraph(struct mix_facgraph *fgraph, int oldNvars, int oldNinteractions, int oldNe);


extern int maxpathlength, submaxpathlength;
extern double epsilon, ptrue_positive;
extern double v1, v2, v3, koinv;


// Clear a mix_facgraph variable.
void clear_mix_facgraph(struct mix_facgraph *fgraph)
{
  int i, j, k;

  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      //if ((fgraph->varnodes+i)->type==4)
      if ((fgraph->varnodes+i)->type==PATHSEL)
	free((fgraph->varnodes+i)->pathinfo);
    }
  free(fgraph->varnodes);

  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    {
      if ((fgraph->facnodes+i)->Nedges>0)
	{
	  free((fgraph->facnodes+i)->presences);
	  free((fgraph->facnodes+i)->signs);
	}
     if ((fgraph->facnodes+i)->Nsigedges>0) 
       {
	 free((fgraph->facnodes+i)->sigdirvars);
	 free((fgraph->facnodes+i)->sigdirvarvals);
       }
     if ((fgraph->facnodes+i)->phivi->Nvars>0)
       {
	 free((fgraph->facnodes+i)->phivi->varinds);

	 if ((fgraph->facnodes+i)->phivi->Nconfigs>0)
	   {
	     free((fgraph->facnodes+i)->phivi->Nstates);
	     free((fgraph->facnodes+i)->phivi->probs);
	   }
       }
     free((fgraph->facnodes+i)->phivi);
    }
  free(fgraph->facnodes);

  free(fgraph->e); free(fgraph->interactions); free(fgraph->vars);

  free(fgraph);

}


// Retract a factor graph to the original state.
struct mix_facgraph * retract_mix_facgraph(struct mix_facgraph *fgraph, int oldNvars, int oldNinteractions, int oldNe)
{
  int i, j, k, l;
  
  for (i=oldNvars; i<=(fgraph->Nvars-1); i++)
    {
      //if ((fgraph->varnodes+i)->type==4)
      if ((fgraph->varnodes+i)->type==PATHSEL)
	free((fgraph->varnodes+i)->pathinfo);
    }
  fgraph->varnodes=realloc(fgraph->varnodes,sizeof(struct mix_path_var_node)*oldNvars);


  for (i=oldNinteractions; i<=(fgraph->Ninteractions-1); i++)
    {
      if ((fgraph->facnodes+i)->Nedges>0)
	{
	  free((fgraph->facnodes+i)->presences);
	  free((fgraph->facnodes+i)->signs);
	}
     if ((fgraph->facnodes+i)->Nsigedges>0) 
       {
	 free((fgraph->facnodes+i)->sigdirvars);
	 free((fgraph->facnodes+i)->sigdirvarvals);
       }
    
     if ((fgraph->facnodes+i)->phivi->Nvars>0)
       {
	 free((fgraph->facnodes+i)->phivi->varinds);
	 if ((fgraph->facnodes+i)->phivi->Nconfigs>0)
	   {
	     free((fgraph->facnodes+i)->phivi->Nstates);
	     free((fgraph->facnodes+i)->phivi->probs);
	   }
       }
     free((fgraph->facnodes+i)->phivi);
     
    }
  fgraph->facnodes=realloc(fgraph->facnodes,sizeof(struct mix_path_fac_node)*oldNinteractions);

  fgraph->vars=realloc(fgraph->vars,sizeof(int)*oldNvars);
  fgraph->interactions=realloc(fgraph->interactions,sizeof(int)*oldNinteractions);
  fgraph->e=realloc(fgraph->e,sizeof(struct edge)*oldNe);
  fgraph->Nvars=oldNvars; fgraph->Ninteractions=oldNinteractions;
  fgraph->Ne=oldNe;

  return fgraph;
}


// Clear degenerate models.
void cleardegmodels(struct model_configuration *degmodels)
{
  int i, j, k;
  
  free(degmodels->vartypes);
  free(degmodels->presence2var);
  free(degmodels->sign2var);
  free(degmodels->sigedge2edge);
  free(degmodels->sigdir2var);
  free(degmodels->ko2var);
  free(degmodels->var2pathind);
  free(degmodels->usedpathinfo);
  free(degmodels->mapconfigs);
  free(degmodels);
}
