// Utility functions regarding the operations of potential functions defined by both functional forms and lookup tables.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include "integration.h"
# include "integration3.h"
# include "connect.h"
# include "cluster.h"
# include "integration4.h"
# include "integration5.h"
# define MAXLINEWIDTH 1000
# define INFINITY 1e38
# define POSINF 100000
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)

extern void maxsumproduct1(struct mix_facgraph *fgraph, struct message *msg, int *fixedvals, struct potential *belief, int mode);
extern int * recurse_getmapconfigs(struct confignode *configtree, int *Nmapconfigs, int *mapconfigs);
extern void recurse_freeconfignode(struct confignode *configtree);
void recurse_fixval2(struct confignode *configtree, struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int *vartypes);
extern int inlistind(int Nv, int *v, int q);
extern void copypotential(struct potential *phi2, struct potential *phi1);
extern struct cpath * decompresscpath(struct csparsegraph *cgraph, struct compresscpath *path, int ind);
extern long config2number2(int Nvars, int *Nstates, int *configuration);
extern void number2config2(int Nvars, int *Nstates, long val, int *configuration);
extern void number2config(int Nvars, int val, int *configuration);
extern void potentialproduct2(struct potential *phi1, struct potential *phi2);
extern void potentialproduct3(struct potential *phi1, struct potential *phi2);
extern struct potential * createpotential(struct potential *phi1);
extern void maximizepotential2(struct potential *phi, int ind);
extern void maxmarginal3(struct mix_path_fac_node *facnode, int Nmsgs, int *msginds, struct message *msg, struct potential *phi, int deb);
extern void marginalizepotential4(struct potential *phi, int varind, int Nmsgs, int *msginds, struct message *msg, struct potential *newphi);
extern int pair2msg(int Nmsgs, struct message *msg, struct pair *node1, struct pair *node2);
extern void maxmarginal4(struct mix_facgraph *fgraph, int factorind, int varind, int Nmsgs, int *msginds, struct message *msg, struct potential *phi, int deb);


void marginalizepotential5(struct mix_facgraph *fgraph, int factorind, int varind, int Nmsgs, int *msginds, struct message *msg, struct potential *newphi, int deb);
void maxsumproduct3(struct mix_facgraph *fgraph, struct message *msg, int *fixedvals, struct potential *belief, int mode, int updatemode);
struct mix_facgraph * construct_fac_graph2(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko);


extern int maxpathlength, submaxpathlength;
extern double epsilon, ptrue_positive;
extern double v1, v2, v3, koinv;



// Marginalize the product of a potential function and multiple messages.  The output is a message of one variable.
// Difference from marginalizepotential3: the output potential function has one variable, and the input messages are also marginalized.
// Difference from marginalizepotential4: apply to logic potential functions, runs much faster and does not require to store the lookup table.
// Notice the input messages must be normalized.
void marginalizepotential5(struct mix_facgraph *fgraph, int factorind, int varind, int Nmsgs, int *msginds, struct message *msg, struct potential *newphi, int deb)
{
  int i, j, k, *phi2msg, *msg2phi, flag, ind;
  double val;
  struct potential *phi;

  
  // If Nmsgs=0, the newphi=(fgraph->facnodes+factorind)->phivi.
  if (Nmsgs<=0)
    {
      // If factor type is PATHSELOR, there is a single explanatory path.  Assign m(sigma=0)=v3 and m(sigma=1)=1.
      if ((fgraph->facnodes+factorind)->type==PATHSELOR)
	{
	  *(newphi->probs+0)=v3; *(newphi->probs+1)=1;
	}
      else
	copypotential(newphi,(fgraph->facnodes+factorind)->phivi);
      return;
    }

  
  // Find the index of the potential function variable.
  ind=-1; i=0; phi=(fgraph->facnodes+factorind)->phivi;
  while ((i<=(phi->Nvars-1))&&(ind==-1))
    {
      if (*(phi->varinds+i)==varind)
	ind=i;
      i++;
    }

  // Build the lookup tables from potential variables to messages and vice versa.
  phi2msg=malloc(sizeof(int)*phi->Nvars);
  msg2phi=malloc(sizeof(int)*Nmsgs);
  for (i=0; i<=(phi->Nvars-1); i++)
    *(phi2msg+i)=-1;
  
  for (i=0; i<=(Nmsgs-1); i++)
    {
      j=inlistind(phi->Nvars,phi->varinds,*((msg+*(msginds+i))->phi->varinds));
      *(phi2msg+j)=i; *(msg2phi+i)=j;
    }
  
  newphi->Nvars=1; 
  newphi->varinds=realloc(newphi->varinds,sizeof(int));
  *(newphi->varinds)=*(phi->varinds+ind);
  newphi->Nstates=realloc(newphi->Nstates,sizeof(int));

  if ((fgraph->facnodes+factorind)->type==SINGLEPATHEXPLAIN)
    {
      // only ko variable is not binary.
      //if ((fgraph->varnodes+varind)->type==3)
      if ((fgraph->varnodes+varind)->type==KNOCKOUT)
	*newphi->Nstates=3;
      else
	*newphi->Nstates=2;
    }
  else if ((fgraph->facnodes+factorind)->type==PATHSELOR)
    *newphi->Nstates=2;
  else if ((fgraph->facnodes+factorind)->type==SINGLEVAR)
    *newphi->Nstates=*(phi->Nstates+ind);
  else if ((fgraph->facnodes+factorind)->type==PREDICTION)
    *newphi->Nstates=3;

  newphi->Nconfigs=*newphi->Nstates;
  newphi->probs=realloc(newphi->probs,sizeof(double)*newphi->Nconfigs);
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    *(newphi->probs+i)=0.0;
    

  /*
  // Debug
  if (deb==1)
    {
      int cnt1, cnt2, cnt3, cnt4, cnt5, cnt6, ind;
      int *config, sigmabit, kobit;
      double val1, val2, val3, val4;

      printf("phi.varinds=(");
      for (i=0; i<=(phi->Nvars-1); i++)
	printf("%d ",*(phi->varinds+i));
      printf(")\n");

      i=0; ind=-1;
      while ((i<=(phi->Nvars-1))&&(ind==-1))
	{
	  if (*(phi2msg+i)<0)
	    ind=i;
	  i++;
	}
      cnt1=0; cnt2=0; cnt3=0;
      for (i=0; i<=(phi->Nconfigs-1); i++)
	{
	  if (*(phi->probs+i)==v1)
	    cnt1++;
	  else if (*(phi->probs+i)==v2)
	    cnt2++;
	  else if (*(phi->probs+i)==v3)
	    cnt3++;
	}
      printf("%d configs have v1, %d configs have v2, %d configs have v3.\n",cnt1,cnt2,cnt3);
      config=malloc(sizeof(int)*phi->Nvars);
      for (i=0; i<=(phi->Nvars-1); i++)
	*(config+i)=0;
      printf("v1=%f v2=%f v3=%f \n",v1,v2,v3);

      sigmabit=phi->Nvars-2; kobit=phi->Nvars-1;

      cnt1=0; cnt2=0; cnt3=0; cnt4=0; cnt5=0; cnt6=0;
      val=0; val2=0; val3=0; val4=0;
      for (i=0; i<=(phi->Nconfigs-1); i++)
	{
	  number2config2(phi->Nvars,phi->Nstates,i,config);

	  val1=1;
	  for (j=0; j<=(Nmsgs-1); j++)
	    val1*=*((msg+*(msginds+j))->phi->probs+*(config+*(msg2phi+j)));
	  val+=val1;

	  
	  //if (((*(config+sigmabit)==0)||(*(config+kobit)==0))&&(*(config+ind)==0))
	  //val2+=val1;
	  //else if ((*(config+sigmabit)==1)&&(*(config+kobit)!=0)&&(*(config+ind)==0))
	  //val3+=val1;
	  

	  if ((fabs(*(phi->probs+i)-v1)<1e-5)&&(*(config+ind)==1))
	    {
	      val2+=val1;

	      printf("(config=(");
	      //for (j=0; j<=(phi->Nvars-1); j++)
	      //printf("%d ",*(config+j));
	      //printf(")) ");

	      for (j=0; j<=((fgraph->facnodes+factorind)->Nedges-1); j++)
		printf("%d ",*(config+j+(fgraph->facnodes+factorind)->Nedges));
	      printf(", %d), val2=%f) ",*(config+kobit),val1);
	    }


	  else if ((fabs(*(phi->probs+i)-v2)<1e-5)&&(*(config+ind)==1))
	    val3+=val1;
	  else if ((fabs(*(phi->probs+i)-v3)<1e-5)&&(*(config+ind)==1))
	    val4+=val1;


	  if (*(config+ind)==0)
	    {
	      if (fabs(*(phi->probs+i)-v1)<1e-5)
		cnt1++;
	      else if (fabs(*(phi->probs+i)-v2)<1e-5)
		cnt2++;
	      else if (fabs(*(phi->probs+i)-v3)<1e-5)
		cnt3++;
	    }
	  else if (*(config+ind)==1)
	    {
	      if (fabs(*(phi->probs+i)-v1)<1e-5)
		cnt4++;
	      else if (fabs(*(phi->probs+i)-v2)<1e-5)
		cnt5++;
	      else if (fabs(*(phi->probs+i)-v3)<1e-5)
		cnt6++;
	    }
	}
      printf("x=0, %d configs have v1, %d configs have v2, %d configs have v3.\n",cnt1,cnt2,cnt3);
      printf("x=1, %d configs have v1, %d configs have v2, %d configs have v3.\n",cnt4,cnt5,cnt6);

      printf("sum msgs =%f, cond2 sum=%f, cond3 sum=%f, cond4 sum=%f \n",val,val2,val3,val4);

      free(config);
    }
  */
      

  // Single path explanation potential function.
  if ((fgraph->facnodes+factorind)->type==SINGLEPATHEXPLAIN)
    {
      int Nedges, Nsigedges, *presencemsgs, *dirmsgs, *signmsgs, sigmamsg, komsg;
      int flag, *signconfig, carry, N, predsign;
      double non_exp_term, valid_path_term;
      double val0, val1, val2;

      Nedges=(fgraph->facnodes+factorind)->Nedges;
      Nsigedges=(fgraph->facnodes+factorind)->Nsigedges;
      presencemsgs=malloc(sizeof(int)*Nedges);
      signmsgs=malloc(sizeof(int)*Nedges);
      dirmsgs=malloc(sizeof(int)*Nsigedges);
      for (i=0; i<=(Nedges-1); i++)
	{
	  *(presencemsgs+i)=-1; *(signmsgs+i)=-1;
	}
      for (i=0; i<=(Nsigedges-1); i++)
	*(dirmsgs+i)=-1;
      sigmamsg=-1; komsg=-1;
      for (i=0; i<=(Nedges-1); i++)
	{
	  j=0; flag=0;
	  while ((j<=(Nmsgs-1))&&(flag==0))
	    {
	      if (*((fgraph->facnodes+factorind)->presences+i)==*((msg+*(msginds+j))->phi->varinds))
		{  flag=1;  *(presencemsgs+i)=*(msginds+j); }
	      j++;
	    }
	  j=0; flag=0;
	  while ((j<=(Nmsgs-1))&&(flag==0))
	    {
	      if (*((fgraph->facnodes+factorind)->signs+i)==*((msg+*(msginds+j))->phi->varinds))
		{  flag=1;  *(signmsgs+i)=*(msginds+j); }
	      j++;
	    } 
	}
      for (i=0; i<=(Nsigedges-1); i++)
	{
	  j=0; flag=0;
	  while ((j<=(Nmsgs-1))&&(flag==0))
	    {
	      if (*((fgraph->facnodes+factorind)->sigdirvars+i)==*((msg+*(msginds+j))->phi->varinds))
		{  flag=1;  *(dirmsgs+i)=*(msginds+j); }
	      j++;
	    }  
	}
     j=0; flag=0;
     while ((j<=(Nmsgs-1))&&(flag==0))
       {
	 if ((fgraph->facnodes+factorind)->sigma==*((msg+*(msginds+j))->phi->varinds))

	   {  flag=1;  sigmamsg=*(msginds+j); }
	 j++;
       }   
     j=0; flag=0;
     while ((j<=(Nmsgs-1))&&(flag==0))
       {
	 if ((fgraph->facnodes+factorind)->ko==*((msg+*(msginds+j))->phi->varinds))
	   {  flag=1;  komsg=*(msginds+j); }
	 j++;
       }   
     signconfig=malloc(sizeof(int)*Nedges);
     N=1;
     for (i=0; i<=(Nedges-1); i++)
       N*=2;
     
     if ((sigmamsg>=0)&&(komsg>=0))
       {
	 non_exp_term=*((msg+sigmamsg)->phi->probs+0);
	 non_exp_term+=*((msg+komsg)->phi->probs+0);
	 non_exp_term-=(*((msg+sigmamsg)->phi->probs+0)*(*((msg+komsg)->phi->probs+0)));
	 non_exp_term*=(v2-v3); non_exp_term+=v3;
       }
     else if (sigmamsg<0)
       non_exp_term=(v2-v3)*(*((msg+komsg)->phi->probs+0))+v3;
     else if (komsg<0)
       non_exp_term=(v2-v3)*(*((msg+sigmamsg)->phi->probs+0))+v3;


     valid_path_term=(v1-v3);
     for (i=0; i<=(Nedges-1); i++)
       if (*(presencemsgs+i)>=0)
	 valid_path_term*=*((msg+*(presencemsgs+i))->phi->probs+1);
     for (i=0; i<=(Nsigedges-1); i++)
       if (*(dirmsgs+i)>=0)
	 valid_path_term*=*((msg+*(dirmsgs+i))->phi->probs+*((fgraph->facnodes+factorind)->sigdirvarvals+i));
     if (sigmamsg>=0)
       valid_path_term*=*((msg+sigmamsg)->phi->probs+1); 


     // Debug
     if (deb==1)
       {
	 printf("varind=%d, factorind=%d, type=%d, presence=(",varind,factorind,(fgraph->varnodes+varind)->type);
	 for (i=0; i<=(Nedges-1); i++)
	   printf("%d ",*((fgraph->facnodes+factorind)->presences+i));
	 printf("), sign=(");
	 for (i=0; i<=(Nedges-1); i++)
	   printf("%d ",*((fgraph->facnodes+factorind)->signs+i));
	 printf("), dir=(");
	 for (i=0; i<=(Nsigedges-1); i++)
	   printf("%d ",*((fgraph->facnodes+factorind)->sigdirvars+i));
	 printf("), ko=%d, sigma=%d \n",(fgraph->facnodes+factorind)->ko,(fgraph->facnodes+factorind)->sigma);
	 printf("incoming messages:\n");
	 for (i=0; i<=(Nmsgs-1); i++)
	   {
	     printf("msg[%d](%d,%d)=(",*(msginds+i),(msg+*(msginds+i))->from,(msg+*(msginds+i))->to);
	     for (j=0; j<=((msg+*(msginds+i))->phi->Nconfigs-1); j++)
	       printf("%f ",*((msg+*(msginds+i))->phi->probs+j));
	     printf(") ");
	   }
	 printf("\npresencemsgs=(");
	 for (i=0; i<=(Nedges-1); i++)
	   printf("%d ",*(presencemsgs+i));
	 printf("), signmsgs=(");
	 for (i=0; i<=(Nedges-1); i++)
	   printf("%d ",*(signmsgs+i));
	 printf("), dirmsgs=(");
	 for (i=0; i<=(Nsigedges-1); i++)
	   printf("%d ",*(dirmsgs+i));
	 printf("), sigmamsg=%d, komsg=%d \n",sigmamsg,komsg);
       }



     
     // varind is an edge presence variable.
     //if ((fgraph->varnodes+varind)->type==0)
     if ((fgraph->varnodes+varind)->type==EDGEPRESENCE)
       {	 
	 // m(x1=0)=(v2-v3)[m(sigma=0)+m(k=0)-m(sigma=0)m(k=0)]+v3.
	 // m(x1=1)=(v1-v3)m(x2=1)...m(xn=1)m(d1=d1')...m(dm=dm')m(sigma=1)[m(k=-1) \sum_{S: \prod_{S}=+1} \prod m(s) + m(k=+1) \sum_{S: \prod_{S}=-1} \prod m(s)]+(v2-v3)[m(sigma=0)+m(k=0)-m(sigma=0)m(k=0)]+v3.
	 *(newphi->probs+0)=non_exp_term;

	 val1=0;

	 for (i=0; i<=(Nedges-1); i++)
	   *(signconfig+i)=0;

	 
	 for (i=0; i<=(N-1); i++)
	   {
	     predsign=-1;
	     for (j=0; j<=(Nedges-1); j++)
	       predsign*=(*(signconfig+j)*2-1);
	     j=(int)((3-predsign)/2);

	     val2=*((msg+komsg)->phi->probs+j);
	     for (j=0; j<=(Nedges-1); j++)
	       val2*=*((msg+*(signmsgs+j))->phi->probs+*(signconfig+j));
	     val1+=val2;
	     
	     // Debug
	     if (deb==1)
	       {
		 printf("(config=(");
		 for (j=0; j<=(Nedges-1); j++)
		   printf("%d ",*(signconfig+j));
		 printf(", %d), val2=%f) ",(predsign+1)/2,val2*valid_path_term/(v1-v3));
	       }


	     carry=1; j=Nedges-1;

	     while ((carry==1)&&(j>=0))
	       {
		 *(signconfig+j)+=carry;
		 if (*(signconfig+j)<2)
		   carry=0;
		 else
		   *(signconfig+j)=0;
		 j--;
	       }	     
	   }
	 *(newphi->probs+1)=valid_path_term*val1+non_exp_term;	 	 

	 // Debug
	 if (deb==1)
	   printf("first prob term=%f, newphi=(%f %f) \n",valid_path_term*val1/(v1-v3),*(newphi->probs+0),*(newphi->probs+1));

       }
     

     // varind is an edge sign variable.
     //else if ((fgraph->varnodes+varind)->type==1)
     else if ((fgraph->varnodes+varind)->type==EDGESIGN)
       {
	 // m(s1=+1)=(v1-v3)m(x1=1)...m(xn=1)m(d1=d1')...m(dm=dm')m(sigma=1)[m(k=-1) \sum_{S: \prod_{S}=+1, s1=+1} \prod m(s) + m(k=+1) \sum_{S: \prod_{S}=-1, s1=+1} \prod m(s)]+(v2-v3)[m(sigma=0)+m(k=0)-m(sigma=0)m(k=0)]+v3.
	 // m(s1=-1)=(v1-v3)m(x1=1)...m(xn=1)m(d1=d1')...m(dm=dm')m(sigma=1)[m(k=-1) \sum_{S: \prod_{S}=+1, s1=-1} \prod m(s) + m(k=+1) \sum_{S: \prod_{S}=-1, s1=-1} \prod m(s)]+(v2-v3)[m(sigma=0)+m(k=0)-m(sigma=0)m(k=0)]+v3.
	 val1=0; val0=0;
	 for (i=0; i<=(Nedges-1); i++)
	   *(signconfig+i)=0;
	 for (i=0; i<=(N-1); i++)
	   {
	     predsign=-1;
	     for (j=0; j<=(Nedges-1); j++)
	       predsign*=(*(signconfig+j)*2-1);
	     j=(int)((3-predsign)/2);

	     val2=*((msg+komsg)->phi->probs+j);
	     for (j=0; j<=(Nedges-1); j++)
	       {
		 if (*(signmsgs+j)>=0)
		   val2*=*((msg+*(signmsgs+j))->phi->probs+*(signconfig+j));
		 else
		   k=j;
	       }
	     if (*(signconfig+k)==0)
	       val0+=val2;
	     else
	       val1+=val2;

	     carry=1; j=Nedges-1;
	     while ((carry==1)&&(j>=0))
	       {
		 *(signconfig+j)+=carry;
		 if (*(signconfig+j)<2)
		   carry=0;
		 else
		   *(signconfig+j)=0;
		 j--;
	       }
	   }
	 *(newphi->probs+0)=valid_path_term*val0+non_exp_term;
	 *(newphi->probs+1)=valid_path_term*val1+non_exp_term;
       }
     
     // varind is an edge direction variable.
     //else if ((fgraph->varnodes+varind)->type==2)
     else if ((fgraph->varnodes+varind)->type==EDGEDIR)
       {
	 int sigedgeind=-1;

	 i=0;
	 while ((i<=(Nsigedges-1))&&(sigedgeind==-1))
	   {
	     if (*(dirmsgs+i)==-1)
	       sigedgeind=i;
	     i++;
	   }

	 // m(d1!=d1')=(v2-v3)[m(sigma=0)+m(k=0)-m(sigma=0)m(k=0)]+v3.
	 // m(d1=d1')=(v1-v3)m(x1=1)...m(xn=1)m(d2=d2')...m(dm=dm')m(sigma=1)[m(k=-1) \sum_{S: \prod_{S}=+1} \prod m(s) + m(k=+1) \sum_{S: \prod_{S}=-1} \prod m(s)]+(v2-v3)[m(sigma=0)+m(k=0)-m(sigma=0)m(k=0)]+v3.

	 j=(*((fgraph->facnodes+factorind)->sigdirvarvals+sigedgeind)==0 ? 1 : 0);
	 *(newphi->probs+j)=non_exp_term;
	 val1=0;
	 for (i=0; i<=(Nedges-1); i++)
	   *(signconfig+i)=0;
	 for (i=0; i<=(N-1); i++)
	   {
	     predsign=-1;
	     for (j=0; j<=(Nedges-1); j++)
	       predsign*=(*(signconfig+j)*2-1);
	     j=(int)((3-predsign)/2);

	     val2=*((msg+komsg)->phi->probs+j);
	     for (j=0; j<=(Nedges-1); j++)
	       val2*=*((msg+*(signmsgs+j))->phi->probs+*(signconfig+j));
	     val1+=val2;

	     carry=1; j=Nedges-1;
	     while ((carry==1)&&(j>=0))
	       {
		 *(signconfig+j)+=carry;
		 if (*(signconfig+j)<2)
		   carry=0;
		 else
		   *(signconfig+j)=0;
		 j--;
	       }
	   }

	 j=*((fgraph->facnodes+factorind)->sigdirvarvals+sigedgeind);

	 *(newphi->probs+j)=valid_path_term*val1+non_exp_term;

	 // Debug
	 if (deb==1)
	   printf("newphi=(%f %f) \n",*(newphi->probs+0),*(newphi->probs+1));

       }

     // varind is a knockout effect variable.
     //else if ((fgraph->varnodes+varind)->type==3)
     else if ((fgraph->varnodes+varind)->type==KNOCKOUT)
       {
	 // m(k=0)=(v2-v3)[m(sigma=0)+1-m(sigma=0)]+v3=v2.
	 // m(k=+1)=(v1-v3)m(x1=1)...m(xn=1)m(d1=d1')...m(dm=dm')m(sigma=1)[m(k=+1) \sum_{S: \prod_{S}=-1, s1=+1} \prod m(s)]+(v2-v3)m(sigma=0)+v3.
	 // m(k=-1)=(v1-v3)m(x1=1)...m(xn=1)m(d1=d1')...m(dm=dm')m(sigma=1)[m(k=-1) \sum_{S: \prod_{S}=+1, s1=-1} \prod m(s)]+(v2-v3)m(sigma=0)+v3.

	 *(newphi->probs+0)=v2;

	 val1=0; val0=0;
	 for (i=0; i<=(Nedges-1); i++)
	   *(signconfig+i)=0;
	 for (i=0; i<=(N-1); i++)
	   {
	     predsign=-1;
	     for (j=0; j<=(Nedges-1); j++)
	       predsign*=(*(signconfig+j)*2-1);
	     val2=1;
	     for (j=0; j<=(Nedges-1); j++)
	       val2*=*((msg+*(signmsgs+j))->phi->probs+*(signconfig+j));
	     j=(int)((3-predsign)/2);

	     if (j==2)
	       val0+=val2;
	     else
	       val1+=val2;

	     carry=1; j=Nedges-1;
	     while ((carry==1)&&(j>=0))
	       {
		 *(signconfig+j)+=carry;
		 if (*(signconfig+j)<2)
		   carry=0;
		 else
		   *(signconfig+j)=0;
		 j--;
	       }
	   }
	 *(newphi->probs+2)=valid_path_term*val0+non_exp_term;
	 *(newphi->probs+1)=valid_path_term*val1+non_exp_term;
       }

     // varind is a path selection variable.
     //else if ((fgraph->varnodes+varind)->type==4)
     else if ((fgraph->varnodes+varind)->type==PATHSEL)
       {
	 // m(sigma=0)=v3.
	 // m(sigma=1)=(v1-v3)m(x1=1)...m(xn=1)m(d1=d1')...m(dm=dm')m(sigma=1)[m(k=-1) \sum_{S: \prod_{S}=+1} \prod m(s) + m(k=+1) \sum_{S: \prod_{S}=-1} \prod m(s)]+(v2-v3)m(k=0)+v3.

	 *(newphi->probs+0)=v2;
	 val1=0;
	 for (i=0; i<=(Nedges-1); i++)
	   *(signconfig+i)=0;
	 for (i=0; i<=(N-1); i++)
	   {
	     predsign=-1;
	     for (j=0; j<=(Nedges-1); j++)
	       predsign*=(*(signconfig+j)*2-1);
	     j=(int)((3-predsign)/2);

	     val2=*((msg+komsg)->phi->probs+j);
	     for (j=0; j<=(Nedges-1); j++)
	       val2*=*((msg+*(signmsgs+j))->phi->probs+*(signconfig+j));
	     val1+=val2;

	     carry=1; j=Nedges-1;
	     while ((carry==1)&&(j>=0))
	       {
		 *(signconfig+j)+=carry;
		 if (*(signconfig+j)<2)
		   carry=0;
		 else
		   *(signconfig+j)=0;
		 j--;
	       }
	   }
	 *(newphi->probs+1)=valid_path_term*val1;
	 *(newphi->probs+1)+=(v2-v3)*(*((msg+komsg)->phi->probs+0));
	 *(newphi->probs+1)+=v3;
       }
       

     // A singlepath explain potential does not contain output response variables.
     
     free(presencemsgs); free(dirmsgs); free(signmsgs); free(signconfig);
     
     
    }

  // Path selection OR function.
  // m(sigma1=0)=(v3-v1)m(sigma2=0)...m(sigman=0)+v1.
  // m(sigma1=1)=v1.
  // Here v1=1.

  
  else if ((fgraph->facnodes+factorind)->type==PATHSELOR)
    {
      double val1;

      val1=v3-1;
      for (i=0; i<=(Nmsgs-1); i++)
	val1*=*((msg+*(msginds+i))->phi->probs+0);
      val1+=1;
      *(newphi->probs+0)=val1;
      *(newphi->probs+1)=1;
    }

  // Single variable potential function returns the potential function itself.  Already covered.

  // Prediction potential function.  Only marginalize to the output variable.
  else if ((fgraph->facnodes+factorind)->type==PREDICTION)
    {
      int Nedges, Nsigedges, *presencemsgs, *dirmsgs, *signmsgs, sigmamsg, komsg;
      int flag, *signconfig, carry, N, predsign;
      double valid_path_term, val0, val1, val2;

      // Only update the output response variable.
      //if ((fgraph->varnodes+varind)->type!=5)
      if ((fgraph->varnodes+varind)->type!=OUTPUTRESPONSE)
	return;
      

      // If the path is not effective, then m(y=0)=v2, m(y=+1)=m(y=-1)=v3.
      if ((fgraph->facnodes+factorind)->effective==0)
	{
	  *(newphi->probs+0)=v2; *(newphi->probs+1)=v3; *(newphi->probs+2)=v3;
	}
      else
	{      
	  // m(y=0)=(v3-v2)m(x1=1)...m(xn=1)m(d1=d1')...m(dm=dm')+(v2-v3)+v3.
	  // m(y=+1)=(v1-v3)m(x1=1)...m(xn=1)m(d1=d1')...m(dm=dm') \sum_{S: (-1)s1'...sn'=+1}\prod_{si}m(si=si')+v3.
	  // m(y=-1)=(v1-v3)m(x1=1)...m(xn=1)m(d1=d1')...m(dm=dm') \sum_{S: (-1)s1'...sn'=-1}\prod_{si}m(si=si')+v3.
	  Nedges=(fgraph->facnodes+factorind)->Nedges;
	  Nsigedges=(fgraph->facnodes+factorind)->Nsigedges;
       
	  presencemsgs=malloc(sizeof(int)*Nedges);
	  signmsgs=malloc(sizeof(int)*Nedges);
	  dirmsgs=malloc(sizeof(int)*Nsigedges);
	  for (i=0; i<=(Nedges-1); i++)
	    {
	      *(presencemsgs+i)=-1; *(signmsgs+i)=-1;
	    }
	  for (i=0; i<=(Nsigedges-1); i++)
	    *(dirmsgs+i)=-1;
	  sigmamsg=-1; komsg=-1;
	  for (i=0; i<=(Nedges-1); i++)
	    {
	      j=0; flag=0;
	      while ((j<=(Nmsgs-1))&&(flag==0))
		{
		  if (*((fgraph->facnodes+factorind)->presences+i)==*((msg+*(msginds+j))->phi->varinds))
		    {  flag=1;  *(presencemsgs+i)=*(msginds+j); }
		  j++;
		}
	      j=0; flag=0;
	      while ((j<=(Nmsgs-1))&&(flag==0))
		{
		  if (*((fgraph->facnodes+factorind)->signs+i)==*((msg+*(msginds+j))->phi->varinds))
		    {  flag=1;  *(signmsgs+i)=*(msginds+j); }
		  j++;
		} 
	    }
	  for (i=0; i<=(Nsigedges-1); i++)
	    {
	      j=0; flag=0;
	      while ((j<=(Nmsgs-1))&&(flag==0))
		{
		  if (*((fgraph->facnodes+factorind)->sigdirvars+i)==*((msg+*(msginds+j))->phi->varinds))
		    {  flag=1;  *(dirmsgs+i)=*(msginds+j); }
		  j++;
		}  
	    }
	  signconfig=malloc(sizeof(int)*Nedges);
	  N=1;
	  for (i=0; i<=(Nedges-1); i++)
	    N*=2;
     
	  valid_path_term=1;
	  for (i=0; i<=(Nedges-1); i++)
	    if (*(presencemsgs+i)>=0)
	      valid_path_term*=*((msg+*(presencemsgs+i))->phi->probs+1);
	  for (i=0; i<=(Nsigedges-1); i++)
	    if (*(dirmsgs+i)>=0)
	      valid_path_term*=*((msg+*(dirmsgs+i))->phi->probs+*((fgraph->facnodes+factorind)->sigdirvarvals+i));
	  *(newphi->probs+0)=(v3-v2)*valid_path_term+v2;
	
	  // Debug
	  if (deb==1)
	    printf("valid_path_term=%f, v3-v2=%f, newphi[0]=%f \n",valid_path_term,v3-v2,*(newphi->probs+0));

  
	  val1=0; val0=0;
	  for (i=0; i<=(Nedges-1); i++)
	    *(signconfig+i)=0;
	  for (i=0; i<=(N-1); i++)
	    {
	      predsign=-1;
	      for (j=0; j<=(Nedges-1); j++)
		predsign*=(*(signconfig+j)*2-1);	 
	      val2=1;
	      for (j=0; j<=(Nedges-1); j++)
		val2*=*((msg+*(signmsgs+j))->phi->probs+*(signconfig+j));
	      j=(int)((3-predsign)/2);
	      if (j==2)
		val0+=val2;
	      else if (j==1)
		val1+=val2;
	      
	      carry=1; j=Nedges-1;
	      while ((carry==1)&&(j>=0))
		{
		  *(signconfig+j)+=carry;
		  if (*(signconfig+j)<2)
		    carry=0;
		  else
		    *(signconfig+j)=0;
		  j--;
		}
	    }
	  *(newphi->probs+1)=(v1-v3)*valid_path_term*val1+v3;
	  *(newphi->probs+2)=(v1-v3)*valid_path_term*val0+v3;	

	  // Debug
	  if (deb==1)
	    printf("v1-v3=%f val1=%f val2=%f newphi[1]=%f newphi[2]=%f\n",v1-v3,val1,val2,*(newphi->probs+1),*(newphi->probs+2));


	  free(presencemsgs); free(dirmsgs); free(signmsgs); free(signconfig);
	}
    }
  

  free(phi2msg); free(msg2phi);

  
  // Normalize the message.
  val=0;
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    val+=*(newphi->probs+i);
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    *(newphi->probs+i)/=val;
    
}


// Combination of maxsumproduct1 and maxsumproduct2.  Allow to choose either update messages thouroughly or only messages going to output responese nodes.
// Difference from maxsumproduct1 and maxsumproduct2: apply fast algorithms to perform maximization (maxmarginal3) and marginalization (marginalizepotential5) for logic functions.  Boost the speed by 20 folds.
// mode=0: maximization, mode=1: marginalization.
// updatemode=0: regular update, updatemode=1: update messages going to response nodes only.

void maxsumproduct3(struct mix_facgraph *fgraph, struct message *msg, int *fixedvals, struct potential *belief, int mode, int updatemode)
{
  int i, j, k, l, n, *fixedmsg, **adjmsg;
  int Nsteps, terminate, flag, *containy, Nstarty;
  struct message *tmpmsg;
  int *isinitialized, **msgfromvar, **msgtovar, **msgfromfac, **msgtofac;


  // If a variable x is fixed, then the messages going out of x are fixed, i.e. they are not updated according to belief propagation.  

  fixedmsg=malloc(sizeof(int)*fgraph->Ne*2);
  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      if (((msg+i)->dir==0)&&(*(fixedvals+(msg+i)->from)>=0))	
	*(fixedmsg+i)=1;
      else
	*(fixedmsg+i)=0;
    }

  // Initialize template messages.  Needed only when updatemode=0.
  if (updatemode==0)
    {
      tmpmsg=malloc(sizeof(struct message)*fgraph->Ne*2);
      for (i=0; i<=(fgraph->Ne*2-1); i++)
	{
	  (tmpmsg+i)->from=(msg+i)->from;
	  (tmpmsg+i)->to=(msg+i)->to;
	  (tmpmsg+i)->dir=(msg+i)->dir;
	  (tmpmsg+i)->phi=malloc(sizeof(struct potential));
	  (tmpmsg+i)->phi->Nvars=(msg+i)->phi->Nvars; 
	  (tmpmsg+i)->phi->varinds=malloc(sizeof(int)*(tmpmsg+i)->phi->Nvars);
	  for (j=0; j<=((tmpmsg+i)->phi->Nvars-1); j++)
	    *((tmpmsg+i)->phi->varinds+j)=*((msg+i)->phi->varinds+j);
	  (tmpmsg+i)->phi->Nconfigs=(msg+i)->phi->Nconfigs;
	  (tmpmsg+i)->phi->Nstates=malloc(sizeof(int)*(tmpmsg+i)->phi->Nvars);
	  for (j=0; j<=((tmpmsg+i)->phi->Nvars-1); j++)
	    *((tmpmsg+i)->phi->Nstates+j)=*((msg+i)->phi->Nstates+j);
	  (tmpmsg+i)->phi->probs=malloc(sizeof(double)*((tmpmsg+i)->phi->Nconfigs));
	  for (j=0; j<=((tmpmsg+i)->phi->Nconfigs-1); j++)
	    *((tmpmsg+i)->phi->probs+j)=*((msg+i)->phi->probs+j);    
	  
      
	  // Fix the messages going out of fixed variables.
	  if (((msg+i)->dir==0)&&(*(fixedvals+(msg+i)->from)>=0))
	    {
	      for (j=0; j<=((msg+i)->phi->Nconfigs-1); j++)
		*((msg+i)->phi->probs+j)=0.0;
	      *((msg+i)->phi->probs+*(fixedvals+(msg+i)->from))=1.0;
	    }
	}
    }


  // Detect the prediction potential functions which contain output response variables.
  // Needed when updatemode=1.
  Nstarty=fgraph->Nvars;
  if (updatemode==1)
    {
      containy=malloc(sizeof(int)*fgraph->Ninteractions);
      k=0;
      //while ((k<=(fgraph->Nvars-1))&&((fgraph->varnodes+k)->type!=5))
      while ((k<=(fgraph->Nvars-1))&&((fgraph->varnodes+k)->type!=OUTPUTRESPONSE))
	k++;
      for (i=0; i<=(fgraph->Ninteractions-1); i++)
	*(containy+i)=0;
      for (i=0; i<=(fgraph->Ninteractions-1); i++)
	if (*((fgraph->facnodes+i)->phivi->varinds)>=k)
	  *(containy+i)=1;
      Nstarty=k;
    }

  /*
  // Find neighors of messages.
  adjmsg=malloc(sizeof(int)*fgraph->Ne*2);
  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      *(adjmsg+i)=malloc(sizeof(int));
      **(adjmsg+i)=0;
    }
  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      for (j=0; j<=(fgraph->Ne*2-1); j++)
	{
	  if (((msg+i)->dir!=(msg+j)->dir)&&((msg+j)->to==(msg+i)->from)&&((msg+j)->from!=(msg+i)->to))
	    {
	      **(adjmsg+i)+=1;
	      *(adjmsg+i)=realloc(*(adjmsg+i),sizeof(int)*(**(adjmsg+i)+1));
	      *(*(adjmsg+i)+**(adjmsg+i))=j;
	    }
	}
    }
  */

  
  // Linear time algorithm to identify neighbors.
  msgfromvar=malloc(sizeof(int)*fgraph->Nvars);
  msgtovar=malloc(sizeof(int)*fgraph->Nvars);
  msgfromfac=malloc(sizeof(int)*fgraph->Ninteractions);
  msgtofac=malloc(sizeof(int)*fgraph->Ninteractions);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      *(msgfromvar+i)=malloc(sizeof(int));
      **(msgfromvar+i)=0;
      *(msgtovar+i)=malloc(sizeof(int));
      **(msgtovar+i)=0;
    }
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    {
      *(msgfromfac+i)=malloc(sizeof(int));
      **(msgfromfac+i)=0;
      *(msgtofac+i)=malloc(sizeof(int));
      **(msgtofac+i)=0;
    }
  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      if ((msg+i)->dir==0)
	{
	  **(msgfromvar+(msg+i)->from)+=1;
	  *(msgfromvar+(msg+i)->from)=realloc(*(msgfromvar+(msg+i)->from),sizeof(int)*(**(msgfromvar+(msg+i)->from)+1));
	  *(*(msgfromvar+(msg+i)->from)+**(msgfromvar+(msg+i)->from))=i;
	  **(msgtofac+(msg+i)->to)+=1;
	  *(msgtofac+(msg+i)->to)=realloc(*(msgtofac+(msg+i)->to),sizeof(int)*(**(msgtofac+(msg+i)->to)+1));
	  *(*(msgtofac+(msg+i)->to)+**(msgtofac+(msg+i)->to))=i;
	}
      else if ((msg+i)->dir==1)
	{
	  **(msgfromfac+(msg+i)->from)+=1;
	  *(msgfromfac+(msg+i)->from)=realloc(*(msgfromfac+(msg+i)->from),sizeof(int)*(**(msgfromfac+(msg+i)->from)+1));
	  *(*(msgfromfac+(msg+i)->from)+**(msgfromfac+(msg+i)->from))=i;
	  **(msgtovar+(msg+i)->to)+=1;
	  *(msgtovar+(msg+i)->to)=realloc(*(msgtovar+(msg+i)->to),sizeof(int)*(**(msgtovar+(msg+i)->to)+1));
	  *(*(msgtovar+(msg+i)->to)+**(msgtovar+(msg+i)->to))=i;
	}
    }
  

  /*
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      *(msgfromvar+i)=realloc(*(msgfromvar+i),sizeof(int)*(**(msgfromvar+i)+1));
      *(msgtovar+i)=realloc(*(msgtovar+i),sizeof(int)*(**(msgtovar+i)+1));
      **(msgfromvar+i)=0; **(msgtovar+i)=0;
    }
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    {
      *(msgfromfac+i)=realloc(*(msgfromfac+i),sizeof(int)*(**(msgfromfac+i)+1));
      *(msgtofac+i)=realloc(*(msgtofac+i),sizeof(int)*(**(msgtofac+i)+1));
      **(msgfromfac+i)=0; **(msgtofac+i)=0;
    }
  
  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      if ((msg+i)->dir==0)
	{
	  **(msgfromvar+(msg+i)->from)+=1;
	  //*(msgfromvar+(msg+i)->from)=realloc(*(msgfromvar+(msg+i)->from),sizeof(int)*(**(msgfromvar+(msg+i)->from)+1));
	  *(*(msgfromvar+(msg+i)->from)+**(msgfromvar+(msg+i)->from))=i;
	  **(msgtofac+(msg+i)->to)+=1;
	  //*(msgtofac+(msg+i)->to)=realloc(*(msgtofac+(msg+i)->to),sizeof(int)*(**(msgtofac+(msg+i)->to)+1));
	  *(*(msgtofac+(msg+i)->to)+**(msgtofac+(msg+i)->to))=i;
	}
      else if ((msg+i)->dir==1)
	{
	  **(msgfromfac+(msg+i)->from)+=1;
	  //*(msgfromfac+(msg+i)->from)=realloc(*(msgfromfac+(msg+i)->from),sizeof(int)*(**(msgfromfac+(msg+i)->from)+1));
	  *(*(msgfromfac+(msg+i)->from)+**(msgfromfac+(msg+i)->from))=i;
	  **(msgtovar+(msg+i)->to)+=1;
	  //*(msgtovar+(msg+i)->to)=realloc(*(msgtovar+(msg+i)->to),sizeof(int)*(**(msgtovar+(msg+i)->to)+1));
	  *(*(msgtovar+(msg+i)->to)+**(msgtovar+(msg+i)->to))=i;
	}
    }
  */
  


  
  adjmsg=malloc(sizeof(int)*fgraph->Ne*2);
  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      *(adjmsg+i)=malloc(sizeof(int));
      **(adjmsg+i)=0;
    }
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      for (j=1; j<=**(msgfromvar+i); j++)
	{
	  for (k=1; k<=**(msgtovar+i); k++)
	    {
	      if ((msg+*(*(msgfromvar+i)+j))->to!=(msg+*(*(msgtovar+i)+k))->from)
		{
		  **(adjmsg+*(*(msgfromvar+i)+j))+=1;
		  *(adjmsg+*(*(msgfromvar+i)+j))=realloc(*(adjmsg+*(*(msgfromvar+i)+j)),sizeof(int)*(**(adjmsg+*(*(msgfromvar+i)+j))+1));
		  *(*(adjmsg+*(*(msgfromvar+i)+j))+**(adjmsg+*(*(msgfromvar+i)+j)))=*(*(msgtovar+i)+k);
		}
	    }
	}
    }
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    {
      for (j=1; j<=**(msgfromfac+i); j++)
	{
	  for (k=1; k<=**(msgtofac+i); k++)
	    {
	      if ((msg+*(*(msgfromfac+i)+j))->to!=(msg+*(*(msgtofac+i)+k))->from)
		{
		  **(adjmsg+*(*(msgfromfac+i)+j))+=1;
		  *(adjmsg+*(*(msgfromfac+i)+j))=realloc(*(adjmsg+*(*(msgfromfac+i)+j)),sizeof(int)*(**(adjmsg+*(*(msgfromfac+i)+j))+1));
		  *(*(adjmsg+*(*(msgfromfac+i)+j))+**(adjmsg+*(*(msgfromfac+i)+j)))=*(*(msgtofac+i)+k);
		}
	    }
	}
    }
  

  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      free(*(msgfromvar+i)); free(*(msgtovar+i));
    }
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    {
      free(*(msgfromfac+i)); free(*(msgtofac+i));
    }
  free(msgfromvar); free(msgtovar); free(msgfromfac); free(msgtofac);


  
  n=0; 

  // Perform iterations twice if updatemode=1.

  if (updatemode==0)
    Nsteps=maxpathlength*2;
  else if (updatemode==1)
    Nsteps=1;

  
  terminate=0;

  while ((terminate==0)&&(n<=Nsteps))
  //while (n<=Nsteps)
    {
      // Debug
      //printf("n=%d\n",n);

      
      // Copy msg to tmpmsg.
      // Needed only when updatemode=0.
      if (updatemode==0)
	{
	  for (i=0; i<=(fgraph->Ne*2-1); i++)
	    copypotential((tmpmsg+i)->phi,(msg+i)->phi);
	}
      
      // The updated message collects the previous message coming to the from node (except the inverse updated message).
      for (i=0; i<=(fgraph->Ne*2-1); i++)
	{
	  double val;

	  // A variable to function message is the product of all incoming messages.
	  if (((msg+i)->dir==0)&&(*(fixedmsg+i)==0))
	    {	     
	      // If updatemode=0, then update all messages and store the outcomes in tmpmsg.
	      if (updatemode==0)
		{
		  for (j=0; j<=((tmpmsg+i)->phi->Nconfigs-1); j++)
		    *((tmpmsg+i)->phi->probs+j)=1.0;

		  for (l=1; l<=**(adjmsg+i); l++)
		    {
		      j=*(*(adjmsg+i)+l);		      
		      if (((msg+j)->to==(msg+i)->from)&&((msg+j)->from!=(msg+i)->to))
			potentialproduct2((tmpmsg+i)->phi,(msg+j)->phi);
		    }	       

		  // Normalize tmpmsg.  Otherwise marginalizepotential5 won't work.
		  val=0;
		  for (j=0; j<=((tmpmsg+i)->phi->Nconfigs-1); j++)
		    val+=*((tmpmsg+i)->phi->probs+j);
		  for (j=0; j<=((tmpmsg+i)->phi->Nconfigs-1); j++)
		    *((tmpmsg+i)->phi->probs+j)/=val;
		}

	      // If updatemode=1, then only update messages from old variable nodes to new prediction factor nodes.  Store the outcomes in msg directly.
	      //else if ((updatemode==1)&&((fgraph->varnodes+(msg+i)->from)->type!=5)&&(*(containy+(msg+i)->to)==1))
	      else if ((updatemode==1)&&((fgraph->varnodes+(msg+i)->from)->type!=OUTPUTRESPONSE)&&(*(containy+(msg+i)->to)==1))
		{
		  for (j=0; j<=((msg+i)->phi->Nconfigs-1); j++)
		    *((msg+i)->phi->probs+j)=1.0;	      
		  for (l=1; l<=**(adjmsg+i); l++)
		    {
		      j=*(*(adjmsg+i)+l);		  		  
		      if (((msg+j)->to==(msg+i)->from)&&((msg+j)->from!=(msg+i)->to))
			potentialproduct2((msg+i)->phi,(msg+j)->phi);		  
		    }
		  // Normalize msg.  Otherwise marginalizepotential5 won't work.
		  val=0;
		  for (j=0; j<=((msg+i)->phi->Nconfigs-1); j++)
		    val+=*((msg+i)->phi->probs+j);
		  for (j=0; j<=((msg+i)->phi->Nconfigs-1); j++)
		    *((msg+i)->phi->probs+j)/=val;
		}
	    }
	  	  
	  // A function to variable message is the maximization of the product of incoming messages and the function.
	  //else if (((msg+i)->dir==1)&&(*(fixedmsg+i)==0))
	  //else if ((((msg+i)->dir==1)&&(*(fixedmsg+i)==0)&&(updatemode==0))||(((msg+i)->dir==1)&&((fgraph->varnodes+(msg+i)->to)->type==5)&&(updatemode==1)))
	  else if ((((msg+i)->dir==1)&&(*(fixedmsg+i)==0)&&(updatemode==0))||(((msg+i)->dir==1)&&((fgraph->varnodes+(msg+i)->to)->type==OUTPUTRESPONSE)&&(updatemode==1)))
	    {
	      struct potential *lphi;
	      

	      // Apply the maximization.
	      if (mode==0)
		{
		  int Nmsgs=0, *msginds, deb=0, ftype;
		  lphi=malloc(sizeof(struct potential));
		  lphi->Nvars=1; lphi->varinds=malloc(sizeof(int));
		  *(lphi->varinds)=(msg+i)->to;
		  lphi->Nstates=malloc(sizeof(int));
		  *(lphi->Nstates)=2;
		  lphi->Nconfigs=2;
		  lphi->probs=malloc(sizeof(double)*2);
		  
		  Nmsgs=0; msginds=malloc(sizeof(int));
		  for (j=1; j<=**(adjmsg+i); j++)
		    {
		      k=*(*(adjmsg+i)+j);
		      //if (((msg+k)->to==(msg+i)->from)&&((msg+k)->from!=(msg+i)->to)&&((updatemode==0)||((updatemode==1)&&((fgraph->varnodes+(msg+k)->from)->type!=5))))
		      if (((msg+k)->to==(msg+i)->from)&&((msg+k)->from!=(msg+i)->to)&&((updatemode==0)||((updatemode==1)&&((fgraph->varnodes+(msg+k)->from)->type!=OUTPUTRESPONSE))))
			{
			  Nmsgs++; msginds=realloc(msginds,sizeof(int)*Nmsgs);
			  *(msginds+Nmsgs-1)=k;
			}
		    }
		  
		  // If the potential function type is PATHSELOR or SINGLEVAR or PREDICTION, then perform numerical max marginal.		  
		  ftype=(fgraph->facnodes+(msg+i)->from)->type;
		  //if ((ftype==SINGLEVAR)||(ftype==PATHSELOR)||(ftype==PREDICTION))
		  if ((ftype==SINGLEVAR)||(ftype==PREDICTION))
		    {
		      //lphi=createpotential((fgraph->facnodes+(msg+i)->from)->phivi);

		      copypotential(lphi,(fgraph->facnodes+(msg+i)->from)->phivi);

		      for (j=1; j<=**(adjmsg+i); j++)
			{
			  k=*(*(adjmsg+i)+j);
			  if (((msg+k)->to==(msg+i)->from)&&((msg+k)->from!=(msg+i)->to))
			    {		      
			      potentialproduct3(lphi,(msg+k)->phi);
			      maximizepotential2(lphi,*((msg+k)->phi->varinds));
			    }
			}		      		      
		    }		  
		  // If the potential function type is SINGLEPATHEXPLAIN, then perform logical max marginal.
		  //else if (ftype==SINGLEPATHEXPLAIN)
		  else if ((ftype==SINGLEPATHEXPLAIN)||(ftype==PATHSELOR))
		    {
		      // Debug
		      //if (((msg+i)->from==0)&&((msg+i)->to==136)&&(n==2))
		      //deb=1;

		      /*
		      // Debug
		      if (((msg+i)->from==1065)&&((msg+i)->to==49797))
			{
			  deb=1;
			  for (j=1; j<=**(adjmsg+i); j++)
			    {
			      k=*(*(adjmsg+i)+j);
			      printf("msg[%d,%d](%d)=(",(msg+k)->from,(msg+k)->to,*((msg+k)->phi->varinds));
			      for (l=0; l<=((msg+k)->phi->Nconfigs-1); l++)
				printf("%f ",*((msg+k)->phi->probs+l));
			      printf(") ");
			    }
			  printf("\n");
			}
		      */

		      // Debug
		      //if ((msg+i)->to==55568)
		      //deb=1;


		      //maxmarginal1(fgraph->facnodes+(msg+i)->from,Nmsgs,msginds,msg,lphi,deb);
		      //maxmarginal2(fgraph->facnodes+(msg+i)->from,Nmsgs,msginds,msg,lphi,deb);
		      //maxmarginal3(fgraph->facnodes+(msg+i)->from,Nmsgs,msginds,msg,lphi,deb);

		      // Debug
		      //printf("(%d,%d) ",(msg+i)->from,(msg+i)->to);

		      maxmarginal4(fgraph,(msg+i)->from,(msg+i)->to,Nmsgs,msginds,msg,lphi,deb);
		    }		  
		  free(msginds);


		  //  }
		}

	      // Apply marginalization.
	      else
		{
		  int *msginds, Nmsgs=0, deb, ftype;
		  
		  lphi=malloc(sizeof(struct potential));
		  lphi->Nvars=1; lphi->varinds=malloc(sizeof(int));
		  *(lphi->varinds)=(msg+i)->to;
		  j=0; k=-1;
		  while ((j<=((fgraph->facnodes+(msg+i)->from)->phivi->Nvars-1))&&(k==-1))
		    {
		      if ((msg+i)->to==*((fgraph->facnodes+(msg+i)->from)->phivi->varinds+j))
			k=j;
		      j++;
		    }
		  lphi->Nstates=malloc(sizeof(int));
		  //*(lphi->Nstates)=*((fgraph->facnodes+(msg+i)->from)->phivi->Nstates+k);

		  //if (((fgraph->varnodes+(msg+i)->to)->type==3)||((fgraph->varnodes+(msg+i)->to)->type==5))
		  if (((fgraph->varnodes+(msg+i)->to)->type==KNOCKOUT)||((fgraph->varnodes+(msg+i)->to)->type==OUTPUTRESPONSE))
		    *(lphi->Nstates)=3;
		  else
		    *(lphi->Nstates)=2;
		  
		  lphi->Nconfigs=*(lphi->Nstates);		  
		  lphi->probs=malloc(sizeof(double)*lphi->Nconfigs);
		  for (j=0; j<=(lphi->Nconfigs-1); j++)
		    *(lphi->probs+j)=1.0;
		  
		  msginds=malloc(sizeof(int));
		  for (j=1; j<=**(adjmsg+i); j++)
		    {
		      k=*(*(adjmsg+i)+j);
		      if (((msg+k)->to==(msg+i)->from)&&((msg+k)->from!=(msg+i)->to)) 
			{
			  Nmsgs++; msginds=realloc(msginds,sizeof(int)*Nmsgs);
			  *(msginds+Nmsgs-1)=k;
			}
		    }
	
		  ftype=(fgraph->facnodes+(msg+i)->from)->type;

		  
		  // Debug
		  deb=0;


		  // If function type is SINGLEPATHEXPLAIN, PATHSELOR or PREDICTION, then apply marginalizepotential5.  Otherwise apply marginalizepotential4.
		  //if (ftype==SINGLEVAR)
		  //if ((ftype==PATHSELOR)||(ftype==SINGLEVAR)||(ftype==PREDICTION))
		  //if ((ftype==PATHSELOR)||(ftype==SINGLEVAR))
		  if (ftype==SINGLEVAR)
		    marginalizepotential4((fgraph->facnodes+(msg+i)->from)->phivi,(msg+i)->to,Nmsgs,msginds,msg,lphi);		  
		  //else if (ftype==SINGLEPATHEXPLAIN)
		  //else if ((ftype==SINGLEPATHEXPLAIN)||(ftype==PREDICTION))
		  else if ((ftype==SINGLEPATHEXPLAIN)||(ftype==PREDICTION)||(ftype==PATHSELOR))
		    marginalizepotential5(fgraph,(msg+i)->from,(msg+i)->to,Nmsgs,msginds,msg,lphi,deb);

		  
                  free(msginds);		  

		}
	      
	      // If updatemode=0, then copy it to tmpmsg, otherwise copy it to msg.
	      if (updatemode==0)
		copypotential((tmpmsg+i)->phi,lphi);
	      else
		copypotential((msg+i)->phi,lphi);

	      free(lphi->varinds); free(lphi->Nstates); free(lphi->probs);
	      free(lphi); 
	    }	        	  	    
	}
    
      /*
      // Debug
      if ((n==Nsteps)&&(updatemode==0))
      //if ((n==0)&&(updatemode==0))
	{
	  for (i=0; i<=(fgraph->Ne*2-1); i++)
	    {
	      printf("msg[%d,%d](",(tmpmsg+i)->from,(tmpmsg+i)->to);
	      for (j=0; j<=((tmpmsg+i)->phi->Nvars-1); j++)
		printf("%d ",*((tmpmsg+i)->phi->varinds+j));
	      printf(")=(");
	      for (j=0; j<=((tmpmsg+i)->phi->Nconfigs-1); j++)
		printf("%f ",*((tmpmsg+i)->phi->probs+j));
	      printf(")\n");
	    }
	}
      */

      /*
      // Debug
      if ((n==1)&&(updatemode==1))
	{
	  for (i=0; i<=(fgraph->Ne*2-1); i++)
	    {
	      if ((((msg+i)->dir==0)&&(*(fixedmsg+i)==0)&&((fgraph->varnodes+(msg+i)->from)->type!=OUTPUTRESPONSE)&&(*(containy+(msg+i)->to)==1))||(((msg+i)->dir==1)&&((fgraph->varnodes+(msg+i)->to)->type==OUTPUTRESPONSE)))
	      //if ((((msg+i)->dir==0)&&(*(fixedmsg+i)==0)&&((fgraph->varnodes+(msg+i)->from)->type!=5)&&(*(containy+(msg+i)->to)==1))||(((msg+i)->dir==1)&&((fgraph->varnodes+(msg+i)->to)->type==5)))
		{
		  printf("msg[%d,%d](",(msg+i)->from,(msg+i)->to);
		  for (j=0; j<=((msg+i)->phi->Nvars-1); j++)
		    printf("%d ",*((msg+i)->phi->varinds+j));
		  printf(")=(");
		  for (j=0; j<=((msg+i)->phi->Nconfigs-1); j++)
		    printf("%f ",*((msg+i)->phi->probs+j));
		  printf(")\n");
		}
	    }
	}
      */
      
      
      // Terminate belief propagation if the messages are converged.
      // Only valid when updatemode=0.
      if (updatemode==0)
	{
	  flag=0; i=0;
	  while ((i<=(fgraph->Ne*2-1))&&(flag==0))
	    {
	      j=0;
	      while ((j<=(msg+i)->phi->Nconfigs-1)&&(flag==0))
		{
		  double val=*((msg+i)->phi->probs+j)-*((tmpmsg+i)->phi->probs+j);
		  
		  if (fabs(val)>1e-10)
		    flag=1;
		  
		  j++;
		}
	      i++;
	    }
	  if (flag==0)
	    terminate=1;
	}
	
      
      // Copy the tmpmsg to msg.  If there is one message which does not have repeated terms, then keep on updating.  Only needed when updatemode=0.
      if (updatemode==0)
	{
	  for (i=0; i<=(fgraph->Ne*2-1); i++)
	    copypotential((msg+i)->phi,(tmpmsg+i)->phi);
      	}

      n++;
      
    }

  /*
  // Update belief functions differently for output response nodes.
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      int flag;
      (belief+i)->Nvars=1;
      (belief+i)->varinds=realloc((belief+i)->varinds,sizeof(int));
      (belief+i)->Nstates=realloc((belief+i)->Nstates,sizeof(int));
      flag=0; j=0;
      while ((j<=(fgraph->Ne-1))&&(flag==0))
	{
	  if ((msg+j)->from==i)
	    {
	      *((belief+i)->varinds)=*((msg+j)->phi->varinds);
	      *((belief+i)->Nstates)=*((msg+j)->phi->Nstates);
	      (belief+i)->Nconfigs=(msg+j)->phi->Nconfigs;
	      (belief+i)->probs=realloc((belief+i)->probs,sizeof(double)*(belief+i)->Nconfigs);
	      for (k=0; k<=((belief+i)->Nconfigs-1); k++)
		*((belief+i)->probs+k)=1.0;
	      flag=1;
	    }
	  j++;
	}
    }     
  */
  

  isinitialized=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(isinitialized+i)=0;
  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      if (((msg+i)->dir==0)&&(*(isinitialized+(msg+i)->from)==0))
	{
	  *((belief+(msg+i)->from)->varinds)=*((msg+i)->phi->varinds);
	  *((belief+(msg+i)->from)->Nstates)=*((msg+i)->phi->Nstates);
	  (belief+(msg+i)->from)->Nconfigs=(msg+i)->phi->Nconfigs;
	  (belief+(msg+i)->from)->probs=realloc((belief+(msg+i)->from)->probs,sizeof(double)*(belief+(msg+i)->from)->Nconfigs);
	  for (j=0; j<=((belief+(msg+i)->from)->Nconfigs-1); j++)
	    *((belief+(msg+i)->from)->probs+j)=1.0;
	  *(isinitialized+(msg+i)->from)=1;
	}
    }
  free(isinitialized);



  // Update non-output response variables.
  // Multiply each message going to var node.
  for (i=0; i<=(fgraph->Ne*2-1); i++)
    if (((msg+i)->dir==1)&&((msg+i)->to<Nstarty))
      potentialproduct2(belief+(msg+i)->to,(msg+i)->phi);

  
  // Debug: the fixed variables have deterministic beliefs.
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      if (*(fixedvals+i)>=0)
	{
	  for (j=0; j<=((belief+i)->Nconfigs-1); j++)
	    *((belief+i)->probs+j)=0.0;
	  *((belief+i)->probs+*(fixedvals+i))=1.0;
	}
    }



  /*
  // Update non-output response variables.
  for (i=0; i<=(Nstarty-1); i++)
    {
      struct pair *node1, *node2;

      if (*(fixedvals+i)>=0)
	{
	  for (j=0; j<=((belief+i)->Nconfigs-1); j++)
	    *((belief+i)->probs+j)=0.0;
	  *((belief+i)->probs+*(fixedvals+i))=1.0;
	}
      else
	{	  
	  node1=malloc(sizeof(struct pair)); node2=malloc(sizeof(struct pair));
	  for (j=0; j<=(fgraph->Ne-1); j++)
	    {
	      if ((fgraph->e+j)->from==i)
		{
		  node1->i=(fgraph->e+j)->to; node1->j=1;
		  node2->i=i; node2->j=0;
		  potentialproduct2(belief+i,(msg+pair2msg(fgraph->Ne*2,msg,node1,node2))->phi);  
		}
	    }

	  free(node1); free(node2);
	}
    }
  */


  // Debug
  //printf("Nstarty=%d, Nvars=%d.\n",Nstarty,fgraph->Nvars);

  /*
  // Update output-response variables.
  // P(Y=1):P(Y=-1) is equal to P^{max}(Y=1):P^{max}(Y=-1). P^{max}(Y=1) and P^{max}(Y=-1) are the path prediction probabilities which are maximized at Y=1 and Y=-1.
  for (i=Nstarty; i<=(fgraph->Nvars-1); i++)
    {
      if (*(fixedvals+i)<0)
	{
 	  int Nexppaths=0, ind1, ind2;
	  double Z, *pys, *py, p1, p2, p01, p02;
	  struct pair *node1, *node2;

	  pys=malloc(sizeof(double)*3);
	  node1=malloc(sizeof(struct pair)); node2=malloc(sizeof(struct pair));
	  for (j=0; j<=(fgraph->Ne-1); j++)
	    {
	      if ((fgraph->e+j)->from==i)
		{
		  node1->i=(fgraph->e+j)->to; node1->j=1;
		  node2->i=i; node2->j=0;
		  Nexppaths++; Z=0;

		  pys=realloc(pys,sizeof(double)*(Nexppaths*3));
		  for (k=0; k<=2; k++)
		    {
		      *(pys+(Nexppaths-1)*3+k)=*((msg+pair2msg(fgraph->Ne*2,msg,node1,node2))->phi->probs+k);
		      Z+=*(pys+(Nexppaths-1)*3+k);
		    }
		  for (k=0; k<=2; k++)
		    *(pys+(Nexppaths-1)*3+k)/=Z; 
		}
	    }
	  free(node1); free(node2);

	  p1=0; p2=0; ind1=-1; ind2=-1;
	  for (j=0; j<=(Nexppaths-1); j++)
	    {
	      if (*(pys+j*3+1)>p1)
		{
		  p1=*(pys+j*3+1); ind1=j;
		}
	      if (*(pys+j*3+2)>p2)
		{
		  p2=*(pys+j*3+2); ind2=j;
		}
	    }
	  p01=*(pys+ind1*3); p02=*(pys+ind2*3);
	  py=malloc(sizeof(double)*3);
	  *py=(p01+p02)/2;
	  *(py+1)=(1-*py)*p1/(p1+p2);
	  *(py+2)=(1-*py)*p2/(p1+p2);

	  for (j=0; j<=2; j++)
	    *((belief+i)->probs+j)=*(py+j);

	  
	  // Debug
	  //printf("(%d,%f,%f,%f), ",(fgraph->varnodes+i)->geneind,*((belief+i)->probs+0),*((belief+i)->probs+1),*((belief+i)->probs+2));
	  //for (j=0; j<=(Nexppaths-1); j++)
	  //printf("(%f,%f,%f)",*(pys+j*3+0),*(pys+j*3+1),*(pys+j*3+2));
	  //printf("\n");
	  
	  

	  free(pys); free(py);
	}
    }
  
  // Debug
  //printf("\n");
  */

  
  // Update output-response variables.
  // When there are multiple paths, check if there predictions are consistent (either 1s or 0s or -1s or 0s).  If yes then output the corresponding prediction.  Otherwise output 0.
  // P(Y=+1)=P(Y1=0,+1)\cdotsP(Ym=0,+1)-P(Y1=0)\cdotsP(Ym=0).
  // P(Y=-1)=P(Y1=0,-1)\cdotsP(Ym=0,-1)-P(Y1=0)\cdotsP(Ym=0).
  // P(Y=0)=1-P(Y=+1)-P(Y=-1).
  for (i=Nstarty; i<=(fgraph->Nvars-1); i++)
    {
      if (*(fixedvals+i)<0)
	{
	  int Nexppaths=0, ind1, ind2;
	  double Z, *pys, *py, p0, p1, p2, p01, p02, val;
	  struct pair *node1, *node2;

	  pys=malloc(sizeof(double)*3);
	  node1=malloc(sizeof(struct pair)); node2=malloc(sizeof(struct pair));
	  for (j=0; j<=(fgraph->Ne-1); j++)
	    {
	      if ((fgraph->e+j)->from==i)
		{
		  node1->i=(fgraph->e+j)->to; node1->j=1;
		  node2->i=i; node2->j=0;
		  Nexppaths++; Z=0;

		  pys=realloc(pys,sizeof(double)*(Nexppaths*3));
		  for (k=0; k<=2; k++)
		    {
		      *(pys+(Nexppaths-1)*3+k)=*((msg+pair2msg(fgraph->Ne*2,msg,node1,node2))->phi->probs+k);
		      Z+=*(pys+(Nexppaths-1)*3+k);
		    }
		  for (k=0; k<=2; k++)
		    *(pys+(Nexppaths-1)*3+k)/=Z; 
		}
	    }
	  free(node1); free(node2);

	  p1=1;
	  for (j=0; j<=(Nexppaths-1); j++)
	    {
	      val=*(pys+j*3+0)+*(pys+j*3+1);
	      p1*=val;
	    }
	  val=1;
	  for (j=0; j<=(Nexppaths-1); j++)
	    val*=*(pys+j*3+0);
	  p1-=val;

	  p2=1;
	  for (j=0; j<=(Nexppaths-1); j++)
	    {
	      val=*(pys+j*3+0)+*(pys+j*3+2);
	      p2*=val;
	    }
	  val=1;
	  for (j=0; j<=(Nexppaths-1); j++)
	    val*=*(pys+j*3+0);
	  p2-=val;

	  p0=1-p1-p2;

	  *((belief+i)->probs+0)=p0;
	  *((belief+i)->probs+1)=p1;
	  *((belief+i)->probs+2)=p2;

	  //
	  // Debug
	  //printf("(%d,%f,%f,%f), ",(fgraph->varnodes+i)->geneind,p0,p1,p2);
	  //for (j=0; j<=(Nexppaths-1); j++)
	  //printf("(%f,%f,%f)",*(pys+j*3+0),*(pys+j*3+1),*(pys+j*3+2));
	  //printf("\n");
	  //

	  
	  free(pys);
	}
    }
  
  // Debug
  //printf("\n");
  
 
  if (updatemode==0)
    {
      for (i=0; i<=(fgraph->Ne*2-1); i++)
	{
	  free((tmpmsg+i)->phi->varinds);      
	  free((tmpmsg+i)->phi->Nstates);      
	  free((tmpmsg+i)->phi->probs);
	  free((tmpmsg+i)->phi);
	  
	}     
      free(tmpmsg);
    }
  else if (updatemode==1)
    free(containy);

  for (i=0; i<=(fgraph->Ne*2-1); i++)
    free(*(adjmsg+i));
  free(adjmsg);
  free(fixedmsg);
  
}


  
// Marginalize the product of a potential function and multiple messages.
// Difference from marginalizepotential3: the output potential function has one variable, and the input messages are also marginalized.
void marginalizepotential4(struct potential *phi, int varind, int Nmsgs, int *msginds, struct message *msg, struct potential *newphi)
{
  int i, j, k, *phi2msg, *msg2phi, *config, flag, ind;
  double val;

  
  // If Nmsgs=0 then newphi=phi.
  if (Nmsgs<=0)
    {
      copypotential(newphi,phi);
      return;
    }

  // Find the index of the potential function variable.
  ind=-1; i=0;
  while ((i<=(phi->Nvars-1))&&(ind==-1))
    {
      if (*(phi->varinds+i)==varind)
	ind=i;
      i++;
    }

  
  // Build the lookup tables from potential variables to messages and vice versa.
  phi2msg=malloc(sizeof(int)*phi->Nvars);
  msg2phi=malloc(sizeof(int)*Nmsgs);
  for (i=0; i<=(phi->Nvars-1); i++)
    *(phi2msg+i)=-1;

  
  for (i=0; i<=(Nmsgs-1); i++)
    {
      j=inlistind(phi->Nvars,phi->varinds,*((msg+*(msginds+i))->phi->varinds));
      *(phi2msg+j)=i; *(msg2phi+i)=j;
    }
  
  
  config=malloc(sizeof(int)*phi->Nconfigs);
  for (i=0; i<=(phi->Nconfigs-1); i++)
    *(config+i)=0;

  
  newphi->Nvars=1; newphi->varinds=realloc(newphi->varinds,sizeof(int));
  *(newphi->varinds)=*(phi->varinds+ind);
  newphi->Nstates=realloc(newphi->Nstates,sizeof(int));
  *newphi->Nstates=*(phi->Nstates+ind);
  newphi->Nconfigs=*newphi->Nstates;
  newphi->probs=realloc(newphi->probs,sizeof(double)*newphi->Nconfigs);
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    *(newphi->probs+i)=0.0;
  
  // Sum over the values of all configurations.
  flag=0; k=0;
  while (flag==0)
    {
      int carry=0;
      
      // Add the product of potential and messages of a particular configuration.
      val=*(phi->probs+k);

      
      for (i=0; i<=(Nmsgs-1); i++)
	{
	  j=*(config+*(msg2phi+i));
	  val*=*((msg+*(msginds+i))->phi->probs+j);
	}
      
      *(newphi->probs+*(config+ind))+=val;
      

      // Increment the configuration.
      i=phi->Nvars-1;
      *(config+i)+=1;
      if (*(config+i)>=*(phi->Nstates+i))
	{
	  carry=1; *(config+i)=0;
	}
      i--;
      while ((carry==1)&&(i>=0))
	{
	  *(config+i)+=carry;
	  if (*(config+i)<*(phi->Nstates+i))
	    carry=0;
	  else
	    *(config+i)=0;
	  i--;
	}
      k++;
      flag=1; i=0;
      while ((i<=(phi->Nvars-1))&&(flag==1))
	{
	  if (*(config+i)>0)
	    flag=0;
	  i++;
	}
    }

  // Normalize the prob values.
  val=0.0;
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    val+=*(newphi->probs+i);
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    *(newphi->probs+i)/=val;
    
  free(phi2msg); free(msg2phi); free(config);

}
