// Utility functions regarding the operations of potential functions defined by both functional forms and lookup tables.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include "integration.h"
# include "integration3.h"
# include "connect.h"
# include "cluster.h"
# include "integration4.h"
# include "integration5.h"
# define MAXLINEWIDTH 1000
# define INFINITY 1e38
# define POSINF 100000
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)

extern void maxsumproduct1(struct mix_facgraph *fgraph, struct message *msg, int *fixedvals, struct potential *belief, int mode);
extern int * recurse_getmapconfigs(struct confignode *configtree, int *Nmapconfigs, int *mapconfigs);
extern void recurse_freeconfignode(struct confignode *configtree);
void recurse_fixval2(struct confignode *configtree, struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int *vartypes);
extern int inlistind(int Nv, int *v, int q);
extern void copypotential(struct potential *phi2, struct potential *phi1);
extern struct cpath * decompresscpath(struct csparsegraph *cgraph, struct compresscpath *path, int ind);
extern long config2number2(int Nvars, int *Nstates, int *configuration);
extern void number2config2(int Nvars, int *Nstates, long val, int *configuration);
extern void number2config(int Nvars, int val, int *configuration);
extern void potentialproduct2(struct potential *phi1, struct potential *phi2);
extern void potentialproduct3(struct potential *phi1, struct potential *phi2);
extern struct potential * createpotential(struct potential *phi1);
extern void maximizepotential2(struct potential *phi, int ind);
extern void maxmarginal3(struct mix_path_fac_node *facnode, int Nmsgs, int *msginds, struct message *msg, struct potential *phi, int deb);
extern void marginalizepotential4(struct potential *phi, int varind, int Nmsgs, int *msginds, struct message *msg, struct potential *newphi);
extern int pair2msg(int Nmsgs, struct message *msg, struct pair *node1, struct pair *node2);


void maxmarginal4(struct mix_facgraph *fgraph, int factorind, int varind, int Nmsgs, int *msginds, struct message *msg, struct potential *phi, int deb);


extern int maxpathlength, submaxpathlength;
extern double epsilon, ptrue_positive;
extern double v1, v2, v3, koinv;



// Compute the max marginal of a mixed_path factor and all incoming messages to this factor except the variable which appears in the max marginal message.
// m(x1) = \max_{U \ x1} \sum_{\sigma} f(x1,\cdots,xn,\sigma) m(\sigma) m(x2) \cdots m(xn).
// Difference from maxmarginal1: simplifies the code and does not use the lookup table.
// Difference from maxmarginal2: perform max marginalization for all variables.
// Difference from maxmarginal3 and maxmarginal2: also applies to PREDICTION and PATHSELOR function types.
void maxmarginal4(struct mix_facgraph *fgraph, int factorind, int varind, int Nmsgs, int *msginds, struct message *msg, struct potential *phi, int deb)
{
  int i, j, k, l, n, N, *ismarginalized, type, ind, flag;
  int *presence2phi, *sign2phi, sigma2phi, ko2phi;
  int *phi2msg, *msg2phi, *maskval, *sigdir2phi, targetphiind, targetedgeind;
  int *defconfig, *config, *signconfig, maxind, *degcnt;
  double maxval, val, val1, val2, val3, optvals[3][3], defval;
  struct mix_path_fac_node *facnode;


  facnode=(fgraph->facnodes+factorind);

  // In case Nmsgs=0, return the potential function term.
  if (Nmsgs<=0)
    {
      val=0;
      phi->Nvars=1;
      phi->varinds=realloc(phi->varinds,sizeof(int));
      *(phi->varinds)=*(facnode->phivi->varinds);

      if (facnode->type==PATHSELOR)
	{
	  *(phi->probs+0)=v3; *(phi->probs+1)=1;
	}
      else
	copypotential(phi,facnode->phivi);

      return;
    }
  

  // Build lookup tables from presence, sign, selection and ko variables to vars in potential.
  if (facnode->Nedges>0)
    {
      presence2phi=malloc(sizeof(int)*facnode->Nedges);
      for (i=0; i<=(facnode->Nedges-1); i++)
	*(presence2phi+i)=inlistind(facnode->phivi->Nvars,facnode->phivi->varinds,*(facnode->presences+i));
      sign2phi=malloc(sizeof(int)*facnode->Nedges);
      for (i=0; i<=(facnode->Nedges-1); i++)
	*(sign2phi+i)=inlistind(facnode->phivi->Nvars,facnode->phivi->varinds,*(facnode->signs+i));
    }
  if (facnode->type==SINGLEPATHEXPLAIN)
    {
      sigma2phi=inlistind(facnode->phivi->Nvars,facnode->phivi->varinds,facnode->sigma);
      ko2phi=inlistind(facnode->phivi->Nvars,facnode->phivi->varinds,facnode->ko);
    }
  
  // Build a lookup table from signal edge dir to vars in the potential.
  if (facnode->Nsigedges>0)
    {
      sigdir2phi=malloc(sizeof(int)*facnode->Nsigedges);
      for (i=0; i<=(facnode->Nsigedges-1); i++)
	*(sigdir2phi+i)=inlistind(facnode->phivi->Nvars,facnode->phivi->varinds,*(facnode->sigdirvars+i));
    }
  

  // Build a lookup table from potential function vars to msg vars.
  phi2msg=malloc(sizeof(int)*facnode->phivi->Nvars);
  for (i=0; i<=(facnode->phivi->Nvars-1); i++)
    *(phi2msg+i)=-1;
  msg2phi=malloc(sizeof(int)*Nmsgs);
  for (i=0; i<=(Nmsgs-1); i++)
    *(msg2phi+i)=-1;
  for (i=0; i<=(Nmsgs-1); i++)
    {
      j=inlistind(facnode->phivi->Nvars,facnode->phivi->varinds,*((msg+*(msginds+i))->phi->varinds));
      if (j>=0)
	{
	  *(phi2msg+j)=i; *(msg2phi+i)=j;
	}
    }

  // Construct the mask value table.  If maskval>=0 then the value of the variable is fixed.
  maskval=malloc(sizeof(int)*facnode->phivi->Nvars);
  for (i=0; i<=(facnode->phivi->Nvars-1); i++)
    *(maskval+i)=-1;  
  
  for (i=0; i<=(Nmsgs-1); i++)
    {
      int cnt1=0, cnt2=0, ind=-1;
      for (j=0; j<=((msg+*(msginds+i))->phi->Nconfigs-1); j++)
	{
	  if (*((msg+*(msginds+i))->phi->probs+j)==1.0)
	    {  cnt1++; ind=j; }
	  else if (*((msg+*(msginds+i))->phi->probs+j)==0.0)
	    cnt2++;
	}
      if ((cnt1==1)&&(cnt2==((msg+*(msginds+i))->phi->Nconfigs-1)))
	*(maskval+*(msg2phi+i))=ind;
    }
  
  
  type=(fgraph->varnodes+varind)->type;

  targetphiind=inlistind(facnode->phivi->Nvars,facnode->phivi->varinds,varind);
  targetedgeind=-1;
  if (type==0)
    targetedgeind=inlistind(facnode->Nedges,facnode->presences,varind);
  else if (type==1)
    targetedgeind=inlistind(facnode->Nedges,facnode->signs,varind);
  else if (type==2)
    targetedgeind=inlistind(facnode->Nedges,facnode->sigdirvars,varind);

  phi->Nvars=1;
  phi->varinds=realloc(phi->varinds,sizeof(int));
  *(phi->varinds)=varind;

  if ((type==3)||(type==5))
    phi->Nconfigs=3;
  else
    phi->Nconfigs=2;

  phi->Nstates=realloc(phi->Nstates,sizeof(int));
  *(phi->Nstates)=phi->Nconfigs;
  phi->probs=realloc(phi->probs,sizeof(double)*phi->Nconfigs);
  for (i=0; i<=(phi->Nconfigs-1); i++)
    *(phi->probs+i)=1.0;


 
  // Identify the best value of each variable according to its message.
  // Check if certain variables are fixed.
  // Also check whether some variables yield the same probability for diff values.
  defconfig=malloc(sizeof(int)*facnode->phivi->Nvars);
  for (i=0; i<=(facnode->phivi->Nvars-1); i++)
    *(defconfig+i)=-1;
  degcnt=malloc(sizeof(int)*facnode->phivi->Nvars);
  for (i=0; i<=(facnode->phivi->Nvars-1); i++)
    *(degcnt+i)=0;

  
  for (i=0; i<=(Nmsgs-1); i++)
    {
      if (*(maskval+*(msg2phi+i))>=0)
	maxind=*(maskval+*(msg2phi+i));
      else
	{
	  maxind=-1; maxval=0;
	  for (j=0; j<=((msg+*(msginds+i))->phi->Nconfigs-1); j++)
	    {
	      if (*((msg+*(msginds+i))->phi->probs+j)>maxval)
		{
		  maxind=j; maxval=*((msg+*(msginds+i))->phi->probs+j);
		}
	    }
	}
      *(defconfig+*(msg2phi+i))=maxind;
    }
 

  for (i=0; i<=(Nmsgs-1); i++)
    {
      if (*(maskval+*(msg2phi+i))>=0)
	*(degcnt+*(msg2phi+i))=1;
      else 
	for (j=0; j<=((msg+*(msginds+i))->phi->Nconfigs-1); j++)
	  if (*((msg+*(msginds+i))->phi->probs+j)==*((msg+*(msginds+i))->phi->probs+*(defconfig+*(msg2phi+i))))
	    *(degcnt+*(msg2phi+i))+=1;
    }


  // Maximize a single path explanation function.
  if (facnode->type==SINGLEPATHEXPLAIN)
    {  
      config=malloc(sizeof(int)*facnode->phivi->Nvars);
      signconfig=malloc(sizeof(int)*facnode->Nedges);
      N=1;
      for (i=0; i<=(facnode->Nedges-1); i++)
	N*=2;

      // If there are multiple configurations which yield the same max message product, choose the one which is consistent with the ko data.
      
      for (i=0; i<=(facnode->Nedges-1); i++)
	if (*(degcnt+*(presence2phi+i))==2)
	  *(defconfig+*(presence2phi+i))=1;
      if (type==0)
	*(defconfig+targetphiind)=1;
      if (*(degcnt+sigma2phi)==2)
	*(defconfig+sigma2phi)=1;
      if ((*(defconfig+ko2phi)==1)||(*(defconfig+ko2phi)==2))
	{
	  int maxko=0, flag2=1;
	  if (*(defconfig+ko2phi)==1)
	    maxko=1;
	  else if (*(defconfig+ko2phi)==2)
	    maxko=-1;
	  for (i=0; i<=(facnode->Nedges-1); i++)
	    *(signconfig+i)=0;
	  i=0;
	  while ((i<=(N-1))&&(flag2==1))
	    {
	      int carry;
	      
	      flag=1; j=0;
	      while ((j<=(facnode->Nedges-1))&&(flag==1))
		{
		  if ((*(signconfig+j)!=*(defconfig+*(sign2phi+j)))&&(*(degcnt+*(sign2phi+j))!=2))
		    flag=0;
		  j++;
		}
	      j=1;
	      for (k=0; k<=(facnode->Nedges-1); k++)
		j*=(*(signconfig+k)*2-1);
	      j*=(-1);
	      if ((j==maxko)&&(flag==1))
		{
		  for (k=0; k<=(facnode->Nedges-1); k++)
		    *(defconfig+*(sign2phi+k))=*(signconfig+k);
		  flag2=0;
		}
	      carry=1; j=facnode->Nedges-1; i++;
	      while ((carry==1)&&(j>=0))
		{
		  *(signconfig+j)+=carry;
		  if (*(signconfig+j)<2)
		    carry=0;
		  else
		    *(signconfig+j)=0;
		  j--;
		}
	    }
	}  

      // Debug
      if (deb==1)
	{
	  printf("vars=(");
	  for (i=0; i<=(facnode->phivi->Nvars-1); i++)
	    printf("%d ",*(facnode->phivi->varinds+i));
	  printf("), defconfig=(");
	  for (i=0; i<=(facnode->phivi->Nvars-1); i++)
	    printf("%d ",*(defconfig+i));
	  printf(")\n");
	}

      // optvals[i][j]: the optimal value when variable has value i and scenario j.
      for (i=0; i<=2; i++)
	for (j=0; j<=2; j++)
	  optvals[i][j]=0;


      // scenario 0: k=0, other variables (except the target var) are determined by messages.
      defval=v2;
      for (i=0; i<=(Nmsgs-1); i++)
	{
	  //if (*(msg2phi+i)!=sigma2phi)
	  {
	    val=*((msg+*(msginds+i))->phi->probs+*(defconfig+*(msg2phi+i)));
	    defval*=val;
	  }
	}
  
      // If defconfig[ko]=0 or target variable is ko, then opt value is defval.
      // Otherwise the opt value is adjusted by the message value of k=0.
      val1=defval;
      if ((*(defconfig+ko2phi)!=0)&&(type!=3))
	{
	  val1=*((msg+*(msginds+*(phi2msg+ko2phi)))->phi->probs+0);
	  val=*((msg+*(msginds+*(phi2msg+ko2phi)))->phi->probs+*(defconfig+ko2phi));
	  val1=defval*val1/val;
	}
  
      // scenario 0 when type=0, 1, 2, 4, and when maskval[ko] is not fixed to -1 or +1.
      if ((type!=3)&&((*(maskval+ko2phi)!=1)&&(*(maskval+ko2phi)!=2)))
	for (i=0; i<=1; i++)
	  optvals[i][0]=val1;
  
      // scenario 1 of sigma=0: other variables are determined by messages, including ko.
      //if (type==2)
      if (type==4)
	optvals[0][1]=defval;

      // scenario 1 of ko=0: other variables are determined by messages.
      if (type==3)
	optvals[0][1]=defval;

      // scenario 1 of x=0, x=1, s=+1, s=-1, ko!=0, sigma=1, edge dir=preferred value, edge dir!=preferred value: k!=0, the path cannot explain the ko effect.
      // simply check if the default config satisfies the explanation constraint.
      // if yes, then optval=0.
      if ((*(defconfig+ko2phi)!=0)||(type==3))
	{
	  int maxko=0;

	  if (type!=3)
	    n=2;
	  else
	    n=3;
	  
	  for (i=0; i<=(n-1); i++)
	    {
	      for (j=0; j<=(facnode->phivi->Nvars-1); j++)
		*(config+j)=*(defconfig+j);
	      // Set the target value.
	      *(config+targetphiind)=i;
	      j=0; flag=1;
	      while ((j<=(facnode->Nedges-1))&&(flag==1))
		{
		  if (*(config+*(presence2phi+j))==0)
		    flag=0;
		  j++;
		}
	      
	      j=0;
	      while ((j<=(facnode->Nsigedges-1))&&(flag==1))
		{
		  if (*(config+*(sigdir2phi+j))!=*(facnode->sigdirvarvals+j))
		    flag=0;
		  j++;
		}

	      // Debug
	      if (deb==1)
		{
		  printf("i=%d, config sigdirvars=(",i);
		  for (j=0; j<=(facnode->Nsigedges-1); j++)
		    printf("%d ",*(config+*(sigdir2phi+j)));
		  printf("), preferred sigdirvars=(");
		  for (j=0; j<=(facnode->Nsigedges-1); j++)
		    printf("%d ",*(facnode->sigdirvarvals+j));
		  printf(")\n");
		}

	  
	      if (*(config+ko2phi)==0)
		maxko=0;
	      else if (*(config+ko2phi)==1)
		maxko=1;
	      else if (*(config+ko2phi)==2)
		maxko=-1;	
	      j=1;
	      for (k=0; k<=(facnode->Nedges-1); k++)
		j*=(*(config+*(sign2phi+k))*2-1);
	      j*=(-1);
	      if (j!=maxko)
		flag=0;

	      // scenario 1 of sigma=0 or ko=0 has been evaluated.
	      if ((type==0)||(type==1)||((type==4)&&(i!=0))||((type==3)&&(i!=0))||(type==2))
	      //if ((type==0)||(type==1)||((type==2)&&(i!=0))||((type==3)&&(i!=0))||(type==4))
		//if ((type==0)||(type==1)||((type==2)&&(i!=0))||((type==3)&&(i!=0)))
		//if ((~((type==2)&&(i==0)))&&(~((type==3)&&(i==0))))
		{
		  // The best config according to messages is consistent with ko, scenario 1 is always smaller than scenario 2. 
		  // The logic potential term of scenario 2 is v1 if the default config explains ko.
		  if (flag==1)
		    {
		      optvals[i][1]=0; val1=v1;
		      

		      //if (type!=2)
		      //val1=*((msg+*(msginds+*(phi2msg+sigma2phi)))->phi->probs+0)*v2+*((msg+*(msginds+*(phi2msg+sigma2phi)))->phi->probs+1)*v1;		    
		      //else
		      //val1=v1;
		      val1=val1*defval/v2;
		      optvals[i][2]=val1;


		      // Debug
		      if (deb==1)
			{
			  printf("optvals[%d][2]=%f, config=(",i,optvals[i][2]);
			  for (k=0; k<=(facnode->phivi->Nvars-1); k++)
			    printf("%d ",*(config+k));
			  printf(")\n");
			}
		    }
		  // The best config according to messages is not consistent with ko.  
		  // The potential value is v2, same as scenario 0.
		  // If default sigma=1, then the opt val is adjusted by setting it to 0.
		  
		  // There are two possible conditions in this case.
		  // 1. If config sigmal=0, then optval does not need to be adjusted.  Logic potential term is v2.
		  // 2. If config sigma=1, then optmal is either v3 with sigma=1, or v2 with sigma=0.  The greatest of the two.  If sigma is the target variable, then there is only one choice: v3 with sigma=1.
		  else
		    {
		      if (*(config+sigma2phi)==0)
			val1=defval;
		      else
			{
			  val1=v3*defval/v2;
			  //if (type!=2)
			  if (type!=4)
			    {
			      val2=defval;
			      val2*=*((msg+*(msginds+*(phi2msg+sigma2phi)))->phi->probs+0);
			      val2/=*((msg+*(msginds+*(phi2msg+sigma2phi)))->phi->probs+1);
			    }
			  else
			    val2=0;
			  val1=max(val1,val2);
			  if (val1==val2)
			    *(config+sigma2phi)=0;
			}

		  
		      /*
			if ((type!=2)&&(*(defconfig+sigma2phi)==1))
			{
			val1=*((msg+*(msginds+*(phi2msg+sigma2phi)))->phi->probs+0);
			val1/=*((msg+*(msginds+*(phi2msg+sigma2phi)))->phi->probs+1);
			}
			else
			val1=1;
			val1=val1*defval;
		      */


		      optvals[i][1]=val1;
										 
		  
		      // Debug
		      if (deb==1)
			{
			  printf("optvals[%d][1]=%f, flag=%d, config=(",i,optvals[i][1],flag);
			  for (k=0; k<=(facnode->phivi->Nvars-1); k++)
			    printf("%d ",*(config+k));
			  printf(")\n");
			}
		      
		      
		      //if (type!=2)
		      //val1=*((msg+*(msginds+*(phi2msg+sigma2phi)))->phi->probs+0)*v2+*((msg+*(msginds+*(phi2msg+sigma2phi)))->phi->probs+1)*v3;
		      //else
		      //val1=v3;
		      //val1=val1*defval/v2;
		      //optvals[i][1]=val1;
		    }
		}
	    }
	}
 
      // scenario 2 of x=1, s=+1, s=-1, ko!=0, sigma=1, edge dir=preferred value: ko!=0, the path explains the ko effect.
      //if (((*(defconfig+ko2phi)!=0)||(type==3))&&((*(maskval+sigma2phi)!=0)||(type==2)))
      if (((*(defconfig+ko2phi)!=0)||(type==3))&&((*(maskval+sigma2phi)!=0)||(type==4)))
	{
	  int maxko;
	  
	  if (type!=3)
	    n=2;
	  else
	    n=3;
	  
	  // Find the max configuration.
	  for (i=0; i<=(n-1); i++)
	    {
	      // If x=0 or ko=0 or sigma=0 or edge dir!=preferred val, then scenario 2 is invalid.
	      // Skip if the value of scenario 2 is already evaluated.
	      //if ((~((type==0)&&(i==0)))&&(~((type==2)&&(i==0)))&&(~((type==3)&&(i==0)))&&(optvals[i][2]<=0))
	      //if ((((type==0)&&(i!=0))||(type==1)||((type==2)&&(i!=0))||((type==3)&&(i!=0)))&&(optvals[i][2]<=0))
	      //if ((((type==0)&&(i!=0))||(type==1)||((type==2)&&(i!=0))||((type==3)&&(i!=0))||((facnode->Nsigedges>0)&&(type==4)&&(i==*(facnode->sigdirvarvals+targetedgeind))))&&(optvals[i][2]<=0))
	      if ((((type==0)&&(i!=0))||(type==1)||((type==4)&&(i!=0))||((type==3)&&(i!=0))||((facnode->Nsigedges>0)&&(type==2)&&(i==*(facnode->sigdirvarvals+targetedgeind))))&&(optvals[i][2]<=0)) 
		{
		  // Set the target variable.
		  *(config+targetphiind)=i;
		  // If ko is not the target variable, set its value to optimal ko.
		  if (type!=3)
		    *(config+ko2phi)=*(defconfig+ko2phi);
		  // Set the path selection variable.
		  *(config+sigma2phi)=1;
		  
		  // Set all edge presence vars to be 1.
		  flag=1; j=0;
		  while ((j<=(facnode->Nedges-1))&&(flag==1))
		    {
		      if (*(maskval+*(presence2phi+j))!=0)
			*(config+*(presence2phi+j))=1;
		      else
			flag=0;
		      j++;
		    }
		  
		  // Set all edge dir vars to preferred values.
		  j=0;
		  while ((j<=(facnode->Nsigedges-1))&&(flag==1))
		    {
		      if ((*(maskval+*(sigdir2phi+j))==-1)||(*(maskval+*(sigdir2phi+j))==*(facnode->sigdirvarvals+j)))
			*(config+*(sigdir2phi+j))=*(facnode->sigdirvarvals+j);
		      else
			flag=0;
		      j++;
		    }
		  
		  
		  // If type=0 and i=0, then scenario 2 is invalid.
		  if ((type==0)&&(i==0))
		    flag=0;

		  // If type==2 and edge dir value != preferred value, then scenario 2 is invalid.
		  //if ((facnode->Nsigedges>0)&&(type==4)&&(i!=*(facnode->sigdirvarvals+targetedgeind)))
		  if ((facnode->Nsigedges>0)&&(type==2)&&(i!=*(facnode->sigdirvarvals+targetedgeind)))
		    flag=0;
		  
		  // Debug
		  //if (deb==1)
		  //printf("varind=%d, ind=%d, flag=%d, i=%d, preferred val=%d \n",varind,ind,flag,i,*(facnode->sigdirvarvals+ind));


		  if (flag==1)
		    {
		      int maxko=0;
		      if (*(config+ko2phi)==0)
			maxko=0;
		      else if (*(config+ko2phi)==1)
			maxko=1;
		      else if (*(config+ko2phi)==2)
			maxko=-1;

		      // Identify consistent sign configurations.
		      for (j=0; j<=(facnode->Nedges-1); j++)
			*(signconfig+j)=0;
		      maxind=-1; maxval=0; j=0;
		      
		      while (j<=(N-1))
			{
			  int flag2, carry=0;
			  // Filter out the sign configurations which are inconsistent with fixed values and ko effect.
			  flag2=1; k=0;
			  while (k<=(facnode->Nedges-1))
			    {
			      if ((*(maskval+*(sign2phi+k))>=0)&&(*(maskval+*(sign2phi+k))!=*(signconfig+k)))
				flag2=0;
			      k++;
			    }
			  // If type=1 and the sign config value is not identical to the set value, then invalid.
			  if ((type==1)&&(*(signconfig+targetedgeind)!=i))
			    flag2=0;

			  k=1;
			  for (l=0; l<=(facnode->Nedges-1); l++)
			    k*=(*(signconfig+l)*2-1);
			  k*=(-1);

			  if ((k==maxko)&&(flag2==1))
			    {
			      
			      // Notice if the target variable is a sign var, then it does not have message.
			      val=1;
			      for (l=0; l<=(facnode->Nedges-1); l++)
				if (*(phi2msg+*(sign2phi+l))>=0)
				  val*=*((msg+*(msginds+*(phi2msg+*(sign2phi+l))))->phi->probs+*(signconfig+l));
			      if (val>maxval)
				{
				  maxind=j; maxval=val;
				}			  
			    }

			  // Increment the sign configuration by one.
			  carry=1; k=facnode->Nedges-1; j++;
			  while ((carry==1)&&(k>=0))
			    {
			      *(signconfig+k)+=carry;
			      if (*(signconfig+k)<2)
				carry=0;
			      else
				*(signconfig+k)=0;
			      k--;
			    }
			}

		      // maxind>=0: there exists a sign configuration which is consistent with ko.
		      if (maxind>=0)
			{
			  number2config(facnode->Nedges,maxind,signconfig);

			  for (j=0; j<=(facnode->Nedges-1); j++)
			    *(config+*(sign2phi+j))=*(signconfig+j);
			  val1=1;
			  for (j=0; j<=(Nmsgs-1); j++)
			    //if (*(msg2phi+j)!=sigma2phi)
			    val1*=*((msg+*(msginds+j))->phi->probs+*(config+*(msg2phi+j)));
			  
			  // The logic potential function term is v1.
			  val1*=v1;

		      
			  // If type!=2, then the potential function avg value is m(sigma=0)v2+m(sigma=1)v1.
			  // Otherwise it is v1.
			  //if (type==2)
			  //val1*=v1;
			  //else
			  //{
			  //  val=*((msg+*(msginds+*(phi2msg+sigma2phi)))->phi->probs+0)*v2+*((msg+*(msginds+*(phi2msg+sigma2phi)))->phi->probs+1)*v1;
			  //  val1*=val;
			  //}

			  optvals[i][2]=val1;
			  

			  // Debug
			  if (deb==1)
			    {
			      printf("optvals[%d][2]=%f, config=(",i,optvals[i][2]);
			      for (k=0; k<=(facnode->phivi->Nvars-1); k++)
				printf("%d ",*(config+k));
			      printf(")\n");
			    }

			}
		    }
		}
	    }
	}


      // Compare optval and obtain the probability values.
      if (type!=3)
	n=2;
      else
	n=3;
      val1=0;
      for (i=0; i<=(n-1); i++)
	{
	  val=max(optvals[i][0],optvals[i][1]);
	  val=max(val,optvals[i][2]);
	  *(phi->probs+i)=val;
	  val1+=val;
	}

      // Normalize the probability table.
      for (i=0; i<=(n-1); i++)
	*(phi->probs+i)/=val1;

      // Debug
      if (deb==1)
	{
	  for (i=0; i<=1; i++)
	    for (j=0; j<=2; j++)
	      printf("optvals[%d][%d]=%f ",i,j,optvals[i][j]);
	  printf("\n");
	}

      free(config); free(signconfig); 	

    }

  
  // The factor type is PATHSELOR.
  else if (facnode->type==PATHSELOR)
    {
      // Check if defconfig satisfies the OR constraint, i.e., there is at least one sigma variable (!=varind) whose argmax value is 1.
      for (i=0; i<=(facnode->phivi->Nvars-1); i++)
	if ((*(defconfig+i)==0)&&(*(degcnt+i)==2)&&(i!=ind))
	  *(defconfig+i)=1;
      flag=0; i=0;
      while ((i<=(facnode->phivi->Nvars-1))&&(flag==0))
	{
	  if ((i!=ind)&&(*(defconfig+i)==1))
	    flag=1;
	  i++;
	}
      defval=1.0;
      for (i=0; i<=(Nmsgs-1); i++)
	defval*=*((msg+*(msginds+i))->phi->probs+*(defconfig+*(msg2phi+i)));
      // If the default configuration satisifies the constraint, then returns the message product.
      if (flag==1)
	{
	  *(phi->probs+0)=defval; *(phi->probs+1)=defval;
	}
      // If the default configuration does not satisfy the constraint, then phi[1] is the same as previous and phi[0] is obtained by the max of (1)replacing the value of a variable with 1, (2)default configuration*v3.
      else
	{
	  *(phi->probs+1)=defval;
	  val1=defval*v3;
	  maxind=-1; maxval=0;
	  for (i=0; i<=(Nmsgs-1); i++)
	    {
	      val=(*((msg+*(msginds+i))->phi->probs+1))/(*((msg+*(msginds+i))->phi->probs+0)+1e-20);
	      if (val>maxval)
		{
		  maxind=i; maxval=val;
		}
	    }
	  val2=defval*(*((msg+*(msginds+maxind))->phi->probs+1))/(*((msg+*(msginds+maxind))->phi->probs+0));
	  *(phi->probs+0)=max(val1,val2);
	}

      val=*(phi->probs+0)+*(phi->probs+1);
      *(phi->probs+0)/=val; *(phi->probs+1)/=val; 
    }
  
  /*
  // The factor type is PREDICTION.
  // If the target variable is not the output response variable, then do nothing.
  else if ((facnode->type==PREDICTION)&&((fgraph->varnodes+varind)->type==5))
    {
      // If the path is not effective, then m(y=0)=v2*default messages, m(y=+1)=m(y=-1)=v3*default messages.
      if (facnode->effective==0)
	{
	  defval=1.0;
	  for (i=0; i<=(Nmsgs-1); i++)
	    defval*=*((msg+*(msginds+i))->phi->probs+*(defconfig+*(msg2phi+i)));
	  *(phi->probs+0)=v2*defval;
	  *(phi->probs+1)=v3*defval;
	  *(phi->probs+2)=v3*defval;
	}
      
      else
	{
	  int flag2, maxind1, maxind2, carry;
	  double maxval1, maxval2;

	  signconfig=malloc(sizeof(int)*facnode->Nedges);
	  N=1;
	  for (i=0; i<=(facnode->Nedges-1); i++)
	    N*=2;

	  // Find the max sign configurations of aggregate effect +1 and -1.
	  maxind1=-1; maxval1=0; maxind2=-1; maxval2=0;
	  for (i=0; i<=(N-1); i++)
	    {
	      flag2=1; carry=1;

	      // Filter out the sign configs which are not consistent with mask values.
	      j=0;
	      while (j<=(facnode->Nedges-1))
		{
		  if ((*(maskval+*(sign2phi+j))>=0)&&(*(maskval+*(sign2phi+j))!=*(signconfig+j)))
		    flag2=0;
		  j++;
		}
	      if (flag2==1)
		{
		  k=-1;
		  for (j=0; j<=(facnode->Nedges-1); j++)
		    k*=(*(signconfig+j)*2-1);
		  val=1.0;
		  for (j=0; j<=(facnode->Nedges-1); j++)
		    val*=*((msg+*(msginds+*(phi2msg+*(sign2phi+j))))->phi->probs+*(signconfig+j));
		  if ((k==+1)&&(val>maxval1))
		    {
		      maxind1=i; maxval1=val;
		    }
		  else if ((k==-1)&&(val>maxval2))
		    {
		      maxind2=i; maxval2=val;
		    }
		}
	      carry=1; j=facnode->Nedges-1;
	      while ((carry==1)&&(j>=0))
		{
		  *(signconfig+j)+=carry;
		  if (*(signconfig+j)<2)
		    carry=0;
		  else
		    *(signconfig+j)=0;
		  j--;
		}
	    }

	  // If there are multiple configurations which yield the same max message product, choose the one which yields a valid explanatory path.
	  for (i=0; i<=(facnode->Nedges-1); i++)
	    if (*(degcnt+*(presence2phi+i))==2)
	      *(defconfig+*(presence2phi+i))=1;
	  for (i=0; i<=(facnode->Nsigedges-1); i++)
	    if (*(degcnt+*(sigdir2phi+i))==2)
	      *(defconfig+*(sigdir2phi+i))=*(facnode->sigdirvarvals+i);
	  
	  defval=1.0;
	  for (i=0; i<=(Nmsgs-1); i++)
	    defval*=*((msg+*(msginds+i))->phi->probs+*(defconfig+*(msg2phi+i)));

	  // Check if the default configuration yields a valid path.
	  flag=1; i=0;
	  while ((i<=(facnode->Nedges-1))&&(flag==1))
	    {
	      if (*(defconfig+*(presence2phi+i))==0)
		flag=0;
	      i++;
	    }
	  i=0;
	  while ((i<=(facnode->Nsigedges-1))&&(flag==1))
	    {
	      if (*(defconfig+*(sigdir2phi+i))!=*(facnode->sigdirvarvals+i))
		flag=0;
	      i++;
	    }
	  
	  // The default configuration is not a valid path.
	  if (flag==0)
	    {
	      // m(y=0)=v2*defval.
	      *(phi->probs+0)=v2*defval;
	      // m(y=+1)=max{v3*defval,v1*\prod_{i}m(x_{i}=1)*\prod_{i}m(d_{i}=1)*m(max sign configuration such that the aggregate sign is +1)}.
	      // m(y=-1)=max{v3*defval,v1*\prod_{i}m(x_{i}=1)*\prod_{i}m(d_{i}=1)*m(max sign configuration such that the aggregate sign is -1)}. 
	      
	      val=1.0;
	      for (i=0; i<=(facnode->Nedges-1); i++)
		val*=*((msg+*(msginds+*(phi2msg+*(presence2phi+i))))->phi->probs+1);
	      for (i=0; i<=(facnode->Nsigedges-1); i++)
		val*=*((msg+*(msginds+*(phi2msg+*(sigdir2phi+i))))->phi->probs+*(facnode->sigdirvarvals+i));
	      *(phi->probs+1)=max(v3*defval,v1*val*maxval1);
	      *(phi->probs+2)=max(v3*defval,v1*val*maxval2);
	    }
		    
	  // The default configuration is a valid path.
	  else
	    {
	      k=-1;
	      for (i=0; i<=(facnode->Nedges-1); i++)
		k*=(*(defconfig+*(sign2phi+i))*2-1);
	      
	      // If the aggregate sign is +1, then m(y=+1)=v1*defval, same for y=-1.
	      if (k==1)
		*(phi->probs+1)=v1*defval;
	      else if (k==-1)
		*(phi->probs+0)=v1*defval;

	      // If the aggregate sign is +1, then m(y=-1)=max{v3*defval,v1*\prod_{i}m(x_{i}=1)*\prod_{i}m(d_{i}=1)*m(max sign configuration such that the aggregate sign is -1)}. 
	      // Same for y=-1.
	      val=1.0;
	      for (i=0; i<=(facnode->Nedges-1); i++)
		val*=*((msg+*(msginds+*(phi2msg+*(presence2phi+i))))->phi->probs+1);
	      for (i=0; i<=(facnode->Nsigedges-1); i++)
		val*=*((msg+*(msginds+*(phi2msg+*(sigdir2phi+i))))->phi->probs+*(facnode->sigdirvarvals+i));
	      if (k==1)
		*(phi->probs+2)=max(v3*defval,v1*val*maxval2);
	      else if (k==-1)
		*(phi->probs+1)=max(v3*defval,v1*val*maxval1);


	      // m(y=0)=max{v3*defval, v2*m(max configuration of breaking one edge presence or dir var)}.
	      maxind=-1; maxval=0;
	      for (i=0; i<=(facnode->Nedges-1); i++)
		{
		  val=(*((msg+*(msginds+*(phi2msg+*(presence2phi+i))))->phi->probs+0))/(*((msg+*(msginds+*(phi2msg+*(presence2phi+i))))->phi->probs+1)+1e-20);
		  if (val>maxval)
		    {
		      maxind=i; maxval=val;
		    }
		}
	      for (i=0; i<=(facnode->Nsigedges-1); i++)
		{
		  j=*(facnode->sigdirvarvals+i);
		  k=(j==0 ? 1 : 0);
		  val=(*((msg+*(msginds+*(phi2msg+*(sigdir2phi+i))))->phi->probs+k))/(*((msg+*(msginds+*(phi2msg+*(sigdir2phi+i))))->phi->probs+j)+1e-20);
		  if (val>maxval)
		    {
		      maxind=i+facnode->Nedges; maxval=val;
		    }
		}
	      maxval*=defval;
	      *(phi->probs+0)=max(defval*v3,maxval*v2);
	    }
	  free(signconfig);
	}

      // Normalize the probability.
      val=0;
      for (i=0; i<=2; i++)
	val+=*(phi->probs+i);
      for (i=0; i<=2; i++)
	*(phi->probs+i)/=val;
    }
  */


  free(degcnt); free(defconfig);
  

  if (facnode->Nsigedges>0)
    free(sigdir2phi);
  
  if (facnode->Nedges>0)
    {
      free(presence2phi); free(sign2phi); 
    }
  free(phi2msg); free(msg2phi); free(maskval);

}
  
