// Utility functions regarding the operations of potential functions defined by both functional forms and lookup tables.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include "integration.h"
# include "integration3.h"
# include "connect.h"
# include "cluster.h"
# include "integration4.h"
# include "integration5.h"
# define MAXLINEWIDTH 1000
# define INFINITY 1e38
# define POSINF 100000
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)

extern void maxsumproduct1(struct mix_facgraph *fgraph, struct message *msg, int *fixedvals, struct potential *belief, int mode);
extern int * recurse_getmapconfigs(struct confignode *configtree, int *Nmapconfigs, int *mapconfigs);
extern void recurse_freeconfignode(struct confignode *configtree);
void recurse_fixval2(struct confignode *configtree, struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int *vartypes);
extern int inlistind(int Nv, int *v, int q);
extern void copypotential(struct potential *phi2, struct potential *phi1);
extern struct cpath * decompresscpath(struct csparsegraph *cgraph, struct compresscpath *path, int ind);
extern long config2number2(int Nvars, int *Nstates, int *configuration);
extern void number2config2(int Nvars, int *Nstates, long val, int *configuration);
extern void number2config(int Nvars, int val, int *configuration);
extern void potentialproduct2(struct potential *phi1, struct potential *phi2);
extern void potentialproduct3(struct potential *phi1, struct potential *phi2);
extern struct potential * createpotential(struct potential *phi1);
extern void maximizepotential2(struct potential *phi, int ind);
extern void maxmarginal3(struct mix_path_fac_node *facnode, int Nmsgs, int *msginds, struct message *msg, struct potential *phi, int deb);
extern void marginalizepotential4(struct potential *phi, int varind, int Nmsgs, int *msginds, struct message *msg, struct potential *newphi);
extern int pair2msg(int Nmsgs, struct message *msg, struct pair *node1, struct pair *node2);
extern void maxsumproduct3(struct mix_facgraph *fgraph, struct message *msg, int *fixedvals, struct potential *belief, int mode, int updatemode);
extern struct mix_facgraph * construct_fac_graph2(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko);
extern int inlist(int Nv, int *v, int q);
extern void recursecpath(struct compresscpath *path, int ind, int *Nv, int *nodes, int *dirs, int *signs);

struct cpath * decompresscpath2(int Nselgenes, int *ref2sel, int Nlocs, struct triplet *loc, int Npps, struct triplet *pp, struct compresscpath *path, int ind, int deb);
int containnode2(int Nselgenes, int *ref2sel, int Nlocs, struct triplet *loc, int Npps, struct triplet *pp, struct cpath *path, int node);
int * findmaxconfig11(struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int *Nmapconfigs);
void recurse_fixval4(struct confignode *configtree, struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief);
struct compresscpath *findvalidcpaths2(int Nselgenes, int *ref2sel, int Nlocs, struct triplet *loc, int Npps, struct triplet *pp, int Nkos, struct triplet *ko, int pathlength, int *Nvalidcpaths);
void explainknockout4(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, struct model_configuration *degmodels, int configind, int Nsubkos, int *subkoinds, int *Nexplained, int *explained, int *edgeexplainedfreq, int *nodeexplainedfreq);
void explainknockout5(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, struct model_configuration *degmodels, int configind, int Nsubkos, int *subkoinds, int *Nexplained, int *explained, int *edgeexplainedfreq, int *nodeexplainedfreq);
struct compresscpath *augment_validcpaths(int Nselgenes, int *ref2sel, struct csparsegraph *model, int Nkos, struct triplet *ko, int *Ncpaths, struct compresscpath *compresscpaths, int *isvalid);


extern int maxpathlength, submaxpathlength;
extern double epsilon, ptrue_positive;
extern double v1, v2, v3, koinv;


// Find valid paths connecting source and destination genes of knockout pairs up to a fixed length.
// Difference from findvalidcpaths: only consider the paths eminating from the ko source nodes.
struct compresscpath *findvalidcpaths2(int Nselgenes, int *ref2sel, int Nlocs, struct triplet *loc, int Npps, struct triplet *pp, int Nkos, struct triplet *ko, int pathlength, int *Nvalidcpaths)
{
  int i, j, k, n, lowerbd, uppbd;
  struct compresscpath *validpaths, *newvalidpaths;
  struct cpath *path;
  long curalloc=0, block=1000, pathcnt;
  int Nsources, *issource, *isdest;
  int Neffective, *iseffective, *path2effectivepath;
  int **loclookup, **pplookup1, **pplookup2;

  
  // Build the lookup table of location pairs.  Each selected gene contains entries of location pairs whose end gene is the selected gene.
  loclookup=malloc(sizeof(int)*Nselgenes);
  for (i=0; i<=(Nselgenes-1); i++)
    {
      *(loclookup+i)=malloc(sizeof(int));
      **(loclookup+i)=0;
    }
  for (i=0; i<=(Nlocs-1); i++)
    {
      //if ((*(ref2sel+(loc+i)->i)>=0)&&(*(ref2sel+(loc+i)->j)>=0))
	{
	  **(loclookup+*(ref2sel+(loc+i)->i))+=1;
	  *(loclookup+*(ref2sel+(loc+i)->i))=realloc(*(loclookup+*(ref2sel+(loc+i)->i)),sizeof(int)*(**(loclookup+*(ref2sel+(loc+i)->i))+1));
	  *(*(loclookup+*(ref2sel+(loc+i)->i))+**(loclookup+*(ref2sel+(loc+i)->i)))=i;
	}
    }

  // Build the lookup tables of pp pairs.  One maps to (pp+i)->i and the other maps to (pp+i)->j.
  pplookup1=malloc(sizeof(int)*Nselgenes);
  pplookup2=malloc(sizeof(int)*Nselgenes);
  for (i=0; i<=(Nselgenes-1); i++)
    {
      *(pplookup1+i)=malloc(sizeof(int));
      **(pplookup1+i)=0;
      *(pplookup2+i)=malloc(sizeof(int));
      **(pplookup2+i)=0;
    }
  
  for (i=0; i<=(Npps-1); i++)
    {
      //if ((*(ref2sel+(pp+i)->i)>=0)&&(*(ref2sel+(pp+i)->j)>=0))
	{
	  **(pplookup1+*(ref2sel+(pp+i)->i))+=1;
	  *(pplookup1+*(ref2sel+(pp+i)->i))=realloc(*(pplookup1+*(ref2sel+(pp+i)->i)),sizeof(int)*(**(pplookup1+*(ref2sel+(pp+i)->i))+1));
	  *(*(pplookup1+*(ref2sel+(pp+i)->i))+**(pplookup1+*(ref2sel+(pp+i)->i)))=i;
	  **(pplookup2+*(ref2sel+(pp+i)->j))+=1;
	  *(pplookup2+*(ref2sel+(pp+i)->j))=realloc(*(pplookup2+*(ref2sel+(pp+i)->j)),sizeof(int)*(**(pplookup2+*(ref2sel+(pp+i)->j))+1));
	  *(*(pplookup2+*(ref2sel+(pp+i)->j))+**(pplookup2+*(ref2sel+(pp+i)->j)))=i;
	}
    }

  // Identify source nodes.
  issource=malloc(sizeof(int)*Nselgenes);
  for (i=0; i<=(Nselgenes-1); i++)
    *(issource+i)=0;
  for (i=0; i<=(Nkos-1); i++)
    *(issource+*(ref2sel+(ko+i)->i))=1;
  isdest=malloc(sizeof(int)*Nselgenes);
  for (i=0; i<=(Nselgenes-1); i++)
    *(isdest+i)=0;
  for (i=0; i<=(Nkos-1); i++)
    *(isdest+*(ref2sel+(ko+i)->j))=1;
  

  // Debug
  j=0; k=0;
  for (i=0; i<=(Nselgenes-1); i++)
    {
      if (*(issource+i)==1)
	j++;
      if (*(isdest+i)==1)
	k++;
    }
  printf("%d sources, %d destinations.\n",j,k);

  
  // Build all 1-paths.  Only consider edges eminating from source nodes.
  // pp edges have two directions.
  // Avoid self-crossing edges.
  pathcnt=0;
  validpaths=malloc(sizeof(struct compresspath)*block);
  curalloc=block;
  for (i=0; i<=(Nlocs-1); i++)
    {
      //if (*(issource+*(ref2sel+(loc+i)->i))==1)
      if ((*(issource+*(ref2sel+(loc+i)->i))==1)&&((loc+i)->i!=(loc+i)->j))
	{
	  pathcnt++;
	  if (pathcnt>curalloc)
	    {
	      curalloc+=block;
	      validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
	    }
	  (validpaths+pathcnt-1)->pathid=-1;
	  (validpaths+pathcnt-1)->start=*(ref2sel+(loc+i)->i);
	  (validpaths+pathcnt-1)->end=*(ref2sel+(loc+i)->j);
	  (validpaths+pathcnt-1)->length=1;
	  (validpaths+pathcnt-1)->lastdir=+1;
	  (validpaths+pathcnt-1)->lastsign=0;
	  (validpaths+pathcnt-1)->lasttype=0;
	}
    }
  
  /*
  for (i=0; i<=(Npps-1); i++)
    {
      //if (*(issource+*(ref2sel+(pp+i)->i))==1)
      if ((*(issource+*(ref2sel+(pp+i)->i))==1)||(*(issource+*(ref2sel+(pp+i)->j))==1))
	{
	  pathcnt++;
	  if (pathcnt>curalloc)
	    {
	      curalloc+=block;
	      validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
	    }
	  (validpaths+pathcnt-1)->pathid=-1;
	  (validpaths+pathcnt-1)->start=*(ref2sel+(pp+i)->i);
	  (validpaths+pathcnt-1)->end=*(ref2sel+(pp+i)->j);
	  (validpaths+pathcnt-1)->length=1;
	  (validpaths+pathcnt-1)->lastdir=+1;
	  (validpaths+pathcnt-1)->lastsign=0;
	  (validpaths+pathcnt-1)->lasttype=1;

	  pathcnt++;
	  if (pathcnt>curalloc)
	    {
	      curalloc+=block;
	      validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
	    }
	  (validpaths+pathcnt-1)->pathid=-1;
	  (validpaths+pathcnt-1)->start=*(ref2sel+(pp+i)->j);
	  (validpaths+pathcnt-1)->end=*(ref2sel+(pp+i)->i);
	  (validpaths+pathcnt-1)->length=1;
	  (validpaths+pathcnt-1)->lastdir=+1;
	  (validpaths+pathcnt-1)->lastsign=0;
	  (validpaths+pathcnt-1)->lasttype=1;

	}

    }
  */
  
  for (i=0; i<=(Npps-1); i++)
    {
      if ((*(issource+*(ref2sel+(pp+i)->i))==1)&&((pp+i)->i!=(pp+i)->j))
	{
	  pathcnt++;
	  if (pathcnt>curalloc)
	    {
	      curalloc+=block;
	      validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
	    }
	  (validpaths+pathcnt-1)->pathid=-1;
	  (validpaths+pathcnt-1)->start=*(ref2sel+(pp+i)->i);
	  (validpaths+pathcnt-1)->end=*(ref2sel+(pp+i)->j);
	  (validpaths+pathcnt-1)->length=1;
	  (validpaths+pathcnt-1)->lastdir=+1;
	  (validpaths+pathcnt-1)->lastsign=0;
	  (validpaths+pathcnt-1)->lasttype=1;
	}

      if ((*(issource+*(ref2sel+(pp+i)->j))==1)&&((pp+i)->i!=(pp+i)->j))
	{
	  pathcnt++;
	  if (pathcnt>curalloc)
	    {
	      curalloc+=block;
	      validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
	    }
	  (validpaths+pathcnt-1)->pathid=-1;
	  (validpaths+pathcnt-1)->start=*(ref2sel+(pp+i)->j);
	  (validpaths+pathcnt-1)->end=*(ref2sel+(pp+i)->i);
	  (validpaths+pathcnt-1)->length=1;
	  (validpaths+pathcnt-1)->lastdir=+1;
	  (validpaths+pathcnt-1)->lastsign=0;
	  (validpaths+pathcnt-1)->lasttype=1;
	}
    }
    
  
  // Recursively build k-paths from (k-1)-paths and edges.
  lowerbd=0; uppbd=pathcnt-1; n=1;


  /*
  while (n<pathlength)
    {
      // For each (n-1)-path connecting xi to xk, create n-paths by extending an edge (xk,xj).
      for (i=lowerbd; i<=uppbd; i++)
	{
	  // Add a directed edge to a path.
	  // Consider the direction consistency and avoid loops.
	  for (j=0; j<=(Nlocs-1); j++)
	    {
	      if (*(ref2sel+(loc+j)->i)==(validpaths+i)->end)
		{
		  path=decompresscpath2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,validpaths,i,0);
		  if (containnode2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,path,*(ref2sel+(loc+j)->j))==0)
		    {
		      pathcnt++;
		      if (pathcnt>curalloc)
			{
			  curalloc+=block;
			  validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
			}
		      (validpaths+pathcnt-1)->pathid=i;
		      (validpaths+pathcnt-1)->start=(validpaths+i)->start;
		      (validpaths+pathcnt-1)->end=*(ref2sel+(loc+j)->j);
		      (validpaths+pathcnt-1)->lastdir=+1;
		      (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;
		      (validpaths+pathcnt-1)->lastsign=0;
		      (validpaths+pathcnt-1)->lasttype=0;
		    }
		  free(path->e); free(path);
		}
	    }
	  
	  // Add an undirected edge to path.
	  // No problem with direction consistency.  Avoid loops.
	  for (j=0; j<=(Npps-1); j++)
	    {
	      if ((*(ref2sel+(pp+j)->i)==(validpaths+i)->end)||(*(ref2sel+(pp+j)->j)==(validpaths+i)->end))
		{
		  path=decompresscpath2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,validpaths,i,0);

		  //if (containnode2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,path,*(ref2sel+(pp+j)->j))==0)
		  if (((*(ref2sel+(pp+j)->i)==(validpaths+i)->end)&&(containnode2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,path,*(ref2sel+(pp+j)->j))==0))||((*(ref2sel+(pp+j)->j)==(validpaths+i)->end)&&(containnode2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,path,*(ref2sel+(pp+j)->i))==0)))
		    {
		      pathcnt++;
		      if (pathcnt>curalloc)
			{
			  curalloc+=block;
			  validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
			}
		      // (xk,xj) points away from xk.
		      if (*(ref2sel+(pp+j)->i)==(validpaths+i)->end)
			{
			  (validpaths+pathcnt-1)->pathid=i;
			  (validpaths+pathcnt-1)->start=(validpaths+i)->start;
			  (validpaths+pathcnt-1)->end=*(ref2sel+(pp+j)->j);
			  (validpaths+pathcnt-1)->lastdir=+1;
			  (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;
			  (validpaths+pathcnt-1)->lastsign=0;
			  (validpaths+pathcnt-1)->lasttype=1;
			}

		      // (xk,xj) points toward xk.
		      if (*(ref2sel+(pp+j)->j)==(validpaths+i)->end)
			{
			  (validpaths+pathcnt-1)->pathid=i;
			  (validpaths+pathcnt-1)->start=(validpaths+i)->start;
			  (validpaths+pathcnt-1)->end=*(ref2sel+(pp+j)->i);
			  (validpaths+pathcnt-1)->lastdir=+1;
			  (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;
			  (validpaths+pathcnt-1)->lastsign=0;
			  (validpaths+pathcnt-1)->lasttype=1;
			}
 	      
		    }
		  free(path->e); free(path);
		}
	    }
	}
      lowerbd=uppbd+1; uppbd=pathcnt-1; n++;
    }
  */  

  
  while (n<pathlength)
    {
      // Debug
      printf("n=%d, lowerbd=%d, uppbd=%d\n",n,lowerbd,uppbd);

      // For each (n-1)-path connecting xi to xk, create n-paths by extending an edge (xk,xj).
      for (i=lowerbd; i<=uppbd; i++)
	{
	  int ind;

	  // Debug
	  if ((i%1000)==0)
	    printf("i=%d\n",i);

	  // Add a directed edge to a path.
	  // Consider the direction consistency and avoid loops.
	  for (ind=1; ind<=**(loclookup+(validpaths+i)->end); ind++)
	    {
	      j=*(*(loclookup+(validpaths+i)->end)+ind);

	      if (*(ref2sel+(loc+j)->i)==(validpaths+i)->end)
		{
		  path=decompresscpath2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,validpaths,i,0);
		  if (containnode2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,path,*(ref2sel+(loc+j)->j))==0)
		    {
		      pathcnt++;
		      if (pathcnt>curalloc)
			{
			  curalloc+=block;
			  validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
			}
		      (validpaths+pathcnt-1)->pathid=i;
		      (validpaths+pathcnt-1)->start=(validpaths+i)->start;
		      (validpaths+pathcnt-1)->end=*(ref2sel+(loc+j)->j);
		      (validpaths+pathcnt-1)->lastdir=+1;
		      (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;
		      (validpaths+pathcnt-1)->lastsign=0;
		      (validpaths+pathcnt-1)->lasttype=0;
		    }
		  free(path->e); free(path);
		}
	    }
	  
	  // Add an undirected edge to path.
	  // No problem with direction consistency.  Avoid loops.
	  for (ind=1; ind<=**(pplookup1+(validpaths+i)->end); ind++)
	    {
	      j=*(*(pplookup1+(validpaths+i)->end)+ind);

	      if (*(ref2sel+(pp+j)->i)==(validpaths+i)->end)
		{
		  path=decompresscpath2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,validpaths,i,0);

		  if ((*(ref2sel+(pp+j)->i)==(validpaths+i)->end)&&(containnode2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,path,*(ref2sel+(pp+j)->j))==0))
		    {
		      pathcnt++;
		      if (pathcnt>curalloc)
			{
			  curalloc+=block;
			  validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
			}
		      // (xk,xj) points away from xk.
		      if (*(ref2sel+(pp+j)->i)==(validpaths+i)->end)
			{
			  (validpaths+pathcnt-1)->pathid=i;
			  (validpaths+pathcnt-1)->start=(validpaths+i)->start;
			  (validpaths+pathcnt-1)->end=*(ref2sel+(pp+j)->j);
			  (validpaths+pathcnt-1)->lastdir=+1;
			  (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;
			  (validpaths+pathcnt-1)->lastsign=0;
			  (validpaths+pathcnt-1)->lasttype=1;
			}
		    }
		  free(path->e); free(path);
		}
	    }	    

	  for (ind=1; ind<=**(pplookup2+(validpaths+i)->end); ind++)
	    {
	      j=*(*(pplookup2+(validpaths+i)->end)+ind);

	      if (*(ref2sel+(pp+j)->j)==(validpaths+i)->end)
		{
		  path=decompresscpath2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,validpaths,i,0);

		  if ((*(ref2sel+(pp+j)->j)==(validpaths+i)->end)&&(containnode2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,path,*(ref2sel+(pp+j)->i))==0))
		    {
		      pathcnt++;
		      if (pathcnt>curalloc)
			{
			  curalloc+=block;
			  validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
			}
		      // (xk,xj) points toward xk.
		      if (*(ref2sel+(pp+j)->j)==(validpaths+i)->end)
			{
			  (validpaths+pathcnt-1)->pathid=i;
			  (validpaths+pathcnt-1)->start=(validpaths+i)->start;
			  (validpaths+pathcnt-1)->end=*(ref2sel+(pp+j)->i);
			  (validpaths+pathcnt-1)->lastdir=+1;
			  (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;
			  (validpaths+pathcnt-1)->lastsign=0;
			  (validpaths+pathcnt-1)->lasttype=1;
			} 
		    }
		  free(path->e); free(path);
		}
	    }
	}
      lowerbd=uppbd+1; uppbd=pathcnt-1; n++;
    }
  

  // Debug
  printf("pathcnt=%d \n",pathcnt);


  // Consider the paths which connect knockout pairs or are subpaths connecting knockout pairs.
  iseffective=malloc(sizeof(int)*pathcnt);
  for (i=0; i<=(pathcnt-1); i++)
    *(iseffective+i)=0;

  
  for (i=0; i<=(pathcnt-1); i++)
    {
      if ((*(issource+(validpaths+i)->start)==1)&&(*(isdest+(validpaths+i)->end)==1))
	{
	  struct compresscpath *pnt;
	  // Enable all subpaths along path i.
	  *(iseffective+i)=1; pnt=validpaths+i;
	  while ((j=pnt->pathid)>=0)
	    {
	      *(iseffective+j)=1;
	      pnt=(validpaths+j);
	    }
	}
    }


  // Debug
  //for (i=0; i<=(pathcnt-1); i++)
  //*(iseffective+i)=1;


  Neffective=0; path2effectivepath=malloc(sizeof(int)*pathcnt);
  for (i=0; i<=(pathcnt-1); i++)
    *(path2effectivepath+i)=-1;
  for (i=0; i<=(pathcnt-1); i++)
    {
      if (*(iseffective+i)==1)
	{
	  Neffective++; *(path2effectivepath+i)=Neffective-1;
	}
    }
  newvalidpaths=malloc(sizeof(struct compresscpath)*Neffective);
  k=0;
  for (i=0; i<=(pathcnt-1); i++)
    {
      if (*(iseffective+i)==1)
	{
	  k++;
	  if ((validpaths+i)->pathid==-1)
	    (newvalidpaths+k-1)->pathid=-1;
	  else
	    (newvalidpaths+k-1)->pathid=*(path2effectivepath+(validpaths+i)->pathid);
	  (newvalidpaths+k-1)->start=(validpaths+i)->start;
	  (newvalidpaths+k-1)->end=(validpaths+i)->end;
	  (newvalidpaths+k-1)->lastdir=(validpaths+i)->lastdir;
	  (newvalidpaths+k-1)->length=(validpaths+i)->length;
	  (newvalidpaths+k-1)->lastsign=(validpaths+i)->lastsign;
	  (newvalidpaths+k-1)->lasttype=(validpaths+i)->lasttype;
	}
    }
  
  *Nvalidcpaths=Neffective; 

  //*Nvalidcpaths=pathcnt;
 

  /*
  // Debug
  for (i=0; i<=(Neffective-1); i++)
    {
      if ((*(issource+(newvalidpaths+i)->start)==1)&&(*(isdest+(newvalidpaths+i)->end)==1))
	printf("(%d %d %d) ",i,(newvalidpaths+i)->start,(newvalidpaths+i)->end);
    }
  printf("\n");
  */


  free(issource); free(isdest); 

  free(validpaths); 

  

  free(path2effectivepath);

  
  for (i=0; i<=(Nselgenes-1); i++)
    {
      free(*(loclookup+i)); free(*(pplookup1+i)); free(*(pplookup2+i));
    }
  free(loclookup); free(pplookup1); free(pplookup2);

 
  // Debug
  printf("Neffective=%d ",Neffective);

  return newvalidpaths;
  

  //return validpaths;
}


// Augment valid paths to existing valid paths.  Include the paths which do not start with source nodes of knockout pairs but are sub-paths of existing valid paths.  Is needed in experiment design procedures.
// The last edge of an augmented path is a pd edge.

struct compresscpath *augment_validcpaths(int Nselgenes, int *ref2sel, struct csparsegraph *model, int Nkos, struct triplet *ko, int *Ncpaths, struct compresscpath *compresscpaths, int *isvalid)
{
  int i, j, k, *issource;
  struct cpath *path;
  long curalloc=0, block=1000, pathcnt, oldNcpaths;


  // Identify source and destination nodes.
  issource=malloc(sizeof(int)*Nselgenes);
  for (i=0; i<=(Nselgenes-1); i++)
    *(issource+i)=0;
  for (i=0; i<=(Nkos-1); i++)
    *(issource+*(ref2sel+(ko+i)->i))=1;
  
  
  // For each valid path, find all its subpaths which end at the same destination.
  // If the start of a subpath is a source node, then this subpath is among existing paths.
  // If the start of a subpath is not a source node, then iteratively build a compresscpath data structure of that subpath.

  pathcnt=*Ncpaths; oldNcpaths=*Ncpaths;
  curalloc=oldNcpaths+block;
  compresscpaths=realloc(compresscpaths,sizeof(struct compresscpath)*curalloc);

  
  for (i=0; i<=(oldNcpaths-1); i++)
    {
      if ((*(isvalid+i)==1)&&((compresscpaths+i)->length>1))
	{
	  int *nodelist, flag=1;

	  path=decompresscpath(model,compresscpaths,i);

	  // It is possible for a path to contain edges of ind=-1 since some p-p edges which are overlapped with p-d edges are deleted.
	  j=0;
	  while ((j<=(path->Ne-1))&&(flag==1))
	    {
	      if (*(path->e+j)<0)
		flag=0;
	      j++;
	    }
	  
	  if (flag==1)
	    {	  
	      nodelist=malloc(sizeof(int)*(path->Ne+1));
	      *nodelist=(((model->e+*path->e)->to==(model->e+*(path->e+1))->from)||((model->e+*path->e)->to==(model->e+*(path->e+1))->to) ? (model->e+*path->e)->from : (model->e+*path->e)->to);
	      for (j=1; j<=(path->Ne-1); j++)
		*(nodelist+j)=(((model->e+*(path->e+j))->from==(model->e+*(path->e+j-1))->to)||((model->e+*(path->e+j))->from==(model->e+*(path->e+j-1))->from) ? (model->e+*(path->e+j))->from : (model->e+*(path->e+j))->to);
	      *(nodelist+path->Ne)=(model->e+*(path->e+path->Ne-1))->to;
	      
	      /*
	      for (j=1; j<=(path->Ne-1); j++)
		{
		  // If the starting node is not a source node, then add the subpath to the list.
		  if (*(issource+*(nodelist+j))==0)
		    {
		      // Different from findvalidcpaths or findvalidcpaths2, first create a longer path then iteratively create subpaths.
		      k=path->Ne;
		      while (k>j)
			{
			  pathcnt++;
			  if (pathcnt>curalloc)
			    {
			      curalloc+=block;
			      compresscpaths=realloc(compresscpaths,sizeof(struct compresspath)*curalloc);
			    }
			  
			  // If k-j=1, the subpath is an edge.
			  if ((k-j)==1)
			    (compresscpaths+pathcnt-1)->pathid=-1;
			  // Otherwise create a path and point it to the next subpath.
			  else
			    (compresscpaths+pathcnt-1)->pathid=pathcnt;
			  
			  (compresscpaths+pathcnt-1)->start=*(nodelist+j);
			  (compresscpaths+pathcnt-1)->end=*(nodelist+k);
			  (compresscpaths+pathcnt-1)->length=(k-j);
			  if ((model->e+*(path->e+k-1))->from==*(nodelist+k-1))
			    (compresscpaths+pathcnt-1)->lastdir=+1;
			  else
			    (compresscpaths+pathcnt-1)->lastdir=-1;
			  (compresscpaths+pathcnt-1)->lastsign=0;
			  (compresscpaths+pathcnt-1)->lasttype=(model->e+*(path->e+k-1))->type;

			  
			  k--;
			}
		    }
		}
	      */
		
	      
	      // NEW: Discard the subpaths whose last edges are pp.
	      for (j=1; j<=(path->Ne-1); j++)
		{
		  // If the starting node is not a source node, then add the subpath to the list.
		  if (*(issource+*(nodelist+j))==0)
		    {
		      // Different from findvalidcpaths or findvalidcpaths2, first create a longer path then iteratively create subpaths.
		      k=path->Ne;
		      while (k>j)
			{
			  int v1, v2, edgeind=-1;

			  //
			  //v1=*(nodelist+k-1); v2=*(nodelist+k); l=0;
			  //while ((l<=(path->Ne-1))&&(edgeind==-1))
			  //{
			  //if ((((model->e+*(path->e+l))->from==v1)&&((model->e+*(path->e+l))->to==v2))||(((model->e+*(path->e+l))->to==v1)&&((model->e+*(path->e+l))->from==v2)))
			  //edgeind=*(path->e+l);
			  //l++;
			  //}
			  //if (edgeind==-1)
			  //{
			  //printf("error in augment_validcpaths.\n");
			  //return;
			  //}
			  //

			  edgeind=*(path->e+k-1);

			  if ((model->e+edgeind)->type==0)
			    {

			      pathcnt++;
			      if (pathcnt>curalloc)
				{
				  curalloc+=block;
				  compresscpaths=realloc(compresscpaths,sizeof(struct compresspath)*curalloc);
				}
			  
			      // If k-j=1, the subpath is an edge.
			      if ((k-j)==1)
				(compresscpaths+pathcnt-1)->pathid=-1;
			      // Otherwise create a path and point it to the next subpath.
			      else
				(compresscpaths+pathcnt-1)->pathid=pathcnt;
			      
			      (compresscpaths+pathcnt-1)->start=*(nodelist+j);
			      (compresscpaths+pathcnt-1)->end=*(nodelist+k);
			      (compresscpaths+pathcnt-1)->length=(k-j);
			      if ((model->e+*(path->e+k-1))->from==*(nodelist+k-1))
				(compresscpaths+pathcnt-1)->lastdir=+1;
			      else
				(compresscpaths+pathcnt-1)->lastdir=-1;
			      (compresscpaths+pathcnt-1)->lastsign=0;
			      (compresscpaths+pathcnt-1)->lasttype=(model->e+*(path->e+k-1))->type;
			    }

			  k--;
			}
		    }
		}
	      
	      
	      free(nodelist);
	    } 
	  
	  
	  free(path->e); free(path);
	}
    }

  free(issource); 
  *Ncpaths=pathcnt;
  
  
  return compresscpaths;
}

// Check if a cpath contains a node.
// Difference from containnode: input location and pp pairs instead of model edges.
int containnode2(int Nselgenes, int *ref2sel, int Nlocs, struct triplet *loc, int Npps, struct triplet *pp, struct cpath *path, int node)
{
  int i=0, flag=0;

  while ((flag==0)&&(i<=(path->Ne-1)))
    {
      int from, to;

      if (*(path->e+i)>=Nlocs)
	{
	  from=*(ref2sel+(pp+*(path->e+i)-Nlocs)->i);
	  to=*(ref2sel+(pp+*(path->e+i)-Nlocs)->j);
	}
      else
	{
	  from=*(ref2sel+(loc+*(path->e+i))->i);
	  to=*(ref2sel+(loc+*(path->e+i))->j);
	}
      if ((from==node)||(to==node))
	flag=1;
      i++;
    }
  return flag;
}

// Decompress a path from compressedcpath format to cpath format.
// Difference from decompresscpath: inputs are location and pp edges instead of a model. 
struct cpath * decompresscpath2(int Nselgenes, int *ref2sel, int Nlocs, struct triplet *loc, int Npps, struct triplet *pp, struct compresscpath *path, int ind, int deb)
{
  int i, j, Nv, *nodes, *dirs, *signs;
  struct cpath *decompressedpath;

  i=(path+ind)->length+1;
  nodes=malloc(sizeof(int)*(i+1));
  dirs=malloc(sizeof(int)*(i+1));
  signs=malloc(sizeof(int)*(i+1));

  //nodes=malloc(sizeof(int)*(ind+1));
  //dirs=malloc(sizeof(int)*(ind+1));
  //signs=malloc(sizeof(int)*(ind+1));
  Nv=0;

  recursecpath(path,ind,&Nv,nodes,dirs,signs);

  /*
  // Debug
  if (deb==1)
    {
      printf("nodes: ");
      for (i=0; i<=(Nv-1); i++)
	printf("%d ",*(nodes+i));
      printf("\n");
      printf("dirs: ");
      for (i=0; i<=(Nv-2); i++)
	printf("%d ",*(dirs+i));
      printf("\n");
    }
  */

  /*
  // Debug
  if (deb==1)
    {
      printf("loc: ");
      for (i=0; i<=(Nlocs-1); i++)
	printf("(%d,%d %d)",i,(loc+i)->i,(loc+i)->j);
      printf("\npp: ");
      for (i=0; i<=(Npps-1); i++)
	printf("(%d,%d %d)",i,(pp+i)->i,(pp+i)->j);
      printf("\n");
    }
  */

  decompressedpath=malloc(sizeof(struct cpath));
  decompressedpath->Ne=Nv-1; decompressedpath->sign=0;
  decompressedpath->e=malloc(sizeof(int)*(Nv-1));
  for (i=0; i<=(Nv-2); i++)
    {
      int ind1, ind2, edgeind;
      ind1=*(nodes+Nv-1-i); ind2=*(nodes+Nv-2-i);
      edgeind=-1; j=0;
      while ((j<=(Nlocs-1))&&(edgeind<0))
	{
	  if ((*(ref2sel+(loc+j)->i)==ind1)&&(*(ref2sel+(loc+j)->j)==ind2))
	    edgeind=j;
	  j++;
	}
      j=0;
      while ((j<=(Npps-1))&&(edgeind<0))
	{
	  if (((*(ref2sel+(pp+j)->i)==ind1)&&(*(ref2sel+(pp+j)->j)==ind2))||((*(ref2sel+(pp+j)->j)==ind1)&&(*(ref2sel+(pp+j)->i)==ind2)))
	    edgeind=j+Nlocs;
	  j++;
	}
	
      // Debug
      if (deb==1)
	printf("i=%d ind1=%d ind2=%d Nlocs=%d Npps=%d edgeind=%d \n",i,ind1,ind2,Nlocs,Npps,edgeind);

      *(decompressedpath->e+i)=edgeind;
    }

  free(nodes);
  free(dirs);
  free(signs);
  return decompressedpath;
}
