/* Generate subnetworks based on rosetta knockout data.
   Different from explain26: revise the function of subnetwork generation.
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "connect.h"
# include "integration.h"
# include "cluster.h"
# include "matrix.h"
# include "integration3.h"
# include "integration4.h"
# include "integration5.h"
# define MAXLINEWIDTH 1000
# define GENENAMELENGTH 9
# define PHASELENGTH 4
# define INFINITY 1e10
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)

extern char * getitemval(char *s, int ind, int *length);
extern char * getnextitemval(char *s, int *length, int curpos, int *nextpos);
extern char * getnextitemval2(char *s, int *length, int curpos, int *nextpos, int *end);
extern double atof2(char *s, int length);
extern long * loadmatrix(char *filename);
extern double * loadmatrix2(char *filename);
extern struct compresscpath * findvalidcpaths(struct csparsegraph *cgraph, int pathlength, int *Nvalidcpaths);
extern struct cpath * decompresscpath(struct csparsegraph *cgraph, struct compresscpath *path, int ind);
extern void explainknockout(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, int *Nexplained, int *explained);
extern void loadgenenames(char *filename, int *N, int *genenamespnt, int *anothergenenamespnt);
extern void loadselgenes(int N, char *genenames, char *anothergenenames, char *selfilename, int *Nselgenes, int *ref2selpnt, int *sel2refpnt);
extern void loadppdata2(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *ppdata, int *Npps, int *pppnt);
extern void loadkodata(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *kogenefilename, char *expfilename, char *kologratiofilename, char *kopvalfilename, double kopthre, int *isknockedpnt, int *Nkos, int *kopnt);
extern void determinevalidpaths(int Ncpaths, struct compresscpath *compresscpaths, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, struct csparsegraph *model, int N, int *isknocked, int *isvalid, int *kopairs);
extern struct compresscpath *findvalidcpaths2(int Nselgenes, int *ref2sel, int Nlocs, struct triplet *locs, int Npps, struct triplet *pps, int Nkos, struct triplet *ko, int pathlength, int *Nvalidcpaths);
extern struct compresscpath *augment_validcpaths(int Nselgenes, int *ref2sel, struct csparsegraph *model, int Nkos, struct triplet *ko, int *Ncpaths, struct compresscpath *compresscpaths, int *isvalid);
extern void loadlocdata3(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *locfilename, double locpthre, int *Nfactors, int *ref2facpnt, int *fac2refpnt, int *Nlocs, int *locpnt);
extern double invcchi2cdf(double p, int D, double lower, double upper, double precision);
extern struct chi2cdf_lookup * loadchi2table(char *filename, int *Nentries);
extern double chi2cdf_deg1_2(double x, int mode, int Nentries, struct chi2cdf_lookup *chi2table);
extern double invchi2cdf_deg1_2(double val, int mode, int Nentries, struct chi2cdf_lookup *chi2table);
extern struct model_configuration * generate_subnetworks1(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, char *anothergenenames, int *Ndecomposedmodels, int *decomposedmodelpnt);


int maxpathlength=5;
int submaxpathlength=4;

//int maxpathlength=3;
//int submaxpathlength=3;
int Nreplicates=3;

double v1=0.7, v2=0.299, v3=0.001, koinv=1e-20;
extern char *chi2lookuptable;


void main(int argc, char *argv[])
{
  // *(argv+1) -- reference table.
  // *(argv+2) -- genes of interest.
  // *(argv+3) -- location pair data.
  // *(argv+4) -- threshold on location pvalues.
  // *(argv+5) -- p-p interaction pairs.
  // *(argv+6) -- rosetta knockout gene table.
  // *(argv+7) -- rosetta knockout experiment list.
  // *(argv+8) -- rosetta knockout pvalue threshold.
  // *(argv+9) -- rosetta knockout pvalues.
  // *(argv+10) -- rosetta logratios.
  // *(argv+11) -- subnetwork sif file.
  // *(argv+12) -- subnetwork edge dir file.
  // *(argv+13) -- subnetwork edge sign file.
  // *(argv+14) -- subnetwork knockout file.
  // *(argv+15) -- subnetwork knockout pairs file.
  

  int i, j, k, l, n, N, Nselgenes, Nfactors, *ref2sel, *sel2ref;
  int *ref2fac, *fac2ref, Nlocs, Nkos, Npps, pnt1, pnt2, pnt3;
  int oldNcpaths, Ncpaths, *isvalid, *kopairs, Nconnected, *isconnected;
  double locpthre, kopthre, *errors, error_rate;
  struct triplet *loc, *ko, *pp;
  char *genenames, *anothergenenames, ch, *s;
  struct csparsegraph *model;
  struct compresscpath *compresscpaths;
  struct model_configuration *degmodels;
  struct decomposed_model *decomposedmodels;
  FILE *fp1, *fp2, *fp3, *fp4, *fp5;
  int Nvalidpaths, *validpathinds, *isknocked, Nexplained, *koinds, Nleaveout;
  int *explained, *conn2ko, *var2ko, *isaffected, *edge2sigedge;
  struct chi2cdf_lookup *chi2table;
  int Nentries, Nleaveoutexps, *leaveoutexps, Ndecomposedmodels;
  int *isused, N1, N2, kopairs_pathlength[6], minplength;
  int *isedgeconnected, Ndegpre, Ndegsign, Ndegdir, Ndegko, Ndegpathsel;
  int *connectivity;
  

  // Load reference table.
  genenames=malloc(sizeof(char));
  anothergenenames=malloc(sizeof(char));
  loadgenenames(*(argv+1),&N,&pnt1,&pnt2);
  genenames=pnt1; anothergenenames=pnt2;

  // Debug
  printf("N=%d, ",N);
  
  // Load genes of interest.
  if (strncmp(*(argv+1),*(argv+2),max(strlen(*(argv+1)),strlen(*(argv+2))))!=0)
    {
      loadselgenes(N,genenames,anothergenenames,*(argv+2),&Nselgenes,&pnt1,&pnt2);
      ref2sel=pnt1; sel2ref=pnt2;
    }
  else
    {  
      Nselgenes=N; ref2sel=malloc(sizeof(int)*N); sel2ref=malloc(sizeof(int)*N);
      for (i=0; i<=(N-1); i++)
	{
	  *(ref2sel+i)=i; *(sel2ref+i)=i;
	}
    }
  
  // Debug
  printf("Nselgenes=%d, ",Nselgenes);

  // Load the chi2lookup table.
  chi2table=loadchi2table(chi2lookuptable,&Nentries);

  // Debug
  //maxpathlength=3;
  
  // Load location data.
  locpthre=atof(*(argv+4));
  loadlocdata3(N,genenames,anothergenenames,Nselgenes,ref2sel,sel2ref,*(argv+3),locpthre,&Nfactors,&pnt1,&pnt2,&Nlocs,&pnt3);
  ref2fac=pnt1; fac2ref=pnt2; loc=pnt3;

  // Determine the likelihood ratios of pd edges.
  for (i=0; i<=(Nlocs-1); i++)
    {
      double val, val2;
      val=(loc+i)->s;
      //val2=0.5*invcchi2cdf(val,1,0,100,0.01);
      val2=0.5*invchi2cdf_deg1_2(val,1,Nentries,chi2table);
      val2-=0.5*log(Nreplicates);
      (loc+i)->s=exp(val2);
    }

  // Debug
  printf("Nlocs=%d, Nfactors=%d, ",Nlocs,Nfactors);
  
  // Load protein-protein interaction data.
  // (pp+i)->s returns the likelihood ratio of the interaction.
  loadppdata2(N,genenames,anothergenenames,Nselgenes,ref2sel,sel2ref,*(argv+5),&Npps,&pnt1);
  pp=pnt1;

  // Debug
  printf("Npps=%d, ",Npps);
  
  
  // Load knockout gene table. 
  kopthre=atof(*(argv+8));
  loadkodata(N,genenames,anothergenenames,Nselgenes,ref2sel,sel2ref,*(argv+6),*(argv+7),*(argv+9),*(argv+10),kopthre,&pnt1,&Nkos,&pnt2);
  isknocked=pnt1; ko=pnt2;
    
  // Determine the likelihood ratios of ko data.
  for (i=0; i<=(Nkos-1); i++)
    {
      double val, val2;
      val=(ko+i)->s;
      //val2=0.5*invcchi2cdf(val,1,0,100,0.1);
      val2=0.5*invchi2cdf_deg1_2(val,1,Nentries,chi2table);
      val2-=0.5*log(Nreplicates);
      (ko+i)->s=exp(val2);
    }
  
  // Debug
  printf("Nkos=%d \n",Nkos);
  

  compresscpaths=findvalidcpaths2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,Nkos,ko,maxpathlength,&Ncpaths);
  
  // Debug
  printf("Ncpaths=%d \n",Ncpaths);

  
  // Construct the physical model.
  model=malloc(sizeof(struct csparsegraph));
  model->Nv=Nselgenes;
  model->v=malloc(sizeof(int)*model->Nv);
  for (i=0; i<=(Nselgenes-1); i++)
    *(model->v+i)=i;
  model->Ne=Nlocs+Npps; model->e=malloc(sizeof(struct cedge)*model->Ne);
  model->Ne=0;

  // Delete self-crossing edges.
  // Add directed edges.
  for (i=0; i<=(Nlocs-1); i++)
    {
      if ((loc+i)->i!=(loc+i)->j)
	{
	  model->Ne+=1;
	  (model->e+model->Ne-1)->type=0;
	  (model->e+model->Ne-1)->from=*(ref2sel+(loc+i)->i);
	  (model->e+model->Ne-1)->to=*(ref2sel+(loc+i)->j);
	  (model->e+model->Ne-1)->sign=0;
	  (model->e+model->Ne-1)->priorp=(loc+i)->s;
	}
    }
  // Add undirected edges.  A directed edge overwrites an undirected edge.
  for (i=0; i<=(Npps-1); i++)
    {
      int flag=1;
      
      if ((pp+i)->i!=(pp+i)->j)
	{
	  j=0;
	  while ((j<=(Nlocs-1))&&(flag==1))
	    {
	      if (((pp+i)->i==(loc+j)->i)&&((pp+i)->j==(loc+j)->j))
		flag=0;
	      j++;
	    }
	  
	  if (flag==1)
	    {
	      model->Ne+=1;
	      (model->e+model->Ne-1)->type=1;
	      (model->e+model->Ne-1)->from=*(ref2sel+(pp+i)->i);
	      (model->e+model->Ne-1)->to=*(ref2sel+(pp+i)->j);
	      (model->e+model->Ne-1)->sign=0;
	      (model->e+model->Ne-1)->priorp=(pp+i)->s;
	    }
	}
    }  
  
  // Debug
  printf("model->Nv=%d, model->Ne=%d \n",model->Nv,model->Ne);


  // Identify valid paths according to exlanation criteria.
  isvalid=malloc(sizeof(int)*Ncpaths);
  kopairs=malloc(sizeof(int)*Ncpaths);
  determinevalidpaths(Ncpaths,compresscpaths,ref2sel,sel2ref,Nkos,ko,model,N,isknocked,isvalid,kopairs);

  // Augment valid paths to incorporate subpaths of existing valid paths.  Necessary because existing paths do not include genes which are not knocked out as sources.
  oldNcpaths=Ncpaths;
  compresscpaths=augment_validcpaths(Nselgenes,ref2sel,model,Nkos,ko,&Ncpaths,compresscpaths,isvalid);
  isvalid=realloc(isvalid,sizeof(int)*Ncpaths);
  for (i=oldNcpaths; i<=(Ncpaths-1); i++)
    *(isvalid+i)=1;
  
  // Debug
  printf("oldNcpaths=%d, newNcpaths=%d\n",oldNcpaths,Ncpaths);
  free(kopairs);
  

  // Remove the paths which contain pp edges overwritten by location edges.
  for (i=0; i<=(Ncpaths-1); i++)
    {
      if (*(isvalid+i)==1)
	{
	  struct cpath *path;
	  path=decompresscpath(model,compresscpaths,i);
	  for (j=0; j<=(path->Ne-1); j++)
	    if (*(path->e+j)<0)
	      *(isvalid+i)=0;
	  free(path->e); free(path);
	}
    }
  
  Nvalidpaths=0; validpathinds=malloc(sizeof(int));
  for (i=0; i<=(Ncpaths-1); i++)
    {
      if (*(isvalid+i)==1)
	{
	  Nvalidpaths++; validpathinds=realloc(validpathinds,sizeof(int)*Nvalidpaths);
	  *(validpathinds+Nvalidpaths-1)=i;
	}
    }
      
  // Debug
  printf("There are %d valid paths, model->Ne=%d.\n",Nvalidpaths,model->Ne);

  
  // Find the knockout pairs which are connected in the physical model.
  isconnected=malloc(sizeof(int)*Nkos);
  for (i=0; i<=(Nkos-1); i++)
    *(isconnected+i)=0;
  Nconnected=0;
  for (i=0; i<=(Nkos-1); i++)
    {
      j=0;
      //while ((j<=(Ncpaths-1))&&(*(isconnected+i)==0))
      while ((j<=(Nvalidpaths-1))&&(*(isconnected+i)==0))
	{
	  //if ((*(isvalid+j)==1)&&((compresscpaths+j)->start==*(ref2sel+(ko+i)->i))&&((compresscpaths+j)->end==*(ref2sel+(ko+i)->j)))
	  if (((compresscpaths+*(validpathinds+j))->start==*(ref2sel+(ko+i)->i))&&((compresscpaths+*(validpathinds+j))->end==*(ref2sel+(ko+i)->j)))
	    {
	      *(isconnected+i)=1; Nconnected++;
	    }
	  j++;
	}
    }

  // Debug
  printf("Nconnected=%d \n",Nconnected);


  
  // Construct MAP models from the data.
  degmodels=generate_subnetworks1(model,Ncpaths,compresscpaths,isvalid,ref2sel,sel2ref,Nkos,ko,anothergenenames,&Ndecomposedmodels,&pnt1);
  decomposedmodels=pnt1;


  connectivity=malloc(sizeof(int)*N);
  for (i=0; i<=(N-1); i++)
    *(connectivity+i)=0;
  for (i=0; i<=(Npps-1); i++)
    {
      *(connectivity+(pp+i)->i)+=1;
      *(connectivity+(pp+i)->j)+=1;
    }



  // Debug
  printf("Nvars=%d, Ndecomposedmodels=%d\n",degmodels->Nvars,Ndecomposedmodels);

  
  var2ko=malloc(sizeof(int)*degmodels->Nvars);
  for (i=0; i<=(degmodels->Nvars-1); i++)
    *(var2ko+i)=-1;
  for (i=0; i<=(degmodels->Nkos-1); i++)
    *(var2ko+*(degmodels->ko2var+i))=i;
  
  // Report the subnet attributes files.
  if ((fp1=fopen(*(argv+11),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+11));
      return;
    }
  fprintf(fp1,"There are %d decomposed models.\n",Ndecomposedmodels);
  if ((fp2=fopen(*(argv+12),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+12));
      return;
    }
  fprintf(fp2,"There are %d decomposed models.\n",Ndecomposedmodels);
  fprintf(fp2,"direction\n");
  if ((fp3=fopen(*(argv+13),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+13));
      return;
    }
  fprintf(fp3,"There are %d decomposed models.\n",Ndecomposedmodels);
  fprintf(fp3,"edgesign\n");
  if ((fp4=fopen(*(argv+14),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+14));
      return;
    }
  fprintf(fp4,"There are %d decomposed models.\n",Ndecomposedmodels);
  fprintf(fp4,"knockout\n");
  if ((fp5=fopen(*(argv+15),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+15));
      return;
    }
  fprintf(fp5,"There are %d decomposed models.\n",Ndecomposedmodels);


  fprintf(fp1,"location thre=%f, rosetta ko thre=%f\n",locpthre,kopthre);
  fprintf(fp2,"location thre=%f, rosetta ko thre=%f\n",locpthre,kopthre);
  fprintf(fp3,"location thre=%f, rosetta ko thre=%f\n",locpthre,kopthre);
  fprintf(fp4,"location thre=%f, rosetta ko thre=%f\n",locpthre,kopthre);
  fprintf(fp5,"location thre=%f, rosetta ko thre=%f\n",locpthre,kopthre);
  


  
  isknocked=realloc(isknocked,sizeof(int)*model->Nv);
  for (i=0; i<=(model->Nv-1); i++)
    *(isknocked+i)=0;
  for (i=0; i<=(Nkos-1); i++)
    if (*(isconnected+i)==1)
      *(isknocked+*(ref2sel+(ko+i)->i))=1;
  isaffected=malloc(sizeof(int)*model->Nv);
  edge2sigedge=malloc(sizeof(int)*model->Ne);
  for (i=0; i<=(model->Ne-1); i++)
    *(edge2sigedge+i)=-1;
  for (i=0; i<=(degmodels->Nsigedges-1); i++)
    *(edge2sigedge+*(degmodels->sigedge2edge+i))=i;

  for (i=0; i<=(Ndecomposedmodels-1); i++)
    {
      int Nnoninvariant=0;
           
      for (j=0; j<=((decomposedmodels+i)->Nvars-1); j++)
	if (*((decomposedmodels+i)->vartypes+j)!=0)
	  Nnoninvariant++;

      //if (Nnoninvariant>3)
      if ((decomposedmodels+i)->Nedges>0)
	{
	  int *isedgesel, *isnodesel;
	  
	  fprintf(fp1,"model %d\n",i);
	  fprintf(fp2,"model %d\n",i);
	  fprintf(fp3,"model %d\n",i);
	  fprintf(fp4,"model %d\n",i);
	  fprintf(fp5,"model %d\n",i);

	  isedgesel=malloc(sizeof(int)*model->Ne);
	  isnodesel=malloc(sizeof(int)*model->Nv);
	  for (j=0; j<=(model->Ne-1); j++)
	    *(isedgesel+j)=0;
	  for (j=0; j<=(model->Nv-1); j++)
	    *(isnodesel+j)=0;

	  for (j=0; j<=((decomposedmodels+i)->Nedges-1); j++)
	    {
	      int edgeind=*((decomposedmodels+i)->edgeinds+j);
	      int varind=-1;

	      *(isedgesel+edgeind)=1;

	      // Check if this edge contains invariant edge sign.
	      k=0;
	      while ((k<=((decomposedmodels+i)->Nvars-1))&&(varind==-1))
		{
		  if (*((decomposedmodels+i)->varinds+k)==*(degmodels->sign2var+edgeind))
		    varind=k;
		  k++;
		}
	      if ((varind>=0)&&(*((decomposedmodels+i)->vartypes+varind)==0))
		{
		  fprintf(fp1,"* "); 
		  fprintf(fp2,"* ");  
		  fprintf(fp3,"* "); 
		}


	      k=0;
	      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+edgeind)->from)*GENENAMELENGTH+k))!=' '))
		{
		  fprintf(fp1,"%c",ch); 
		  fprintf(fp2,"%c",ch); 
		  fprintf(fp3,"%c",ch); 
		  k++;
		}
	      if ((model->e+edgeind)->type==0)
		{
		  fprintf(fp1," pd ");
		  fprintf(fp2," (pd) ");
		  fprintf(fp3," (pd) ");
		}
	      else
		{
		  fprintf(fp1," pp ");
		  fprintf(fp2," (pp) ");
		  fprintf(fp3," (pp) ");
		}
	      k=0;
	      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+edgeind)->to)*GENENAMELENGTH+k))!=' '))
		{
		  fprintf(fp1,"%c",ch);
		  fprintf(fp2,"%c",ch); 
		  fprintf(fp3,"%c",ch); 
		  k++;
		}
	      fprintf(fp1,"\n");
	      if ((model->e+edgeind)->type==0)
		fprintf(fp2," = forw");
	      else
		{
		  if (*(degmodels->mapconfigs+*(degmodels->sigdir2var+*(edge2sigedge+edgeind)))==1)
		    fprintf(fp2," = forw");
		  else if (*(degmodels->mapconfigs+*(degmodels->sigdir2var+*(edge2sigedge+edgeind)))==0)
		    fprintf(fp2," = back");
		  else
		    fprintf(fp2," = unknown");
		}
	      fprintf(fp2,"\n");

	      if (*(degmodels->mapconfigs+*(degmodels->sign2var+edgeind))==1)
		fprintf(fp3," = positive");
	      else if (*(degmodels->mapconfigs+*(degmodels->sign2var+edgeind))==0)
		fprintf(fp3," = negative");
	      else
		fprintf(fp3," = unknown");
	      fprintf(fp3,"\n");
	    }
	  fprintf(fp1,"\n"); fprintf(fp2,"\n"); fprintf(fp3,"\n");

	  for (j=0; j<=(model->Ne-1); j++)
	    {
	      if (*(isedgesel+j)==1)
		{
		  *(isnodesel+(model->e+j)->from)=1;
		  *(isnodesel+(model->e+j)->to)=1;
		}
	    }
	  
	  for (j=0; j<=(model->Nv-1); j++)
	    *(isaffected+j)=0;
	  for (j=0; j<=(Nkos-1); j++)
	    if ((*(isconnected+j)==1)&&(*(isnodesel+*(ref2sel+(ko+j)->i))==1)&&(*(isnodesel+*(ref2sel+(ko+j)->j))==1))
	      *(isaffected+*(ref2sel+(ko+j)->j))=1;

	  for (j=0; j<=(model->Nv-1); j++)
	    {
	      if (*(isnodesel+j)==1)
		{
		  k=0;
		  while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+j)*GENENAMELENGTH+k))!=' '))
		    {
		      fprintf(fp4,"%c",ch); k++;
		    }
		  fprintf(fp4," = ");
		  if (*(isknocked+j)==1)
		    fprintf(fp4,"deleted");
		  else if (*(isaffected+j)==1)
		    fprintf(fp4,"affected");
		  else
		    fprintf(fp4,"others");

		  fprintf(fp4,"  %d",*(connectivity+*(sel2ref+j)));

		  fprintf(fp4,"\n");
		}
	    }
	  fprintf(fp4,"\n");
	  free(isnodesel); free(isedgesel);


	  for (j=0; j<=((decomposedmodels+i)->Nvars-1); j++)
	    {
	      if (*(degmodels->vartypes+*((decomposedmodels+i)->varinds+j))==KNOCKOUT)
		{
		  int koind=*(var2ko+*((decomposedmodels+i)->varinds+j));
		  k=0;
		  while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+(ko+koind)->i*GENENAMELENGTH+k))!=' '))
		    {
		      fprintf(fp5,"%c",ch); k++;
		    }
		  fprintf(fp5," ");
		  k=0;
		  while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+(ko+koind)->j*GENENAMELENGTH+k))!=' '))
		    {
		      fprintf(fp5,"%c",ch); k++;
		    }
		  fprintf(fp5," ");
		  if ((ko+koind)->k>0)
		    fprintf(fp5,"+");
		  else
		    fprintf(fp5,"-");
		  fprintf(fp5,"\n");
		}
	    }
	  fprintf(fp5,"\n");

	}
    }
  fclose(fp1); fclose(fp2); fclose(fp3); fclose(fp4);

  
}
