// Utility functions of graphs.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include "integration.h"
# include "connect.h"
# include "cluster.h"
# define MAXLINEWIDTH 1000
# define POSINF 100000

extern struct cliquelineage * findcomponents2(long Nnodes, long *nodes, double *graph, double thre, long *Ncomponents);
extern long conv_ind(long i, long j, long N);

void recursepath(struct compresspath *path, int ind, int *Nv, int *nodes, int *dirs, int *signs);
void copypathway(struct pathway *inpath, struct pathway *outpath);
struct compresspath * findvalidpaths(struct sparsegraph *graph, int pathlength, int *Nvalidpaths);
int inlist(int Nv, int *v, int q);
void rmvlist(int *Nv, int *v, int q);

// Decompress a path from compresspath format to bipath format.
// Recursively identify the last node in a path and the subpath in this path.  Stop when the path becomes a 0-path (no subpath).
// Only need to look at the paths of smaller indices.
struct bipath * decompress(struct compresspath *path, int ind)
{
  int i, Nv, *nodes, *dirs, *signs;
  struct bipath *decompressedpath;

  nodes=malloc(sizeof(int)*(ind+1));
  dirs=malloc(sizeof(int)*(ind+1));
  signs=malloc(sizeof(int)*(ind+1));
  Nv=0;

  recursepath(path,ind,&Nv,nodes,dirs,signs);

  /*
  // Debug
  printf("nodes: ");
  for (i=0; i<=(Nv-1); i++)
    printf("%d ",*(nodes+i));
  printf("\n");
  printf("dirs: ");
  for (i=0; i<=(Nv-2); i++)
    printf("%d ",*(dirs+i));
  printf("\n");
  */

  decompressedpath=malloc(sizeof(struct bipath));
  decompressedpath->Nv=Nv;
  decompressedpath->v=malloc(sizeof(int)*Nv);
  decompressedpath->dir=malloc(sizeof(int)*(Nv-1));
  decompressedpath->sign=malloc(sizeof(int)*(Nv-1));
  for (i=(Nv-1); i>=0; i--)
    {
      *(decompressedpath->v+Nv-1-i)=*(nodes+i);
      if (i<(Nv-1))
	{
	  *(decompressedpath->dir+Nv-2-i)=*(dirs+i);
	  *(decompressedpath->sign+Nv-2-i)=*(signs+i);
	}
    }

  free(nodes);
  free(dirs);
  free(signs);
  return decompressedpath;
}

// Verify if a node is on the list of nodes.
int contained(long Nv, long *v, long ind)
{
  int i, flag;
  
  flag=0;
  for (i=0; i<=(Nv-1); i++)
    {
      if (*(v+i)==ind)
	{
	  flag=1;
	  break;
	}
    }
  
  return flag;
}

void recursepath(struct compresspath *path, int ind, int *Nv, int *nodes, int *dirs, int *signs)
{
  int i;

  // If (path+ind)->pathid=-1, then stop.  Return both ends.
  if ((path+ind)->pathid==-1)
    {
      if ((path+ind)->start!=(path+ind)->end)
	{
	  *(nodes+*Nv)=(path+ind)->end;
	  *(nodes+*Nv+1)=(path+ind)->start;
	  *(dirs+*Nv)=(path+ind)->lastdir;
	  *(signs+*Nv)=(path+ind)->lastsign;
	  *Nv=*Nv+2;
	}
	else
	  {
	    *(nodes+*Nv)=(path+ind)->end;
	    *Nv=*Nv+1;
	  }
      return;
    }

  // If (path+ind)->pathid!=-1, then keep on parsing the paths.
  else
    {
      *(nodes+*Nv)=(path+ind)->end;
      *(dirs+*Nv)=(path+ind)->lastdir;
      *(signs+*Nv)=(path+ind)->lastsign;
      *Nv=*Nv+1;
      recursepath(path,(path+ind)->pathid,Nv,nodes,dirs,signs);
    }
}


// Find all valid paths in a graph up to certain path length.
// A valid path is either unidirectional or divergent.
struct compresspath * findvalidpaths(struct sparsegraph *graph, int pathlength, int *Nvalidpaths)
{
  int i, j, n, lowerbd, uppbd;
  struct compresspath *validpaths;
  struct bipath *decompressedpath;
  long curalloc=0, block=1000, pathcnt;

  /*
  // Debug
  printf("graph: \n");
  for (i=0; i<=(graph->Nv-1); i++)
    printf("%d ",*(graph->v+i));
  printf("\n");
  for (i=0; i<=(graph->Ne-1); i++)
    printf("(%d %d %d)",(graph->e+i)->from,(graph->e+i)->to,(graph->e+i)->delay);
  printf("\n");
  */

  // Build all 0-paths (1-node paths).
  pathcnt=0;
  validpaths=malloc(sizeof(struct compresspath)*(graph->Ne*2+graph->Nv));
  curalloc=graph->Ne*2+graph->Nv;
  for (i=0; i<=(graph->Nv-1); i++)
    {
      (validpaths+pathcnt)->start=i;
      (validpaths+pathcnt)->pathid=-1;
      (validpaths+pathcnt)->end=i;
      (validpaths+pathcnt)->lastdir=0;
      (validpaths+pathcnt)->delay=0;
      (validpaths+pathcnt)->weight=0;
      (validpaths+pathcnt)->length=0;
      pathcnt++;
    }


  // Build all 1-paths.
  //pathcnt=0;
  //validpaths=malloc(sizeof(struct compresspath)*graph->Ne*2);
  //curalloc=graph->Ne*2;
  for (i=0; i<=(graph->Ne-1); i++)
    {
      (validpaths+pathcnt)->start=(graph->e+i)->from;
      (validpaths+pathcnt)->pathid=-1;
      (validpaths+pathcnt)->end=(graph->e+i)->to;
      (validpaths+pathcnt)->lastdir=+1;
      (validpaths+pathcnt)->delay=(graph->e+i)->delay;
      (validpaths+pathcnt)->weight=0;
      (validpaths+pathcnt)->length=1;
      pathcnt++;
    }
  for (i=0; i<=(graph->Ne-1); i++)
    {
      (validpaths+pathcnt)->start=(graph->e+i)->to;
      (validpaths+pathcnt)->pathid=-1;
      (validpaths+pathcnt)->end=(graph->e+i)->from;
      (validpaths+pathcnt)->lastdir=-1;
      (validpaths+pathcnt)->delay=-1*(graph->e+i)->delay;
      (validpaths+pathcnt)->weight=0;
      (validpaths+pathcnt)->length=1;
      pathcnt++;
    }

  // Recursively build k-paths from (k-1)-paths and edges.
  //n=0; lowerbd=0; uppbd=pathcnt-1;
  n=0; lowerbd=graph->Nv; uppbd=pathcnt-1;
  while (n<pathlength)
    {
      // For each (n-1)-path connecting xi to xk, create n-paths by extending an edge (xk,xj) if the (n-1)-path does not converge with the edge (xk,xj) at xk.
      for (i=lowerbd; i<=uppbd; i++)
	{
	  for (j=0; j<=(graph->Ne-1); j++)
	    {
	      // (xk,xj) points away from xk, always valid.
	      if ((graph->e+j)->from==(validpaths+i)->end)
		{
		  // Remove the loopy paths, i.e., a valid path does not cross itself.
		  decompressedpath=decompress(validpaths,i);
		  if (contained(decompressedpath->Nv,decompressedpath->v,(graph->e+j)->to)==0)
		    {
		      pathcnt++;
		      if (pathcnt>curalloc)
			{
			  curalloc+=block;
			  validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
			}
		      
		      (validpaths+pathcnt-1)->start=(validpaths+i)->start;
		      (validpaths+pathcnt-1)->pathid=i;
		      (validpaths+pathcnt-1)->end=(graph->e+j)->to;
		      (validpaths+pathcnt-1)->lastdir=+1;
		      (validpaths+pathcnt-1)->delay=(validpaths+i)->delay+(graph->e+j)->delay;
		      (validpaths+pathcnt-1)->weight=0;
		      (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;

		      // Debug
		      //if (pathcnt==92)
		      //{
		      //  printf("id=%d, edge=(%d %d), start=%d end=%d lastdir=%d delay1=%d delay2=%d delay=%d\n",i,(graph->e+j)->from,(graph->e+j)->to,(validpaths+pathcnt-1)->start,(validpaths+pathcnt-1)->end,(validpaths+pathcnt-1)->lastdir,(validpaths+i)->delay,(graph->e+j)->delay,(validpaths+pathcnt-1)->delay);
		      //}

		    }
		  free(decompressedpath->v);
		  free(decompressedpath->dir);
		  free(decompressedpath->sign);
		  free(decompressedpath);
		}
	      
	      // (xk,xj) points toward xk, only valid when the (n-1)-path points away from xk.
	      else if (((graph->e+j)->to==(validpaths+i)->end) && ((validpaths+i)->lastdir==-1))
		{
		  // Remove the loopy paths, i.e., a valid path does not cross itself.
		  decompressedpath=decompress(validpaths,i);
		  if (contained(decompressedpath->Nv,decompressedpath->v,(graph->e+j)->from)==0)
		    {
		      pathcnt++;
		      if (pathcnt>curalloc)
			{
			  curalloc+=block;
			  validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
			}

		      (validpaths+pathcnt-1)->start=(validpaths+i)->start;
		      (validpaths+pathcnt-1)->pathid=i;
		      (validpaths+pathcnt-1)->end=(graph->e+j)->from;
		      (validpaths+pathcnt-1)->lastdir=-1;
		      (validpaths+pathcnt-1)->delay=(validpaths+i)->delay-(graph->e+j)->delay;
		      (validpaths+pathcnt-1)->weight=0;
		      (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;


		      // Debug
		      //if (pathcnt==92)
		      //{
		      //  printf("id=%d, edge=(%d %d), start=%d end=%d lastdir=%d delay1=%d delay2=%d delay=%d\n",i,(graph->e+j)->from,(graph->e+j)->to,(validpaths+pathcnt-1)->start,(validpaths+pathcnt-1)->end,(validpaths+pathcnt-1)->lastdir,(validpaths+i)->delay,(graph->e+j)->delay*(-1),(validpaths+pathcnt-1)->delay);
		      //}

		    }
		  free(decompressedpath->v);
		  free(decompressedpath->dir);
		  free(decompressedpath->sign);
		  free(decompressedpath);
		}
	    }
	}

      // Update lower bound and upper bound of n-paths.
      lowerbd=uppbd+1; uppbd=pathcnt-1; n++;
    }

  *Nvalidpaths=pathcnt;

  return validpaths;
}


// Find all valid paths in a graph up to certain path length.
// Different from findvalidpaths: a valid path is unidirectional, and only the last edge can have negative sign.
struct compresspath * findvalidpaths2(struct sparsegraph *graph, int pathlength, int *Nvalidpaths)
{
  int i, j, n, lowerbd, uppbd;
  struct compresspath *validpaths;
  struct bipath *decompressedpath;
  long curalloc=0, block=1000, pathcnt;

  // Build all 0-paths (1-node paths).
  pathcnt=0;
  validpaths=malloc(sizeof(struct compresspath)*(graph->Ne*2+graph->Nv));
  curalloc=graph->Ne*2+graph->Nv;
  for (i=0; i<=(graph->Nv-1); i++)
    {
      (validpaths+pathcnt)->start=i;
      (validpaths+pathcnt)->pathid=-1;
      (validpaths+pathcnt)->end=i;
      (validpaths+pathcnt)->lastdir=0;
      (validpaths+pathcnt)->delay=0;
      (validpaths+pathcnt)->weight=0;
      (validpaths+pathcnt)->length=0;
      (validpaths+pathcnt)->lastsign=0;
      pathcnt++;
    }


  // Build all 1-paths.
  for (i=0; i<=(graph->Ne-1); i++)
    {
      (validpaths+pathcnt)->start=(graph->e+i)->from;
      (validpaths+pathcnt)->pathid=-1;
      (validpaths+pathcnt)->end=(graph->e+i)->to;
      (validpaths+pathcnt)->lastdir=+1;
      (validpaths+pathcnt)->delay=(graph->e+i)->delay;
      (validpaths+pathcnt)->weight=0;
      (validpaths+pathcnt)->length=1;
      (validpaths+pathcnt)->lastsign=(graph->e+i)->sign;
      pathcnt++;
    }
  
  // Recursively build k-paths from (k-1)-paths and edges.
  //n=0; lowerbd=0; uppbd=pathcnt-1;
  n=0; lowerbd=graph->Nv; uppbd=pathcnt-1;
  while (n<pathlength)
    {
      // For each (n-1)-path connecting xi to xk, create n-paths by extending an edge (xk,xj) if the (n-1)-path does not converge with the edge (xk,xj) at xk.
      for (i=lowerbd; i<=uppbd; i++)
	{
	  for (j=0; j<=(graph->Ne-1); j++)
	    {
	      // (xk,xj) points away from xk, always valid.
	      if (((graph->e+j)->from==(validpaths+i)->end)&&((validpaths+i)->lastsign==1))
		{
		  // Remove the loopy paths, i.e., a valid path does not cross itself.
		  decompressedpath=decompress(validpaths,i);
		  if (contained(decompressedpath->Nv,decompressedpath->v,(graph->e+j)->to)==0)
		    {
		      pathcnt++;
		      if (pathcnt>curalloc)
			{
			  curalloc+=block;
			  validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
			}
		      
		      (validpaths+pathcnt-1)->start=(validpaths+i)->start;
		      (validpaths+pathcnt-1)->pathid=i;
		      (validpaths+pathcnt-1)->end=(graph->e+j)->to;
		      (validpaths+pathcnt-1)->lastdir=+1;
		      (validpaths+pathcnt-1)->delay=(validpaths+i)->delay+(graph->e+j)->delay;
		      (validpaths+pathcnt-1)->weight=0;
		      (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;
		      (validpaths+pathcnt-1)->lastsign=(graph->e+j)->sign;

		    }
		  free(decompressedpath->v);
		  free(decompressedpath->dir);
		  free(decompressedpath->sign);
		  free(decompressedpath);
		}	      	      
	    }
	}

      // Update lower bound and upper bound of n-paths.
      lowerbd=uppbd+1; uppbd=pathcnt-1; n++;
    }

  *Nvalidpaths=pathcnt;

  return validpaths;
}


// Find all valid paths in a graph up to certain path length.
// Different from findvalidpaths: a valid path is unidirectional.
// Different from findvalidpaths2: a path is valid when knocking out an upstream gene always down-regulates non-terminal downstream genes and up or down regulates downstream terminal gene.
struct compresspath * findvalidpaths3(struct sparsegraph *graph, int pathlength, int Nkogenes, double *effects, int *Nvalidpaths)
{
  int i, j, k, l, m, n, lowerbd, uppbd, flag;
  struct compresspath *validpaths;
  struct bipath *decompressedpath;
  long curalloc=0, block=1000, pathcnt;

  // Build all 0-paths (1-node paths).
  pathcnt=0;
  validpaths=malloc(sizeof(struct compresspath)*(graph->Ne*2+graph->Nv));
  curalloc=graph->Ne*2+graph->Nv;
  for (i=0; i<=(graph->Nv-1); i++)
    {
      (validpaths+pathcnt)->start=i;
      (validpaths+pathcnt)->pathid=-1;
      (validpaths+pathcnt)->end=i;
      (validpaths+pathcnt)->lastdir=0;
      (validpaths+pathcnt)->delay=0;
      (validpaths+pathcnt)->weight=0;
      (validpaths+pathcnt)->length=0;
      (validpaths+pathcnt)->lastsign=0;
      pathcnt++;
    }


  // Build all 1-paths.
  for (i=0; i<=(graph->Ne-1); i++)
    {
      (validpaths+pathcnt)->start=(graph->e+i)->from;
      (validpaths+pathcnt)->pathid=-1;
      (validpaths+pathcnt)->end=(graph->e+i)->to;
      (validpaths+pathcnt)->lastdir=+1;
      (validpaths+pathcnt)->delay=(graph->e+i)->delay;
      (validpaths+pathcnt)->weight=0;
      (validpaths+pathcnt)->length=1;
      (validpaths+pathcnt)->lastsign=(graph->e+i)->sign;
      pathcnt++;
    }
  
  // Recursively build k-paths from (k-1)-paths and edges.
  //n=0; lowerbd=0; uppbd=pathcnt-1;
  n=0; lowerbd=graph->Nv; uppbd=pathcnt-1;
  while (n<pathlength)
    {
      // For each (n-1)-path connecting xi to xk, create n-paths by extending an edge (xk,xj) if the (n-1)-path does not converge with the edge (xk,xj) at xk.
      for (i=lowerbd; i<=uppbd; i++)
	{
	  for (j=0; j<=(graph->Ne-1); j++)
	    {
	      // (xk,xj) points away from xk, always valid.
	      if (((graph->e+j)->from==(validpaths+i)->end)&&((validpaths+i)->lastsign==1))
		{
		  // Remove the loopy paths, i.e., a valid path does not cross itself.
		  decompressedpath=decompress(validpaths,i);
		  if (contained(decompressedpath->Nv,decompressedpath->v,(graph->e+j)->to)==0)
		    {
		      int sign=*(effects+(graph->e+j)->from*Nkogenes+(graph->e+j)->to);

		      // Check if knocking out upstream genes always down-regulates intermediate genes and regulates the end gene.

		      flag=1;
		      for (k=0; k<=(decompressedpath->Nv-2); k++)
			{
			  for (l=(k+1); l<=(decompressedpath->Nv-1); l++)
			    {
			      if (*(effects+*(decompressedpath->v+k)*Nkogenes+*(decompressedpath->v+l))!=+1)
				{
				  flag=0;
				  break;
				}
			    }
			  l=(graph->e+j)->to;
			  if ((*(effects+*(decompressedpath->v+k)*Nkogenes+l)==0)||(*(effects+*(decompressedpath->v+k)*Nkogenes+l)!=sign))
			    flag=0;
			  
			  
			  if (flag==0)
			    break;
			}
		      


		      if (flag==1)
			{

			  pathcnt++;
			  if (pathcnt>curalloc)
			    {
			      curalloc+=block;
			      validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
			    }
			  
			  (validpaths+pathcnt-1)->start=(validpaths+i)->start;
			  (validpaths+pathcnt-1)->pathid=i;
			  (validpaths+pathcnt-1)->end=(graph->e+j)->to;
			  (validpaths+pathcnt-1)->lastdir=+1;
			  (validpaths+pathcnt-1)->delay=(validpaths+i)->delay+(graph->e+j)->delay;
			  (validpaths+pathcnt-1)->weight=0;
			  (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;
			  (validpaths+pathcnt-1)->lastsign=(graph->e+j)->sign;

			}
		    }
		  free(decompressedpath->v);
		  free(decompressedpath->dir);
		  free(decompressedpath->sign);
		  free(decompressedpath);
		}	      	      
	    }
	}

      // Update lower bound and upper bound of n-paths.
      lowerbd=uppbd+1; uppbd=pathcnt-1; n++;
    }

  *Nvalidpaths=pathcnt;

  return validpaths;
}

// Check if a number is in a list.
int inlist(int Nv, int *v, int q)
{
  int i, flag=0;
  int cnt=0;
  
  for (i=0; i<=(Nv-1); i++)
    {
      if (*(v+i)==q)
	{
	  flag=1; break;
	}

      cnt++;

    }

  //printf("cnt=%d\n",cnt);

  return flag;
}

// Remove an element from a list.
void rmvlist(int *Nv, int *v, int q)
{
  int i, ind=-1;

  if (inlist(*Nv,v,q)==1)
    {
      for (i=0; i<=(*Nv-1); i++)
	{
	  if (*(v+i)==q)
	    {
	      ind=i; break;
	    }
	}
      
      if (ind>=0)
	{
	  for (i=ind; i<=(*Nv-2); i++)
	    *(v+i)=*(v+i+1);
	  *Nv-=1;
	  v=realloc(v,sizeof(int)*(*Nv));
	}
      
    }
  

  return;
}


// Check if a number is in a list.
// Assume the elements in the list are sorted in the ascending order.  Hence can apply binary search to boost process.
int inlist2(int Nv, int *v, int q)
{
  int i, flag=0, ind, entry, start, end, osc=0;
  double tmpval, val;
  int cnt=0;

  tmpval=floor((double)Nv/(double)2); ind=(int)tmpval;
  start=0; end=Nv;
    
  while ((flag==0)&&(ind>=start)&&(ind<=end)&&(start<end)&&(osc==0))
    {
      cnt++;

      //printf("(%d %d %d %d)",start,end,ind,*(v+ind));

      entry=*(v+ind);
      if (entry==q)
	flag=1;
      else if (((end-start)==1)&&(*(v+start)!=q)&&(*(v+end)!=q))
	osc=1;
      else if (entry>q)
	{
	  end=ind;
	  val=(double)((double)(start+ind)/(double)2);
	  tmpval=floor(val);
	  ind=(int)tmpval;
	}
      else if (entry<q)
	{
	  start=ind;
	  val=(double)((double)(ind+end)/(double)2);
	  tmpval=ceil(val);
	  ind=(int)tmpval;
	}
      
    }

  //printf("cnt=%d\n",cnt);

  return flag;
}


// Return the index of a number in a list.  If there are multiple identical numbers, then return the first.
int inlistind(int Nv, int *v, int q)
{
  int i=0, ind=-1;

  while ((i<=(Nv-1))&&(ind==-1))
    {
      if (*(v+i)==q)
	ind=i;
      i++;
    }

  return ind;

}
