// Utility functions of graphs.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include "integration.h"
# include "integration3.h"
# include "connect.h"
# include "cluster.h"
# define MAXLINEWIDTH 1000
# define POSINF 100000
# define max(i,j) (i>j ? i : j)

struct compresscpath * findvalidcpaths(struct csparsegraph *cgraph, int pathlength, int *Nvalidcpaths);
int containnode(struct csparsegraph *cgraph, struct cpath *path, int node);
int containpath(struct cpath *lpath, struct cpath *spath);
struct cpath * decompresscpath(struct csparsegraph *cgraph, struct compresscpath *path, int ind);
void recursecpath(struct compresscpath *path, int ind, int *Nv, int *nodes, int *dirs, int *signs);
int nodes2edgeind(struct csparsegraph *cgraph, int ind1, int ind2);
void constructconcattree(struct cpath *mainpath, int Nannotatedpaths, struct cpath *annotatedpaths, struct concatenationnode *concattree, int curnode, int *Nnodes, int *Nvleaves, int *vleaves);
void findseqinconcattree(struct concatenationnode *concattree, int curnode, int *Nseq, int *seq);
extern int * intersection(int N1, int *v1, int N2, int *v2, int *N);
extern struct cliquelineage * findcomponents2(long Nnodes, long *nodes, double *graph, double thre, long *Ncomponents);
extern int inlist(int Nv, int *v, int q);
extern struct bipath * decompress(struct compresspath *path, int ind);
void number2config(int Nvars, long val, int *configuration);
long config2number(int Nvars, int *configuration);
void number2config2(int Nvars, int *Nstates, long val, int *configuration);
long config2number2(int Nvars, int *Nstates, int *configuration);
struct potential * createpotential(struct potential *phi1);
void copypotential(struct potential *phi2, struct potential *phi1);
void potentialproduct(struct potential *phi1, struct potential *phi2);
void potentialproduct2(struct potential *phi1, struct potential *phi2);
void potentialproduct3(struct potential *phi1, struct potential *phi2);
void maximizepotential(struct potential *phi, int Nv, int *v);
void maximizepotential2(struct potential *phi, int ind);
void maximizepotential3(struct potential *phi, int varind);
void marginalizepotential(struct potential *phi, int Nv, int *v);
void marginalizepotential2(struct potential *phi, int varind);
void marginalizepotential3(struct potential *phi, int varind);

// Find all valid cpaths in a graph.  A valid cpath contains uni-directional or undirectional edges.  Remove the paths of 0 length.
struct compresscpath * findvalidcpaths(struct csparsegraph *cgraph, int pathlength, int *Nvalidcpaths)
{
  int i, j, n, lowerbd, uppbd;
  struct compresscpath *validpaths;
  struct cpath *decompressedpath;
  long curalloc=0, block=1000, pathcnt;

  // Build all 1-paths.
  pathcnt=0;
  validpaths=malloc(sizeof(struct compresspath)*cgraph->Ne*2);

  curalloc=cgraph->Ne*2;
  for (i=0; i<=(cgraph->Ne-1); i++)
    {
      (validpaths+pathcnt)->pathid=-1;
      (validpaths+pathcnt)->start=(cgraph->e+i)->from;
      (validpaths+pathcnt)->end=(cgraph->e+i)->to;
      (validpaths+pathcnt)->length=1;
      (validpaths+pathcnt)->lastdir=+1;
      (validpaths+pathcnt)->lastsign=0;
      (validpaths+pathcnt)->lasttype=(cgraph->e+i)->type;
      pathcnt++;

    }
  for (i=0; i<=(cgraph->Ne-1); i++)
    {
      if ((cgraph->e+i)->type==1)
	{
	  (validpaths+pathcnt)->pathid=-1;
	  (validpaths+pathcnt)->start=(cgraph->e+i)->to;
	  (validpaths+pathcnt)->end=(cgraph->e+i)->from;
	  (validpaths+pathcnt)->length=1;
	  (validpaths+pathcnt)->lastdir=+1;
	  (validpaths+pathcnt)->lastsign=0;
	  (validpaths+pathcnt)->lasttype=(cgraph->e+i)->type;
	  pathcnt++;
	}
    }


  // Recursively build k-paths from (k-1)-paths and edges.
  n=1; 

  // lowerbd=cgraph->Nv; uppbd=pathcnt-1;
  lowerbd=0; uppbd=pathcnt-1;

  while (n<pathlength)
    {
      // For each (n-1)-path connecting xi to xk, create n-paths by extending an edge (xk,xj).
      for (i=lowerbd; i<=uppbd; i++)
	{
	  for (j=0; j<=(cgraph->Ne-1); j++)
	    {
	      // Add an edge to a path.
	      if (((cgraph->e+j)->from==(validpaths+i)->end) || (((cgraph->e+j)->to==(validpaths+i)->end)&&((cgraph->e+j)->type==1)))
		{

		  // Remove the loopy paths, i.e., a valid path does not cross itself.
		  decompressedpath=decompresscpath(cgraph,validpaths,i);
		  if ((((cgraph->e+j)->from==(validpaths+i)->end)&&(containnode(cgraph,decompressedpath,(cgraph->e+j)->to)==0)) || (((cgraph->e+j)->to==(validpaths+i)->end)&&((cgraph->e+j)->type==1)&&(containnode(cgraph,decompressedpath,(cgraph->e+j)->from)==0)))
		    {
		      pathcnt++;
		      if (pathcnt>curalloc)
			{
			  curalloc+=block;
			  validpaths=realloc(validpaths,sizeof(struct compresspath)*curalloc);
			}

		      // (xk,xj) points away from xk.
		      if ((cgraph->e+j)->from==(validpaths+i)->end)
			{		      
			  (validpaths+pathcnt-1)->pathid=i;
			  (validpaths+pathcnt-1)->start=(validpaths+i)->start;
			  (validpaths+pathcnt-1)->end=(cgraph->e+j)->to;
			  (validpaths+pathcnt-1)->lastdir=+1;
			  (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;
			  (validpaths+pathcnt-1)->lastsign=0;
			  (validpaths+pathcnt-1)->lasttype=(cgraph->e+j)->type;

			  // Debug
			  //printf("(%d %d)",pathcnt-1,(validpaths+pathcnt-1)->lasttype);
			}
		      
		      // (xk,xj) points toward xk
		      else if (((cgraph->e+j)->to==(validpaths+i)->end)&&((cgraph->e+j)->type==1))
			{
			  (validpaths+pathcnt-1)->pathid=i;
			  (validpaths+pathcnt-1)->start=(validpaths+i)->start;
			  (validpaths+pathcnt-1)->end=(cgraph->e+j)->from;
			  (validpaths+pathcnt-1)->lastdir=+1;
			  (validpaths+pathcnt-1)->length=(validpaths+i)->length+1;
			  (validpaths+pathcnt-1)->lastsign=0;
			  (validpaths+pathcnt-1)->lasttype=(cgraph->e+j)->type;

			  // Debug
			  //printf("(%d %d)",pathcnt-1,(validpaths+pathcnt-1)->lasttype);
			}

		    }
		  free(decompressedpath->e);
		  free(decompressedpath);
		}	      	      
	    }
	}

      // Update lower bound and upper bound of n-paths.
      lowerbd=uppbd+1; uppbd=pathcnt-1; n++;
    }

  *Nvalidcpaths=pathcnt;
  

  /*
  // Debug  
  printf("pathcnt=%d\n",pathcnt);

  for (i=0; i<=(pathcnt-1); i++)
    {
      decompressedpath=decompresscpath(cgraph,validpaths,i);
      printf("(%d ",i);
      for (j=0; j<=(decompressedpath->Ne-1); j++)
	printf("%d ",*(decompressedpath->e+j));
      printf(")");
      free(decompressedpath->e);
      free(decompressedpath);
    }
  printf("\n");
  */


  return validpaths;
}


// Check if a cpath contains a node.
int containnode(struct csparsegraph *cgraph, struct cpath *path, int node)
{
  int i=0, flag=0;

  while ((flag==0)&&(i<=(path->Ne-1)))
    {
      if (((cgraph->e+*(path->e+i))->from==node)||((cgraph->e+*(path->e+i))->to==node))
	flag=1;
      i++;
    }
  return flag;
}


// Check if a cpath contains another cpath.
int containpath(struct cpath *lpath, struct cpath *spath)
{
  int i=0, j, flag=0;
  
  while ((flag==0)&&(i<=(lpath->Ne-1)))
    {
      if (*(lpath->e+i)==*(spath->e+i))
	flag=1;
      i++;
    }
  
  if (flag==1)
    {
      i-=1; j=0;
      while ((j<=(spath->Ne-1))&&(flag==1))
	{
	  if (*(lpath->e+i)!=*(spath->e+j))
	    flag=0;
	  i++; j++;
	}
    }

  return flag;
}


// Decompress a path from compressedcpath format to cpath format.
struct cpath * decompresscpath(struct csparsegraph *cgraph, struct compresscpath *path, int ind)
{
  int i, Nv, *nodes, *dirs, *signs;
  struct cpath *decompressedpath;

  nodes=malloc(sizeof(int)*(ind+1));
  dirs=malloc(sizeof(int)*(ind+1));
  signs=malloc(sizeof(int)*(ind+1));
  Nv=0;

  recursecpath(path,ind,&Nv,nodes,dirs,signs);

  /*
  // Debug
  printf("nodes: ");
  for (i=0; i<=(Nv-1); i++)
    printf("%d ",*(nodes+i));
  printf("\n");
  printf("dirs: ");
  for (i=0; i<=(Nv-2); i++)
    printf("%d ",*(dirs+i));
  printf("\n");
  */


  decompressedpath=malloc(sizeof(struct cpath));
  decompressedpath->Ne=Nv-1; decompressedpath->sign=0;
  decompressedpath->e=malloc(sizeof(int)*(Nv-1));
  for (i=0; i<=(Nv-2); i++)
    {
      int ind1, ind2, edgeind;
      ind1=*(nodes+Nv-1-i); ind2=*(nodes+Nv-2-i);
      edgeind=nodes2edgeind(cgraph,ind1,ind2);
      *(decompressedpath->e+i)=edgeind;
    }

  free(nodes);
  free(dirs);
  free(signs);
  return decompressedpath;
}


void recursecpath(struct compresscpath *path, int ind, int *Nv, int *nodes, int *dirs, int *signs)
{
  // If (path+ind)->pathid=-1, then stop.  Return both ends.
  if ((path+ind)->pathid==-1)
    {
      if ((path+ind)->start!=(path+ind)->end)
	{
	  *(nodes+*Nv)=(path+ind)->end;
	  *(nodes+*Nv+1)=(path+ind)->start;
	  *(dirs+*Nv)=(path+ind)->lastdir;
	  *(signs+*Nv)=(path+ind)->lastsign;
	  *Nv=*Nv+2;
	}
	else
	  {
	    *(nodes+*Nv)=(path+ind)->end;
	    *Nv=*Nv+1;
	  }
      return;
    }

  // If (path+ind)->pathid!=-1, then keep on parsing the paths.
  else
    {
      *(nodes+*Nv)=(path+ind)->end;
      *(dirs+*Nv)=(path+ind)->lastdir;
      *(signs+*Nv)=(path+ind)->lastsign;
      *Nv=*Nv+1;
      recursecpath(path,(path+ind)->pathid,Nv,nodes,dirs,signs);
    }
}


int nodes2edgeind(struct csparsegraph *cgraph, int ind1, int ind2)
{
  int i=0, edgeind=-1;

  while ((i<=(cgraph->Ne-1))&&(edgeind<0))
    {
      if (((cgraph->e+i)->from==ind1)&&((cgraph->e+i)->to==ind2))
	edgeind=i;
      else if (((cgraph->e+i)->to==ind1)&&((cgraph->e+i)->from==ind2)&&((cgraph->e+i)->type==1))
	edgeind=i;
      i++;
    }

  return edgeind;
}



// Convert an integer into a binary configuration.  0 means -1.
void number2config(int Nvars, long val, int *configuration)
{
  int q, r, tmpval, i;

  tmpval=val; i=Nvars-1;
  while (i>=0)
    {
      q=tmpval/2; r=tmpval%2;
      
      //if (r==0)
      //*(configuration+i)=-1;
      //else
      //*(configuration+i)=1;
      
      *(configuration+i)=r;

      tmpval=q; i--;
    }
}


// Convert a configuration into a number.
long config2number(int Nvars, int *configuration)
{
  int i, val, c;
  
  i=0; val=0;
  while (i<=(Nvars-1))
    {
      //if (*(configuration+i)==-1)
      //c=0;
      //else
      //c=1;

      c=*(configuration+i);

      val=val*2+c;
      i++;
    }
  return val;
}


// Convert an integer into a configuration of states.  Each variable can have more than two states.
void number2config2(int Nvars, int *Nstates, long val, int *configuration)
{
  int q, r, tmpval, i;

  
  tmpval=val; i=Nvars-1;
  while (i>=0)
    {
      
      q=tmpval/(*(Nstates+i)); r=tmpval%(*(Nstates+i));

      //if (r==0)
      //*(configuration+i)=-1;
      //else
      //*(configuration+i)=1;
      
      *(configuration+i)=r;
      
      tmpval=q; i--;
    }
  
}


// Convert a configuration into a number, allowing multiple states.
long config2number2(int Nvars, int *Nstates, int *configuration)
{
  int i, val, c;
  
  i=0; val=0;
  while (i<=(Nvars-1))
    {
      //if (*(configuration+i)==-1)
      //c=0;
      //else
      //c=1;
      
      c=*(configuration+i);

      //val=val*2+c;

      val=val*(*(Nstates+i))+c;

      i++;
    }
  return val;
}


// Create a potential.
struct potential * createpotential(struct potential *phi1)
{
  int i;
  struct potential *phi2;

  phi2=malloc(sizeof(struct potential));
  phi2->Nvars=phi1->Nvars;
  phi2->varinds=malloc(sizeof(int)*phi2->Nvars);
  for (i=0; i<=(phi2->Nvars-1); i++)
    *(phi2->varinds+i)=*(phi1->varinds+i);
  phi2->Nconfigs=phi1->Nconfigs;
  phi2->Nstates=malloc(sizeof(int)*phi2->Nvars);
  for (i=0; i<=(phi2->Nvars-1); i++)
    *(phi2->Nstates+i)=*(phi1->Nstates+i);
  phi2->probs=malloc(sizeof(double)*phi2->Nconfigs);
  for (i=0; i<=(phi2->Nconfigs-1); i++)
    *(phi2->probs+i)=*(phi1->probs+i);

  return phi2;
}

// Copy a potential from phi1 to phi2.
void copypotential(struct potential *phi2, struct potential *phi1)
{
  int i;
  
  phi2->Nvars=phi1->Nvars;  
  phi2->varinds=realloc(phi2->varinds,sizeof(int)*phi2->Nvars);
  for (i=0; i<=(phi2->Nvars-1); i++)
    *(phi2->varinds+i)=*(phi1->varinds+i);
  phi2->Nconfigs=phi1->Nconfigs;
  phi2->Nstates=realloc(phi2->Nstates,sizeof(int)*phi2->Nvars);
  for (i=0; i<=(phi2->Nvars-1); i++)
    *(phi2->Nstates+i)=*(phi1->Nstates+i);
  phi2->probs=realloc(phi2->probs,sizeof(double)*phi2->Nconfigs);
  for (i=0; i<=(phi2->Nconfigs-1); i++)
    *(phi2->probs+i)=*(phi1->probs+i);
}


// Compute the product of potential functions.
void potentialproduct(struct potential *phi1, struct potential *phi2)
{
  int i, j, Nv, Nv1, Nv2, *ref2v1, *ref2v2, *ref2v, *v2ref, *config1, *config2, *config, maxind;
  double Z;
  struct potential *phi3;
  int *isinv1, *isinv2, *isinv;
  
  // Find the union of potential variables and the corresponding indices of variables in each potential function.
  Nv=0; Nv1=phi1->Nvars; Nv2=phi2->Nvars; 
  
  maxind=0;
  for (i=0; i<=(phi1->Nvars-1); i++)
    if (*(phi1->varinds+i)>maxind)
      maxind=*(phi1->varinds+i);
  for (i=0; i<=(phi2->Nvars-1); i++)
    if (*(phi2->varinds+i)>maxind)
      maxind=*(phi2->varinds+i);
  
  maxind++;
  v2ref=malloc(sizeof(int));
  ref2v=malloc(sizeof(int)*maxind);
  ref2v1=malloc(sizeof(int)*maxind);
  ref2v2=malloc(sizeof(int)*maxind);
  for (i=0; i<=(maxind-1); i++)
    {
      *(ref2v+i)=-1; *(ref2v1+i)=-1; *(ref2v2+i)=-1; 
    }
  
  isinv1=malloc(sizeof(int)*maxind);
  isinv2=malloc(sizeof(int)*maxind);
  isinv=malloc(sizeof(int)*maxind);
  for (i=0; i<=(maxind-1); i++)
    {
      *(isinv1+i)=0;
      *(isinv2+i)=0;
      *(isinv+i)=0;
    }
  for (i=0; i<=(phi1->Nvars-1); i++)
    {
      *(isinv1+*(phi1->varinds+i))=1;
      *(isinv+*(phi1->varinds+i))=1;
    }
  for (i=0; i<=(phi2->Nvars-1); i++)
    {
      *(isinv2+*(phi2->varinds+i))=1;
      *(isinv+*(phi2->varinds+i))=1;
    }
  Nv1=0; Nv2=0; Nv=0;
  for (i=0; i<=(maxind-1); i++)
    {
      if (*(isinv1+i)==1)
	{
	  Nv1++; *(ref2v1+i)=Nv1-1;
	}
      if (*(isinv2+i)==1)
	{
	  Nv2++; *(ref2v2+i)=Nv2-1;
	}
      if (*(isinv+i)==1)
	{
	  Nv++;
	  v2ref=realloc(v2ref,sizeof(int)*Nv);
	  *(v2ref+Nv-1)=i; *(ref2v+i)=Nv-1;
	}
    }
  free(isinv1); free(isinv2); free(isinv);
  

  /*
  Nv1=0; Nv2=0;
  for (i=0; i<=(maxind-1); i++)
    {
      if (inlist(phi1->Nvars,phi1->varinds,i)==1)
	{
	  Nv1++; *(ref2v1+i)=Nv1-1;
	}
      if (inlist(phi2->Nvars,phi2->varinds,i)==1)
	{
	  Nv2++; *(ref2v2+i)=Nv2-1;
	}
      if ((inlist(phi1->Nvars,phi1->varinds,i)==1)||(inlist(phi2->Nvars,phi2->varinds,i)==1))
	{
	  Nv++;
	  v2ref=realloc(v2ref,sizeof(int)*Nv);
	  *(v2ref+Nv-1)=i; *(ref2v+i)=Nv-1;
	}
    }
  */

  
  phi3=malloc(sizeof(struct potential));
  phi3->Nvars=Nv;
  phi3->varinds=malloc(sizeof(int)*Nv);
  for (i=0; i<=(Nv-1); i++)
    *(phi3->varinds+i)=*(v2ref+i);
  phi3->Nconfigs=1;
  //for (i=1; i<=Nv; i++)
  //phi3->Nconfigs*=2;


  phi3->Nstates=malloc(sizeof(int)*phi3->Nvars);
  for (i=0; i<=(Nv1-1); i++)
    *(phi3->Nstates+*(ref2v+*(phi1->varinds+i)))=*(phi1->Nstates+i);
  for (i=0; i<=(Nv2-1); i++)
    *(phi3->Nstates+*(ref2v+*(phi2->varinds+i)))=*(phi2->Nstates+i);

  phi3->Nconfigs=1;
  for (i=0; i<=(Nv-1); i++)
    phi3->Nconfigs*=*(phi3->Nstates+i);


  phi3->probs=malloc(sizeof(double)*phi3->Nconfigs);
  
  config=malloc(sizeof(int)*Nv);
  config1=malloc(sizeof(int)*Nv1);
  config2=malloc(sizeof(int)*Nv2);
  Z=0;
  for (i=0; i<=(phi3->Nconfigs-1); i++)
    {
      int ind1, ind2;

      number2config2(phi3->Nvars,phi3->Nstates,i,config);

      for (j=0; j<=(Nv1-1); j++)
	*(config1+j)=*(config+*(ref2v+*(phi1->varinds+j)));
      for (j=0; j<=(Nv2-1); j++)
	*(config2+j)=*(config+*(ref2v+*(phi2->varinds+j)));
      ind1=config2number2(phi1->Nvars,phi1->Nstates,config1);
      ind2=config2number2(phi2->Nvars,phi2->Nstates,config2);

      *(phi3->probs+i)=*(phi1->probs+ind1)*(*(phi2->probs+ind2));
      Z+=*(phi3->probs+i);
    }

  // Normalize the table.  May not be needed if we are not interested in true probability.
  for (i=0; i<=(phi3->Nconfigs-1); i++)
    *(phi3->probs+i)/=Z;      

  copypotential(phi1,phi3);
  
  free(ref2v1); free(ref2v2); free(ref2v); free(v2ref); 
  free(config1); free(config2);
  free(config); free(phi3->varinds); free(phi3->Nstates); free(phi3->probs);
  free(phi3);
  
}
 

// Compute the product of potential functions.
// Difference from potentialproduct: single and identical variables.
void potentialproduct2(struct potential *phi1, struct potential *phi2)
{
  int i;
  double Z, val;
  
  Z=0;
  for (i=0; i<=(phi1->Nconfigs-1); i++)
    {
      val=*(phi1->probs+i);
      val*=(*(phi2->probs+i));
      *(phi1->probs+i)=val;
      Z+=val;
    }
  for (i=0; i<=(phi1->Nconfigs-1); i++)
    *(phi1->probs+i)/=Z;
  
}


// Compute the product of potential functions.
// Difference from potentialproduct2: phi1 contains multiple variables and phi2 contains a single variable.
void potentialproduct3(struct potential *phi1, struct potential *phi2)
{
  int i, j, ind, *config;
  double Z;

  // Find the index of phi2's only variable in phi1.
  ind=-1; i=0;
  while ((i<=(phi1->Nvars-1))&&(ind==-1))
    {
      if (*(phi1->varinds+i)==*(phi2->varinds))
	ind=i;
      i++;
    }

  if ((ind==-1)||(phi2->Nvars>1))
    {
      printf("Problem in multiplication. ");
      return;
    }

  // Multiply phi1 by phi2.
  config=malloc(sizeof(int)*phi1->Nvars);
  for (i=0; i<=(phi1->Nvars-1); i++)
    *(config+i)=0;
  i=0; Z=0;
  while (i<=(phi1->Nconfigs-1))
    {
      int indval, carry;

      // Check the value of v_{ind} in config.
      indval=*(config+ind);
      *(phi1->probs+i)*=*(phi2->probs+indval);
      Z+=*(phi1->probs+i);

      // Increment the configuration.
      carry=1; j=phi1->Nvars-1;
      do
	{
	  *(config+j)+=carry;
	  carry=0;
	  if (*(config+j)>(*(phi1->Nstates+j)-1))
	    {
	      *(config+j)=0; carry=1;
	    }
	  j--;
	} while ((carry==1)&&(j>=0));
      
      i++;
    }

  // Renormalize the entries.
  for (i=0; i<=(phi1->Nconfigs-1); i++)
    *(phi1->probs+i)/=Z;

  free(config);

  return;
}




// Maximize a potential function over the configurations of some variables.
// Notice the sum over remaining variables is not equal to 1.
void maximizepotential(struct potential *phi, int Nv, int *v)
{
  int i, j, k, *old2new, *new2old, *sconfig1, *sconfig2, *lconfig; 
  int Nrmvconfigs, *Nrmvstates, *old2rmv, *rmv2old;
  double Z;
  struct potential *newphi;

  newphi=malloc(sizeof(struct potential));
  newphi->Nvars=phi->Nvars-Nv;
  newphi->varinds=malloc(sizeof(int)*newphi->Nvars);
  old2new=malloc(sizeof(int)*phi->Nvars);
  new2old=malloc(sizeof(int)*newphi->Nvars);
  for (i=0; i<=(phi->Nvars-1); i++)
    *(old2new+i)=-1;
  j=0; Nrmvconfigs=1; Nrmvstates=malloc(sizeof(int)*Nv); k=0;
  old2rmv=malloc(sizeof(int)*phi->Nvars);
  rmv2old=malloc(sizeof(int)*Nv);
  for (i=0; i<=(phi->Nvars-1); i++)
    *(old2rmv+i)=-1;

  for (i=0; i<=(phi->Nvars-1); i++)
    {
      if (inlist(Nv,v,*(phi->varinds+i))==0)
	{
	  j++; *(old2new+i)=j-1; *(new2old+j-1)=i;
	}
      else
	{
	  k++;
	  Nrmvconfigs*=*(phi->Nstates+i);
	  *(Nrmvstates+k-1)=*(phi->Nstates+i);
	  *(old2rmv+i)=k-1; *(rmv2old+k-1)=i;
	}
    }

  for (i=0; i<=(newphi->Nvars-1); i++)
    *(newphi->varinds+i)=*(phi->varinds+*(new2old+i));
  newphi->Nstates=malloc(sizeof(int)*newphi->Nvars);
  for (i=0; i<=(newphi->Nvars-1); i++)
    *(newphi->Nstates+i)=*(phi->Nstates+*(new2old+i));
  newphi->Nconfigs=1;
  for (i=0; i<=(newphi->Nvars-1); i++)
    newphi->Nconfigs*=*(newphi->Nstates+i);
  newphi->probs=malloc(sizeof(double)*newphi->Nconfigs);

  sconfig1=malloc(sizeof(int)*newphi->Nvars);
  sconfig2=malloc(sizeof(int)*Nv);
  lconfig=malloc(sizeof(int)*phi->Nvars);
  
  // For each configuration in newphi, find all consistent configurations in phi and the max potential value over these configurations.
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    {
      int ind;
      double maxval, val;

      // Fix sconfig1 and vary sconfig2.
      number2config2(newphi->Nvars,newphi->Nstates,i,sconfig1);
      for (j=0; j<=(newphi->Nvars-1); j++)
	*(lconfig+*(new2old+j))=*(sconfig1+j);
      maxval=-1;
      for (j=0; j<=(Nrmvconfigs-1); j++)
	{
	  number2config2(Nv,Nrmvstates,j,sconfig2);
	  for (k=0; k<=(Nv-1); k++)
	    *(lconfig+*(rmv2old+k))=*(sconfig2+k);
	  ind=config2number2(phi->Nvars,phi->Nstates,lconfig);
	  val=*(phi->probs+ind);
	  if (val>maxval)
	    maxval=val;

	}
      *(newphi->probs+i)=maxval;
    }


  // Normalize the values.
  Z=0;
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    Z+=*(newphi->probs+i);
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    *(newphi->probs+i)/=Z;


  copypotential(phi,newphi);

  free(old2new); free(new2old); free(sconfig1); free(sconfig2);
  free(lconfig); free(Nrmvstates); free(old2rmv); free(rmv2old);
  free(newphi->varinds); free(newphi->Nstates); free(newphi->probs);
}




// Maximize a potential function over one variable.
// Difference from maximizepotential: marginalize over one variable.
void maximizepotential2(struct potential *phi, int varind)
{
  int i, j, k, ind, *sconfig, Nv1, Nv2, num, num1, num2, shift1, shift2, flag;
  double Z, val, maxval;
  struct potential *newphi;


  ind=-1; i=0;
  while ((ind<0)&&(i<=(phi->Nvars-1)))
    {
      if (*(phi->varinds+i)==varind)
	ind=i;
      i++;
    }
  
  if (ind<0)
    return;

  Nv1=0; Nv2=0; 
  for (i=0; i<=(phi->Nvars-1); i++)
    {
      if (i<ind)
	{
	  Nv1++; 
	}
      else if (i>ind)
	{
	  Nv2++; 
	}
    }

  shift1=1;
  for (i=ind; i<=(phi->Nvars-1); i++)
    shift1*=*(phi->Nstates+i);
  shift2=shift1/(*(phi->Nstates+ind));

  sconfig=malloc(sizeof(int)*(phi->Nvars-1));
  for (i=0; i<=(phi->Nvars-2); i++)
    *(sconfig+i)=0;

  newphi=malloc(sizeof(struct potential));
  newphi->Nvars=phi->Nvars-1;
  newphi->varinds=malloc(sizeof(int)*newphi->Nvars);
  newphi->Nstates=malloc(sizeof(int)*newphi->Nvars);
  j=0;
  for (i=0; i<=(phi->Nvars-1); i++)
    {
      if (i!=ind)
	{
	  *(newphi->varinds+j)=*(phi->varinds+i);
	  *(newphi->Nstates+j)=*(phi->Nstates+i);
	  j++;
	}
    }
  newphi->Nconfigs=phi->Nconfigs/(*(phi->Nstates+ind));
  newphi->probs=malloc(sizeof(double)*newphi->Nconfigs);

  // Debug
  if (newphi->Nconfigs<=0)
    {
      printf("phi->varinds=");
      for (i=0; i<=(phi->Nvars-1); i++)
	printf("%d ",*(phi->varinds+i));
      printf("varind=%d ",varind);
    }


  flag=0; Z=0;
  
  while (flag==0)
    {
      int carry=0;
            
      // Find the first value with the designated marginal values.
      if (Nv1>0)
	num1=config2number2(Nv1,newphi->Nstates,sconfig);
      else
	num1=0;
      if (Nv2>0)
	num2=config2number2(Nv2,newphi->Nstates+Nv1,sconfig+Nv1);
      else 
	num2=0;
      num=num1*shift1+num2;
      maxval=-1;
      

      for (i=0; i<=(*(phi->Nstates+ind)-1); i++)
	{	  
	  val=*(phi->probs+num);
	  if (val>maxval)
	    maxval=val;
	  num+=shift2;
	}
      num=num1*shift2+num2;

      
      // Debug
      if ((num>=newphi->Nconfigs)||(num<0))
	{
	  printf("phi->varinds=");
	  for (i=0; i<=(phi->Nvars-1); i++)
	    printf("%d ",*(phi->varinds+i));
	  printf("varind=%d ",varind);
	  printf("Nconfigs=%d, num=%d \n",newphi->Nconfigs,num);
	}


      *(newphi->probs+num)=maxval;

      Z+=maxval;
      

      i=newphi->Nvars-1;
      *(sconfig+i)+=1;
      if (*(sconfig+i)>=*(newphi->Nstates+i))
	{
	  carry=1;
	  *(sconfig+i)=0;
	}
	

      i--;
      while ((carry==1)&&(i>=0))
	{
	  *(sconfig+i)+=carry;
	  if (*(sconfig+i)<*(newphi->Nstates+i))
	    carry=0;
	  else
	    *(sconfig+i)=0;
	  i--;
	}


      flag=1; i=0;
      while ((i<=(newphi->Nvars-1))&&(flag==1))
	{
	  if (*(sconfig+i)>0)
	    flag=0;
	  i++;
	}

    }

  
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    *(newphi->probs+i)/=Z;
    

  copypotential(phi,newphi);
  free(newphi->varinds); free(newphi->Nstates); free(newphi->probs);
  free(newphi);
  free(sconfig);
}


// Maximize a potential function over (n-1) variables, which makes a potential function of one variable.
// Difference from maximizepotential2: the output potential function has one variable, the varind is the variable remained after maximization.
void maximizepotential3(struct potential *phi, int varind)
{
  int i, j, ind, *config;
  double Z;
  struct potential *newphi;

  // Find the index of varind in phi.
  ind=-1; i=0;
  while ((i<=(phi->Nvars-1))&&(ind==-1))
    {
      if (*(phi->varinds+i)==varind)
	ind=i;
      i++;
    }
  if (ind==-1)
    return;
  
  newphi=malloc(sizeof(struct potential));
  newphi->Nvars=1; newphi->varinds=malloc(sizeof(int));
  *(newphi->varinds)=varind;
  newphi->Nstates=malloc(sizeof(int)); *(newphi->Nstates)=*(phi->Nstates+ind);
  newphi->Nconfigs=*(newphi->Nstates);
  newphi->probs=malloc(sizeof(double)*newphi->Nconfigs);
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    *(newphi->probs+i)=0.0;

  config=malloc(sizeof(int)*phi->Nvars);
  for (i=0; i<=(phi->Nvars-1); i++)
    *(config+i)=0;
  i=0;
  while (i<=(phi->Nconfigs-1))
    {
      int indval, carry;

      // Check the value of v_{ind} in config.
      indval=*(config+ind);
      
      // If the potential value > newphi.probs[indval], then update it.
      if (*(phi->probs+i)>*(newphi->probs+indval))
	*(newphi->probs+indval)=*(phi->probs+i);

      // Increment the configuration.
      carry=1; j=phi->Nvars-1;
      do
	{
	  *(config+j)+=carry;
	  carry=0;
	  if (*(config+j)>(*(phi->Nstates+j)-1))
	    {
	      *(config+j)=0; carry=1;
	    }
	  j--;
	} while ((carry==1)&&(j>=0));
      
      i++;
    }

  // Renormalize the entries.
  Z=0;
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    Z+=*(newphi->probs+i);
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    *(newphi->probs+i)/=Z;

  copypotential(phi,newphi);
  free(newphi->varinds); free(newphi->Nstates); free(newphi->probs);
  free(newphi); free(config);
}



// Marginalize a potential function over the configurations of some variables.
void marginalizepotential(struct potential *phi, int Nv, int *v)
{
  int i, j, k, *old2new, *new2old, *sconfig1, *sconfig2, *lconfig; 
  int Nrmvconfigs, *Nrmvstates, *old2rmv, *rmv2old;
  double Z;
  struct potential *newphi;

  newphi=malloc(sizeof(struct potential));
  newphi->Nvars=phi->Nvars-Nv;
  newphi->varinds=malloc(sizeof(int)*newphi->Nvars);
  old2new=malloc(sizeof(int)*phi->Nvars);
  new2old=malloc(sizeof(int)*newphi->Nvars);
  for (i=0; i<=(phi->Nvars-1); i++)
    *(old2new+i)=-1;
  j=0; Nrmvconfigs=1; Nrmvstates=malloc(sizeof(int)*Nv); k=0;
  old2rmv=malloc(sizeof(int)*phi->Nvars);
  rmv2old=malloc(sizeof(int)*Nv);
  for (i=0; i<=(phi->Nvars-1); i++)
    *(old2rmv+i)=-1;

  for (i=0; i<=(phi->Nvars-1); i++)
    {
      if (inlist(Nv,v,*(phi->varinds+i))==0)
	{
	  j++; *(old2new+i)=j-1; *(new2old+j-1)=i;
	}
      else
	{
	  k++;
	  Nrmvconfigs*=*(phi->Nstates+i);
	  *(Nrmvstates+k-1)=*(phi->Nstates+i);
	  *(old2rmv+i)=k-1; *(rmv2old+k-1)=i;
	}
    }

  for (i=0; i<=(newphi->Nvars-1); i++)
    *(newphi->varinds+i)=*(phi->varinds+*(new2old+i));
  newphi->Nstates=malloc(sizeof(int)*newphi->Nvars);
  for (i=0; i<=(newphi->Nvars-1); i++)
    *(newphi->Nstates+i)=*(phi->Nstates+*(new2old+i));
  newphi->Nconfigs=1;
  for (i=0; i<=(newphi->Nvars-1); i++)
    newphi->Nconfigs*=*(newphi->Nstates+i);
  newphi->probs=malloc(sizeof(double)*newphi->Nconfigs);

  sconfig1=malloc(sizeof(int)*newphi->Nvars);
  sconfig2=malloc(sizeof(int)*Nv);
  lconfig=malloc(sizeof(int)*phi->Nvars);
  
  // For each configuration in newphi, find all consistent configurations in phi and the sum of potential value over these configurations.
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    {
      int ind;
      double sumval, val;

      // Fix sconfig1 and vary sconfig2.
      number2config2(newphi->Nvars,newphi->Nstates,i,sconfig1);
      for (j=0; j<=(newphi->Nvars-1); j++)
	*(lconfig+*(new2old+j))=*(sconfig1+j);
      sumval=0; 
      for (j=0; j<=(Nrmvconfigs-1); j++)
	{
	  number2config2(Nv,Nrmvstates,j,sconfig2);
	  for (k=0; k<=(Nv-1); k++)
	    *(lconfig+*(rmv2old+k))=*(sconfig2+k);
	  ind=config2number2(phi->Nvars,phi->Nstates,lconfig);
	  val=*(phi->probs+ind);
	  sumval+=val;
	}
      *(newphi->probs+i)=sumval;
    }


  // Normalize the values.
  Z=0;
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    Z+=*(newphi->probs+i);
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    *(newphi->probs+i)/=Z;


  copypotential(phi,newphi);

  free(old2new); free(new2old); free(sconfig1); free(sconfig2);
  free(lconfig); free(Nrmvstates); free(old2rmv); free(rmv2old);
  free(newphi->varinds); free(newphi->Nstates); free(newphi->probs); free(newphi);
}


// Mariginalize a potential function over one variable.
// Difference from marginalizepotential: marginalize over one variable.
void marginalizepotential2(struct potential *phi, int varind)
{
  int i, j, k, ind, *sconfig, Nv1, Nv2, num, num1, num2, shift1, shift2, flag;
  double Z, val, sumval;
  struct potential *newphi;

  ind=-1; i=0;
  while ((ind<0)&&(i<=(phi->Nvars-1)))
    {
      if (*(phi->varinds+i)==varind)
	ind=i;
      i++;
    }
  
  if (ind<0)
    return;

  Nv1=0; Nv2=0; 
  for (i=0; i<=(phi->Nvars-1); i++)
    {
      if (i<ind)
	{
	  Nv1++; 
	}
      else if (i>ind)
	{
	  Nv2++; 
	}
    }

  shift1=1;
  for (i=ind; i<=(phi->Nvars-1); i++)
    shift1*=*(phi->Nstates+i);
  shift2=shift1/(*(phi->Nstates+ind));

  sconfig=malloc(sizeof(int)*(phi->Nvars-1));
  for (i=0; i<=(phi->Nvars-2); i++)
    *(sconfig+i)=0;

  newphi=malloc(sizeof(struct potential));
  newphi->Nvars=phi->Nvars-1;
  newphi->varinds=malloc(sizeof(int)*newphi->Nvars);
  newphi->Nstates=malloc(sizeof(int)*newphi->Nvars);
  j=0;
  for (i=0; i<=(phi->Nvars-1); i++)
    {
      if (i!=ind)
	{
	  *(newphi->varinds+j)=*(phi->varinds+i);
	  *(newphi->Nstates+j)=*(phi->Nstates+i);
	  j++;
	}
    }
  newphi->Nconfigs=phi->Nconfigs/(*(phi->Nstates+ind));
  newphi->probs=malloc(sizeof(double)*newphi->Nconfigs);

  flag=0; Z=0;
  
  while (flag==0)
    {
      int carry=0;
            
      // Find the first value with the designated marginal values.
      if (Nv1>0)
	num1=config2number2(Nv1,newphi->Nstates,sconfig);
      else
	num1=0;
      if (Nv2>0)
	num2=config2number2(Nv2,newphi->Nstates+Nv1,sconfig+Nv1);
      else 
	num2=0;
      num=num1*shift1+num2;
      
      sumval=0;
      for (i=0; i<=(*(phi->Nstates+ind)-1); i++)
	{	  
	  val=*(phi->probs+num);
	  sumval+=val;
	  num+=shift2;
	}
      num=num1*shift2+num2;


      *(newphi->probs+num)=sumval;

      Z+=sumval;
      

      i=newphi->Nvars-1;
      *(sconfig+i)+=1;
      if (*(sconfig+i)>=*(newphi->Nstates+i))
	{
	  carry=1;
	  *(sconfig+i)=0;
	}
	

      i--;
      while ((carry==1)&&(i>=0))
	{
	  *(sconfig+i)+=carry;
	  if (*(sconfig+i)<*(newphi->Nstates+i))
	    carry=0;
	  else
	    *(sconfig+i)=0;
	  i--;
	}


      flag=1; i=0;
      while ((i<=(newphi->Nvars-1))&&(flag==1))
	{
	  if (*(sconfig+i)>0)
	    flag=0;
	  i++;
	}

    }

  
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    *(newphi->probs+i)/=Z;
    

  copypotential(phi,newphi);
  free(newphi->varinds); free(newphi->Nstates); free(newphi->probs);
  free(sconfig); free(newphi);
}
  

// Marginalize a potential function over (n-1) variables, which makes a potential function of one variable.
// Difference from maximizepotential2: the output potential function has one variable, the varind is the variable remained after marginalization.
void marginalizepotential3(struct potential *phi, int varind)
{
  int i, j, ind, *config;
  double Z;
  struct potential *newphi;

  // Find the index of varind in phi.
  ind=-1; i=0;
  while ((i<=(phi->Nvars-1))&&(ind==-1))
    {
      if (*(phi->varinds+i)==varind)
	ind=i;
      i++;
    }
  if (ind==-1)
    return;
  
  newphi=malloc(sizeof(struct potential));
  newphi->Nvars=1; newphi->varinds=malloc(sizeof(int));
  *(newphi->varinds)=varind;
  newphi->Nstates=malloc(sizeof(int)); *(newphi->Nstates)=*(phi->Nstates+ind);
  newphi->Nconfigs=*(newphi->Nstates);
  newphi->probs=malloc(sizeof(double)*newphi->Nconfigs);
  for (i=0; i<=(newphi->Nconfigs-1); i++)
    *(newphi->probs+i)=0.0;

  config=malloc(sizeof(int)*phi->Nvars);
  for (i=0; i<=(phi->Nvars-1); i++)
    *(config+i)=0;
  i=0;
  while (i<=(phi->Nconfigs-1))
    {
      int indval, carry;

      // Check the value of v_{ind} in config.
      indval=*(config+ind);
      
      // If the potential value > newphi.probs[indval], then update it.
      //if (*(phi->probs+i)>*(newphi->probs+indval))
      //*(newphi->probs+indval)=*(phi->probs+i);

      // Add the new potential value.
      *(newphi->probs+indval)+=*(phi->probs+i);

      // Increment the configuration.
      carry=1; j=phi->Nvars-1;
      do
	{
	  *(config+j)+=carry;
	  carry=0;
	  if (*(config+j)>(*(phi->Nstates+j)-1))
	    {
	      *(config+j)=0; carry=1;
	    }
	  j--;
	} while ((carry==1)&&(j>=0));
      
      i++;
    }

  // No need to normalize a marginal probability.

  copypotential(phi,newphi);
  free(newphi->varinds); free(newphi->Nstates); free(newphi->probs);
  free(newphi); free(config);
}


// Convert a pair of nodes to the message index.
int pair2msg(int Nmsgs, struct message *msg, struct pair *node1, struct pair *node2)
{
  int i, flag;

  // If node directions are not alternating, then return -1.
  if (node1->j==node2->j)
    return -1;

  i=0; flag=0;
  while ((i<=(Nmsgs-1))&&(flag==0))
    {
      if (((msg+i)->dir==0)&&(node1->j==0)&&((msg+i)->from==node1->i)&&((msg+i)->to==node2->i))
	flag=1;
      else if (((msg+i)->dir==1)&&(node1->j==1)&&((msg+i)->from==node1->i)&&((msg+i)->to==node2->i))
	flag=1;
      i++;
    }
  i--;

  if (flag==1)
    return i;
  else
    return -1;
}
