// Utility functions of factor graph belief propagation and max configuration determination.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include "integration.h"
# include "integration3.h"
# include "connect.h"
# include "cluster.h"
# define MAXLINEWIDTH 1000
# define INFINITY 1e38
# define POSINF 100000
# define max(i,j) (i>j ? i : j)

extern int inlist(int Nv, int *v, int q);
extern struct potential * createpotential(struct potential *phi1);
extern void copypotential(struct potential *phi2, struct potential *phi1);
extern void potentialproduct(struct potential *phi1, struct potential *phi2);
extern void potentialproduct2(struct potential *phi1, struct potential *phi2);
extern void maximizepotential(struct potential *phi, int Nv, int *v);
extern int isintersect(int N1, int *v1, int N2, int *v2);
extern struct compresspath * findfacgpath(struct facgraph *fgraph, int pathlength, int *Nvalidpaths);
extern int pair2msg(int Nmsgs, struct message *msg, struct pair *node1, struct pair *node2);
extern int issubset(int Nv1, int *v1, int Nv2, int *v2);
extern void maxproduct2(struct facgraph *fgraph, struct message *msg, struct potential *belief);
extern void maxproduct3(struct facgraph *fgraph, struct message *msg, struct potential *belief);
extern void maxproduct4(struct facgraph *fgraph, struct message *msg, int *isfixed, struct potential *belief);
extern long config2number2(int Nvars, int *Nstates, int *configuration);
extern void number2config2(int Nvars, int *Nstates, long val, int *configuration);
extern void maximizepotential2(struct potential *phi, int ind);
extern void marginalizepotential(struct potential *phi, int Nv, int *v);
extern void marginalizepotential2(struct potential *phi, int varind);

extern void maxproduct7(struct facgraph *fgraph, int Nhiddenvars, int *Nhiddenvarstates, struct message *msg, int *fixedvals, struct potential *belief);
extern void sumproduct1(struct facgraph *fgraph, int Nhiddenvars, int *Nhiddenvarstates, struct message *msg, int *fixedvals, struct potential *belief);
extern int * findmaxconfig5(struct facgraph *fgraph, struct message *msg, int Nedgevars, struct potential *belief);
extern int * findmaxconfig6(struct facgraph *fgraph, struct message *maxmsg, struct message *summsg, int Nedgevars, struct potential *maxbelief, struct potential *sumbelief);
extern int maxpathlength;
extern void maxproduct8(struct facgraph *fgraph, int Nhiddenvars, int *Nhiddenvarstates, struct message *msg, int *fixedvals, struct potential *belief);

extern void debmaxproduct7(struct facgraph *fgraph, int Nhiddenvars, int *Nhiddenvarstates, struct message *msg, int *fixedvals, struct potential *belief, int debug);


void recurse_fixval(struct confignode *configtree, struct facgraph *fgraph, int Nhiddenvars, int *Nhiddenvarstates, struct message *maxmsg, struct potential *maxbelief, struct message *summsg, struct potential *sumbelief, int Nedgevars);
int * recurse_getmapconfigs(struct confignode *configtree, int *Nmapconfigs, int *mapconfigs);
void recurse_freeconfignode(struct confignode *configtree);


// Recursively obtain the MAP configurations from the configtree.
// Only report the configuration when reaching the leaf nodes.
int * recurse_getmapconfigs(struct confignode *configtree, int *Nmapconfigs, int *mapconfigs)
{
  int i, Nvars, *configuration;

  Nvars=configtree->Nvars; configuration=configtree->config;
  
  // If the leaf node is reached, then report the configuration and return.
  if (configtree->Nchildren==0)
    {
      *Nmapconfigs+=1;

      mapconfigs=realloc(mapconfigs,sizeof(int)*(*Nmapconfigs)*Nvars);
      for (i=0; i<=(Nvars-1); i++)
	*(mapconfigs+(*Nmapconfigs-1)*Nvars+i)=*(configuration+i);

      // Debug
      //printf("recurse_getmapconfigs node ind %d \n",configtree->ind);

      return mapconfigs;
    }


  // Recursively go to its children.
  //mapconfigs=recurse_getmapconfigs(configtree->children,Nmapconfigs,mapconfigs);
  //mapconfigs=recurse_getmapconfigs(configtree->children+1,Nmapconfigs,mapconfigs);


  for (i=0; i<=(configtree->Nchildren-1); i++)
    mapconfigs=recurse_getmapconfigs(configtree->children+i,Nmapconfigs,mapconfigs);

  return mapconfigs;

}



// Recursively free the memory of a confignode structure. 
void recurse_freeconfignode(struct confignode *configtree)
{
  int i;

  free(configtree->config);

  if (configtree->Nchildren==0)
    return;

  for (i=0; i<=(configtree->Nchildren-1); i++)
    recurse_freeconfignode(configtree->children+i);  

  //recurse_freeconfignode(configtree->children);
  // recurse_freeconfignode(configtree->children+1);
  
  free(configtree->children);

  return;
}


