struct edge {
  long from;  // one end of the edge.
  long to;    // the other end of the edge.
  long weight; // edge weight.
  int delay;  // edge delay.
  int sign;   // the sign of edge interaction.
};

struct sparsegraph {
  long Nv;  // number of nodes.
  long *v;  // nodes.
  long Ne;  // number of edges.
  struct edge *e;  // edges.
  long *nodeweight; // node weights.
  long *nodetime;  // node time.
  struct clique *nodeclus; // clusters associated with each node.
};

struct bipath {
  long Nv;  // number of nodes in a path.
  long *v;  // nodes in order of a path.
  int *dir; // directionality of each edge along the path.
  double *weight; // weight score of each edge along the path.
  int *sign; // sign of each edge.
};

struct compresspath {
  long start;   // start node.
  long pathid;  // the index of the intermediate path.
  long end;     // end node.
  int  lastdir; // the direction of the last edge along the path, +1 if point toward the end node, -1 is point away from the end node.
  double weight;  // the weight score of the path.
  int delay;   // the delay of the path.
  int length;  // the length of the path.
  int lastsign;    // the sign of the last edge in the path.
};

struct pathway {
  long Nsupernodes;  // Number of supernodes.
  long *supernodelookup;  // The mapping from node indices to supernode indices.
  struct sparsegraph *graph;  // The graph embedded in the pathway.
  struct cluster *regulated;  // Genes regulated by each supernode.
  long Nexplained;   // Number of explained pairs.
  struct pair *explained;  // Explained pairs.
};

struct pathequation {
  long ind;    // Index of path equation.  Equals to node index if path length is 0.
  long Nvars;  // Number of variables in the equation.
  int *coeff;  // Coefficients of the equation.
  long Nexplained; // Number of correlated pairs explained by this equation.
  long *explained; // Indices of explained pairs.
};

struct pathequationset {
  long Nequations;  // Number of equations.
  long *eqnind;     // Equation indices.
  long Nvars;       // Number of variables.
  struct matrix *coeffmat;   // Coefficient matrix.
  long Nexplained;  // Number of explained pairs.
  long *explained; // Indices of explained pairs.
};

/*
struct sparsegraphcluster {
  long Nv;  // number of nodes.
  long *v;  // nodes.
  long Ne;  // number of edges.
  struct edge *e;  // edges.
  long *nodeweight; // node weights.
  long *nodetime;  // node time.
  struct clique *nodeclus; // clusters associated with each node.
};
*/

struct equation {
  long ind;    // Index of path equation.  Equals to node index if path length is 0.
  long pathind; // Index of path.
  long Nvars;  // Number of variables in the equation.
  int *coeff;  // Coefficients of the equation.
  long Nexplained; // Number of correlated pairs explained by this equation.
  struct pair *explainedpairs; // Explained pairs.
  double covsum; // Sum of square covariances corresponding to the equation.
};

struct equationset {
  long Nequations;  // Number of equations.
  long *eqnind;     // Equation indices.
  long Nvars;       // Number of variables.
  struct matrix *coeffmat;   // Coefficient matrix.
  long Nexplained;  // Number of explained pairs.
  struct pair *explainedpairs; // Explained pairs.
  double covsum;  // Sum of square covariances corresponding to the equation.
};

struct pairs {
  int Nv;      // number of pairs.
  struct pair *v;  // pairs.
};
