# define PHYSICAL 1
# define FUNCTIONAL 0
# define LOGICAL 0
# define NUMERICAL 1

# define SINGLEPATHEXPLAIN 0
# define PATHSELOR 2
# define SINGLEVAR 1
# define PREDICTION 3

# define EDGEPRESENCE 0
# define EDGESIGN 1
# define EDGEDIR 2
# define KNOCKOUT 3
# define PATHSEL 4
# define OUTPUTRESPONSE 5

/*
// The factor graph data structure including both lookup table and explanation constraints in each factor.
struct mix_facgraph {
  int Nvars;         // number of variable nodes.
  int *vars;         // variable nodes.
  int Ninteractions; // number of factor nodes.
  int *interactiontypes; // types of factor nodes. 
  // 0: explanatory_path type, 1: mutual enhancement type.
  int *interactions; // factor nodes.
  int Ne;            // number of edges.
  struct edge *e;    // edges (from variable nodes to factor nodes).
  struct potential *phivi; // potential functions of factor nodes.
  void *factors;     // factor node functions.  can be either explanatory_path type of mutual_enhance_factor type.
};
*/

// The mixed factor node data structure.  Each factor node contains a potential function of lookup table and logic conditions for explanation.
// This data structure is valid for the factor term of a path.
struct mix_path_fac_node {
  struct potential *phivi;  // potential function of the factor node.
  int type;        // type=0: logical potential function, type=1: numerical potential function, 2: path or function, 3: prediction.
  int Nedges;      // number of edges along the path.
  int Nsigedges;   // number of signal edges along the path.
  int *presences;  // edge presence variables.
  int *signs;      // edge sign variables.
  int *sigdirvars;  // signal edge direction variables.
  int *sigdirvarvals; // preferred signal edge direction variable values.
  int sigma;       // index of the path selection variable.
  int ko;          // index of the knockout effect variable.
  int effective;   // effective=0 if some edges along the prediction path are not selected to explain existing data.
};


// The mixed variable node data structure.  Each variable node contains the following information: the type of the variable, the index of the gene, edge, path or knockout effect of the variable.
struct mix_path_var_node {
  int type;         // 0:edge presence, 1:edge sign, 2:edge direction, 3:knockout effect, 4:path selection, 5:predicted outcome.
  int geneind;      // gene index, valid only for output response var.
  int edgeind;      // edge index, valid for edge presence, sign and dir vars.
  int koind;        // knockout index, valid for ko effect var.
  struct triplet *pathinfo; // ko ind (->i), path var ind (->j) and path ind (->k) of a path selection var.
};


// The mixed factor graph data structure.  A bipartite graph of variable and factor nodes.
struct mix_facgraph {
  int Nvars;          // number of variable nodes.
  int *vars;          // variable nodes.
  int Ninteractions;  // number of factor (interaction) nodes.
  int *interactions;  // interaction nodes.
  int Ne;             // number of edges.
  struct edge *e;     // edges (from variable nodes to factor nodes).
  struct mix_path_fac_node *facnodes;  // factor nodes.

  struct mix_path_var_node *varnodes;  // variable nodes.
};


// The model configurations data structure.  
struct model_configuration {
  int Nvars;          // number of variables.
  int Nedges;         // number of edges.
  int Nsigedges;      // number of signal edges.
  int Nkos;           // number of ko effects.
  int Nusedpaths;     // number of paths used to explain kos.
  int *vartypes;      // variable types.
  int *presence2var;  // lookup table from edge presences to variables.
  int *sign2var;      // lookup table from edge signs to variables.
  int *sigedge2edge;  // lookup table from signal edges to edges.
  int *sigdir2var;    // lookup table from signal edge dirs to variables.
  int *ko2var;        // lookup table from ko indices to variables.
  int *var2pathind;   // lookup table from variables to path indices.
  struct pair *usedpathinfo; // the ko index and the variable index of a used path.
  int Nmapconfigs;    // number of MAP configurations.
  int *mapconfigs;    // MAP configurations.
};

