// The data structure of suggested experiments.
struct experiment {
  int type;       // type of experiments, 0:knockout, 1:over expression.
  int Ngenes;     // number of genes to be perturbed.
  int *perturbed; // indices of perturbed genes.
  int Ndownstreams;  // number of downstream genes.
  int *downstreams;  // downstream genes.
  int Naffected;  // number of affected explained ko pairs.
  double score; // score of the perturbation experiment (here the entropy of output response).
  int Nusedpaths; // number of effective paths linking perturbed genes to downstream genes.
  struct triplet *usedpathinfo; // information about effective paths, ->i: perturbed gene, ->j: affected gene, ->k: path ind.
};


// The data structure of subnetwork model configurations.
struct subnet_model_configuration {
  // Global information about the model variables.
  int Nvars;          // number of variables.
  int Nedges;         // number of edges.
  int Nsigedges;      // number of signal edges.
  int Nkos;           // number of ko effects.
  int Nusedpaths;     // number of paths used to explain kos.
  int *vartypes;      // variable types.
  int *presence2var;  // lookup table from edge presences to variables.
  int *sign2var;      // lookup table from edge signs to variables.
  int *sigedge2edge;  // lookup table from signal edges to edges.
  int *sigdir2var;    // lookup table from signal edge dirs to variables.
  int *ko2var;        // lookup table from ko indices to variables.
  int *var2pathind;   // lookup tablle from variables to path indices.
  struct pair *usedpathinfo; // the ko index and the variable index of a used path.
  // Primary variables -- invariant variables across all MAP configurations.
  int Nprimary;       // number of primary variables.
  int *primaryvar2var;   // lookup table from primary variables to variables.
  int *primaryconfig;    // primary variables configuration, common to all MAP configurations.
  // Subnetwork information.
  int Nsubnets;       // number of subnetworks.
  int **subnetvar2var;  // lookup table of subnetwork variables.
  int *var2subnetmember; // subnet membership indices of model variables.
  int **subnetvar2edge;  // lookup table of subnetwork vars to edge indices.
  int *Nsubnetmapconfigs;  // number of map configurations for each subnet.
  int **subnetmapconfigs;  // subnetwork map configurations.
  // MAP configuration.
  int Nmapconfigs;    // number of MAP configurations.
  int *mapconfigs;    // MAP configurations.
};


// The data structure of directed subgraphs.
struct dirsubgraph {
  int Nv;  // number of nodes.
  int *v;  // node list in topological order.
  int Ne;  // number of edges.
  struct edge *e;  // edge list.
};


// The data structure of decomposed physical models.
// There are three types of variables: invariant, independent, dependent.
struct decomposed_model {
  int Nvars;     // number of variables.
  int *varinds;  // variable indices.
  int *vartypes; // variable types. 0:invariant, 1:independent, 2:dependent.
  int Nedges;    // number of edges in the decomposed graph.
  int *edgeinds; // edge indices.
};


// The data structure of a chi2cdf lookup table.
struct chi2cdf_lookup {
  int ind;      // index.
  double x;     // input value.
  double p;     // chi2cdf value.
  double q;     // 1-p.
};


// The data structure of decomposed models and subconfigurations.
// There are three types of variables: invariant, independent, dependent.
struct decomposed_model_config {
  int Nvars;     // number of variables.
  int *varinds;  // variable indices.
  int *vartypes; // variable types. 0:invariant, 1:independent, 2:dependent.
  int Nedges;    // number of edges in the decomposed graph.
  int *edgeinds; // edge indices.
  int Nmapconfigs; // number of optimal configurations within the submodel.
  int *mapconfigs; // optimal configurations.
};
