// Utility functions of loading various types of data.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include "integration.h"
# include "integration3.h"
# include "connect.h"
# include "cluster.h"
# include "integration5.h"
# define GENENAMELENGTH 9
# define MAXLINEWIDTH 1000
# define INFINITY 1e38
# define POSINF 100000
# define max(i,j) (i>j ? i : j)

extern int inlist(int Nv, int *v, int q);
extern char * getitemval(char *s, int ind, int *length);
extern double atof2(char *s, int length);
extern long * loadmatrix(char *filename);
extern double * loadmatrix2(char *filename);
extern double atof2(char *s, int length);
extern char * getitemval(char *s, int ind, int *length);
void loadgenenames(char *filename, int *N, int *genenamespnt, int *anothergenenamespnt);
void loadselgenes(int N, char *genenames, char *anothergenenames, char *selfilename, int *Nselgenes, int *ref2selpnt, int *sel2refpnt);
void loadlocdata(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *gene2spot, char *explist, double locpthre, char *locdata, int *Nfactors, int *ref2facpnt, int *fac2refpnt, int *Nlocs, int *locpnt);
void loadppdata(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *ppdata, int *Npps, int *pppnt);
void loadkodata(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *kogenefilename, char *expfilename, char *kologratiofilename, char *kopvalfilename, double kopthre, int *isknockedpnt, int *Nkos, int *kopnt);
int ** locdegeneratespots(int N, char *genenames, char *gene2spot);
void loadlocdata3(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *locfilename, double locpthre, int *Nfactors, int *ref2facpnt, int *fac2refpnt, int *Nlocs, int *locpnt);


// These values are determined from the paper "Protein interactions: two methods for assessment of the reliability of high-throughput observations" by Deane et al., MCP.
double ppratio=5, alpha_epr_true_all=0.5, alpha_epr_true_s=0.9, alpha_epr_false_s=0.5, alpha_epr_true_m=0.85, alpha_epr_false_m=0.5, pvm_fp=0.05, pvm_fn=0.5;

//double ppratio=1, alpha_epr_true_all=0.5, alpha_epr_true_s=0.9, alpha_epr_false_s=0.5, alpha_epr_true_m=0.85, alpha_epr_false_m=0.5, pvm_fp=0.05, pvm_fn=0.5;


// Load gene names table.  The outputs are the pointer addresses to genenames and anothergenenames.
void loadgenenames(char *filename, int *N, int *genenamespnt, int *anothergenenamespnt)
{
  int i, j, length, flag;
  char ch, *s, *item, *genenames, *anothergenenames;
  FILE *fp;

  if ((fp=fopen(filename,"r"))==NULL)
    {
      printf("Cannot open %s.\n",filename);
      return;
    }

  // Count the number of rows.
  s=malloc(sizeof(char)*MAXLINEWIDTH);

  *N=0; fseek(fp,0,SEEK_SET);
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    *N+=1;
   
  genenames=malloc(sizeof(char)*(*N)*GENENAMELENGTH);
  anothergenenames=malloc(sizeof(char)*(*N)*GENENAMELENGTH);
  fseek(fp,0,SEEK_SET); i=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      int cnt;
      item=getitemval(s,0,&length);
      strncpy(genenames+i*GENENAMELENGTH,item,length);
      if (length<GENENAMELENGTH)
	for (j=length; j<=(GENENAMELENGTH-1); j++)
	  *(genenames+i*GENENAMELENGTH+j)=' ';
      free(item);

      // If the second name is blank, then fill it with the first name.
      cnt=0; j=0;
      while ((j<=(MAXLINEWIDTH-1))&&((ch=*(s+j))!=0x0a))
	{
	  if (ch==0x09)
	    cnt++;
	  j++;
	}
      if (cnt<2)
	strncpy(anothergenenames+i*GENENAMELENGTH,genenames+i*GENENAMELENGTH,GENENAMELENGTH);


      item=getitemval(s,1,&length);
      j=0;
      while ((j<=(length-1))&&((ch=*(item+j))!='/'))
	{
	  *(anothergenenames+i*GENENAMELENGTH+j)=ch;
	  j++;
	}
      while (j<=(GENENAMELENGTH-1))
	{
	  *(anothergenenames+i*GENENAMELENGTH+j)=' ';
	  j++;
	}

      // If the second name is blank, then fill it with the first name.
      flag=0;
      for (j=0; j<=(GENENAMELENGTH-1); j++)
	if (*(anothergenenames+i*GENENAMELENGTH+j)!=' ')
	  flag=1;
      if (flag==0)
	strncpy(anothergenenames+i*GENENAMELENGTH,genenames+i*GENENAMELENGTH,GENENAMELENGTH);

      i++;
    }
  fclose(fp);

  free(s); free(item);
  
  *genenamespnt=genenames;
  *anothergenenamespnt=anothergenenames;

}


// Load reference numbers of a list of gene names.
void loadselgenes(int N, char *genenames, char *anothergenenames, char *selfilename, int *Nselgenes, int *ref2selpnt, int *sel2refpnt)
{
  int i, j, length;
  char *item, *s;
  int *ref2sel, *sel2ref;
  FILE *fp;

  
  if ((fp=fopen(selfilename,"r"))==NULL)
    {
      printf("Cannot open %s.\n",selfilename);
      return;
    }


  s=malloc(sizeof(char)*MAXLINEWIDTH);
  ref2sel=malloc(sizeof(int)*N); 
  for (i=0; i<=(N-1); i++)
    *(ref2sel+i)=-1;
  *Nselgenes=0;

  /*
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      item=getitemval(s,0,&length);
      i=0;
      
      // Resolve the problem that one gene name is the prefix of another.
      while (((strncmp(item,genenames+i*GENENAMELENGTH,length)!=0)||(*(genenames+i*GENENAMELENGTH+length)!=' ')||(length>GENENAMELENGTH))&&((strncmp(item,anothergenenames+i*GENENAMELENGTH,length)!=0)||(*(anothergenenames+i*GENENAMELENGTH+length)!=' ')||(length>GENENAMELENGTH))&&(i<N))
	i++;

      // Debug
      //while ((strncmp(item,genenames+i*GENENAMELENGTH,length)!=0)&&(i<N))
      //i++;

      if (((strncmp(item,genenames+i*GENENAMELENGTH,length)==0))||((strncmp(item,anothergenenames+i*GENENAMELENGTH,length)==0)))
	{	  
	  *(ref2sel+i)=*Nselgenes; *Nselgenes+=1;
	}
      
      free(item);
    }
  */


  // Allow the case when selected genes are not in reference gene table.
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      item=getitemval(s,0,&length);
            
      // Resolve the problem that one gene name is the prefix of another.
      i=0;
      while (((strncmp(item,genenames+i*GENENAMELENGTH,length)!=0)||(*(genenames+i*GENENAMELENGTH+length)!=' ')||(length>GENENAMELENGTH))&&((strncmp(item,anothergenenames+i*GENENAMELENGTH,length)!=0)||(*(anothergenenames+i*GENENAMELENGTH+length)!=' ')||(length>GENENAMELENGTH))&&(i<N))
	i++;
      
      if ((i<N)&&(((strncmp(item,genenames+i*GENENAMELENGTH,length)==0))||((strncmp(item,anothergenenames+i*GENENAMELENGTH,length)==0))))
	{
	  *(ref2sel+i)=*Nselgenes;
	}
      *Nselgenes+=1;

      /*
      if (((strncmp(item,genenames+i*GENENAMELENGTH,length)==0))||((strncmp(item,anothergenenames+i*GENENAMELENGTH,length)==0)))
	{	  
	  *(ref2sel+i)=*Nselgenes; *Nselgenes+=1;
	}
      */
      
      free(item);
    }
  


  fclose(fp);
  
  sel2ref=malloc(sizeof(int)*(*Nselgenes));
  for (i=0; i<=(*Nselgenes-1); i++)
    *(sel2ref+i)=-1;
  for (i=0; i<=(N-1); i++)
    if (*(ref2sel+i)>=0)
      *(sel2ref+*(ref2sel+i))=i;

  free(s);
  
  *ref2selpnt=ref2sel; *sel2refpnt=sel2ref;
  
}


// Load location data.
void loadlocdata(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *gene2spot, char *explist, double locpthre, char *locdata, int *Nfactors, int *ref2facpnt, int *fac2refpnt, int *Nlocs, int *locpnt)
{
  int i, j, k, length, Nlocgenes, *multigene2spot, *ref2locgene, *locgene2ref;
  int Nlocexps, *locexp2fac, **fac2locexp, *template, flag, *ref2fac, *fac2ref;
  char *s, *item, *facnames;
  double *locpvals;
  struct triplet *loc;
  FILE *fp;

  
  // Load location gene-to-spot map table.
  if ((fp=fopen(gene2spot,"r"))==NULL)
    {
      printf("Cannot open %s.\n",gene2spot);
      return;
    }
  
  s=malloc(sizeof(char)*MAXLINEWIDTH);
  Nlocgenes=0; fgets(s,MAXLINEWIDTH,fp);
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    Nlocgenes++;
  
  multigene2spot=malloc(sizeof(int)*Nlocgenes*3);
  for (i=0; i<=(Nlocgenes*3-1); i++)
    *(multigene2spot+i)=-1;

  
  ref2locgene=malloc(sizeof(int)*N);
  for (i=0; i<=(N-1); i++)
    *(ref2locgene+i)=-1;
  locgene2ref=malloc(sizeof(int)*Nlocgenes);
  for (i=0; i<=(Nlocgenes-1); i++)
    *(locgene2ref+i)=-1;

  fseek(fp,0,SEEK_SET); 
  i=0; fgets(s,MAXLINEWIDTH,fp);
  
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      int val1, val2, val3;
      
      item=getitemval(s,0,&length);

      j=0;
      while ((j<N)&&(strncmp(item,genenames+j*GENENAMELENGTH,length)!=0))
	j++;
      free(item);
      
      if (j<N)
	{
	  *(locgene2ref+i)=j;
	  *(ref2locgene+j)=i;
	  item=getitemval(s,2,&length);
	  val1=(int)(atof2(item,length)-1);
	  free(item);
	  item=getitemval(s,3,&length);
	  val2=(int)(atof2(item,length)-1);
	  free(item);
	  item=getitemval(s,4,&length);
	  val3=(int)(atof2(item,length)-1);
	  free(item);
      
	  if ((val1!=val2)||(val2!=val3))
	    {
	      *(multigene2spot+i*3)=val1;
	      *(multigene2spot+i*3+1)=val2;
	      *(multigene2spot+i*3+2)=val3;
	    }
	  else
	    {
	      *(multigene2spot+i*3)=val1;
	      *(multigene2spot+i*3+1)=val1;
	      *(multigene2spot+i*3+2)=val1;
	    }
	}      
      i++;
    }
    
  fclose(fp);


  // Load experiments list.
  if ((fp=fopen(explist,"r"))==NULL)
    {
      printf("Cannot open %s.\n",explist);
      return;
    }
  fgets(s,MAXLINEWIDTH,fp); Nlocexps=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    Nlocexps++;

  locexp2fac=malloc(sizeof(int)*Nlocexps); fac2locexp=malloc(sizeof(int)*Nlocexps);
  for (i=0; i<=(Nlocexps-1); i++)
    {
      *(fac2locexp+i)=malloc(sizeof(int)*10);
      **(fac2locexp+i)=0;
    }
  template=malloc(sizeof(int)*Nlocexps);

  *Nfactors=0; i=0;
  facnames=malloc(sizeof(char)*GENENAMELENGTH);
  fseek(fp,0,SEEK_SET); fgets(s,MAXLINEWIDTH,fp); 
  
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      double val;

      item=getitemval(s,0,&length);
      j=0;
      while ((j<*Nfactors)&&(strncmp(item,facnames+j*GENENAMELENGTH,length)!=0))
	j++;
      
      if (j>=*Nfactors)
	{
	  *Nfactors+=1;
	  facnames=realloc(facnames,sizeof(char)*GENENAMELENGTH*(*Nfactors));
	  strncpy(facnames+(*Nfactors-1)*GENENAMELENGTH,item,length);
	  if (length<GENENAMELENGTH)
	    for (k=length; k<=(GENENAMELENGTH-1); k++)
	      *(facnames+(*Nfactors-1)*GENENAMELENGTH+k)=' ';
	  *(locexp2fac+i)=*Nfactors-1;
	  **(fac2locexp+*Nfactors-1)=1;
	  *(*(fac2locexp+*Nfactors-1)+1)=i;
	}
      else
	{
	  *(locexp2fac+i)=j;
	  **(fac2locexp+j)+=1;
	  *(*(fac2locexp+j)+**(fac2locexp+j))=i;
	}
      
      free(item);

      item=getitemval(s,4,&length);

      
      // Accurate version
      val=atof2(item,length);
      free(item);
      if ((val>=3.99)&&(val<=4.01))
	*(template+i)=0;
      else if ((val>=5.09)&&(val<=5.31))
	*(template+i)=1;
      else
	*(template+i)=2;
	
      
      /*
      // Previous version, wrong.
      val=atof2(s,length);
      free(item);
      if (val==4)
	*(template+i)=0;
      else if ((val>=5.1)&&(val<=5.3))
	*(template+i)=1;
      else
	*(template+i)=2;      
      */


      i++;
    }
  fclose(fp);

  
  ref2fac=malloc(sizeof(int)*N); fac2ref=malloc(sizeof(int)*(*Nfactors));
  for (i=0; i<=(N-1); i++)
    *(ref2fac+i)=-1;
  for (i=0; i<=(*Nfactors-1); i++)
    *(fac2ref+i)=-1;

  for (i=0; i<=(*Nfactors-1); i++)
    {
      k=0;
      while ((k<GENENAMELENGTH)&&(*(facnames+i*GENENAMELENGTH+k)!=' '))
	k++;
      flag=1; j=0;
      while ((flag==1)&&(j<=(N-1)))
	{
	  if ((strncmp(genenames+j*GENENAMELENGTH,facnames+i*GENENAMELENGTH,k)==0)||(strncmp(anothergenenames+j*GENENAMELENGTH,facnames+i*GENENAMELENGTH,k)==0))
	    flag=0;
	  j++;
	}
      if (flag==0)
	{
	  *(ref2fac+j-1)=i;
	  *(fac2ref+i)=j-1;
	}
    }

  
  // Load log location pvalues.  Extract the factor-gene pairs which fall in the scope of interested genes.
  locpvals=loadmatrix2(locdata); locpvals+=2;
  loc=malloc(sizeof(struct triplet)); *Nlocs=0;
  
  for (i=0; i<=(*Nfactors-1); i++)
    {
      int flag, spotind, expind;
      double val, minval=1;
 
      if (*(ref2sel+*(fac2ref+i))>=0)
	{
	  for (j=0; j<=(Nlocgenes-1); j++)
	    {
	      // Select an edge if the end genes are on the list of argv+2, and no self-bindings are selected.
	      //if (*(ref2sel+*(locgene2ref+j))>=0)	      
	      //if ((*(ref2sel+*(locgene2ref+j))>=0)&&(*(ref2sel+*(fac2ref+i))!=*(ref2sel+*(locgene2ref+j))))
	      if ((*(locgene2ref+j)>=0)&&(*(ref2sel+*(locgene2ref+j))>=0)&&(*(ref2sel+*(fac2ref+i))!=*(ref2sel+*(locgene2ref+j))))
		{
		  
		  flag=0; minval=1;
		  for (k=1; k<=**(fac2locexp+i); k++)
		    {		      
		      expind=*(*(fac2locexp+i)+k);
		      spotind=*(multigene2spot+j*3+*(template+expind));	       
      
		      val=*(locpvals+spotind*Nlocexps+expind);
		      
		      if (val<locpthre)
			{
			  flag=1;
			  if (val<minval)
			    minval=val;
			}		      
		    }

		  
		  if (flag==1)
		    {		      
		      *Nlocs+=1;
		      loc=realloc(loc,sizeof(struct triplet)*(*Nlocs));
		      (loc+*Nlocs-1)->i=*(fac2ref+i);
		      (loc+*Nlocs-1)->j=*(locgene2ref+j);
		      (loc+*Nlocs-1)->s=minval;		      
		    }
		  
		}
	    }
	}
    }

  
  
  free(multigene2spot); free(ref2locgene); free(locgene2ref);
  free(locexp2fac);
  for (i=0; i<=(Nlocexps-1); i++)
    free(*(fac2locexp+i));
  free(fac2locexp);
  free(template); free(s); free(facnames);
  locpvals-=2; free(locpvals);  
  *ref2facpnt=ref2fac; *fac2refpnt=fac2ref; *locpnt=loc;
  

}


// Detect the ORFs corresponding to the same spot.
int ** locdegeneratespots(int N, char *genenames, char *gene2spot)
{
  int i, j, k, length, Nlocgenes;
  int *spotnum, **degspots;
  char *s, *item;
  FILE *fp;

  // Load location gene-to-spot map table.
  if ((fp=fopen(gene2spot,"r"))==NULL)
    {
      printf("Cannot open %s.\n",gene2spot);
      return -1;
    }
  
  s=malloc(sizeof(char)*MAXLINEWIDTH);
  Nlocgenes=0; fgets(s,MAXLINEWIDTH,fp);
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    Nlocgenes++;
  
  
  fseek(fp,0,SEEK_SET); 
  i=0; fgets(s,MAXLINEWIDTH,fp);
  
  spotnum=malloc(sizeof(int)*N*3);
  
  for (i=0; i<=(N-1); i++)
    for (j=0; j<=2; j++)
      *(spotnum+i*3+j)=-1;

  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      int val1, val2, val3;
      
      item=getitemval(s,0,&length);      
      j=0;
      while ((j<N)&&(strncmp(item,genenames+j*GENENAMELENGTH,length)!=0))
	j++;
      free(item);
      
      if (j<N)
	{
	  item=getitemval(s,2,&length);
	  val1=(int)(atof2(item,length)-1);
	  free(item);
	  item=getitemval(s,3,&length);
	  val2=(int)(atof2(item,length)-1);
	  free(item);
	  item=getitemval(s,4,&length);
	  val3=(int)(atof2(item,length)-1);
	  free(item);
	  *(spotnum+j*3+0)=val1; *(spotnum+j*3+1)=val2; *(spotnum+j*3+2)=val3;
	  
	}      
      i++;
    }
      
  fclose(fp);

  
  degspots=malloc(sizeof(int)*N);
  for (i=0; i<=(N-1); i++)
    {
      *(degspots+i)=malloc(sizeof(int));
      **(degspots+i)=0;
    }
  for (i=0; i<=(N-1); i++)
    {
      for (j=0; j<=(N-1); j++)
	{
	  if ((i!=j)&&(*(spotnum+i*3)==*(spotnum+j*3))&&(*(spotnum+i*3+1)==*(spotnum+j*3+1))&&(*(spotnum+i*3+2)==*(spotnum+j*3+2)))
	    {
	      **(degspots+i)+=1; *(degspots+i)=realloc(*(degspots+i),sizeof(int)*(**(degspots+i)+1));
	      *(*(degspots+i)+**(degspots+i))=j;
	    }
	}
    }

  free(spotnum);

  return degspots;

}




// Load protein-protein interaction data.
void loadppdata(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *ppdata, int *Npps, int *pppnt)
{
  int i, length;
  char *s, *item;
  struct triplet *pp;
  FILE *fp;

  // Load protein-protein interaction data.
  if ((fp=fopen(ppdata,"r"))==NULL)
    {
      printf("Cannot open %s.\n",ppdata);
      return;
    }
  s=malloc(sizeof(char)*MAXLINEWIDTH);
  pp=malloc(sizeof(struct triplet)); *Npps=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      int v1=-1, v2=-1;
            
      item=getitemval(s,0,&length);

      i=0;
      while ((i<=(N-1))&&((strncmp(item,genenames+i*GENENAMELENGTH,length)!=0)||(length>GENENAMELENGTH)||((length<=GENENAMELENGTH)&&(*(genenames+i*GENENAMELENGTH+length)!=' '))))
	i++;

      if (i<=(N-1))
	v1=i;
      else
	{
	  i=0;
	  while ((i<=(N-1))&&((strncmp(item,anothergenenames+i*GENENAMELENGTH,length)!=0)||((length>GENENAMELENGTH)||((length<=GENENAMELENGTH)&&(*(anothergenenames+i*GENENAMELENGTH+length)!=' ')))))
	    i++;
	  if (i<=(N-1))
	    v1=i;

	} 

      free(item);
      
      item=getitemval(s,1,&length);
           
      i=0;
      while ((i<=(N-1))&&((strncmp(item,genenames+i*GENENAMELENGTH,length)!=0)||(*(genenames+i*GENENAMELENGTH+length)!=' ')))
	i++;

      if (i<=(N-1))
	v2=i;
      else
	{
	  i=0;
	  while ((i<=(N-1))&&((strncmp(item,anothergenenames+i*GENENAMELENGTH,length)!=0)||(*(anothergenenames+i*GENENAMELENGTH+length)!=' ')))
	    i++;
	  if (i<=(N-1))
	    v2=i;
	}

      
      free(item);
      
      if ((v1>=0)&&(v2>=0)&&(*(ref2sel+v1)>=0)&&(*(ref2sel+v2)>=0))
	{
	  *Npps+=1;
	  pp=realloc(pp,sizeof(struct triplet)*(*Npps));
	  (pp+*Npps-1)->i=v1; (pp+*Npps-1)->j=v2;

	  // Since the p-p data does not have p-values, arbitrarily assign p-value=0.001.
	  (pp+*Npps-1)->s=0.01;
	}

    }
  
  fclose(fp); free(s); 

  *pppnt=pp;
}


// Load protein-protein interaction data.
// Difference from loadppdata: receives information about small-scale experiments, multiple experiments and PVM validation.  Use these information to compute the likelihood ratios.
// Also incorporate flag information in .k entry.
// bit3: small scale, bit2: multiple exps, bit1: pvm.
void loadppdata2(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *ppdata, int *Npps, int *pppnt)
{
  int i, length, ll, flaginfo=0;
  char *s, *item, ch;
  struct triplet *pp;
  double val;
  FILE *fp;

  // Load protein-protein interaction data.
  if ((fp=fopen(ppdata,"r"))==NULL)
    {
      printf("Cannot open %s.\n",ppdata);
      return;
    }
  s=malloc(sizeof(char)*MAXLINEWIDTH);
  pp=malloc(sizeof(struct triplet)); *Npps=0;
  // Skip the first line
  fgets(s,MAXLINEWIDTH,fp);

  // If the pp file contains confidence information.
  item=getitemval(s,1,&ll); 
  
  if (ll<=0)
    {
      while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
	{
	  int v1=-1, v2=-1;
	  
	  item=getitemval(s,2,&length);
	  i=0;
	  while ((i<=(N-1))&&((strncmp(item,genenames+i*GENENAMELENGTH,length)!=0)||(length>GENENAMELENGTH)||((length<=GENENAMELENGTH)&&(*(genenames+i*GENENAMELENGTH+length)!=' '))))
	    i++;
	  
	  if (i<=(N-1))
	    v1=i;

	  free(item);
      
	  item=getitemval(s,5,&length);
          i=0;
	  while ((i<=(N-1))&&((strncmp(item,genenames+i*GENENAMELENGTH,length)!=0)||(*(genenames+i*GENENAMELENGTH+length)!=' ')))
	    i++;

	  if (i<=(N-1))
	    v2=i;
      
	  free(item);

	  
	  if ((v1>=0)&&(v2>=0)&&(*(ref2sel+v1)>=0)&&(*(ref2sel+v2)>=0))
	    {
	      int f1, f2, f3, f4;
	      
	      *Npps+=1;
	      pp=realloc(pp,sizeof(struct triplet)*(*Npps));
	      (pp+*Npps-1)->i=v1; (pp+*Npps-1)->j=v2;
	  
	      
	      // Check if this pair has the following labels:
	      // S-small-scale exp, M-multiple exps, P-PVM validation.
	      f1=1; f2=0; f3=0; f4=0;
	      item=getitemval(s,7,&length);
	      i=0;
	      while (i<=(length-1))
		{
		  ch=*(item+i);
		  if (ch=='S')
		    f2=1;
		  if (ch=='M')
		    f3=1;
		  if (ch=='P')
		    f4=1;
		  i++;
		}
	      if (length>0)
		free(item);
	      
	      // Incorporate flaginfo.
	      flaginfo=0;
	      if (f2==1)
		flaginfo+=4;
	      if (f3==1)
		flaginfo+=2;
	      if (f4==1)
		flaginfo+=1;
	      (pp+*Npps-1)->k=flaginfo;


	      // Compute likelihood ratio based on verification labels.
	      val=ppratio*ppratio*ppratio;
	      val*=(alpha_epr_true_all/(1-alpha_epr_true_all));

	      
	      if (f2==1)
		val*=(alpha_epr_true_s/(1-alpha_epr_true_s));
	      else
		val*=(alpha_epr_false_s/(1-alpha_epr_false_s));
	      if (f3==1)
		val*=(alpha_epr_true_m/(1-alpha_epr_true_m));
	      else
		val*=(alpha_epr_false_m/(1-alpha_epr_false_m));
	      if (f4==1)
		val*=((1-pvm_fn)/pvm_fp);
	      else
		val*=(pvm_fn/(1-pvm_fp));
	      
	      /*
	      if (f2==1)
		val*=(alpha_epr_true_s/(1-alpha_epr_true_s));
	      if (f3==1)
		val*=(alpha_epr_true_m/(1-alpha_epr_true_m));
	      if (f4==1)
		val*=((1-pvm_fn)/pvm_fp);
	      */

	      (pp+*Npps-1)->s=val;


	      // Debug: include only small-scale exp results.
	      //if (f2==0)
	      //*Npps-=1;

	    }
	}	
    }

  // If the pp file does not contain confidence information.
  else
    {
      free(item);
      fseek(fp,0,SEEK_SET);
      while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
	{
	  int v1=-1, v2=-1;
            
	  item=getitemval(s,0,&length);

	  i=0;
	  while ((i<=(N-1))&&((strncmp(item,genenames+i*GENENAMELENGTH,length)!=0)||(length>GENENAMELENGTH)||((length<=GENENAMELENGTH)&&(*(genenames+i*GENENAMELENGTH+length)!=' '))))
	    i++;
	  
	  if (i<=(N-1))
	    v1=i;
	  else
	    {
	      i=0;
	      while ((i<=(N-1))&&((strncmp(item,anothergenenames+i*GENENAMELENGTH,length)!=0)||((length>GENENAMELENGTH)||((length<=GENENAMELENGTH)&&(*(anothergenenames+i*GENENAMELENGTH+length)!=' ')))))
		i++;
	      if (i<=(N-1))
		v1=i;	      
	    } 

	  free(item);
      
	  item=getitemval(s,1,&length);
           
	  i=0;
	  while ((i<=(N-1))&&((strncmp(item,genenames+i*GENENAMELENGTH,length)!=0)||(*(genenames+i*GENENAMELENGTH+length)!=' ')))
	    i++;

	  if (i<=(N-1))
	    v2=i;
	  else
	    {
	      i=0;
	      while ((i<=(N-1))&&((strncmp(item,anothergenenames+i*GENENAMELENGTH,length)!=0)||(*(anothergenenames+i*GENENAMELENGTH+length)!=' ')))
		i++;
	      if (i<=(N-1))
		v2=i;
	    }      
	  free(item);
      
	  if ((v1>=0)&&(v2>=0)&&(*(ref2sel+v1)>=0)&&(*(ref2sel+v2)>=0))
	    {
	      *Npps+=1;
	      pp=realloc(pp,sizeof(struct triplet)*(*Npps));
	      (pp+*Npps-1)->i=v1; (pp+*Npps-1)->j=v2;
	      
	      // Since the p-p data does not have p-values, arbitrarily assign likelihood ratio=2.0.
	      (pp+*Npps-1)->s=2.0;
	    }
	}
    }  
	  
  fclose(fp); free(s); 
  
  *pppnt=pp;
}




// Load knockout data.
void loadkodata(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *kogenefilename, char *expfilename, char *kopvalfilename, char *kologratiofilename, double kopthre, int *isknockedpnt, int *Nkos, int *kopnt)
{
  int i, j, k, l, length, Nkogenes, *ref2kogene, *kogene2ref;
  int Nkoexps, *koexp2ref, **ref2koexp, *isknocked;
  double *kopvals, *kologratios;
  char *item, *s;
  struct triplet *ko;
  FILE *fp;

  
  s=malloc(sizeof(char)*MAXLINEWIDTH);

  // Load knockout gene table.
  if ((fp=fopen(kogenefilename,"r"))==NULL)
    {
      printf("Cannot open %s.\n",kogenefilename);
      return;
    }
  Nkogenes=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    Nkogenes++;
  
  ref2kogene=malloc(sizeof(int)*N);
  kogene2ref=malloc(sizeof(int)*Nkogenes);
  for (i=0; i<=(N-1); i++)
    *(ref2kogene+i)=-1;
  for (i=0; i<=(Nkogenes-1); i++)
    *(kogene2ref+i)=-1;

  fseek(fp,0,SEEK_SET); i=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      item=getitemval(s,0,&length);
      j=0;
      while ((j<N)&&(strncmp(item,genenames+j*GENENAMELENGTH,length)!=0))
	j++;
      if (j<N)
	{
	  *(ref2kogene+j)=i;
	  *(kogene2ref+i)=j;
	}
      free(item);
      i++;
    }
  fclose(fp);
   
  // Load knockout experiment list.  Ignore double mutants for the time being.
  ref2koexp=malloc(sizeof(int)*N);
  for (i=0; i<=(N-1); i++)
    {
      *(ref2koexp+i)=malloc(sizeof(int));
      **(ref2koexp+i)=0;
    }
  Nkoexps=0;
  if ((fp=fopen(expfilename,"r"))==NULL)
    {
      printf("Cannot open %s.\n",expfilename);
      return;
    }
  
  fgets(s,MAXLINEWIDTH,fp);
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    Nkoexps++;
 
  koexp2ref=malloc(sizeof(int)*Nkoexps);
  for (i=0; i<=(Nkoexps-1); i++)
    *(koexp2ref+i)=-1;

  fseek(fp,0,SEEK_SET); fgets(s,MAXLINEWIDTH,fp);
  i=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      item=getitemval(s,1,&length);

      // The true length of the exp name is the length of the first word in the entry.
      j=0;
      while ((*(item+j)!=' ')&&(j<length))
	j++;

      length=j;

      // Eleminate the cases when " appears.
      if (*item!='"')
	{
	  // Convert the alphabets to capitals.
	  for (j=0; j<=(length-1); j++)
	    {
	      if ((*(item+j)>='a')&&(*(item+j)<='z'))
		*(item+j)+='A'-'a';
	    }
	}

      j=0;
      while ((j<=(N-1))&&((strncmp(item,anothergenenames+j*GENENAMELENGTH,length)!=0)||(length>GENENAMELENGTH)||(*(anothergenenames+j*GENENAMELENGTH+length)!=' ')))
	j++;

      if (j<=(N-1))
	{
	  *(koexp2ref+i)=j;
	  **(ref2koexp+j)+=1;
	  *(ref2koexp+j)=realloc(*(ref2koexp+j),sizeof(int)*(**(ref2koexp+j)+1));
	  *(*(ref2koexp+j)+**(ref2koexp+j))=i;
	}
      free(item);
      
      i++;
    }
  fclose(fp);

  // Find the genes which are knocked out in some experiments.
  isknocked=malloc(sizeof(int)*N);
  for (i=0; i<=(N-1); i++)
    {
      if (**(ref2koexp+i)>0)
	*(isknocked+i)=1;
      else
	*(isknocked+i)=0;
    }

  
  // Load knockout pvalues and logratio data.  If multiple experiments correspond to knocking out the same gene and one experiment yield high score, then incorporate this pair.
  kopvals=loadmatrix2(kopvalfilename); kopvals+=2;
  kologratios=loadmatrix2(kologratiofilename); kologratios+=2;
  ko=malloc(sizeof(struct triplet)); *Nkos=0; 

  for (i=0; i<=(Nselgenes-1); i++)
    {      
      if (**(ref2koexp+*(sel2ref+i))>0)
	{
	  
	  for (j=0; j<=(Nselgenes-1); j++)
	    {
	      if (i!=j)
		{
		  double minval=1.0, val, val2;
		  		  
		  for (k=1; k<=**(ref2koexp+*(sel2ref+i)); k++)
		    {
		      l=*(*(ref2koexp+*(sel2ref+i))+k);		      
		      val=*(kopvals+*(ref2kogene+*(sel2ref+j))*Nkoexps+l); 
		      
		      if (val<=minval)
			{
			  minval=val;
			  val2=*(kologratios+*(ref2kogene+*(sel2ref+j))*Nkoexps+l);
			}
		      
		      
		    }
		    
		  
		  if (minval<=kopthre)
		    {		      		      

		      int kk;

		      *Nkos+=1;		      		      
		      ko=realloc(ko,sizeof(struct triplet)*(*Nkos));
		      (ko+*Nkos-1)->i=*(sel2ref+i);
		      (ko+*Nkos-1)->j=*(sel2ref+j);
		      if (val2>0)
			(ko+*Nkos-1)->k=1;
		      else
			(ko+*Nkos-1)->k=-1;
		      (ko+*Nkos-1)->s=minval;
		      
		    }		  		  
		}
	    }
	  
	}
    }

  free(ref2kogene); free(kogene2ref); free(koexp2ref);
  for (i=0; i<=(N-1); i++)
    free(*(ref2koexp+i));
  free(ref2koexp); free(s);
  kopvals-=2; free(kopvals); kologratios-=2; free(kologratios);
  
  *isknockedpnt=isknocked; *kopnt=ko;
}


// Load the location data of a simple tabular format.
void loadlocdata2(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *explist, double locpthre, char *locdata, int *Nfactors, int *ref2facpnt, int *fac2refpnt, int *Nlocs, int *locpnt)
{
  int i, j, k, length, Nlocexps, *ref2fac, *fac2ref;
  int *locexp2fac, **fac2locexp, flag, cnt;
  char *s, *item, *facnames;
  struct triplet *loc;
  FILE *fp;

  s=malloc(sizeof(char)*MAXLINEWIDTH);

  // Load experiment list.
  if ((fp=fopen(explist,"r"))==NULL)
    {
      printf("Cannot open %s.\n",explist);
      return;
    }
  Nlocexps=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    Nlocexps++;
  locexp2fac=malloc(sizeof(int)*Nlocexps); fac2locexp=malloc(sizeof(int)*Nlocexps);
  for (i=0; i<=(Nlocexps-1); i++)
    {
      *(fac2locexp+i)=malloc(sizeof(int)*10);
      **(fac2locexp+i)=0;
    } 
  *Nfactors=0; i=0;
  facnames=malloc(sizeof(char)*GENENAMELENGTH);
  fseek(fp,0,SEEK_SET); 
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      item=getitemval(s,0,&length);
      j=0;
      while ((j<*Nfactors)&&(strncmp(item,facnames+j*GENENAMELENGTH,length)!=0))
	j++;
      
      if (j>=*Nfactors)
	{
	  *Nfactors+=1;
	  facnames=realloc(facnames,sizeof(char)*GENENAMELENGTH*(*Nfactors));
	  strncpy(facnames+(*Nfactors-1)*GENENAMELENGTH,item,length);
	  if (length<GENENAMELENGTH)
	    for (k=length; k<=(GENENAMELENGTH-1); k++)
	      *(facnames+(*Nfactors-1)*GENENAMELENGTH+k)=' ';
	  *(locexp2fac+i)=*Nfactors-1;
	  **(fac2locexp+*Nfactors-1)=1;
	  *(*(fac2locexp+*Nfactors-1)+1)=i;
	}
      else
	{
	  *(locexp2fac+i)=j;
	  **(fac2locexp+j)+=1;
	  *(*(fac2locexp+j)+**(fac2locexp+j))=i;
	}
      i++;
      free(item);
    }

  ref2fac=malloc(sizeof(int)*N); fac2ref=malloc(sizeof(int)*(*Nfactors));
  for (i=0; i<=(N-1); i++)
    *(ref2fac+i)=-1;
  for (i=0; i<=(*Nfactors-1); i++)
    *(fac2ref+i)=-1;

  for (i=0; i<=(*Nfactors-1); i++)
    {
      k=0;
      while ((k<GENENAMELENGTH)&&(*(facnames+i*GENENAMELENGTH+k)!=' '))
	k++;
      flag=1; j=0;
      while ((flag==1)&&(j<=(N-1)))
	{
	  if ((strncmp(genenames+j*GENENAMELENGTH,facnames+i*GENENAMELENGTH,k)==0)||(strncmp(anothergenenames+j*GENENAMELENGTH,facnames+i*GENENAMELENGTH,k)==0))
	    flag=0;
	  j++;
	}
      if (flag==0)
	{
	  *(ref2fac+j-1)=i;
	  *(fac2ref+i)=j-1;
	}
    }
  fclose(fp);

  
  if ((fp=fopen(locdata,"r"))==NULL)
    {
      printf("Cannot open %s.\n",locdata);
      return;
    }

  fgets(s,MAXLINEWIDTH,fp); fgets(s,MAXLINEWIDTH,fp); *Nlocs=0;
  loc=malloc(sizeof(struct triplet)); cnt=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      int ind=-1, bind[Nlocexps];
      double val, bindvals[Nlocexps], minval;

      item=getitemval(s,1,&length);
      i=0;
      while ((i<=(N-1))&&(strncmp(item,genenames+i*GENENAMELENGTH,length)!=0))
	i++;
      if (i<=(N-1))
	ind=i;
      free(item);

      if ((ind>=0)&&(*(ref2sel+ind)>=0))
	{
	  for (i=0; i<=(Nlocexps-1); i++)
	    bind[i]=0;
	  for (i=0; i<=(Nlocexps-1); i++)
	    {
	      item=getitemval(s,i+4,&length);
	      //val=atof(item);
	      val=atof2(item,length);
	      free(item);
	      bindvals[i]=val;
	      if (val<=locpthre)
		bind[i]=1;
	    }

	  /*
	  // Debug
	  //if (ind==2240)
	  if (ind==3684)
	    {
	      for (i=0; i<=(GENENAMELENGTH-1); i++)
		printf("%c",*(anothergenenames+ind*GENENAMELENGTH+i));
	      printf("\n");

	      for (i=0; i<=(*Nfactors-1); i++)
		{
		  printf("fac[%d]=(",i);
		  for (j=1; j<=**(fac2locexp+i); j++)
		    printf("%d ",*(*(fac2locexp+i)+j));
		  printf(") ");
		}
	      printf("\n");
	      printf("bindvals=(");
	      for (i=0; i<=(Nlocexps-1); i++)
		printf("%f ",bindvals[i]);
	      printf(")\n");
	    }
	  */
	  
	  for (i=0; i<=(*Nfactors-1); i++)
	    {
	      flag=0; minval=1.0;
	      if (**(fac2locexp+i)==1)
		{
		  j=*(*(fac2locexp+i)+1);

		  if (bind[j]==1)
		    {
		      flag=1; minval=bindvals[j];
		    }
		}
	      
	      else if (**(fac2locexp+i)>1)
		{
		  j=1; flag=1;
		  while (j<=**(fac2locexp+i))
		    {
		      k=*(*(fac2locexp+i)+j);
		      if (bindvals[k]<minval)
			minval=bindvals[k];
		      j++;
		    }
		  if (minval>locpthre)
		    flag=0;

		  // Debug
		  //if ((ind==3684)&&((i==7)||(i==8)))
		  //printf("i=%d, minval=%f \n",i,minval);


		  /*
		  flag=1; j=1;
		  while ((j<=**(fac2locexp+i))&&(flag==1))
		    {
		      k=*(*(fac2locexp+i)+j);
		      if (bindvals[k]<minval)
			minval=bindvals[k];
		      if (bind[k]==0)
			flag=0;
		      j++;
		    }
		  */		  
		}

	      if (flag==1)
		{
		  *Nlocs+=1;
		  loc=realloc(loc,sizeof(struct triplet)*(*Nlocs));
		  (loc+*Nlocs-1)->i=*(fac2ref+i);
		  (loc+*Nlocs-1)->j=ind;
		  (loc+*Nlocs-1)->s=minval;


		  // Debug
		  //if ((ind==3684)&&((i==7)||(i==8)))
		  //printf("loc[%d]=(%d %d %f)\n",*Nlocs-1,(loc+*Nlocs-1)->i,(loc+*Nlocs-1)->j,(loc+*Nlocs-1)->s);

		}
	      
	    }
	  
	}
    }
  fclose(fp);

  free(locexp2fac);
  for (i=0; i<=(Nlocexps-1); i++)
    free(*(fac2locexp+i));
  free(fac2locexp);  free(facnames);  free(s);
    
  *ref2facpnt=ref2fac; *fac2refpnt=fac2ref; *locpnt=loc;
}


// Load location data.  The input is a list of location pairs and their p-values.
void loadlocdata3(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *locfilename, double locpthre, int *Nfactors, int *ref2facpnt, int *fac2refpnt, int *Nlocs, int *locpnt)
{
  int i, j, k, *ref2fac, *fac2ref, cnt;
  struct triplet *loc;
  char *s, ch, *facnames;
  FILE *fp;

  // Load the original location pair list.
  if ((fp=fopen(locfilename,"r"))==NULL)
    {
      printf("Cannot open %s.\n",locfilename);
      return;
    }
  
  s=malloc(sizeof(char)*MAXLINEWIDTH);
  loc=malloc(sizeof(struct triplet)); *Nlocs=0;

  
  // Both TF and DNA should be in the selected gene list.
  // And the pvalue <= threshold.
  fseek(fp,0,SEEK_SET); cnt=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      int length, facind, geneind, flag;
      double pval;
      char *item;

      // Get the factor name.
      item=getitemval(s,0,&length);
      i=0; flag=1;
      /*
      while ((i<=(Nselgenes-1))&&(flag==1))
	{
	  if (strncmp(anothergenenames+*(sel2ref+i)*GENENAMELENGTH,item,length)==0)
	    flag=0;
	  i++;
	}
      */
      
      while ((i<=(Nselgenes-1))&&(flag==1))
	{
	  if (strncmp(anothergenenames+*(sel2ref+i)*GENENAMELENGTH,item,length)==0)
	    {
	      // Check if subsequent characters of the gene name are space.
	      if ((length<GENENAMELENGTH)&&(*(anothergenenames+*(sel2ref+i)*GENENAMELENGTH+length)==' '))
		flag=0;
	      else if (length==GENENAMELENGTH)
		flag=0;

	      
	      // Debug
	      //if ((cnt==8027)&&(flag==0))
	      //{
	      //  printf("anothergenename[%d]=",*(sel2ref+i));
	      //  for (j=0; j<=(GENENAMELENGTH-1); j++)
	      //    printf("%c",*(anothergenenames+*(sel2ref+i)*GENENAMELENGTH+j));
	      //  printf(" ");
	      //}
	      
	    }
	  i++;
	}
      

      /*
      // Debug
      if (cnt==8027)
	{
	  for (j=0; j<=(length-1); j++)
	    printf("%c",*(item+j));
	  printf(" ");
	}
      */

      free(item);
      if (flag==0)
	{
	  facind=*(sel2ref+i-1);
	  item=getitemval(s,1,&length);

	  /*
	  // Debug
	  if (cnt==8027)
	    {
	      for (j=0; j<=(length-1); j++)
		printf("%c",*(item+j));
	      printf(" ");
	    }
	  */
	  
	  i=0; flag=1;
	  while ((i<=(Nselgenes-1))&&(flag==1))
	    {
	      if (strncmp(genenames+*(sel2ref+i)*GENENAMELENGTH,item,length)==0)
		flag=0;
	      i++;
	    }
	  free(item);
	  
	  /*
	  // Debug
	  if (cnt==8027)
	    printf("flag=%d, i=%d ",flag,i);
	  */

	  if (flag==0)
	    {
	      geneind=*(sel2ref+i-1);
	      item=getitemval(s,2,&length);
	      //pval=atof(item);

	      pval=atof2(item,length);

	      free(item);  

	      /*
	      // Debug
	      if (cnt==8027)
		printf("geneind=%d, pval=%e ",geneind,pval);
	      */

	      if (pval<=locpthre)
		{
		  // Check if this pair is already on the location list.
		  flag=1; j=0;
		  while ((j<=(*Nlocs-1))&&(flag==1))
		    {
		      if (((loc+j)->i==facind)&&((loc+j)->j==geneind))
			flag=0;
		      j++;
		    }
		  
		  /*
		  // Debug
		  if (cnt==8027)
		    printf("flag=%d ",flag);
		  */

		  if (flag==1)
		    {
		      *Nlocs+=1;
		      loc=realloc(loc,sizeof(struct triplet)*(*Nlocs));
		      (loc+*Nlocs-1)->i=facind;
		      (loc+*Nlocs-1)->j=geneind;
		      (loc+*Nlocs-1)->s=pval;


		      // Debug
		      //if (cnt==8027)
		      //printf("loc[%d]=(%d %d)\n",*Nlocs-1,facind,geneind);
		      
		    }
		}
	    }
	}
      cnt++;
    }


  
  // Obtain a list of factors.
  *Nfactors=0;
  facnames=malloc(sizeof(char)*GENENAMELENGTH*N);
  fseek(fp,0,SEEK_SET); cnt=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      int length, flag;
      char *item;

      item=getitemval(s,0,&length);
      
      i=0;
      while ((i<=*Nfactors)&&(strncmp(item,facnames+i*GENENAMELENGTH,length)!=0))
	i++;
      if (i>=*Nfactors)
	{
	  // Check whether the factor is on the list of selected genes.
	  j=0; flag=1;
	  while ((j<=(Nselgenes-1))&&(flag==1))
	    {
	      if ((strncmp(item,anothergenenames+*(sel2ref+j)*GENENAMELENGTH,length)==0)&&(*(anothergenenames+*(sel2ref+j)*GENENAMELENGTH+length)==' '))
		flag=0;
	      j++;
	    }

	  if (flag==0)
	    {
	      *Nfactors+=1;
	      //facnames=realloc(facnames,sizeof(char)*GENENAMELENGTH*(*Nfactors));
	      strncpy(facnames+(*Nfactors-1)*GENENAMELENGTH,item,length);
	      if (length<GENENAMELENGTH)
		for (j=length; j<=(GENENAMELENGTH-1); j++)
		  *(facnames+(*Nfactors-1)*GENENAMELENGTH+j)=' ';
	    }
	}
      
      free(item);
      cnt++;
    }

  fclose(fp);

  facnames=realloc(facnames,sizeof(char)*GENENAMELENGTH*(*Nfactors));
 

  ref2fac=malloc(sizeof(int)*N); fac2ref=malloc(sizeof(int)*(*Nfactors));
  for (i=0; i<=(N-1); i++)
    *(ref2fac+i)=-1;
  for (i=0; i<=(*Nfactors-1); i++)
    *(fac2ref+i)=-1;

  for (i=0; i<=(*Nfactors-1); i++)
    {
      int flag;
      k=0;
      while ((k<GENENAMELENGTH)&&(*(facnames+i*GENENAMELENGTH+k)!=' '))
	k++;

      flag=1; j=0;
      while ((flag==1)&&(j<=(N-1)))
	{
	  /*
	  if ((strncmp(genenames+j*GENENAMELENGTH,facnames+i*GENENAMELENGTH,k)==0)||(strncmp(anothergenenames+j*GENENAMELENGTH,facnames+i*GENENAMELENGTH,k)==0))
	    flag=0;
	  */

	  if (strncmp(anothergenenames+j*GENENAMELENGTH,facnames+i*GENENAMELENGTH,k)==0)
	    {
	      if ((k<GENENAMELENGTH)&&(*(anothergenenames+j*GENENAMELENGTH+k)==' '))
		flag=0;
	      else if (k==GENENAMELENGTH)
		flag=0;
	    }

	  j++;
	}
      if (flag==0)
	{
	  *(ref2fac+j-1)=i;
	  *(fac2ref+i)=j-1;
	}
    }

  free(s); free(facnames);

  *ref2facpnt=ref2fac; *fac2refpnt=fac2ref; *locpnt=loc;
 
} 


// Load chi2cdf lookup table.
struct chi2cdf_lookup * loadchi2table(char *filename, int *Nentries)
{
  int i, j, length;
  char *s, *item;
  struct chi2cdf_lookup *lookuptable;
  FILE *fp;

  // Open the lookup table.
  if ((fp=fopen(filename,"r"))==NULL)
    {
      printf("Cannot open %s.\n",filename);
      return NULL;
    }
  *Nentries=0; lookuptable=malloc(sizeof(struct chi2cdf_lookup));
  s=malloc(sizeof(char)*MAXLINEWIDTH);
  fgets(s,MAXLINEWIDTH,fp); fgets(s,MAXLINEWIDTH,fp);
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      *Nentries+=1;
      lookuptable=realloc(lookuptable,sizeof(struct chi2cdf_lookup)*(*Nentries));
      (lookuptable+*Nentries-1)->ind=*Nentries-1;
      item=getitemval(s,0,&length);
      (lookuptable+*Nentries-1)->x=atof2(item,length);
      free(item);
      item=getitemval(s,1,&length);
      (lookuptable+*Nentries-1)->p=atof2(item,length);
      free(item);
      item=getitemval(s,2,&length);
      (lookuptable+*Nentries-1)->q=atof2(item,length);
      free(item);
    }
  free(s); fclose(fp);
  return lookuptable;
}


// Load Rosetta data into the format of a matrix with designation of knockout genes.
void loadrosettadata(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *kogenefilename, char *expfilename, char *kopvalfilename, char *kologratiofilename, double kopthre, int *Nrosettaexps, int *rosetta_condspnt, int *qrosettadatapnt)
{
  int i, j, k, l, length, Nkogenes, *ref2kogene, *kogene2ref;
  int **rosetta_conds, *qrosettadata;
  double *kopvals, *kologratios;
  char *item, *s, ch, *arg56="ARG5,6";
  FILE *fp;

  s=malloc(sizeof(char)*MAXLINEWIDTH);

  // Load knockout gene table.
  if ((fp=fopen(kogenefilename,"r"))==NULL)
    {
      printf("Cannot open %s.\n",kogenefilename);
      return;
    }
  Nkogenes=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    Nkogenes++;
  
  ref2kogene=malloc(sizeof(int)*N);
  kogene2ref=malloc(sizeof(int)*Nkogenes);
  for (i=0; i<=(N-1); i++)
    *(ref2kogene+i)=-1;
  for (i=0; i<=(Nkogenes-1); i++)
    *(kogene2ref+i)=-1;

  fseek(fp,0,SEEK_SET); i=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      item=getitemval(s,0,&length);
      j=0;
      while ((j<N)&&(strncmp(item,genenames+j*GENENAMELENGTH,length)!=0))
	j++;
      if (j<N)
	{
	  *(ref2kogene+j)=i;
	  *(kogene2ref+i)=j;
	}
      free(item);
      i++;
    }
  fclose(fp);


  // Load rosetta experiment list.
  *Nrosettaexps=0;
  if ((fp=fopen(expfilename,"r"))==NULL)
    {
      printf("Cannot open %s.\n",expfilename);
      return;
    }
  fgets(s,MAXLINEWIDTH,fp);
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    *Nrosettaexps+=1;
  
  rosetta_conds=malloc(sizeof(int)*(*Nrosettaexps));
  for (i=0; i<=(*Nrosettaexps-1); i++)
    {
      *(rosetta_conds+i)=malloc(sizeof(int)*3);
      **(rosetta_conds+i)=0;
    }

  fseek(fp,0,SEEK_SET); fgets(s,MAXLINEWIDTH,fp);
  i=0;
  while (fgets(s,MAXLINEWIDTH,fp)!=NULL)
    {
      item=getitemval(s,1,&length);
      
      // The item not encircled by "" is either single knockout of drug.
      if (*item!='"')
	{
	  // The true length of the exp name is the length of the first word in the entry.
	  j=0;
	  while ((*(item+j)!=' ')&&(j<length))
	    j++;
	  length=j;

	  // Convert the alphabets to capitals.
	  for (j=0; j<=(length-1); j++)
	    {
	      if ((*(item+j)>='a')&&(*(item+j)<='z'))
		*(item+j)+='A'-'a';
	    }

	  // Check whether it is a single knockout.
	  j=0;
	  while ((j<=(N-1))&&((strncmp(item,anothergenenames+j*GENENAMELENGTH,length)!=0)||(length>GENENAMELENGTH)||(*(anothergenenames+j*GENENAMELENGTH+length)!=' ')))
	    j++;
	  
	  // If it is a single knockout.
	  if (j<=(N-1))
	    {
	      **(rosetta_conds+i)=1; *(*(rosetta_conds+i)+1)=j;
	    }
	  else 
	    {
	      int length1;
	      char *item1;

	      // If it is a single knockout and contains remarks.
	      item1=malloc(sizeof(char)*length); k=0;
	      while ((k<=(length-1))&&((ch=*(item+k))!=' '))
		{
		  *(item1+k)=ch; k++;
		}
	      
	      length1=k;
	      k=0;
	      while ((k<=(N-1))&&((strncmp(item1,anothergenenames+k*GENENAMELENGTH,length1)!=0)||(length1>GENENAMELENGTH)||(*(anothergenenames+k*GENENAMELENGTH+length1)!=' ')))
		k++;
	      if (k<=(N-1))
		{
		  **(rosetta_conds+i)=1; *(*(rosetta_conds+i)+1)=k;
		}            
	      
	      // If it is a drug effect.
	      else
		**(rosetta_conds+i)=0;
	      
	      free(item1);
	    }
	}

      // The item encircled by "" is a candidate for double knockout.
      else
	{
	  // Convert the alphabets to capitals.
	  j=1;
	  while ((*(item+j)!='"')&&(j<length))
	    j++;
	  length=j;
	  for (j=0; j<=(length-1); j++)
	    {
	      if ((*(item+j)>='a')&&(*(item+j)<='z'))
		*(item+j)+='A'-'a';
	    }
	  if (*(item+j)=='"')
	    length+=1;

	  // Check if it is the special case of arg5,6.
	  if (strncmp(item+1,arg56,6)==0)
	    {
	      **(rosetta_conds+i)=1; *(*(rosetta_conds+i)+1)=1639;
	    }
	  
	  // Obtain two substrings separated by space or comma.
	  else
	    {
	      int length1, length2;
	      char *item1, *item2;
	      
	      item1=malloc(sizeof(char)*length);
	      item2=malloc(sizeof(char)*length);
	      length1=0; j=1;
	      while ((j<=(length-1))&&(((ch=*(item+j))!=',')&&(ch!=' ')))
		{
		  j++; length1++; *(item1+length1-1)=ch;
		}
	   
	      // Possible double knockouts.
	      if (*(item+j)==',')
		{
		  j+=2; length2=0;
		  while ((j<=(length-1))&&(((ch=*(item+j))!=' ')&&(ch!='"')))
		    {
		      j++; length2++; *(item2+length2-1)=ch;
		    }

		  k=0;
		  while ((k<=(N-1))&&((strncmp(item1,anothergenenames+k*GENENAMELENGTH,length1)!=0)||(length1>GENENAMELENGTH)||(*(anothergenenames+k*GENENAMELENGTH+length1)!=' ')))
		    k++;
		  l=0;
		  while ((l<=(N-1))&&((strncmp(item2,anothergenenames+l*GENENAMELENGTH,length2)!=0)||(length2>GENENAMELENGTH)||(*(anothergenenames+l*GENENAMELENGTH+length2)!=' ')))
		    l++;

		  if ((k<=(N-1))&&(l<=(N-1)))
		    {
		      **(rosetta_conds+i)=2; 
		      *(*(rosetta_conds+i)+1)=k;
		      *(*(rosetta_conds+i)+2)=l;
		    }
		  else
		    **(rosetta_conds+i)=0; 
		}

	      // Possible single knockouts.
	      else
		{
		  k=0;
		  while ((k<=(N-1))&&((strncmp(item1,anothergenenames+k*GENENAMELENGTH,length1)!=0)||(length1>GENENAMELENGTH)||(*(anothergenenames+k*GENENAMELENGTH+length1)!=' ')))
		    k++;
		  if (k<=(N-1))
		    {
		      **(rosetta_conds+i)=1; 
		      *(*(rosetta_conds+i)+1)=k;
		    }
		  else
		    **(rosetta_conds+i)=0;
		}	      


	      free(item1); free(item2);
	    }
	}
      i++;
      free(item);
    }
  fclose(fp);

  /*
  // Debug
  for (i=0; i<=(*Nrosettaexps-1); i++)
    {
      printf("exp %d, delete gene ",i+1);
      if (**(rosetta_conds+i)<=0)
	printf("none ");
      else
	{
	  for (j=1; j<=**(rosetta_conds+i); j++)
	    {
	      k=0;
	      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(*(rosetta_conds+i)+j)*GENENAMELENGTH+k))!=' '))
		{
		  printf("%c",ch); k++;
		}
	      printf(" ");
	    }
	}
      printf("\n");
    }
  */
  
  // Quantize rosetta data.
  kopvals=loadmatrix2(kopvalfilename); kopvals+=2;
  kologratios=loadmatrix2(kologratiofilename); kologratios+=2;
  qrosettadata=malloc(sizeof(int)*Nselgenes*(*Nrosettaexps));
  for (i=0; i<=(Nselgenes*(*Nrosettaexps)-1); i++)
    *(qrosettadata+i)=0;
    
  // Debug
  printf("Nselgenes=%d, Nrosettaexps=%d.\n",Nselgenes,*Nrosettaexps);
  
  for (i=0; i<=(Nkogenes-1); i++)
    {
      if ((*(kogene2ref+i)>=0)&&(*(ref2sel+*(kogene2ref+i))>=0))
	{
	  for (j=0; j<=(*Nrosettaexps-1); j++)
	    {
	      double pval, logratio;
	      
	      pval=*(kopvals+i*(*Nrosettaexps)+j);
	      logratio=*(kologratios+i*(*Nrosettaexps)+j);

	      
	      if (pval<=kopthre)
		{
		  if (logratio>0)
		    *(qrosettadata+*(ref2sel+*(kogene2ref+i))*(*Nrosettaexps)+j)=1;
		  else if (logratio<0)
		    *(qrosettadata+*(ref2sel+*(kogene2ref+i))*(*Nrosettaexps)+j)=-1;
		}
	      
	      // The deleted gene is set to -1.
	      if (inlist(**(rosetta_conds+j),*(rosetta_conds+j)+1,*(kogene2ref+i))==1)
		*(qrosettadata+*(ref2sel+*(kogene2ref+i))*(*Nrosettaexps)+j)=-1;
	    }
	}
    }
  
  kopvals-=2; kologratios-=2; free(kopvals); free(kologratios);

  free(ref2kogene); free(kogene2ref); free(s);

  *rosetta_condspnt=rosetta_conds; *qrosettadatapnt=qrosettadata;
  
}
