/* Load a file into a matrix (1D or 2D).
   Input: file name.
   Output: size of the matrix, matrix. */

# include <stdio.h>
# include <stdlib.h>

long * loadmatrix(char *filename)
{
  FILE *fp;
  long n=0, N=0, m=0, M=0, count=0;
  long *matrix;
  unsigned char flags=0;
  char ch;
  long *retval;

  if ((fp=fopen(filename, "r")) == NULL)
     {
       printf("Cannot open %s. \n",filename);
       retval=malloc(sizeof(long)*2);
       *retval=0; *(retval+1)=0;
       return retval;
     }

  
  // First run: figure out the size of the matrix.
  // Count the number of lines and space characters in each line.
  /* Flags: bit0: end of file flag.
            bit1: end of number flag.
	    bit2: sign of the number.
	    bit3: 0 -- to use 0x0a as the delimiter between lines.
	          1 -- to use 0x0d, 0x0a as the delimiter between lines.
  */
  while ((flags & 0x01)==0)
    {
      ch=getc(fp);
      if (ch==EOF)
	{
	  flags |= 0x01;
	  N=n;
	  m++;
	  if (m>M)
	    M=m;
	}
      /*else if (ch=='\n')
	{
	  n++;
	  m++;
	  if (m>M)
	    M=m;
	  m=0;
	  }*/
      else if ((ch==0x20)||(ch==0x09))
	  m++;
      else if (ch==0x0d)
	{
	  n++;
	  m++;
	  if (m>M)
	    M=m;
	  m=0;
	  flags |= 0x08;
	}
      else if (ch==0x0a)
	{
	  if ((flags & 0x08)==0x00)
	    {
	      n++;
	      m++;
	      if (m>M)
		M=m;
	      m=0;
	    }
	}

	/*else if ((ch=='\t')||(ch==' '))
	  m++;*/
    }

  //printf("N=%i M=%i\n",N,M);

  
  // Allocate memory space for an NxM matrix (signed integer).
  retval=malloc(sizeof(long)*(N*M+2));

  matrix=retval+2;

 
  // Second run: load the file content into the matrix
  fseek(fp, 0, SEEK_SET);
  n=0; m=0;
  flags &= 0x08;
  while ((flags & 0x01)==0)
    {
      long num=0;
      flags &= 0x08;
      while ((flags & 0x02)==0x00)
	{
	  ch=getc(fp);
          // putchar(ch);
	  // 0x09 -- tab
	  // 0x0d -- line feed
	  // 0x0a -- change line
	  // 0x20 -- space

	  // if ((ch=='\t')||(ch==EOF)||(ch=='\n'))
	  if ((ch==0x09)||(ch==0x0d)||(ch==0x0a)||(ch==0x20)||(ch==EOF))
	    {
	      
	      // Signed number.
	      if ((flags & 0x04)!=0)
		num*=-1;

	      if (ch!=EOF)
		{
		  count++;
		  *(matrix+n*M+m)=num;
		}

	      flags |= 0x02;
	      flags &= ~(0x04);

	      if (ch==EOF)
		{
		  flags |= 0x01;
		}

	      else if (((ch==0x0d)&&((flags & 0x08)!=0))||((ch==0x0a)&&((flags & 0x08)==0)))
		{
		  n++;
		  m=0;
		}

	      else if ((ch==0x09)||(ch==0x20))
		{
		  m++;
		}
	    }

	  else if ((ch=='-')||((ch>='0')&&(ch<='9')))
	    {
	      if (ch=='-')
		{
		  flags |= 0x04;
		}
	      else if ((ch>='0')&&(ch<='9'))
		{
		  // Convert the characters into integer.
		  num=num*10+(ch-'0');
		}
	      else
		{
		  printf("Error.\n");
		  return NULL;
		}
	    }
	}
    }


  fclose(fp);

  (long)*(retval)=N; (long)*(retval+1)=M;

  return retval;

}

