/* Generate subnetworks based on rosetta knockout data.
   Different from explain26: revise the function of subnetwork generation.
   Different from generate_subnetworks1: randomly assign MAP config values during the iterations and count the fraction of network structure preserved in all tests.
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "connect.h"
# include "integration.h"
# include "cluster.h"
# include "matrix.h"
# include "integration3.h"
# include "integration4.h"
# include "integration5.h"
# define MAXLINEWIDTH 1000
# define GENENAMELENGTH 9
# define PHASELENGTH 4
# define INFINITY 1e10
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)

extern char * getitemval(char *s, int ind, int *length);
extern char * getnextitemval(char *s, int *length, int curpos, int *nextpos);
extern char * getnextitemval2(char *s, int *length, int curpos, int *nextpos, int *end);
extern double atof2(char *s, int length);
extern long * loadmatrix(char *filename);
extern double * loadmatrix2(char *filename);
extern struct compresscpath * findvalidcpaths(struct csparsegraph *cgraph, int pathlength, int *Nvalidcpaths);
extern struct cpath * decompresscpath(struct csparsegraph *cgraph, struct compresscpath *path, int ind);
extern void explainknockout(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, int *Nexplained, int *explained);
extern void loadgenenames(char *filename, int *N, int *genenamespnt, int *anothergenenamespnt);
extern void loadselgenes(int N, char *genenames, char *anothergenenames, char *selfilename, int *Nselgenes, int *ref2selpnt, int *sel2refpnt);
extern void loadppdata2(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *ppdata, int *Npps, int *pppnt);
extern void loadkodata(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *kogenefilename, char *expfilename, char *kologratiofilename, char *kopvalfilename, double kopthre, int *isknockedpnt, int *Nkos, int *kopnt);
extern void determinevalidpaths(int Ncpaths, struct compresscpath *compresscpaths, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, struct csparsegraph *model, int N, int *isknocked, int *isvalid, int *kopairs);
extern struct compresscpath *findvalidcpaths2(int Nselgenes, int *ref2sel, int Nlocs, struct triplet *locs, int Npps, struct triplet *pps, int Nkos, struct triplet *ko, int pathlength, int *Nvalidcpaths);
extern struct compresscpath *augment_validcpaths(int Nselgenes, int *ref2sel, struct csparsegraph *model, int Nkos, struct triplet *ko, int *Ncpaths, struct compresscpath *compresscpaths, int *isvalid);
extern void loadlocdata3(int N, char *genenames, char *anothergenenames, int Nselgenes, int *ref2sel, int *sel2ref, char *locfilename, double locpthre, int *Nfactors, int *ref2facpnt, int *fac2refpnt, int *Nlocs, int *locpnt);
extern double invcchi2cdf(double p, int D, double lower, double upper, double precision);
extern struct chi2cdf_lookup * loadchi2table(char *filename, int *Nentries);
extern double chi2cdf_deg1_2(double x, int mode, int Nentries, struct chi2cdf_lookup *chi2table);
extern double invchi2cdf_deg1_2(double val, int mode, int Nentries, struct chi2cdf_lookup *chi2table);
extern struct model_configuration * generate_subnetworks1(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, char *anothergenenames, int *Ndecomposedmodels, int *decomposedmodelpnt);
extern struct model_configuration * rand_generate_subnetworks1(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, char *anothergenenames, int *Ndecomposedmodels, int *decomposedmodelpnt);


int maxpathlength=5;
int submaxpathlength=4;

//int maxpathlength=3;
//int submaxpathlength=3;
int Nreplicates=3;

double v1=0.7, v2=0.299, v3=0.001, koinv=1e-20;
extern char *chi2lookuptable;
int Nrandomtrials=500;


void main(int argc, char *argv[])
{
  // *(argv+1) -- reference table.
  // *(argv+2) -- genes of interest.
  // *(argv+3) -- location pair data.
  // *(argv+4) -- threshold on location pvalues.
  // *(argv+5) -- p-p interaction pairs.
  // *(argv+6) -- rosetta knockout gene table.
  // *(argv+7) -- rosetta knockout experiment list.
  // *(argv+8) -- rosetta knockout pvalue threshold.
  // *(argv+9) -- rosetta knockout pvalues.
  // *(argv+10) -- rosetta logratios.
  // *(argv+11) -- statistics about the preserved part of subnetworks.


  int i, j, k, l, n, N, Nselgenes, Nfactors, *ref2sel, *sel2ref;
  int *ref2fac, *fac2ref, Nlocs, Nkos, Npps, pnt1, pnt2, pnt3;
  int oldNcpaths, Ncpaths, *isvalid, *kopairs, Nconnected, *isconnected;
  double locpthre, kopthre, *errors, error_rate;
  struct triplet *loc, *ko, *pp;
  char *genenames, *anothergenenames, ch, *s;
  struct csparsegraph *model;
  struct compresscpath *compresscpaths;
  struct model_configuration *degmodels;
  struct decomposed_model *decomposedmodels;
  FILE *fp1, *fp2, *fp3, *fp4, *fp5;
  int Nvalidpaths, *validpathinds, *isknocked, Nexplained, *koinds, Nleaveout;
  int *explained, *conn2ko, *var2ko, *isaffected, *edge2sigedge;
  struct chi2cdf_lookup *chi2table;
  int Nentries, Nleaveoutexps, *leaveoutexps, Ndecomposedmodels;
  int *isused, N1, N2, kopairs_pathlength[6], minplength;
  int *isedgeconnected, Ndegpre, Ndegsign, Ndegdir, Ndegko, Ndegpathsel;
  int *connectivity;
  double *unpreservededges;
  

  // Load reference table.
  genenames=malloc(sizeof(char));
  anothergenenames=malloc(sizeof(char));
  loadgenenames(*(argv+1),&N,&pnt1,&pnt2);
  genenames=pnt1; anothergenenames=pnt2;

  // Debug
  printf("N=%d, ",N);
  
  // Load genes of interest.
  /*
  loadselgenes(N,genenames,anothergenenames,*(argv+2),&Nselgenes,&pnt1,&pnt2);
  ref2sel=pnt1; sel2ref=pnt2;
  */

  
  Nselgenes=N; ref2sel=malloc(sizeof(int)*N); sel2ref=malloc(sizeof(int)*N);
  for (i=0; i<=(N-1); i++)
    {
      *(ref2sel+i)=i; *(sel2ref+i)=i;
    }
  

  // Debug
  printf("Nselgenes=%d, ",Nselgenes);

  // Load the chi2lookup table.
  chi2table=loadchi2table(chi2lookuptable,&Nentries);

  // Debug
  maxpathlength=3;
  
  // Load location data.
  locpthre=atof(*(argv+4));
  loadlocdata3(N,genenames,anothergenenames,Nselgenes,ref2sel,sel2ref,*(argv+3),locpthre,&Nfactors,&pnt1,&pnt2,&Nlocs,&pnt3);
  ref2fac=pnt1; fac2ref=pnt2; loc=pnt3;

  // Determine the likelihood ratios of pd edges.
  for (i=0; i<=(Nlocs-1); i++)
    {
      double val, val2;
      val=(loc+i)->s;
      //val2=0.5*invcchi2cdf(val,1,0,100,0.01);
      val2=0.5*invchi2cdf_deg1_2(val,1,Nentries,chi2table);
      val2-=0.5*log(Nreplicates);
      (loc+i)->s=exp(val2);
    }

  // Debug
  printf("Nlocs=%d, Nfactors=%d, ",Nlocs,Nfactors);

  
  // Load protein-protein interaction data.
  // (pp+i)->s returns the likelihood ratio of the interaction.
  loadppdata2(N,genenames,anothergenenames,Nselgenes,ref2sel,sel2ref,*(argv+5),&Npps,&pnt1);
  pp=pnt1;

  // Debug
  printf("Npps=%d, ",Npps);

  
  // Load knockout gene table.
  kopthre=atof(*(argv+8));
  loadkodata(N,genenames,anothergenenames,Nselgenes,ref2sel,sel2ref,*(argv+6),*(argv+7),*(argv+9),*(argv+10),kopthre,&pnt1,&Nkos,&pnt2);
  isknocked=pnt1; ko=pnt2;
  
  
  // Determine the likelihood ratios of ko data.
  for (i=0; i<=(Nkos-1); i++)
    {
      double val, val2;
      val=(ko+i)->s;
      //val2=0.5*invcchi2cdf(val,1,0,100,0.1);
      val2=0.5*invchi2cdf_deg1_2(val,1,Nentries,chi2table);
      val2-=0.5*log(Nreplicates);
      (ko+i)->s=exp(val2);
    }
  
  // Debug
  printf("Nkos=%d \n",Nkos);

  compresscpaths=findvalidcpaths2(Nselgenes,ref2sel,Nlocs,loc,Npps,pp,Nkos,ko,maxpathlength,&Ncpaths);
  
  // Debug
  printf("Ncpaths=%d \n",Ncpaths);

  
  // Construct the physical model.
  model=malloc(sizeof(struct csparsegraph));
  model->Nv=Nselgenes;
  model->v=malloc(sizeof(int)*model->Nv);
  for (i=0; i<=(Nselgenes-1); i++)
    *(model->v+i)=i;
  model->Ne=Nlocs+Npps; model->e=malloc(sizeof(struct cedge)*model->Ne);
  model->Ne=0;

  // Delete self-crossing edges.
  // Add directed edges.
  for (i=0; i<=(Nlocs-1); i++)
    {
      if ((loc+i)->i!=(loc+i)->j)
	{
	  model->Ne+=1;
	  (model->e+model->Ne-1)->type=0;
	  (model->e+model->Ne-1)->from=*(ref2sel+(loc+i)->i);
	  (model->e+model->Ne-1)->to=*(ref2sel+(loc+i)->j);
	  (model->e+model->Ne-1)->sign=0;
	  (model->e+model->Ne-1)->priorp=(loc+i)->s;
	}
    }
  // Add undirected edges.  A directed edge overwrites an undirected edge.
  for (i=0; i<=(Npps-1); i++)
    {
      int flag=1;
      
      if ((pp+i)->i!=(pp+i)->j)
	{
	  j=0;
	  while ((j<=(Nlocs-1))&&(flag==1))
	    {
	      if (((pp+i)->i==(loc+j)->i)&&((pp+i)->j==(loc+j)->j))
		flag=0;
	      j++;
	    }
	  
	  if (flag==1)
	    {
	      model->Ne+=1;
	      (model->e+model->Ne-1)->type=1;
	      (model->e+model->Ne-1)->from=*(ref2sel+(pp+i)->i);
	      (model->e+model->Ne-1)->to=*(ref2sel+(pp+i)->j);
	      (model->e+model->Ne-1)->sign=0;
	      (model->e+model->Ne-1)->priorp=(pp+i)->s;
	    }
	}
    }  
  
  // Debug
  printf("model->Nv=%d, model->Ne=%d \n",model->Nv,model->Ne);


  // Identify valid paths according to exlanation criteria.
  isvalid=malloc(sizeof(int)*Ncpaths);
  kopairs=malloc(sizeof(int)*Ncpaths);
  determinevalidpaths(Ncpaths,compresscpaths,ref2sel,sel2ref,Nkos,ko,model,N,isknocked,isvalid,kopairs);

  // Augment valid paths to incorporate subpaths of existing valid paths.  Necessary because existing paths do not include genes which are not knocked out as sources.
  oldNcpaths=Ncpaths;
  compresscpaths=augment_validcpaths(Nselgenes,ref2sel,model,Nkos,ko,&Ncpaths,compresscpaths,isvalid);
  isvalid=realloc(isvalid,sizeof(int)*Ncpaths);
  for (i=oldNcpaths; i<=(Ncpaths-1); i++)
    *(isvalid+i)=1;
  
  // Debug
  printf("oldNcpaths=%d, newNcpaths=%d\n",oldNcpaths,Ncpaths);
  free(kopairs);
  

  // Remove the paths which contain pp edges overwritten by location edges.
  for (i=0; i<=(Ncpaths-1); i++)
    {
      if (*(isvalid+i)==1)
	{
	  struct cpath *path;
	  path=decompresscpath(model,compresscpaths,i);
	  for (j=0; j<=(path->Ne-1); j++)
	    if (*(path->e+j)<0)
	      *(isvalid+i)=0;
	  free(path->e); free(path);
	}
    }
  
  Nvalidpaths=0; validpathinds=malloc(sizeof(int));
  for (i=0; i<=(Ncpaths-1); i++)
    {
      if (*(isvalid+i)==1)
	{
	  Nvalidpaths++; validpathinds=realloc(validpathinds,sizeof(int)*Nvalidpaths);
	  *(validpathinds+Nvalidpaths-1)=i;
	}
    }
      
  // Debug
  printf("There are %d valid paths, model->Ne=%d.\n",Nvalidpaths,model->Ne);

  
  // Find the knockout pairs which are connected in the physical model.
  isconnected=malloc(sizeof(int)*Nkos);
  for (i=0; i<=(Nkos-1); i++)
    *(isconnected+i)=0;
  Nconnected=0;
  for (i=0; i<=(Nkos-1); i++)
    {
      j=0;
      //while ((j<=(Ncpaths-1))&&(*(isconnected+i)==0))
      while ((j<=(Nvalidpaths-1))&&(*(isconnected+i)==0))
	{
	  //if ((*(isvalid+j)==1)&&((compresscpaths+j)->start==*(ref2sel+(ko+i)->i))&&((compresscpaths+j)->end==*(ref2sel+(ko+i)->j)))
	  if (((compresscpaths+*(validpathinds+j))->start==*(ref2sel+(ko+i)->i))&&((compresscpaths+*(validpathinds+j))->end==*(ref2sel+(ko+i)->j)))
	    {
	      *(isconnected+i)=1; Nconnected++;
	    }
	  j++;
	}
    }

  // Debug
  printf("Nconnected=%d \n",Nconnected);


  // Obtain the reference decomposed models.
  degmodels=generate_subnetworks1(model,Ncpaths,compresscpaths,isvalid,ref2sel,sel2ref,Nkos,ko,anothergenenames,&Ndecomposedmodels,&pnt1);
  decomposedmodels=pnt1;

  unpreservededges=malloc(sizeof(double)*Nrandomtrials*Ndecomposedmodels);
  for (i=0; i<=(Nrandomtrials*Ndecomposedmodels-1); i++)
    *(unpreservededges+i)=0;

  // Assign random MAP config values and run it many times.
  for (n=0; n<=(Nrandomtrials-1); n++)
    {
      struct model_configuration *rand_degmodels;
      struct decomposed_model *rand_decomposedmodels;
      int rand_Ndecomposedmodels;

      rand_degmodels=rand_generate_subnetworks1(model,Ncpaths,compresscpaths,isvalid,ref2sel,sel2ref,Nkos,ko,anothergenenames,&rand_Ndecomposedmodels,&pnt1);
      rand_decomposedmodels=pnt1;
     
      // Count the overlap with the reference model.
      for (i=0; i<=(Ndecomposedmodels-1); i++)
	{
	  int Nv1, Nv2, *v1, *v2;
	  double rate1=0, rate2=0;

	  Nv1=(decomposedmodels+i)->Nedges;
	  v1=malloc(sizeof(int)*Nv1);
	  for (j=0; j<=(Nv1-1); j++)
	    *(v1+j)=*((decomposedmodels+i)->edgeinds+j);

	  if (i<rand_Ndecomposedmodels)
	    {
	      Nv2=(rand_decomposedmodels+i)->Nedges;
	      v2=malloc(sizeof(int)*Nv2);
	      for (j=0; j<=(Nv2-1); j++)
		*(v2+j)=*((rand_decomposedmodels+i)->edgeinds+j);
	      
	      // Include both false positive and false negative rates.
	      for (j=0; j<=(Nv2-1); j++)
		if (inlist(Nv1,v1,*(v2+j))==0)
		  rate1+=1;
	      for (j=0; j<=(Nv1-1); j++)
		if (inlist(Nv2,v2,*(v1+j))==0)
		  rate2+=1;
	      
	      rate1/=(double)Nv2; rate2/=(double)Nv1; 
	      *(unpreservededges+n*Ndecomposedmodels+i)=(rate1+rate2)/2;
	    }

	  else
	    *(unpreservededges+n*Ndecomposedmodels+i)=1.0;
	    
	}	  
      

      // Debug
      printf("n=%d\n",n);
	
      
      // Free models.
      free(rand_degmodels->vartypes);
      free(rand_degmodels->presence2var);
      free(rand_degmodels->sign2var);
      free(rand_degmodels->sigedge2edge);
      free(rand_degmodels->sigdir2var);
      free(rand_degmodels->ko2var);
      free(rand_degmodels->var2pathind);
      free(rand_degmodels->usedpathinfo);
      free(rand_degmodels->mapconfigs);
      free(rand_degmodels);

      for (i=0; i<=(rand_Ndecomposedmodels-1); i++)
	{
	  free((rand_decomposedmodels+i)->varinds);
	  free((rand_decomposedmodels+i)->vartypes);
	  free((rand_decomposedmodels+i)->edgeinds);
	}
      free(rand_decomposedmodels);
      
    }

  // Report the overlap rates.
  if ((fp1=fopen(*(argv+11),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+11));
      return;
    }
  fprintf(fp1,"raw: random trials, coulmn: subnetworks.\n");
  for (i=0; i<=(Nrandomtrials-1); i++)
    {
      for (j=0; j<=(Ndecomposedmodels-1); j++)
	fprintf(fp1,"%f ",*(unpreservededges+i*Ndecomposedmodels+j));
      fprintf(fp1,"\n");
    }
  fclose(fp1);

}

  
