// Utility functions of subnetwork generation.

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "integration.h"
# include "integration3.h"
# include "connect.h"
# include "cluster.h"
# include "integration4.h"
# include "integration5.h"
# define MAXLINEWIDTH 1000
# define INFINITY 1e38
# define POSINF 100000
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)
# define GENENAMELENGTH 9

extern int * recurse_getmapconfigs(struct confignode *configtree, int *Nmapconfigs, int *mapconfigs);
extern void recurse_freeconfignode(struct confignode *configtree);
extern struct mix_facgraph * construct_fac_graph2(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko); 
extern void maxsumproduct3(struct mix_facgraph *fgraph, struct message *msg, int *fixedvals, struct potential *belief, int mode, int updatemode);
extern void clear_mix_facgraph(struct mix_facgraph *fgraph);;
extern struct cpath * decompresscpath(struct csparsegraph *cgraph, struct compresscpath *path, int ind);
extern int inlist(int Nv, int *v, int q);
extern int inlistind(int Nv, int *v, int q);
extern void findcomponents4(int Nv, int **adjmat, int *membership);
struct mix_facgraph * construct_fac_graph4(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko);
//int * findmaxconfig15(struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int mode, int *Nmapconfigs);
void recurse_fixval9(struct confignode *configtree, struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int *isolated);
//void recurse_fixval10(struct confignode *configtree, struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int *isolated);
struct decomposed_model * findmaxconfig22(struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int *invariant, int *Ndecomposedmodels, struct decomposed_model *decomposedmodels, int *mapconfigs, struct csparsegraph *model, int *sel2ref, char *anothergenenames);
struct decomposed_model * recurse_fixval20(struct confignode *configtree, struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int *isolated, int *invariant, int *Ndecomposedmodels, struct decomposed_model *decomposedmodels, struct csparsegraph *model, int *sel2ref, char *anothergenenames);

extern int maxpathlength, submaxpathlength;
extern double epsilon, ptrue_positive;
extern double v1, v2, v3, koinv;


// Decompose the physical model into submodels.
// Different from computeconfigs9: exclude the subnetworks which do not explain knockout pairs, merge submodels which share non-invariant variables.
// Different from generate_subnetworks1: randomly fixing one MAP config value at each iteration.

struct model_configuration * rand_generate_subnetworks1(struct csparsegraph *model, int Ncpaths, struct compresscpath *compresscpaths, int *isvalid, int *ref2sel, int *sel2ref, int Nkos, struct triplet *ko, char *anothergenenames, int *Ndecomposedmodels, int *decomposedmodelpnt)
{
  int i, j, k, *fixedvals, *mapconfigurations, *Nmapconfigs;
  int Nsigedges, *sigdirvar2edge, *invariant;
  double val;
  struct mix_facgraph *fgraph;
  struct potential *maxbelief;
  struct message *maxmsg;
  struct model_configuration *degmodels;
  struct decomposed_model *decomposedmodels;
  char ch;
  
  // Construct the factor graph of model inference.
  fgraph=construct_fac_graph4(model,Ncpaths,compresscpaths,isvalid,ref2sel,sel2ref,Nkos,ko);

  Nsigedges=0; sigdirvar2edge=malloc(sizeof(int));
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      if ((fgraph->varnodes+i)->type==EDGEDIR)
	{
	  Nsigedges++; 
	  sigdirvar2edge=realloc(sigdirvar2edge,sizeof(int)*Nsigedges);
	  *(sigdirvar2edge+Nsigedges-1)=(fgraph->varnodes+i)->edgeind;
	}
    }

  // Debug
  printf("Nvars=%d, Ninteractions=%d, Ne=%d \n",fgraph->Nvars,fgraph->Ninteractions,fgraph->Ne);

  /*
  // Debug
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    {
      if (inlist((fgraph->facnodes+i)->Nedges,(fgraph->facnodes+i)->presences,3226)==1)
	{
	  printf("(");
	  for (j=0; j<=((fgraph->facnodes+i)->Nedges-1); j++)
	    {
	      printf("(");
	      k=0;
	      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+*((fgraph->facnodes+i)->presences+j))->from)*GENENAMELENGTH+k))!=' '))
		{
		  printf("%c",ch); k++;
		}
	      printf(",");
	      k=0;
	      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+*((fgraph->facnodes+i)->presences+j))->to)*GENENAMELENGTH+k))!=' '))
		{
		  printf("%c",ch); k++;
		}
	      printf(")");
	    }
	  printf(") ");
	}
    }
  printf("\n");
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    {
      if (inlist((fgraph->facnodes+i)->Nedges,(fgraph->facnodes+i)->presences,5157)==1)
	{
	  printf("(");
	  for (j=0; j<=((fgraph->facnodes+i)->Nedges-1); j++)
	    {
	      printf("(");
	      k=0;
	      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+*((fgraph->facnodes+i)->presences+j))->from)*GENENAMELENGTH+k))!=' '))
		{
		  printf("%c",ch); k++;
		}
	      printf(",");
	      k=0;
	      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+*((fgraph->facnodes+i)->presences+j))->to)*GENENAMELENGTH+k))!=' '))
		{
		  printf("%c",ch); k++;
		}
	      printf(")");
	    }
	  printf(") ");
	}
    }
  printf("\n");
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    {
      if (inlist((fgraph->facnodes+i)->Nedges,(fgraph->facnodes+i)->presences,5186)==1)
	{
	  printf("(");
	  for (j=0; j<=((fgraph->facnodes+i)->Nedges-1); j++)
	    {
	      printf("(");
	      k=0;
	      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+*((fgraph->facnodes+i)->presences+j))->from)*GENENAMELENGTH+k))!=' '))
		{
		  printf("%c",ch); k++;
		}
	      printf(",");
	      k=0;
	      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+*((fgraph->facnodes+i)->presences+j))->to)*GENENAMELENGTH+k))!=' '))
		{
		  printf("%c",ch); k++;
		}
	      printf(")");
	    }
	  printf(") ");
	}
    }
  printf("\n");
  
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    {
      if (inlist((fgraph->facnodes+i)->Nedges,(fgraph->facnodes+i)->presences,5191)==1)
	{
	  printf("(");
	  for (j=0; j<=((fgraph->facnodes+i)->Nedges-1); j++)
	    {
	      printf("(");
	      k=0;
	      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+*((fgraph->facnodes+i)->presences+j))->from)*GENENAMELENGTH+k))!=' '))
		{
		  printf("%c",ch); k++;
		}
	      printf(",");
	      k=0;
	      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+*((fgraph->facnodes+i)->presences+j))->to)*GENENAMELENGTH+k))!=' '))
		{
		  printf("%c",ch); k++;
		}
	      printf(")");
	    }
	  printf(") ");
	}
    }
  printf("\n");
  */


  
  
  // Initialize messages, beliefs and fixedvals.  
  maxmsg=malloc(sizeof(struct message)*fgraph->Ne*2);  
  for (i=0; i<=(fgraph->Ne-1); i++)
    {
      // The from node is a variable node and the to node is a factor node.
      (maxmsg+i)->from=(fgraph->e+i)->from;
      (maxmsg+i)->to=(fgraph->e+i)->to;
      (maxmsg+i)->dir=0;
      (maxmsg+i)->phi=malloc(sizeof(struct potential));     
      (maxmsg+i)->phi->Nvars=1;
      (maxmsg+i)->phi->varinds=malloc(sizeof(int));
      *((maxmsg+i)->phi->varinds)=(maxmsg+i)->from;	      
      (maxmsg+i)->phi->Nstates=malloc(sizeof(int));
      if (((fgraph->varnodes+*((maxmsg+i)->phi->varinds))->type==KNOCKOUT)||((fgraph->varnodes+*((maxmsg+i)->phi->varinds))->type==OUTPUTRESPONSE))
	*((maxmsg+i)->phi->Nstates)=3;
      else
	*((maxmsg+i)->phi->Nstates)=2;
      (maxmsg+i)->phi->Nconfigs=*((maxmsg+i)->phi->Nstates);
      (maxmsg+i)->phi->probs=malloc(sizeof(double)*(maxmsg+i)->phi->Nconfigs);
      val=(double)1/(double)(maxmsg+i)->phi->Nconfigs;
      for (j=0; j<=((maxmsg+i)->phi->Nconfigs-1); j++)     
	*((maxmsg+i)->phi->probs+j)=val;
    }
  
  for (i=0; i<=(fgraph->Ne-1); i++)
    {
      // The from node is a factor node and the to node is a variable node.
      (maxmsg+i+fgraph->Ne)->from=(fgraph->e+i)->to;
      (maxmsg+i+fgraph->Ne)->to=(fgraph->e+i)->from;
      (maxmsg+i+fgraph->Ne)->dir=1;
      (maxmsg+i+fgraph->Ne)->phi=malloc(sizeof(struct potential));
      (maxmsg+i+fgraph->Ne)->phi->Nvars=1;
      (maxmsg+i+fgraph->Ne)->phi->varinds=malloc(sizeof(int));
      *((maxmsg+i+fgraph->Ne)->phi->varinds)=(maxmsg+i+fgraph->Ne)->to;
      (maxmsg+i+fgraph->Ne)->phi->Nstates=malloc(sizeof(int));
      if (((fgraph->varnodes+*((maxmsg+i+fgraph->Ne)->phi->varinds))->type==KNOCKOUT)||((fgraph->varnodes+*((maxmsg+i+fgraph->Ne)->phi->varinds))->type==OUTPUTRESPONSE))
	*((maxmsg+i+fgraph->Ne)->phi->Nstates)=3;
      else
	*((maxmsg+i+fgraph->Ne)->phi->Nstates)=2;
      (maxmsg+i+fgraph->Ne)->phi->Nconfigs=*((maxmsg+i+fgraph->Ne)->phi->Nstates);	      
      (maxmsg+i+fgraph->Ne)->phi->probs=malloc(sizeof(double)*(maxmsg+i+fgraph->Ne)->phi->Nconfigs);
      val=(double)1/(double)(maxmsg+i+fgraph->Ne)->phi->Nconfigs;
      for (j=0; j<=((maxmsg+i+fgraph->Ne)->phi->Nconfigs-1); j++)
	*((maxmsg+i+fgraph->Ne)->phi->probs+j)=val;
    }

  maxbelief=malloc(sizeof(struct potential)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      (maxbelief+i)->Nvars=1;
      (maxbelief+i)->varinds=malloc(sizeof(int));
      *((maxbelief+i)->varinds)=i;
      (maxbelief+i)->Nstates=malloc(sizeof(int));

      if ((fgraph->varnodes+i)->type!=KNOCKOUT)
	{
	  (maxbelief+i)->Nconfigs=2;      
	  *((maxbelief+i)->Nstates)=2;
	  (maxbelief+i)->probs=malloc(sizeof(double)*2);
	  *((maxbelief+i)->probs)=1.0;
	  *((maxbelief+i)->probs+1)=1.0;
	}
      else
	{
	  (maxbelief+i)->Nconfigs=3;      
	  *((maxbelief+i)->Nstates)=3;
	  (maxbelief+i)->probs=malloc(sizeof(double)*3);
	  for (j=0; j<=2; j++)
	    *((maxbelief+i)->probs+j)=1.0;
	}
    }
  
  fixedvals=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(fixedvals+i)=-1;

  // Identify the variables of decomposed models.
  Nmapconfigs=malloc(sizeof(int)); *Nmapconfigs=1;
  mapconfigurations=malloc(sizeof(int)*fgraph->Nvars);
  *Ndecomposedmodels=0;
  decomposedmodels=malloc(sizeof(struct decomposed_model));
  invariant=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(invariant+i)=0;
  decomposedmodels=findmaxconfig22(fgraph,maxmsg,maxbelief,invariant,Ndecomposedmodels,decomposedmodels,mapconfigurations,model,sel2ref,anothergenenames);
  free(invariant);
  

  // Build degenerate models from MAP configuration values.
  degmodels=malloc(sizeof(struct model_configuration));
  degmodels->Nvars=fgraph->Nvars;
  degmodels->Nedges=model->Ne;
  degmodels->Nsigedges=Nsigedges;
  degmodels->Nkos=Nkos;
  degmodels->vartypes=malloc(sizeof(int)*degmodels->Nvars);
  for (i=0; i<=(degmodels->Nvars-1); i++)
    *(degmodels->vartypes+i)=(fgraph->varnodes+i)->type;

  degmodels->presence2var=malloc(sizeof(int)*model->Ne);
  degmodels->sign2var=malloc(sizeof(int)*model->Ne);
  for (i=0; i<=(model->Ne-1); i++)
    {
      *(degmodels->presence2var+i)=i;
      *(degmodels->sign2var+i)=i+model->Ne;
    }
  if (Nsigedges>0)
    {
      degmodels->sigedge2edge=malloc(sizeof(int)*Nsigedges);
      degmodels->sigdir2var=malloc(sizeof(int)*Nsigedges);
    }
  else
    {
      degmodels->sigedge2edge=malloc(sizeof(int));
      degmodels->sigdir2var=malloc(sizeof(int));
    }
  for (i=0; i<=(Nsigedges-1); i++)
    {
      *(degmodels->sigedge2edge+i)=*(sigdirvar2edge+i);
      *(degmodels->sigdir2var+i)=i+model->Ne*2;
    }
  degmodels->ko2var=malloc(sizeof(int)*Nkos);
  for (i=0; i<=(Nkos-1); i++)
    *(degmodels->ko2var+i)=-1;
  for (i=0; i<=(fgraph->Nvars-1); i++)
    if ((fgraph->varnodes+i)->type==KNOCKOUT)
      *(degmodels->ko2var+(fgraph->varnodes+i)->koind)=i;
   
  degmodels->var2pathind=malloc(sizeof(int)*degmodels->Nvars);
  for (i=0; i<=(degmodels->Nvars-1); i++)
    {
      if ((fgraph->varnodes+i)->type==PATHSEL)
	*(degmodels->var2pathind+i)=(fgraph->varnodes+i)->pathinfo->k;
      else
	*(degmodels->var2pathind+i)=-1;
    }
  
  degmodels->Nusedpaths=0; degmodels->usedpathinfo=malloc(sizeof(struct pair));
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      if ((fgraph->varnodes+i)->type==PATHSEL)
	{
	  degmodels->Nusedpaths+=1;
	  degmodels->usedpathinfo=realloc(degmodels->usedpathinfo,sizeof(struct pair)*degmodels->Nusedpaths);
	  (degmodels->usedpathinfo+degmodels->Nusedpaths-1)->i=(fgraph->varnodes+i)->pathinfo->i;
	  (degmodels->usedpathinfo+degmodels->Nusedpaths-1)->j=(fgraph->varnodes+i)->pathinfo->j;
	}
    }
  
  degmodels->Nmapconfigs=*Nmapconfigs;
  degmodels->mapconfigs=mapconfigurations;

  free(fixedvals); free(Nmapconfigs);

  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      free((maxmsg+i)->phi->varinds);
      free((maxmsg+i)->phi->Nstates);
      free((maxmsg+i)->phi->probs);
      free((maxmsg+i)->phi);
    }
  free(maxmsg);

  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      free((maxbelief+i)->varinds);
      free((maxbelief+i)->Nstates);
      free((maxbelief+i)->probs);
      
    }
  free(maxbelief);

  clear_mix_facgraph(fgraph);

  if (Nsigedges>0)
    free(sigdirvar2edge);

  *decomposedmodelpnt=decomposedmodels;
 
 
  return degmodels;
}


// Find a list of MAP configurations by recursively setting values of the variables with high connection to undetermined variables.
// Difference from findmaxconfig8: operate on mix_facgraph data structure.
// Difference from findmaxconfig9: use maxsumproduct3 function.
// Notice path selection variables are not fixed during the iteration, even though the beliefs are uneven.
// This is because there is only one factor node connect to each path selection variable.
// Difference from findmaxconfig10: ignore isolated variables when setting values.  An isolated variable does not connect to any factor node, i.e., is not associated with ko effects.
// Difference from previous findmaxconfig functions: can choose enumerating all MAP configs or narrowing down to one MAP config.
// mode=0: enumerate all MAP configs.  mode=1: choose one MAP config.
// Difference from findmaxconfig15: generate decomposed models.
// Difference from findmaxconfig16: exclude submodels which do not explain ko pairs, combine submodels which share non-invariant variables.
// Difference from findmaxconfig21: randomly fix a MAP config value.

struct decomposed_model * findmaxconfig22(struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int *invariant, int *Ndecomposedmodels, struct decomposed_model *decomposedmodels, int *mapconfigs, struct csparsegraph *model, int *sel2ref, char *anothergenenames)
{
  int i, j, k, l, *configuration, *isolated, Nclu, *validsubmodel, cnt;
  struct confignode *configtree;
  struct clique *clu;
  struct decomposed_model *newdecomposedmodels;
  
  configuration=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(configuration+i)=-1;

  // Find the max configs of variables.
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      int Nv=0, v[(maxbelief+i)->Nconfigs];
      double sumval, val, maxval=-1;
      
      sumval=0;
      for (j=0; j<=((maxbelief+i)->Nconfigs-1); j++)
	sumval+=*((maxbelief+i)->probs+j);
      for (j=0; j<=((maxbelief+i)->Nconfigs-1); j++)
	if ((val=*((maxbelief+i)->probs+j)/sumval)>maxval)
	  maxval=val;
      for (j=0; j<=((maxbelief+i)->Nconfigs-1); j++)
	{
	  if ((maxval-(val=*((maxbelief+i)->probs+j)/sumval))<1e-20)
	    {
	      Nv++; v[Nv-1]=j;
	    }
	}
      if (Nv==1)
	*(configuration+i)=v[0];
    }
  
  // Set the variables which are already fixed.
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      if (*(configuration+i)!=-1)
	{
	  int newfix=i;

	  for (j=0; j<=(fgraph->Ne*2-1); j++)
	    {
	      if (((maxmsg+j)->dir==0)&&((maxmsg+j)->from==newfix))
		{
		  int ind=*(configuration+newfix);
		  
		  for (k=0; k<=((maxmsg+j)->phi->Nconfigs-1); k++)
		    *((maxmsg+j)->phi->probs+k)=0.0;
		  *((maxmsg+j)->phi->probs+ind)=1.0;
		}
	    }
	}
    }

  // Identify isolated variables.
  // An isolated variable does not connect to any factor node.
  // Since path selection variables should not be fixed, they are also considered as isolated variables. 
  isolated=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(isolated+i)=1;
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    for (j=0; j<=((fgraph->facnodes+i)->phivi->Nvars-1); j++)
      *(isolated+*((fgraph->facnodes+i)->phivi->varinds+j))=0;
  for (i=0; i<=(fgraph->Nvars-1); i++)
    if ((fgraph->varnodes+i)->type==PATHSEL)
      *(isolated+i)=1;

  // Fix one variable each time until all configuration values are determined.
  // Build the root configuration node.
  configtree=malloc(sizeof(struct confignode));
  configtree->Nvars=fgraph->Nvars; configtree->level=0;
  configtree->config=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(configtree->config+i)=*(configuration+i);
  configtree->Nchildren=0;

  // Debug
  //configtree->ind=nodeind; nodeind++;

  configtree->ind=0;

  // The first decomposed submodel is the invariant submodel.
  
  decomposedmodels=recurse_fixval20(configtree,fgraph,maxmsg,maxbelief,isolated,invariant,Ndecomposedmodels,decomposedmodels,model,sel2ref,anothergenenames);

  /*
  // Merge decomposed models when they share non-invariant variables.
  Nclu=1; clu=malloc(sizeof(struct clique));
  clu->Nv=1; clu->v=malloc(sizeof(int)); *(clu->v)=0;
  for (i=1; i<=(*Ndecomposedmodels-1); i++)
    {
      int flag2=0;
      for (j=1; j<=(Nclu-1); j++)
	{
	  int flag=0; 
	  k=0;
	  while ((k<=(clu+j)->Nv-1)&&(flag==0))
	    {
	      int intersectflag=0;
	      l=0;
	      while ((l<=((decomposedmodels+i)->Nvars-1))&&(intersectflag==0))
		{
		  if ((*(invariant+*((decomposedmodels+i)->varinds+l))==0)&&(inlist((decomposedmodels+*((clu+j)->v+k))->Nvars,(decomposedmodels+*((clu+j)->v+k))->varinds,*((decomposedmodels+i)->varinds+l))==1))
		    intersectflag=1;


		  // Debug
		  if ((i==3)&&(j==1)&&(intersectflag==1))
		    {
		      int edgeind=*((decomposedmodels+i)->varinds+l)-model->Ne;
		      char ch;

		      printf("var[%d].type=%d\n",*((decomposedmodels+i)->varinds+l),(fgraph->varnodes+*((decomposedmodels+i)->varinds+l))->type);
		      printf("(");
		      k=0;
		      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+edgeind)->from)*GENENAMELENGTH+k))!=' '))
			{
			  printf("%c",ch); k++;
			}
		      printf(",");
		      k=0;
		      while ((k<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+edgeind)->to)*GENENAMELENGTH+k))!=' '))
			{
			  printf("%c",ch); k++;
			}
		      printf(")\n");

		      

		    }


		  l++;
		}
	      if (intersectflag==1)
		flag=1;
	      k++;
	    }
	  // Merge it with previous models.
	  if (flag==1)
	    {
	      flag2=1;
	      (clu+j)->Nv+=1;
	      (clu+j)->v=realloc((clu+j)->v,sizeof(int)*(clu+j)->Nv);
	      *((clu+j)->v+(clu+j)->Nv-1)=i;
	    }
	}
      // Create a new decomposedmodel.
      if (flag2==0)
	{
	  Nclu++; clu=realloc(clu,sizeof(struct clique)*Nclu);
	  (clu+Nclu-1)->Nv=1; (clu+Nclu-1)->v=malloc(sizeof(int));
	  *((clu+Nclu-1)->v)=i;
	}
    }
  */
  

  // Merge two decomposed models if the latter contains the independent variable of the former.
  Nclu=0; clu=malloc(sizeof(struct clique));
  for (i=0; i<=(*Ndecomposedmodels-1); i++)
    {
      int flag2=0;
      for (j=0; j<=(Nclu-1); j++)
	{
	  int flag=0; 
	  k=0;
	  while ((k<=(clu+j)->Nv-1)&&(flag==0))
	    {
	      //if (inlist((decomposedmodels+i)->Nvars,(decomposedmodels+i)->varinds,*((decomposedmodels+*((clu+j)->v+k))->varinds))==1)
	      if ((*((decomposedmodels+*((clu+j)->v+k))->varinds)>=0)&&(inlist((decomposedmodels+i)->Nvars,(decomposedmodels+i)->varinds,*((decomposedmodels+*((clu+j)->v+k))->varinds))==1))
		{
		  flag=1;
		}

	      k++;
	    }

	  // Merge it with previous models.
	  if (flag==1)
	    {
	      flag2=1;
	      (clu+j)->Nv+=1;
	      (clu+j)->v=realloc((clu+j)->v,sizeof(int)*(clu+j)->Nv);
	      *((clu+j)->v+(clu+j)->Nv-1)=i;
	    }
	}
      // Create a new decomposedmodel.
      if (flag2==0)
	{
	  Nclu++; clu=realloc(clu,sizeof(struct clique)*Nclu);
	  (clu+Nclu-1)->Nv=1; (clu+Nclu-1)->v=malloc(sizeof(int));
	  *((clu+Nclu-1)->v)=i;
	}
    }
 
  // Remove the submodel which contains less than 3 ko pairs.
  validsubmodel=malloc(sizeof(int)*Nclu);
  for (i=0; i<=(Nclu-1); i++)
    *(validsubmodel+i)=1;
  
  for (i=0; i<=(Nclu-1); i++)
    {
      int kocnt=0;

      for (j=0; j<=((clu+i)->Nv-1); j++)
	{
	  int ind=*((clu+i)->v+j);

	  for (k=0; k<=((decomposedmodels+ind)->Nvars-1); k++)
	    {
	      int varind=*((decomposedmodels+ind)->varinds+k);

	      //if ((fgraph->varnodes+varind)->type==KNOCKOUT)
	      if ((varind>=0)&&((fgraph->varnodes+varind)->type==KNOCKOUT))
		kocnt++;
	    }
	}
      if (kocnt<=3)
	*(validsubmodel+i)=0;
    }
  

  newdecomposedmodels=malloc(sizeof(struct decomposed_model)*Nclu); cnt=0;
  for (i=0; i<=(Nclu-1); i++)
    {
      if (*(validsubmodel+i)==1)
	{
	  cnt++;

	  (newdecomposedmodels+cnt-1)->Nvars=0;
	  k=0;
	  for (j=0; j<=((clu+i)->Nv-1); j++)
	    k+=(decomposedmodels+*((clu+i)->v+j))->Nvars;
	  (newdecomposedmodels+cnt-1)->varinds=malloc(sizeof(int)*k);
	  (newdecomposedmodels+cnt-1)->vartypes=malloc(sizeof(int)*k);
	  
	  for (j=0; j<=((clu+i)->Nv-1); j++)
	    {
	      for (k=0; k<=((decomposedmodels+*((clu+i)->v+j))->Nvars-1); k++)
		{
		  if (inlist((newdecomposedmodels+cnt-1)->Nvars,(newdecomposedmodels+cnt-1)->varinds,*((decomposedmodels+*((clu+i)->v+j))->varinds+k))!=1)
		    {
		      (newdecomposedmodels+cnt-1)->Nvars+=1;
		      *((newdecomposedmodels+cnt-1)->varinds+(newdecomposedmodels+cnt-1)->Nvars-1)=*((decomposedmodels+*((clu+i)->v+j))->varinds+k);
		      *((newdecomposedmodels+cnt-1)->vartypes+(newdecomposedmodels+cnt-1)->Nvars-1)=*((decomposedmodels+*((clu+i)->v+j))->vartypes+k);
		    }
		}
	    }
	  (newdecomposedmodels+cnt-1)->varinds=realloc((newdecomposedmodels+cnt-1)->varinds,sizeof(int)*(newdecomposedmodels+cnt-1)->Nvars);
	  (newdecomposedmodels+cnt-1)->vartypes=realloc((newdecomposedmodels+cnt-1)->vartypes,sizeof(int)*(newdecomposedmodels+cnt-1)->Nvars);
	}
    }

  for (i=0; i<=(*Ndecomposedmodels-1); i++)
    {
      if ((decomposedmodels+i)->Nvars>0)
	{
	  free((decomposedmodels+i)->varinds);
	  free((decomposedmodels+i)->vartypes);
	}      
      if ((decomposedmodels+i)->Nedges>0)
	free((decomposedmodels+i)->edgeinds);      
    }

  free(decomposedmodels);
  //*Ndecomposedmodels=Nclu;
  

  *Ndecomposedmodels=cnt;

 
  for (i=0; i<=(*Ndecomposedmodels-1); i++)
    {
      (newdecomposedmodels+i)->Nedges=0;
      (newdecomposedmodels+i)->edgeinds=malloc(sizeof(int)*(newdecomposedmodels+i)->Nvars);
      for (j=0; j<=((newdecomposedmodels+i)->Nvars-1); j++)
	{
	  int varind=*((newdecomposedmodels+i)->varinds+j);
	  int edgeind=(fgraph->varnodes+varind)->edgeind;
	  if (((fgraph->varnodes+varind)->type==EDGESIGN)&&(inlist((newdecomposedmodels+i)->Nedges,(newdecomposedmodels+i)->edgeinds,edgeind)==0))
	    {
	      (newdecomposedmodels+i)->Nedges+=1;
	      *((newdecomposedmodels+i)->edgeinds+(newdecomposedmodels+i)->Nedges-1)=edgeind;
	    }
	}
      (newdecomposedmodels+i)->edgeinds=realloc((newdecomposedmodels+i)->edgeinds,sizeof(int)*(newdecomposedmodels+i)->Nedges);

    }
  decomposedmodels=newdecomposedmodels;
  for (i=0; i<=(Nclu-1); i++)
    free((clu+i)->v);
  free(clu);
  
  // Obtain the MAP configurations from the configtree.
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(mapconfigs+i)=*(configtree->config+i);

  // Free the space allocated to configtree.
  recurse_freeconfignode(configtree); free(validsubmodel);
  free(configtree); free(configuration); free(isolated);
  
  
  return decomposedmodels;
}


// Recursively fix values of the variables with the highest connection to undetermined variables.
// Difference from recurse_fixval: operates on mix_facgraph structure and does not require sumproduct.
// Difference from recurse_fixval2: use maxsumproduct3 function to perform max product or sum product.
// Difference from recurse_fixval3: ignore isolated variables when setting values.
// Difference from recurse_fixval9: fix variables to one value.
// Difference from recurse_fixval10: update decomposed models.
// Difference from recurse_fixval11: create invariant submodel.
// Difference from recurse_fixval19: randomly fix a MAP config value.

struct decomposed_model * recurse_fixval20(struct confignode *configtree, struct mix_facgraph *fgraph, struct message *maxmsg, struct potential *maxbelief, int *isolated, int *invariant, int *Ndecomposedmodels, struct decomposed_model *decomposedmodels, struct csparsegraph *model, int *sel2ref, char *anothergenenames)
{
  int i, j, k, l, *influencecnt, undetermined, maxconnnode;
  struct message *tmpmsg;
  struct potential *tmpbelief;
  int Nnewdetermined, *newdetermined, *configuration, *fixedvals;
    
  configuration=configtree->config;

  // Debug
  printf("level=%d, \n",configtree->level);

 
  // If all non-isolated variables are fixed, then return.
  undetermined=0; i=0;
  while ((i<=(fgraph->Nvars-1))&&(undetermined==0))
    {
      if ((*(isolated+i)==0)&&(*(configuration+i)==-1))
	undetermined=1;
      i++;
    }
  if (undetermined==0)
    return decomposedmodels;
    

  // Perform maxproduct.
  tmpmsg=malloc(sizeof(struct message)*fgraph->Ne*2);
  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      (tmpmsg+i)->from=(maxmsg+i)->from;
      (tmpmsg+i)->to=(maxmsg+i)->to;
      (tmpmsg+i)->dir=(maxmsg+i)->dir;
      (tmpmsg+i)->phi=malloc(sizeof(struct potential));
      (tmpmsg+i)->phi->Nvars=1; 
      (tmpmsg+i)->phi->varinds=malloc(sizeof(int));
      *((tmpmsg+i)->phi->varinds)=*((maxmsg+i)->phi->varinds);
      (tmpmsg+i)->phi->Nconfigs=(maxmsg+i)->phi->Nconfigs;
      (tmpmsg+i)->phi->Nstates=malloc(sizeof(int));
      *((tmpmsg+i)->phi->Nstates)=*((maxmsg+i)->phi->Nstates);
      (tmpmsg+i)->phi->probs=malloc(sizeof(double)*(tmpmsg+i)->phi->Nconfigs);
      for (j=0; j<=((tmpmsg+i)->phi->Nconfigs-1); j++)
	*((tmpmsg+i)->phi->probs+j)=*((maxmsg+i)->phi->probs+j);

      // Reset the messages going out of path sel vars to be even.
      if (((tmpmsg+i)->dir==0)&&((fgraph->varnodes+(tmpmsg+i)->from)->type==PATHSEL))
	for (j=0; j<=((tmpmsg+i)->phi->Nconfigs-1); j++)
	  *((tmpmsg+i)->phi->probs+j)=1.0;    
    }

  tmpbelief=malloc(sizeof(struct potential)*fgraph->Nvars);
  
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      (tmpbelief+i)->Nvars=1;
      (tmpbelief+i)->varinds=malloc(sizeof(int));
      *((tmpbelief+i)->varinds)=*((maxbelief+i)->varinds);
      (tmpbelief+i)->Nconfigs=(maxbelief+i)->Nconfigs;
      (tmpbelief+i)->Nstates=malloc(sizeof(int));
      *((tmpbelief+i)->Nstates)=*((maxbelief+i)->Nstates);
      (tmpbelief+i)->probs=malloc(sizeof(double)*(tmpbelief+i)->Nconfigs);
      for (j=0; j<=((tmpbelief+i)->Nconfigs-1); j++)
	*((tmpbelief+i)->probs+j)=*((maxbelief+i)->probs+j);
    }

  // Do not fix the path selection variables.
  fixedvals=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)  
    *(fixedvals+i)=-1;
  for (i=0; i<=(fgraph->Nvars-1); i++) 
    if ((fgraph->varnodes+i)->type!=PATHSEL)
      *(fixedvals+i)=*(configuration+i);

  maxsumproduct3(fgraph,tmpmsg,fixedvals,tmpbelief,0,0);

  // Detect newly determined vars.
  Nnewdetermined=0; newdetermined=malloc(sizeof(int)*fgraph->Nvars);
  

  // Determine the configuration value of previously undetermined variables or path sel variables.
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      if ((*(configuration+i)==-1)||((fgraph->varnodes+i)->type==PATHSEL))
	{
	  int Nv=0, v[(maxbelief+i)->Nconfigs];
	  double sumval, val, maxval=-1;
      
	  sumval=0;
	  for (j=0; j<=((tmpbelief+i)->Nconfigs-1); j++)
	    sumval+=*((tmpbelief+i)->probs+j);
	  for (j=0; j<=((tmpbelief+i)->Nconfigs-1); j++)
	    if ((val=*((tmpbelief+i)->probs+j)/sumval)>maxval)
	      maxval=val;
	  for (j=0; j<=((tmpbelief+i)->Nconfigs-1); j++)
	    {
	      //if ((maxval-(val=*((tmpbelief+i)->probs+j)/sumval))<1e-20)
	      if ((maxval-(val=*((tmpbelief+i)->probs+j)/sumval))<1e-6)
		{
		  Nv++; v[Nv-1]=j;
		}
	    }
	  if (Nv==1)
	    {
	      if ((*(configuration+i)==-1)&&(v[0]!=-1))
		{
		  Nnewdetermined++; *(newdetermined+Nnewdetermined-1)=i;
		}

	      *(configuration+i)=v[0];
	    }
	}
    }

  // Identify invariant variables.
  if (configtree->level==0)
    {
      for (i=0; i<=(fgraph->Nvars-1); i++)
	{
	  if (*(configuration+i)!=-1)
	    *(invariant+i)=1;
	  else
	    *(invariant+i)=0;
	}
    }
  
  // Identify variales in the current decomposed model.
  // The independent variable is given.  The dependent variables are the newly determined variables.  Also add variables in previous decomposed models essential for the determinations of dependent variable values.

  // Generate the first submodel.
  if (*Ndecomposedmodels==0)
    {
      *Ndecomposedmodels+=1;
      //(decomposedmodels+*Ndecomposedmodels-1)->Nvars+=Nnewdetermined;
      //(decomposedmodels+*Ndecomposedmodels-1)->Nvars=Nnewdetermined+1;
      
      (decomposedmodels+*Ndecomposedmodels-1)->Nvars=Nnewdetermined;
      (decomposedmodels+*Ndecomposedmodels-1)->Nedges=0;


      //(decomposedmodels+*Ndecomposedmodels-1)->varinds=realloc((decomposedmodels+*Ndecomposedmodels-1)->varinds,sizeof(int)*fgraph->Nvars);
      //(decomposedmodels+*Ndecomposedmodels-1)->vartypes=realloc((decomposedmodels+*Ndecomposedmodels-1)->vartypes,sizeof(int)*fgraph->Nvars);


      (decomposedmodels+*Ndecomposedmodels-1)->varinds=malloc(sizeof(int)*fgraph->Nvars);
      (decomposedmodels+*Ndecomposedmodels-1)->vartypes=malloc(sizeof(int)*fgraph->Nvars);

      
      bzero((decomposedmodels+*Ndecomposedmodels-1)->varinds,sizeof(int)*fgraph->Nvars);
      bzero((decomposedmodels+*Ndecomposedmodels-1)->vartypes,sizeof(int)*fgraph->Nvars);
      for (i=0; i<=(fgraph->Nvars-1); i++)
	{
	  *((decomposedmodels+*Ndecomposedmodels-1)->varinds+i)=-1;
	  *((decomposedmodels+*Ndecomposedmodels-1)->vartypes+i)=-1;
	}
      


      //*((decomposedmodels+*Ndecomposedmodels-1)->varinds)=-1;

      l=0;

      for (i=0; i<=(Nnewdetermined-1); i++)
	{
	  int isinclude=0;

	  // Only include edge signs, dirs and knockout effects vars.
	  // Include knockout effects whose all explanatory vars are determined. 
	  
	  if (((fgraph->varnodes+*(newdetermined+i))->type==EDGESIGN)||((fgraph->varnodes+*(newdetermined+i))->type==EDGEDIR))
	    isinclude=1;
	  else if ((fgraph->varnodes+*(newdetermined+i))->type==KNOCKOUT)
	    {
	      isinclude=0; j=0;
	      while ((j<=(fgraph->Ninteractions-1))&&(isinclude==0))
		{
		  if (((fgraph->facnodes+j)->phivi->Nvars>1)&&(inlist((fgraph->facnodes+j)->phivi->Nvars,(fgraph->facnodes+j)->phivi->varinds,*(newdetermined+i))==1))
		    {
		      k=0; isinclude=1;
		      while ((k<=((fgraph->facnodes+j)->phivi->Nvars-1))&&(isinclude==1))
			{
			  if (*(invariant+*((fgraph->facnodes+j)->phivi->varinds+k))==0)
			    isinclude=0;
			  k++;
			}
		    }
		  j++;
		}
	    }
	      
 
	  /*
	  if (isinclude==1)
	    {
	      *((decomposedmodels+*Ndecomposedmodels-1)->varinds+i+1)=*(newdetermined+i);
	      *((decomposedmodels+*Ndecomposedmodels-1)->vartypes+i+1)=0;
	    }
	  */

	  if (isinclude==1)
	    {
	      *((decomposedmodels+*Ndecomposedmodels-1)->varinds+l)=*(newdetermined+i);
	      *((decomposedmodels+*Ndecomposedmodels-1)->vartypes+l)=0;
	      l++;
	    }

	}

      (decomposedmodels+*Ndecomposedmodels-1)->varinds=realloc((decomposedmodels+*Ndecomposedmodels-1)->varinds,sizeof(int)*l);
      (decomposedmodels+*Ndecomposedmodels-1)->vartypes=realloc((decomposedmodels+*Ndecomposedmodels-1)->vartypes,sizeof(int)*l);
      (decomposedmodels+*Ndecomposedmodels-1)->Nvars=l;


    }
         
  // Generate submodels containing variant variables.
  else if (*Ndecomposedmodels>0)
    {
      int *inprevious;

      inprevious=malloc(sizeof(int)*fgraph->Nvars);
      for (i=0; i<=(fgraph->Nvars-1); i++)
	*(inprevious+i)=0;
      //for (i=0; i<=(*Ndecomposedmodels-2); i++)
      //*(inprevious+*((decomposedmodels+i)->varinds+0))=1;
            
      for (i=0; i<=(*Ndecomposedmodels-2); i++)
	for (j=0; j<=((decomposedmodels+i)->Nvars-1); j++)
	  *(inprevious+*((decomposedmodels+i)->varinds+j))=1;
      for (i=0; i<=(fgraph->Nvars-1); i++)
	if (*(invariant+i)==1)
	  *(inprevious+i)=0;
       

      (decomposedmodels+*Ndecomposedmodels-1)->Nvars+=Nnewdetermined;

      j=*((decomposedmodels+*Ndecomposedmodels-1)->varinds);
      k=*((decomposedmodels+*Ndecomposedmodels-1)->vartypes);

      (decomposedmodels+*Ndecomposedmodels-1)->varinds=realloc((decomposedmodels+*Ndecomposedmodels-1)->varinds,sizeof(int)*fgraph->Nvars);
      (decomposedmodels+*Ndecomposedmodels-1)->vartypes=realloc((decomposedmodels+*Ndecomposedmodels-1)->vartypes,sizeof(int)*fgraph->Nvars);

      
      //(decomposedmodels+*Ndecomposedmodels-1)->varinds=malloc(sizeof(int)*fgraph->Nvars);
      //(decomposedmodels+*Ndecomposedmodels-1)->vartypes=malloc(sizeof(int)*fgraph->Nvars);

      bzero((decomposedmodels+*Ndecomposedmodels-1)->varinds,sizeof(int)*fgraph->Nvars);
      bzero((decomposedmodels+*Ndecomposedmodels-1)->vartypes,sizeof(int)*fgraph->Nvars);
      for (i=0; i<=(fgraph->Nvars-1); i++)
	{
	  *((decomposedmodels+*Ndecomposedmodels-1)->varinds+i)=-1;
	  *((decomposedmodels+*Ndecomposedmodels-1)->vartypes+i)=-1;
	}
      *((decomposedmodels+*Ndecomposedmodels-1)->varinds)=j;
      *((decomposedmodels+*Ndecomposedmodels-1)->vartypes)=k;


      for (i=0; i<=(Nnewdetermined-1); i++)
	{
	  *((decomposedmodels+*Ndecomposedmodels-1)->varinds+i+1)=*(newdetermined+i);
	  *((decomposedmodels+*Ndecomposedmodels-1)->vartypes+i+1)=2;
	}

      for (i=0; i<=(Nnewdetermined-1); i++)
	{
	  // Find the factor nodes which contain the independent variable and the newly determined variable.
	  for (j=0; j<=(fgraph->Ninteractions-1); j++)
	    {
	      if ((inlist((fgraph->facnodes+j)->phivi->Nvars,(fgraph->facnodes+j)->phivi->varinds,*((decomposedmodels+*Ndecomposedmodels-1)->varinds))==1)&&(inlist((fgraph->facnodes+j)->phivi->Nvars,(fgraph->facnodes+j)->phivi->varinds,*(newdetermined+i))==1))
		{
		  int flag=1;

		  // If all other variables connecting to this factor are either invariant, or contained in previous decomposed models, or are newly determined variables, then incorporate these variables.
		  // Don't include edge presence, path sel vars.
		  k=0;
		  while ((k<=((fgraph->facnodes+j)->phivi->Nvars-1))&&(flag==1))
		    {
		      int varind=*((fgraph->facnodes+j)->phivi->varinds+k);
		      if (((fgraph->varnodes+varind)->type!=PATHSEL)&&(varind!=*((decomposedmodels+*Ndecomposedmodels-1)->varinds))&&(inlist(Nnewdetermined,newdetermined,varind)==0)&&(*(invariant+varind)!=1)&&(*(inprevious+varind)!=1))
			flag=0;
		      k++;
		    }

		  if (flag==1)
		    {
		      for (k=0; k<=((fgraph->facnodes+j)->phivi->Nvars-1); k++)
			{
			  int varind=*((fgraph->facnodes+j)->phivi->varinds+k);
			  if (((fgraph->varnodes+varind)->type!=PATHSEL)&&((fgraph->varnodes+varind)->type!=EDGEPRESENCE)&&(inlist((decomposedmodels+*Ndecomposedmodels-1)->Nvars,(decomposedmodels+*Ndecomposedmodels-1)->varinds,varind)==0))
			    {
			      (decomposedmodels+*Ndecomposedmodels-1)->Nvars+=1;
			      *((decomposedmodels+*Ndecomposedmodels-1)->varinds+(decomposedmodels+*Ndecomposedmodels-1)->Nvars-1)=varind;
			      if (*(invariant+varind)==1)
				*((decomposedmodels+*Ndecomposedmodels-1)->vartypes+(decomposedmodels+*Ndecomposedmodels-1)->Nvars-1)=0;
			      else if (*(inprevious+varind)==1)
				*((decomposedmodels+*Ndecomposedmodels-1)->vartypes+(decomposedmodels+*Ndecomposedmodels-1)->Nvars-1)=1;
			    }
			}
		    }
		}
	    }
	}

      (decomposedmodels+*Ndecomposedmodels-1)->varinds=realloc((decomposedmodels+*Ndecomposedmodels-1)->varinds,sizeof(int)*(decomposedmodels+*Ndecomposedmodels-1)->Nvars);
      (decomposedmodels+*Ndecomposedmodels-1)->vartypes=realloc((decomposedmodels+*Ndecomposedmodels-1)->vartypes,sizeof(int)*(decomposedmodels+*Ndecomposedmodels-1)->Nvars);
      

      free(inprevious);
  

      // Identify the edges in the current decomposed model.
      (decomposedmodels+*Ndecomposedmodels-1)->Nedges=0;
      (decomposedmodels+*Ndecomposedmodels-1)->edgeinds=realloc((decomposedmodels+*Ndecomposedmodels-1)->edgeinds,sizeof(int)*(decomposedmodels+*Ndecomposedmodels-1)->Nvars);
      for (i=0; i<=((decomposedmodels+*Ndecomposedmodels-1)->Nvars-1); i++)
	{
	  int varind=*((decomposedmodels+*Ndecomposedmodels-1)->varinds+i);
	  int edgeind=(fgraph->varnodes+varind)->edgeind;
	  if (((fgraph->varnodes+varind)->type==EDGESIGN)&&(inlist((decomposedmodels+*Ndecomposedmodels-1)->Nedges,(decomposedmodels+*Ndecomposedmodels-1)->edgeinds,edgeind)==0))
	    {
	      (decomposedmodels+*Ndecomposedmodels-1)->Nedges+=1;
	      *((decomposedmodels+*Ndecomposedmodels-1)->edgeinds+(decomposedmodels+*Ndecomposedmodels-1)->Nedges-1)=edgeind;
	    }
	}
      (decomposedmodels+*Ndecomposedmodels-1)->edgeinds=realloc((decomposedmodels+*Ndecomposedmodels-1)->edgeinds,sizeof(int)*(decomposedmodels+*Ndecomposedmodels-1)->Nedges);

    }

  
  // Debug
  if (configtree->level>0)
    {
      char ch;
      printf("decomposedmodels[%d] contains %d vars: ",*Ndecomposedmodels-1,(decomposedmodels+*Ndecomposedmodels-1)->Nvars);

      for (i=0; i<=((decomposedmodels+*Ndecomposedmodels-1)->Nvars-1); i++)
	{
	  int edgeind=-1, koind=-1, ind=*((decomposedmodels+*Ndecomposedmodels-1)->varinds+i);
	  if ((fgraph->varnodes+ind)->type!=KNOCKOUT)
	    {
	      printf("var[%d]=%d, type=%d, ",ind,*(configuration+ind),*((decomposedmodels+*Ndecomposedmodels-1)->vartypes+i));
	      if ((fgraph->varnodes+ind)->type==0)
		{
		  edgeind=(fgraph->varnodes+ind)->edgeind;
		  printf("edge presence ");
		}
	      else if ((fgraph->varnodes+ind)->type==1)
		{
		  edgeind=(fgraph->varnodes+ind)->edgeind;
		  printf("edge sign ");
		}
	      else if ((fgraph->varnodes+ind)->type==2)
		{
		  edgeind=(fgraph->varnodes+ind)->edgeind;
		  printf("edge dir ");
		}
	      else if ((fgraph->varnodes+ind)->type==3)
		{
		  koind=(fgraph->varnodes+ind)->koind;
		  printf("ko ");
		}
	  
	      if (edgeind>=0)
		{
		  printf("(%d,",edgeind);
		  if ((model->e+edgeind)->type==0)
		    printf("pd,");
		  else
		    printf("pp,");
		  j=0;
		  while ((j<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+edgeind)->from)*GENENAMELENGTH+j))!=' '))
		    {
		      printf("%c",ch); j++;
		    }
		  printf(",");
		  j=0;
		  while ((j<=(GENENAMELENGTH-1))&&((ch=*(anothergenenames+*(sel2ref+(model->e+edgeind)->to)*GENENAMELENGTH+j))!=' '))
		    {
		      printf("%c",ch); j++;
		    }
		  printf("); ");
		}
	      else if (koind>=0)
		{
		  printf("%d; ",koind);
		}
	    }
	}
      
      printf("\n");
    }

  free(newdetermined);

  for (i=0; i<=(fgraph->Ne*2-1); i++)
    {
      free((tmpmsg+i)->phi->varinds);
      free((tmpmsg+i)->phi->Nstates);
      free((tmpmsg+i)->phi->probs);
      free((tmpmsg+i)->phi);
    }
  free(tmpmsg); free(fixedvals);
  

  // Again, if all non-isolated variables are fixed, then return.
  undetermined=0; i=0;
  while ((i<=(fgraph->Nvars-1))&&(undetermined==0))
    {
      if ((*(isolated+i)==0)&&(*(configuration+i)==-1))
	undetermined=1;
      i++;
    }


  // Debug
  //if (configtree->level==4)
  //undetermined=0;


  if (undetermined==0)
    {
      for (i=0; i<=(fgraph->Nvars-1); i++)
	{
	  free((tmpbelief+i)->varinds);
	  free((tmpbelief+i)->Nstates);
	  free((tmpbelief+i)->probs);
	}
      
      free(tmpbelief);  
      return decomposedmodels;
    }   


  // Otherwise choose the (edge) variable which has the highest connection to other undetermined edge variables.
  influencecnt=malloc(sizeof(int)*fgraph->Nvars);
  for (i=0; i<=(fgraph->Nvars-1); i++)
    *(influencecnt+i)=0; 

  // Find the number of one-neighbors of each variable nodes.
  // Do not count the neighbors of variable nodes if the path corresponding to the factor is not selected.
  for (i=0; i<=(fgraph->Ninteractions-1); i++)
    {
      int pathselind=(fgraph->facnodes+i)->sigma;
      if ((pathselind<0)||((pathselind>=0)&&(*(configuration+pathselind)==1)))
	{
	  for (j=0; j<=((fgraph->facnodes+i)->phivi->Nvars-2); j++)
	    {
	      for (k=(j+1); k<=((fgraph->facnodes+i)->phivi->Nvars-1); k++)
		{
		  int ind1, ind2;
		  ind1=*((fgraph->facnodes+i)->phivi->varinds+j);
		  ind2=*((fgraph->facnodes+i)->phivi->varinds+k);
		  if ((*(configuration+ind1)==-1)&&(*(configuration+ind2)==-1)&&(*(isolated+ind1)==0)&&(*(isolated+ind2)==0))
		    {
		      *(influencecnt+ind1)+=1;
		      *(influencecnt+ind2)+=1;
		    }
		}
	    }
	}
    }
    
  // Fix the variable of the maximum influence and determine other configuration values.
  // Based on sumproduct but check if maxproduct is consistent.
  j=-1; k=0;
  for (i=0; i<=(fgraph->Nvars-1); i++)
    {
      if (*(influencecnt+i)>k)
	{
	  k=*(influencecnt+i); j=i;
	}
    }
  maxconnnode=j; free(influencecnt);

  if (maxconnnode<0)
    return decomposedmodels;

  
  // Debug
  //printf("maxconnnode=%d \n",maxconnnode);


  // Set the value of an undetermined variable to a random MAP config value.
  {
    int Nv, *v;
    double maxval=0;
    time_t *tp;

    v=malloc(sizeof(int)*((tmpbelief+maxconnnode)->Nconfigs));
    for (i=0; i<=((tmpbelief+maxconnnode)->Nconfigs-1); i++)
      if (*((tmpbelief+maxconnnode)->probs+i)>maxval)
	maxval=*((tmpbelief+maxconnnode)->probs+i);
    Nv=0;
    for (i=0; i<=((tmpbelief+maxconnnode)->Nconfigs-1); i++)
      {
	if ((maxval-*((tmpbelief+maxconnnode)->probs+i))<=1e-5)
	  {
	    Nv++; *(v+Nv-1)=i;
	  }
      }
    tp=malloc(sizeof(time_t)); srand((int)(time(tp)));
    k=rand(); k=k%Nv; k=*(v+k);
    *(configuration+maxconnnode)=k;	    
    configtree->Nchildren=0; configtree->level+=1;
    free(v); free(tp);

    for (i=0; i<=(fgraph->Nvars-1); i++)
      {
	free((tmpbelief+i)->varinds);
	free((tmpbelief+i)->Nstates);
	free((tmpbelief+i)->probs);
      }
  
    free(tmpbelief); 

    // Debug
    printf("maxconnnode=%d, val=%d\n",maxconnnode,k);
  }



  // Iteratively set the value of an undetermined variable to 1 until all variables are set.
  //configtree->Nchildren=0; configtree->level+=1;
  //*(configuration+maxconnnode)=1;

  
  // Create a new decomposed model.  The independent variable is maxconnnode.
  *Ndecomposedmodels+=1;
  decomposedmodels=realloc(decomposedmodels,sizeof(struct decomposed_model)*(*Ndecomposedmodels));
  (decomposedmodels+*Ndecomposedmodels-1)->Nvars=1;
  (decomposedmodels+*Ndecomposedmodels-1)->varinds=malloc(sizeof(int));
  *((decomposedmodels+*Ndecomposedmodels-1)->varinds)=maxconnnode;
  (decomposedmodels+*Ndecomposedmodels-1)->vartypes=malloc(sizeof(int));
  *((decomposedmodels+*Ndecomposedmodels-1)->vartypes)=1;
  (decomposedmodels+*Ndecomposedmodels-1)->edgeinds=malloc(sizeof(int));

  
  decomposedmodels=recurse_fixval20(configtree,fgraph,maxmsg,maxbelief,isolated,invariant,Ndecomposedmodels,decomposedmodels,model,sel2ref,anothergenenames);

  return decomposedmodels;

}

